-- Got Lakes Map Script
-- Author: Scrum Lord
-- Ported from Civ V BNW to Civ VI.
-- Purpose: a customizable land + climate shuffler that includes lakes and 
-- many other types of maps.
------------------------------------------------------------------------------
------------------------------------------------------------------------------
include "MapEnums"
include "MapUtilities"
include "MountainsCliffs"
include "RiversLakes"
include "FeatureGenerator"
include "TerrainGenerator"
include "NaturalWonderGenerator"
include "ResourceGenerator"
include "AssignStartingPlots"

------------------------------------------------------------------------------
-- Got Lakes version info for debugging
------------------------------------------------------------------------------
local glVersion = 17;
------------------------------------------------------------------------------
-- Map Option defaults
------------------------------------------------------------------------------
local GOT_ALL  = -1;
local GOT_NONE = -2;

-- Defaults for actually playing the game.
-- Maximizes randomness except for resources.
local MapDefaults = 
{
	WORLDAGE    = 4,
	SEALEVEL    = 4,
	TEMPERATURE = 4,
	RAINFALL    = 4,
	RESOURCES   = 2, -- standard
	START       = 2, -- standard
	WORLDWRAP   = 3,
	MOUNTAINS   = 7,
	EXTRAS      = 8,
	CONTINENTS  = GOT_ALL, -- shuffle all landmasses and climates
	ISLANDS     = GOT_ALL,
	LAKES       = GOT_ALL,
	PANGEA      = GOT_ALL,
	COLD        = GOT_ALL,
	TEMPERATE   = GOT_ALL,
	TILTED      = GOT_ALL,
	WARM        = GOT_ALL
};

--[[
-- Defaults for development & debugging
-- Minimizes randomness
MapDefaults = 
{
	WORLDAGE    = 2, -- 4B yrs
	SEALEVEL    = 2, -- medium
	TEMPERATURE = 2, -- normal
	RAINFALL    = 2, -- normal
	RESOURCES   = 2, -- standard
	START       = 2, -- standard
	WORLDWRAP   = 2, -- region
	MOUNTAINS   = 1, -- tectonic plates
	EXTRAS      = 1, -- none
	CONTINENTS  = GOT_NONE, -- None for most lands and climates.
	ISLANDS     = GOT_NONE,
	LAKES       = 4,        -- Large lakes
	PANGEA      = GOT_NONE,
	COLD        = GOT_NONE,
	TEMPERATE   = 3,        -- Prairie
	TILTED      = GOT_NONE,
	WARM        = GOT_NONE
};
-- ]]
------------------------------------------------------------------------------
-- Map options
--
-- The order and numbering of these option constants must 
-- match the order in which the options appear in the 
-- CustomOptions list in GetMapScriptInfo().
------------------------------------------------------------------------------
local MapOptions = 
{
	-- Core options
	WORLDAGE    = 1, -- plot types (landmasses)
	SEALEVEL    = 2, -- plot types (landmasses)
	TEMPERATURE = 3, -- terrain and features (climates)
	RAINFALL    = 4, -- features (climates)
	RESOURCES   = 5, -- resource abundance
	START       = 6, -- starting plots
	
	-- Got Lakes additional options
	WORLDWRAP   = 7, -- Globe, Region, or Random
	MOUNTAINS   = 8, -- Mountain Pattern
	EXTRAS      = 9, -- Additional islands/lakes
	
	-- Landmass families
	CONTINENTS  = 10,
	ISLANDS     = 11,
	LAKES       = 12,
	PANGEA      = 13,
	
	-- Climate families
	COLD        = 14,
	TEMPERATE   = 15,
	TILTED      = 16,
	WARM        = 17
};

local FIRST_LAND_OPT    = MapOptions.CONTINENTS;
local LAST_LAND_OPT     = MapOptions.PANGEA;

local FIRST_CLIMATE_OPT = MapOptions.COLD;
local LAST_CLIMATE_OPT  = MapOptions.WARM;

-- Map names for Civ 6 option lookup.
-- The order of these names must match the order of the 
--corresponding map option indices.
local MapOptionNames = 
{
	-- Core options
	"world_age",   -- plot types (all landmasses)
	"sea_level",   -- plot types (all landmasses)
	"temperature", -- terrain (all climates)
	"rainfall",    -- features (all climates)
	"resources",   -- starting plots (all maps)	
	"start",       -- starting plots (all maps)
	
	-- Got Lakes additional options
	"world_wrap",  -- Globe, Region, or Random
	"mountains",   -- Mountain Pattern
	"extras",      -- Additional islands/lakes
	
	-- Landmass families
	"continents",
	"islands",
	"lakes",
	"pangea",
	
	-- Climate families
	"cold",
	"temperate",
	"tilted",
	"warm"
};

------------------------------------------------------------------------------
-- Possible non-random values for options
------------------------------------------------------------------------------
-- World age
local WORLDAGE_YOUNG  = 1;
local WORLDAGE_NORMAL = 2;
local WORLDAGE_OLD    = 3;

-- Sea level
local SEALEVEL_LOW    = 1;
local SEALEVEL_NORMAL = 2;
local SEALEVEL_HIGH   = 3;

-- Temperature.
local TEMPERATURE_COOL   = 1;
local TEMPERATURE_NORMAL = 2;
local TEMPERATURE_HOT    = 3;

-- Rainfall.
local RAINFALL_DRY    = 1;
local RAINFALL_NORMAL = 2;
local RAINFALL_WET    = 3;

-- Start Position.
local START_BALANCED = 1;
local START_NORMAL   = 2;
local START_LEGEND   = 3;

-- World Wrap
local WWRAP_GLOBE   = 1; -- wrap-x only
local WWRAP_REGION  = 2; -- no wrap

-- Mountain patterns
local MtnPatterns = 
{
	TECTONIC  = 1, -- tectonic a.k.a. normal
	FRACTAL   = 2, -- fractal, snaky ranges
	SCATTERED = 3, -- fractal, lone mountains and hills
	CLUSTERED = 4, -- fractal, groups of mountains / hills
	UNCANNY   = 5, -- fractal, double-ranges (valleys)
	WINDING   = 6, -- fractal, pockets of non-mountains
};

-- Extras
local ExtraIslesLakes = 
{
	NONE     = 1, -- No extra islands
	CHAINS   = 2, -- Chains of islands (based on "plus" maps)
	CLUSTERS = 3, -- Dense clusters of islands
	DONUTS   = 4, -- Mini Donuts
	TECTONIC = 5, -- Tectonic islands
	TCHAINS  = 6, -- Chains and Tectonic islands
	ALL      = 7, -- All types of extra islands
};
local NUM_EXTRAS_OPTS = ExtraIslesLakes.ALL;

-- Supported landmasses (possible values of UserInput.landmass)
local Landmasses = 
{
	ARCHIPELAGO = 1,  -- fractal
	BAGELS      = 2,  -- MLF bagels
	CONTINENTS  = 3,  -- fractal continents
	DONUT       = 4,  -- MLF donut
	EXTRAS      = 5,  -- MLF extras
	FRACTAL     = 6,  -- fractal
	INLANDSEA   = 7,  -- MLF inland sea
	INVEXTRAS   = 8,  -- MLF inverted extras
	ISLEPLATES  = 9,  -- MLF island plates
	LAKES       = 10, -- MLF lakes
	LRGISLES    = 11, -- MLF large isles
	LRGLAKES    = 12, -- MLF large lakes
	MAZE        = 13, -- MLF Maze
	OVAL        = 14, -- MLF oval
	PANGEA      = 15, -- fractal pangea
	SALTLAKES   = 16, -- MLF large lakes
	SMALLCONTS  = 17, -- fractal
	TINYISLES   = 18, -- fractal
};


-- Supported climates (possible values of UserInput.climate)
local Climates = 
{
	ARBOREA     = 1,  -- 
	EVERGREEN   = 2,  --
	ICEAGECROP  = 3,  -- (cropped)
	ICEAGEFULL  = 4,  -- (full)
	NORMAL      = 5,  -- (no overrides)
	PRAIRIE     = 6,  -- 
	RAINFOREST  = 7,  -- 
	SANDSTORM   = 8,  -- 
	SANDYJUNGLE = 9,  -- 
	TILTEDCROP  = 10, -- (cropped)
	TILTEDFULL  = 11, -- (full)
	TWOSUNS     = 12, --
	WETLAND     = 13, -- 
};

------------------------------------------------------------------------------
-- Misc constants
------------------------------------------------------------------------------
-- Directions for traversing to a hex tile's neighbors.
--
-- For some reason, DirectionTypes becomes undefined
-- after hard-killing WB, so for easier debugging let's
-- use internally-defined constants instead.
local DIRECTION_NE = 0;
local DIRECTION_E  = 1;
local DIRECTION_SE = 2;
local DIRECTION_SW = 3;
local DIRECTION_W  = 4;
local DIRECTION_NW = 5;

-- All six directions in a table for convenience.
local HexDirections = 
{
	DIRECTION_NE, 
	DIRECTION_E, 
	DIRECTION_SE, 
	DIRECTION_SW, 
	DIRECTION_W, 
	DIRECTION_NW 
};

-- Map Size IDs for mapping worldSize hash to a named map size.
local MAPSIZE_DUEL     = "MAPSIZE_DUEL";
local MAPSIZE_TINY     = "MAPSIZE_TINY";
local MAPSIZE_SMALL    = "MAPSIZE_SMALL";
local MAPSIZE_STANDARD = "MAPSIZE_STANDARD";
local MAPSIZE_LARGE    = "MAPSIZE_LARGE";
local MAPSIZE_HUGE     = "MAPSIZE_HUGE";

-- Trig constants
local TAU = 2*math.pi;

-- Possible directions for shifting a pangea.
local PANGEA_SHIFT_NORTH = 1;
local PANGEA_SHIFT_SOUTH = 2;

-- Possible values for donut fillings.
local FILLING_NO  = 0; -- Ignored by donut filling algos
local FILLING_YES = 1; -- Special donut plot

-- The maximum number of tiles that a fresh-water lake can have.
-- Any bigger and the lake becomes a salt-water ocean.
local LAKE_MAXNUMTILES = 9;

-- Special latitude-terrain mapping for tilted axis.
local TILTEDAXIS_SNOWLAT   = 0.95;
local TILTEDAXIS_TUNDRALAT = 0.65; 
local TILTEDAXIS_GRASSLAT  = 0.5;
local TILTEDAXIS_JUNGLELAT = 0.4;
local TILTEDAXIS_DESERTLAT = 0.35; -- a.k.a hot pole sand zone

-- Regional division methods.
local DIVMETHOD_BIGGESTLAND = 1;
local DIVMETHOD_CONTINENTAL = 2; 
local DIVMETHOD_RECTANGULAR = 3; 
local DIVMETHOD_SUBRECT     = 4; 

------------------------------------------------------------------------------
-- User input globals to indicate chosen map configuration
-- These determine various choices such as function args, 
-- method overrides, and conditional statements.
------------------------------------------------------------------------------
-- User selections for map options
local UserInput = 
{
	-- Core map options
	worldAge    = nil,
	seaLevel    = nil,
	temperature = nil,
	rainfall    = nil,
	resources   = nil,
	start       = nil,
	
	-- Got Lakes map options
	worldWrap   = nil,
	mountains   = nil,
	extras      = nil,
	
	-- Shuffle options
	landmass    = nil,
	climate     = nil
};


-- Convenience vars for user's choice of extra islands/lakes.
local ExtrasGot = 
{
	chains       = nil,
	clusters     = nil,
	miniDonuts   = nil,
	tectonic     = nil,
	tecAndChains = nil,
	anything     = nil,
	everything   = nil
};

-- Convenience vars for groups of options 
local gotExtrasLandmass    = nil;
local gotLakesFamily       = nil;
local gotRegionalInlandSea = nil; -- Inland Sea + Region

local gotIceAgeFamily      = nil;
local gotTiltedFamily      = nil;

------------------------------------------------------------------------------
-- Map Script globals - data that may span multiple steps of map generation
------------------------------------------------------------------------------
local ScriptData = 
{
	-- World size passed into GetMapInitData()
	worldSize = nil,
	
	-- Which function to use for randomizing map configuration.
	configRandFcn = nil,

	-- Make a copy of the plot types before adding mountains, so that
	-- we know what to put back if we remove mountains.
	preMtnPlotTypes = nil,

	-- Map of mountains and hills according to the mountain pattern
	mtnPlots = {},
	
	-- Map of trenches: where to put water for inverted tectonic islands.
	trenchPlots = {},

	-- Fractal for determining where to put tectonic islands / submerged mountains.
	tectonicIslesFractal = {},

	-- A map of tectonic islands to avoid when adding mountains and hills.
	-- Nil means no tectonic islands.
	tectonicIslesMap = nil,

	-- A map that tracks which tiles are part of the donut filling.
	donutFillingMap = {},

	-- A climate-specific offset to temperature 
	-- for latitude/terrain/feature overrides.
	temperatureShift = nil,
	
	-- Fractal to make latitude bands less obvious.
	-- Shared between terrain-gen and feature-gen.
	latVariationFrac = nil,
	latVariationMinHeight  = nil,
	latVariationMaxHeight  = nil,
	latVariationHeightDiff = nil,
	
	-- Info for regional terrain and feature calculations.
	-- Initialized during terrain generator init.
	terrainInfo = {},

	-- World wrap configuration.
	wrapX = false,
	wrapY = false, -- actually this is a constant.

	-- World Size configuration
	worldSizeInfo = {},

	-- Landmass configuration
	landInfo = {},

	-- Climate configuration
	climateInfo = {},
	
	-- Detect whether we're running in World Builder instead of the actual game.
	WBDetected = false,
	WBNumCivs  = 8, -- Pretend that there are this many civs in the WB.
};

-- Commonly used vars, following Civ6 naming convention.
local g_iW = 0;
local g_iH = 0;
local g_iFlags = {};
local g_fracXExp = -1;
local g_fracYExp = -1;

-- River vars for the Inland Sea landmass
local g_riverPlots = {};
local g_riverStartPlots = {};
local g_iRiverID = 0;


------------------------------------------------------------------------------
-- Support for overlapping mods
-- Detect other mods that might overlap with this script.
------------------------------------------------------------------------------
--local MODID_EXAMPLE  = "<insert MOD ID here>";
--local gotMod_example = false;

------------------------------------------------------------------------------
-- Misc utilities
------------------------------------------------------------------------------
Stack = {}

function Stack:Create()
    local stack = {}
    setmetatable(stack, self)
    self.__index = self
    return stack
end

function Stack:IsEmpty()
    return #self == 0
end

function Stack:Peek()
    return self[#self]
end

function Stack:Pop()
    return table.remove(self)
end

function Stack:Push(item)
    table.insert(self, item)
end

function CheckForNilPlotTypes(plotTypes)
	-- Check for nil plots in the plot type map.
	for y = 0, g_iH - 1, 1 do
		for x = 0, g_iW - 1, 1 do
			local i = y * g_iW + x + 1;
			local plotType = plotTypes[i];
			
			if (plotType == nil) then
				print("Nil plot type at " .. CoordString(x, y));
				-- Treat nil plots as ocean so that water-heavy maps can ignore them
				-- and land-heavy maps get extra lakes.
				plotTypes[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end
end

function CoordString(x,y)
	return "(" .. x .. "," .. y .. ")";
end

function Distance(x1, y1, x2, y2)
	-- Euclidian distance
	
	-- Shift x by half a plot on odd rows to account for hex grid.
	local x1 = x1;
	local x2 = x2;
	if ((y1 % 2) == 1) then
		x1 = x1 + 0.5;
	end
	
	if ((y2 % 2) == 1) then
		x2 = x2 + 0.5;
	end
	
	-- Works well for things like donut and tilted axis region.
	 return math.sqrt(((x1 - x2)*(x1 - x2)) + ((y1 - y2)*(y1 - y2)));
end

function GetSeaLevel()
    -- Map user's selected sea level from option # to 
    -- water percentage for a fractal world.

    local seas;
    
    if (gotIceAgeFamily) then
        -- Use Ice Age sea levels, which are lower than normal sea levels.
        print("Using ice age sea levels.");
        seas = ScriptData.landInfo.seas_iceage;
    else
        -- Use normal sea levels.
        seas = ScriptData.landInfo.seas_normal;
    end
    
    -- Map sea level option to sea level as water percent.
    local sea_level =  seas[UserInput.seaLevel] or 
                       seas[SEALEVEL_NORMAL]    or 72;
    
    print("Sea level is " .. sea_level); -- For debugging.
    
    return sea_level;
end

function GetPassableLandArea(plot, searchLimitNumTiles)
	-- Count the number of passable land tiles reachable from the specified plot.
	-- Stop searching if we've reached the "big enough" threshold.
	--
	-- Args:
	--  plot: starting plot
	--  searchLimitNumTiles: stop traversing after reaching this many tiles.
	--
	-- Returns: the number of passable land tiles reachable from the specified
	--          plot, or zero if the plot itself is not passable land.

	if ((searchLimitNumTiles == nil) or 
	    (IsPassableLand(plot) == false)) then
		-- Non-passable starting plot.
		return 0;
	end
	
	local visited = table.fill(0, g_iW*g_iH);
	
	local iX = plot:GetX();
	local iY = plot:GetY();
	
	-- Use a stack to search neighboring tiles in DFS order until we've
	-- covered all of the tiles that share the same passable-land area.
	local stack = Stack:Create();
	local count = 0;
	local iVisited = iY * g_iW + iX + 1; -- add one because Lua arrays start at 1
	
	stack:Push({x=iX, y=iY});
	visited[iVisited] = 1;
	
	while ((not stack:IsEmpty()) and (count < searchLimitNumTiles)) do
	
		local coord = stack:Pop();
		local x, y = coord.x, coord.y;
		
		-- Increase the tile count.
		count = count + 1;
		
		for i, direction in ipairs(HexDirections) do
			local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
			
			if (adjacentPlot ~= nil) then
				-- We have an adjacent plot.
				-- Let's see if it's unvisited passable land.
				local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
				local iCouldVisit = aY * g_iW + aX + 1;
			
				if ((IsPassableLand(adjacentPlot) == true) and 
					(visited[iCouldVisit] == 0)) then
					-- Add to the stack of plots to visit.
					stack:Push({x=aX, y=aY});
					visited[iCouldVisit] = 1;
				end
			end
		end
	end

	return count;
end

function GetWaterArea(plotTypes, iX, iY, ignoreIceCaps)
	-- Count the number of water tiles in the body of water that contains
	-- the specified coordinate.
	--
	-- Args:
	--  plotTypes: grid of plot types
	--  iX: coordinate
	--  iY: coordinate
	--  ignoreIceCaps: if true, then polar ice doesn't count toward water area.

	local numIgnoredRows = 0;
	
	if (ignoreIceCaps == true) then
	
		numIgnoredRows = ScriptData.landInfo.polarIceNumRows;
	
		if ((iY < numIgnoredRows) or (iY >= (g_iH - numIgnoredRows))) then
			-- Polar ice doesn't count towards water area.
			return 0;
		end
	end
	
	local visited = table.fill(0, g_iW*g_iH);
	
	-- Use a stack to search neighboring tiles in DFS order until we've
	-- covered all of the water tiles that share the same body of water.
	local stack = Stack:Create();
	local count = 0;
	local iVisited = iY * g_iW + iX + 1; -- add one because Lua arrays start at 1
	
	stack:Push({x=iX, y=iY});
	visited[iVisited] = 1;
	
	while (not stack:IsEmpty()) do
	
		local coord = stack:Pop();
		local x, y = coord.x, coord.y;
		
		-- Increase the water count.
		count = count + 1;
		
		for i, direction in ipairs(HexDirections) do
			local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
			
			if (adjacentPlot ~= nil) then
				-- We have an adjacent plot.
				-- Let's see if it's unvisited water.
				local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
				local iCouldVisit = aY * g_iW + aX + 1;
			
				if ((aY >= numIgnoredRows) and 
				    (aY < (g_iH - numIgnoredRows)) and
					(plotTypes[iCouldVisit] == g_PLOT_TYPE_OCEAN) and 
					(visited[iCouldVisit] == 0)) then
					-- Add to the stack of plots to visit.
					stack:Push({x=aX, y=aY});
					visited[iCouldVisit] = 1;
				end
			end
		end
	end

	return count;
end

function HasCliff(plot)
	if (not plot) then 
		return false;
	end
	
	local x = plot:GetX();
	local y = plot:GetY();
	
	local plotNW = Map.GetAdjacentPlot(x, y, DIRECTION_NW);
	local plotW  = Map.GetAdjacentPlot(x, y, DIRECTION_W);
	local plotNE = Map.GetAdjacentPlot(x, y, DIRECTION_NE);

	local isSEOfCliff = (plotNW ~= nil) and plotNW:IsNWOfCliff();
	local isEOfCliff  = (plotW  ~= nil) and plotW:IsWOfCliff();
	local isSWOfCliff = (plotNE ~= nil) and plotNE:IsNEOfCliff();
	
	-- Why couldn't the Plot API just have getters for all 6 directions?
	return  plot:IsNWOfCliff()    or 
			plot:IsWOfCliff()     or 
			plot:IsNEOfCliff()    or
			(isSEOfCliff == true) or
			(isEOfCliff  == true) or
			(isSWOfCliff == true);
end

function IsAdjacentToLand(plotTypes, iX, iY)
	local adjacentPlot;	

	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		adjacentPlot = Map.GetAdjacentPlot(iX, iY, direction);
		if (adjacentPlot ~= nil) then
	   		local i = adjacentPlot:GetY() * g_iW + adjacentPlot:GetX() + 1;
			if (plotTypes[i] ~= g_PLOT_TYPE_OCEAN) then
				return true;
			end
		end
	end
	return false;
end

function IsAdjacentToRiver(x, y)
	-- Fixes a typo by using x,y everywhere instead of taking iX,iY and using x,y
	local adjacentPlot;	

	if (Map.GetPlot(x, y) == nil) then
			return false;
	end

	if (Map.GetPlot(x, y):IsRiver() == true) then
			return true;
	end

	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
		if (adjacentPlot ~= nil) then
			if (adjacentPlot:IsRiver()) then
				return true;
			end
		end
	end
	return false;
end

function IsAdjacentToShallowWater(terrainTypes, iX, iY)
	local adjacentPlot;	

	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		adjacentPlot = Map.GetAdjacentPlot(iX, iY, direction);
		if (adjacentPlot ~= nil) then
	   		local i = adjacentPlot:GetY() * g_iW + adjacentPlot:GetX() + 1;
			if (terrainTypes[i] == g_TERRAIN_TYPE_COAST) then
				return true;
			end
		end
	end
	return false;
end

function IsAdjacentToWater(x, y)
	-- Checks a plot (x, y) to see if it is any type of land adjacent to at least one body of water.
	local plot = Map.GetPlot(x, y);
	if not plot:IsWater() then -- This plot is land, process it.
		for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
			local testPlot = Map.GetAdjacentPlot(x, y, direction);
			if testPlot ~= nil then
				if testPlot:IsWater() then -- Adjacent plot is water!
					return true
				end
			end
		end
	end
	-- Current plot is itself water, or else no salt water found among adjacent plots.
	return false
end

function IsBaseTerrainType(baseTerrain, terrainType)
	-- Return true iff baseTerrain == terrainType, or 
	-- terrainType matches either the land, hills, or mountains equivalent of the base.
	
	if (baseTerrain == terrainType) then
		return true;
	end
	
	if (baseTerrain == g_TERRAIN_TYPE_GRASS) then
		return (terrainType == g_TERRAIN_TYPE_GRASS_HILLS) or 
			   (terrainType == g_TERRAIN_TYPE_GRASS_MOUNTAIN);
	end
	
	if (baseTerrain == g_TERRAIN_TYPE_PLAINS) then
		return (terrainType == g_TERRAIN_TYPE_PLAINS_HILLS) or 
			   (terrainType == g_TERRAIN_TYPE_PLAINS_MOUNTAIN);
	end
	
	if (baseTerrain == g_TERRAIN_TYPE_DESERT) then
		return (terrainType == g_TERRAIN_TYPE_DESERT_HILLS) or 
			   (terrainType == g_TERRAIN_TYPE_DESERT_MOUNTAIN);
	end
	
	if (baseTerrain == g_TERRAIN_TYPE_TUNDRA) then
		return (terrainType == g_TERRAIN_TYPE_TUNDRA_HILLS) or 
			   (terrainType == g_TERRAIN_TYPE_TUNDRA_MOUNTAIN);
	end
	
	if (baseTerrain == g_TERRAIN_TYPE_SNOW) then
		return (terrainType == g_TERRAIN_TYPE_SNOW_HILLS) or 
			   (terrainType == g_TERRAIN_TYPE_SNOW_MOUNTAIN);
	end

	return false;
end

function IsPassable(plot)
	-- Return true iff the plot is passable by "normal" land units 
	-- that can cross both land and water.
	--
	-- Examples of tiles that are *not* passable:
	--   nil plots (off the edge of the map)
	--   mountains
	--   ice
	--   natural wonders
	
	if (not plot) then
		-- Nil plot: a phantom neighbor of an edge tile.
		return false;
	end
	
	if (plot:IsMountain()) then 
		-- Mountain
		return false;
	end
	
	if (plot:IsImpassable()) then
		-- Ice, natural wonders
		return false;
	end
	
	return true;
end

function IsPassableLand(plot)
	-- Return true iff the plot is passable and is not water.	
	return IsPassable(plot) and (plot:IsWater() == false);
end

function MapGetPlotType(x,y)
	-- Get plot type from the global map.
	local plot = Map.GetPlot(x,y);
	
	if (plot:IsWater()) then
		return g_PLOT_TYPE_OCEAN;
		
	elseif plot:IsMountain() then
		return g_PLOT_TYPE_MOUNTAIN;
	
	elseif plot:IsHills() then
		return g_PLOT_TYPE_HILLS;
		
	else
		return g_PLOT_TYPE_LAND;
	end
end

function PercentString(n,d)
	-- Pretty-percentage utility function.
	--
	-- Args:
	--   n numerator
	--   d denominator (optional); 100 if nil
	-- 
	-- Returns:
	--   a pretty percentage string for the fraction n/d, or "0 %" if 
	--   d is 0.
	
	local n = n or 0;
	local d = d or 100;
	
	if (d == 0) then
		return "0%";
	end
	
	return (Round(100*n/d) .. "%");

end

function RadString(theta)
	-- Pretty-print a value in radians.
	if (not theta) then
		return "nil";
	end
	
	return (theta/math.pi) .. " pi";
end

function RandomRadians(thetaLow, thetaHigh, granularity)
	-- A random float in the range [thetaLow,thetaHigh), 
	-- Args:
	--   thetaLow: lower bound inclusive
    --   thetaHigh: upper bound exclusive
	--   granularity: how many possible values to allow in the 
	--     random number genration; defaults to 100 if 0.
	
	local thetaLow    = thetaLow    or 0;
	local thetaHigh   = thetaHigh   or TAU;
	local granularity = granularity or 100;
	
	if (granularity == 0) then
		granularity = 100;
	end
	
	return thetaLow + (thetaHigh - thetaLow) * 
		TerrainBuilder.GetRandomNumber(granularity,"Random Angle") / granularity;
end

function Round(number)
	-- Round to the nearest integer.
	if number >= 0 then 
		return math.floor(number + 0.5);
	else 
		return math.ceil(number - 0.5);
	end
end

function NonNilStr(str)
	if (str == nil) then
		return "nil";
	end
	return str;
end

------------------------------------------------------------------------------
-- PassMaker: a reusable reachability algorithm
--
-- Ensures that all passable tiles are reachable, regardless of the placement 
-- of impassable tiles. Creates passes if necessary.
--
-- Allows overrides to define what counts as passable or impassable.
------------------------------------------------------------------------------
PassMaker = {}

function PassMaker.Create(args)
	-- Constructor.
	
	local args = args or {};

	local plotTypes       = args.plotTypes;      -- or nil	
	local obstaclesName   = args.obstaclesName   or "mountains";
	local InitReachPlot   = args.InitReachPlot   or PassMaker.Mtn_InitReachPlot;
	local RemoveObstacle  = args.RemoveObstacle  or PassMaker.Mtn_RemoveObstacle;

	-- Default to global map accessors.
	local GetPlotType     = PassMaker.Map_GetPlotType;
	local SetPlotType     = PassMaker.Map_SetPlotType;
	local GetFeatureType  = PassMaker.Map_GetFeatureType;
	local SetFeatureType  = PassMaker.Map_SetFeatureType;
	local HasCliff        = PassMaker.Map_HasCliff;
	local RemoveCliffs    = PassMaker.Map_RemoveCliffs;
	local IsNaturalWonder = PassMaker.Map_IsNaturalWonder;
	
	if (plotTypes ~= nil) then
		-- Injected plot-types array.
		-- Override to use field accessors.
		GetPlotType     = PassMaker.Field_GetPlotType;
		SetPlotType     = PassMaker.Field_SetPlotType;
		GetFeatureType  = PassMaker.Field_GetFeatureType;
		SetFeatureType  = PassMaker.Field_SetFeatureType;
		HasCliff        = PassMaker.Field_HasCliff;
		RemoveCliffs    = PassMaker.Field_RemoveCliffs;
		IsNaturalWonder = PassMaker.Field_IsNaturalWonder;
	end
	
	local data = 
	{
		-- Reachability algorithm.
		Initialize             = PassMaker.Initialize,
		IsRemovableObstacle    = PassMaker.IsRemovableObstacle,
		ReachAroundObstacles   = PassMaker.ReachAroundObstacles,
		FindRemovableObstaclesNearUnreached = 
			PassMaker.FindRemovableObstaclesNearUnreached,
		FindPassStart          = PassMaker.FindPassStart,
		SearchObstaclesForPass = PassMaker.SearchObstaclesForPass,
		FindPass               = PassMaker.FindPass,
		CreatePass             = PassMaker.CreatePass,
		EnsureReachability     = PassMaker.EnsureReachability,
				
		-- Hooks for obstacle-specific logic.
		InitReachPlot  = InitReachPlot,
		RemoveObstacle = RemoveObstacle,
		
		-- Wrappers for plot types and feature types.
		IsMountain      = PassMaker.IsMountain,
		IsIce           = PassMaker.IsIce,
		GetPlotType     = GetPlotType,
		SetPlotType     = SetPlotType,
		GetFeatureType  = GetFeatureType,
		SetFeatureType  = SetFeatureType,
		HasCliff        = HasCliff,
		RemoveCliffs    = RemoveCliffs,
		IsNaturalWonder = IsNaturalWonder,

		-- Misc params.
		plotTypes      = plotTypes,
		obstaclesName  = obstaclesName,
		
		-- Possible values for each plot on a reachability map.
		REACH_NO           = 0, -- Not reachable / not yet reached
		REACH_YES          = 1, -- Reachable
		REACH_SEEN         = 2, -- Dead-end removable obstacle (inside an unreachable area)
		REACH_REMOVABLE    = 3, -- Removable obstacle
		REACH_NONREMOVABLE = 4, -- Non-removable obstacle
	};
	
	data:Initialize();
	
	return data;
end

function PassMaker:Initialize()
	-- Initialize the reachability algorithm.
	--
	-- New fields:
	--   reachMap: map data containing reachability info of each tile.
	--   reachStart: start location for the initial reach scan
	--   numReachable: the number of reachable tiles (non-mountain lands)

	--print("Analyzing plots for reachability.");
	
	local numTiles = g_iW * g_iH;

	-- Create a parallel map table to track reachability of each plot.
	local reachMap = table.fill(self.REACH_NO, numTiles);
	
	-- Scan for mountains and a place to start reaching.
	local reachStart = nil;
	
	local numReachable = 0;
	
	self.allObsRemovable = true;
	
	-- Add some randomness to the start location of the initial scan.
	local iStart = 1 + TerrainBuilder.GetRandomNumber(numTiles, 
		"PassMaker - start location of initial scan.");

	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1;
					
			reachMap[i] = self:InitReachPlot(x,y);
			
			if (reachMap[i] == self.REACH_NO) then
				-- Reachable tile.
				numReachable = numReachable + 1;
		
				if (i > iStart) and (not reachStart) then
					reachStart = {x=x, y=y};
				end
				
			elseif (reachMap[i] == self.REACH_NONREMOVABLE) then
				-- Got at least one non-removable obstacle.
				self.allObsRemovable = false;
			end
		end	
	end
	
	if (not reachStart) then
		-- We missed all of the potential starts in our attempt to randomize the
		-- location of the initial scan!
		--print("We missed the potential start tiles; need to do another pass through plot types.");
		for iTile = 1, iStart, 1 do
			local x = ((iTile-1)  % g_iW);
			local y = (iTile-x-1) / g_iW;
						
			if (reachMap[i] == self.REACH_NO) then
				reachStart = {x=x, y=y};
				break;
			end
		end
	end
	
	if (not reachStart) then
		print("WARNING: no passable plots in the entire map!");
		return;
	end
	
	self.reachMap     = reachMap;
	self.reachStart   = reachStart;
	self.numReachable = numReachable;
	
	--print("Analysis complete.");
end

function PassMaker:IsRemovableObstacle(iPlot)
	-- Determine whether the plot at iPlot is an obstacle that we can
	-- remove with a pass.
	--
	-- Args:
	--   iPlot: map index of the plot to check
	-- Returns:
	--   true iff the plot contains a removable obstacle
	return self.reachMap[iPlot] == self.REACH_REMOVABLE;
end

function PassMaker:ReachAroundObstacles(start, obsCoords)
	-- Visit all non-obstacle tiles that are reachable from the start
	-- tile.
	--
	-- Args:
	--   start: {x,y} coord of starting tile; must not be an obstacle tile.
	--   obsCoords: previous obstacle coords, or nil if this is the first reach.
	--
	-- Returns:
	--   numReached: the number of visited tiles.
	--   obsCoords: a stack containing the {x,y} of every obstacle tile within reach.
	
	--print("Reaching around obstacles.");
	
	-- Use a stack to search neighboring tiles in DFS order.
	local reachStack = Stack:Create();
	local obsCoords  = obsCoords or Stack:Create();
	
	local iVisited = start.y * g_iW + start.x + 1; -- add one because Lua arrays start at 1
	local count = 0;
	
	reachStack:Push(start);
	self.reachMap[iVisited] = self.REACH_YES;
	
	while (not reachStack:IsEmpty()) do
	
		local coord = reachStack:Pop();
		local x, y = coord.x, coord.y;
		
		local plot = Map.GetPlot(x,y);
		
		-- Increase the reach count.
		count = count + 1;
		
		
		for i, direction in ipairs(HexDirections) do
			local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
			
			if (adjacentPlot ~= nil) then
				-- We have an adjacent plot.
				-- Let's see if it's unvisited.
				local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
				local iCouldVisit = aY * g_iW + aX + 1;
							
				if (self.reachMap[iCouldVisit] == self.REACH_NO) then
					-- Add to the stack of plots to reach.
					reachStack:Push({x=aX, y=aY});
					self.reachMap[iCouldVisit] = self.REACH_YES;
					
				elseif (self:IsRemovableObstacle(iCouldVisit)) then
					-- Found a removable obstacle tile.
					-- Add to the stack of tiles to consider for passes
					obsCoords:Push({x=aX, y=aY});
				end
				-- Ignore obstacles that are not removable.
			end
		end
	end
	
	--print("Reached " .. count .. " tile(s).");
	
	return count, obsCoords;
end

function PassMaker:FindRemovableObstaclesNearUnreached()
	-- Find relevant obstacles near unreached tiles.
	--
	-- Returns: {x,y} stack of every relevant obstacle tile that is 
	--   adjacent to an unreached tile

	local numTiles = g_iW * g_iH;
	
	-- Add some randomness to the start location of the scan.
	local iOffset = TerrainBuilder.GetRandomNumber(numTiles, 
		"Reachability - start location of scan for obstacles near unreached tiles.");
	
	local obsCoords = Stack:Create();
	
	for i = 1, numTiles, 1 do
			
		local iTile = (i + iOffset) % numTiles + 1;
			
		local x = ((iTile-1)  % g_iW);
		local y = (iTile-x-1) / g_iW;	
			
		if (self.reachMap[iTile] == self.REACH_NO) then
			-- Found an unreached tile.
			-- Check neighbors for ice.

			for j, direction in ipairs(HexDirections) do
				local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
									
				if (adjacentPlot ~= nil) then
					-- We have an adjacent plot.
					-- Let's see if it's unreached
					local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
					local iCouldVisit = aY * g_iW + aX + 1;
				
					if (self:IsRemovableObstacle(iCouldVisit)) then
						-- Found a removable obstacle next to an unreached tile.
						obsCoords:Push({x=aX,y=aY});
					end
				end
			end
		end	
	end
	
	return obsCoords;
end

function PassMaker:FindPassStart(obsCoords, passStart, passEnd)
	-- Find a candidate starting location for the pass search algorithm.
	--
	-- Args:
	--   obsCoords: a stack containing the {x,y} of obstacle tiles within reach.
	--   passStart: start passes at obstacle tiles next to this reach state
	--   passEnd: declare success when the pass reaches a tile with 
	--     this reach state.
	--
	-- Return a candidate start tile, or nil if no such start tile exists.
	
	local numTiles = g_iW * g_iH;
		
	-- Need to find removable obstacles that are next to an unreached tile.
	local start = nil;
	
	while (not obsCoords:IsEmpty()) do
	
		local obsCoord = obsCoords:Pop();
		local x = obsCoord.x;
		local y = obsCoord.y;
		local iTile = y * g_iW + x + 1; 
	
		if (self:IsRemovableObstacle(iTile)) then
			-- Got an obstacle that we haven't yet seen.
		
			-- Check neighboring tiles to see if this is a good 
			-- place to look for a pass.
			local numPassStart = 0;
			local numPassEnd   = 0;
			local numNeighbors = 6;
			
			for i, direction in ipairs(HexDirections) do
				local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
									
				if (adjacentPlot ~= nil) then
					-- We have an adjacent plot.
					-- Let's see if it's unreached
					local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
					local iCouldVisit = aY * g_iW + aX + 1;
				
					if (self.reachMap[iCouldVisit] == passEnd) then
						-- Found a tile that matches end criteria.
						numPassEnd = numPassEnd + 1;
						
						-- We can stop searching neighbors, since
						-- this obstacle touches both start and 
						-- end terratory.
						break;
						
					elseif (self.reachMap[iCouldVisit] == passStart) then
						-- Found a tile that matches start criteria
						-- But we're not interested in obstacle tiles that
						-- are completely inside the starting zone.
						numPassStart = numPassStart + 1;
					end
				else
					-- At an edge of the map.
					numNeighbors = numNeighbors - 1;
				end
			end
	
			if (numPassEnd > 0) then
				-- Got a removable obstacle that is adjacent to an end tile.
				--print("Definite pass: " .. CoordString(x,y));
				return {x=x, y=y};
				
			elseif (numPassStart < (numNeighbors - 2)) then
				-- Removable obstacle that is not surrounded by start tiles.
				--print("Possible pass: " .. CoordString(x,y));
				return {x=x, y=y};
			end
		end
	end
	
	print("FindPassStart: searched the entire map.");
	return nil;
end

function PassMaker:SearchObstaclesForPass(start, passEnd)
	-- Search through obstacles for a pass between reachable
	-- and unreachable tiles
	--
	-- Args:
	--   start: {x,y} of an obstacle tile from which to start searching obstacles
	--   passEnd: declare success when the pass reaches a tile with 
	--     this reach state.
	--
	-- Returns a stack of {x,y} coords representing a path
	--   or nil if no path found
	
	--print("Searching for a pass, starting at " .. CoordString(start.x,start.y));
	
	-- Use a stack to search neighboring tiles in DFS order.
	local stack = Stack:Create();
	
	
	stack:Push(start);
	
	while (not stack:IsEmpty()) do
	
		-- Don't pop the coord unless it's a dead-end.
		local coord = stack:Peek();
		local x, y = coord.x, coord.y;
		local iVisiting = y * g_iW + x + 1;
		
		-- Mark the currently visited obstacle so that we don't
		-- visit the same tile multiple times.
		self.reachMap[iVisiting] = self.REACH_SEEN;
		
		--print("Scoping out " .. CoordString(x,y));
		
		local numObstacles = 0;
		
		-- First check whether we're at the end of the path.
		for i, direction in ipairs(HexDirections) do
			local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
			
			if (adjacentPlot ~= nil) then
				-- We have an adjacent plot.
				-- Let's see if it's unvisited.
				local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
				local iCouldVisit = aY * g_iW + aX + 1;
			
				if (self.reachMap[iCouldVisit] == passEnd) then
					-- We found a path across the obstacles!
					--print("Successfully crossed obstacles at " .. CoordString(aX,aY));
					return stack;
				
				elseif (self:IsRemovableObstacle(iCouldVisit)) then
					-- An unexplored removable obstacle tile.
					numObstacles = numObstacles + 1;
					
				elseif (self.reachMap[iCouldVisit] == self.REACH_SEEN) then
					-- An already-explored removable obstacle tile.
					numObstacles = numObstacles + 1;
				end
			end
		end
		
		-- Not at the end of the path.
		local foundSomething = false;
		
		if ((self.allObsRemovable == true) and (numObstacles < 3)) then
			-- We're walking along obstacles instead of crossing them!
			--
			-- Treat this as a dead-end.
			-- This optimization only works if all obstacles are removable.
			
			--[[
			print("Found " .. numObstacles .. " adjacent obstacle(s)..." .. 
				  "we're probably not on the shortest path."); 
			--]]
		
		else
			-- Try to extend the path.
			--
			-- Shuffle the directions so that the path isn't
			-- a freakishly straight line or a giant hexagon
			local shuffledDirs = GetShuffledCopyOfTable(HexDirections);
			
			for i, direction in ipairs(shuffledDirs) do
				local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
				
				if (adjacentPlot ~= nil) then
					-- We have an adjacent plot.
					-- Let's see if it's unvisited.
					local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
					local iCouldVisit = aY * g_iW + aX + 1;
				
					if (self:IsRemovableObstacle(iCouldVisit)) then
						-- Add to the stack of plots to reach.
						stack:Push({x=aX, y=aY});
						foundSomething = true;
						break;
					end
				end
			end
		end
		
		if (foundSomething == false) then
			-- Remove this dead-end obstacle from the path.
			--print("Backtracking from dead-end at " .. CoordString(x,y));
			stack:Pop();
		end
	end
	
	print("No pass at " .. CoordString(start.x,start.y));
	
	-- No path from this obstacle tile to a reached tile.
	return nil;
end

function PassMaker:FindPass(obsCoords, passStart, passEnd)
	-- Search for an obstacle path that starts next to anunreachable
	-- tile and ends next to a reachable tile, or vice-versa.
	--
	-- Args:
	--   obsCoords: a stack containing the {x,y} of every 
	--      obstacle tile within reach
	--   passStart: start passes at obstacle tiles next to this reach state
	--   passEnd: declare success when the pass reaches a tile with 
	--     this reach state.
	--
	-- Returns:
	--   the start coord for the next reachability scan, 
	--      or nil if no pass exists
	--   a stack of obstacle tiles representing the pass, 
	--      or nil if no such pass exists.
	
	
	local numTiles = g_iW * g_iH;
	
	-- Search the map for removable obstacles that touch unreached terratory.	
	local obsStart = self:FindPassStart(obsCoords, passStart, passEnd);
	if (not obsStart) then
		print("Unexpected: no initial obstacle start in FindPass()");
		return nil, nil;
	end
	
	-- Traverse obstacles tiles until we find a path to a reachable tile.
	local pathCoords = nil;
	
	while ((not pathCoords) and obsStart) do
		-- Search for a path, starting at an obstacle tile next
		-- to the start tile.
		pathCoords = self:SearchObstaclesForPass(obsStart, passEnd);
		
		-- Reset pass backtracking.
		for i = 1, numTiles, 1 do
			if (self.reachMap[i] == self.REACH_SEEN) then
				self.reachMap[i] = self.REACH_REMOVABLE;
			end
		end
		
		if ((not pathCoords) or (pathCoords:IsEmpty())) then
			-- Unable to find a path out of the unreachable zone.
			-- Need to search for another ice pass start.
			obsStart = self:FindPassStart(obsCoords, passStart, passEnd);
		end
	end

	if (not obsStart) then
		-- No ice pass detected, after searching the entire map.
		print("Failed to create a pass.");
		return nil, nil;
	end
	
	return obsStart, pathCoords;
end

function PassMaker:CreatePass(obsCoords, passStart, passEnd)
	-- Search for an unreachable tile.
	-- Then search for a path from that tile to a reachable
	-- tile and clear any removable obstacles on that path.
	--
	-- Args:
	--   obsCoords: a stack containing the {x,y} of every removable 
	--     obstacle tile within reach
	--   passStart: start passes at removable obstacles next to this 
	--     reach state
	--   passEnd: declare success when the pass reaches a tile with 
	--     this reach state.
	--
	-- Returns:
	--   the start coord for the next reachability scan. 
	--   the number of removable obstacle tiles altered by the pass.
	
	-- Search for a pass.
	local obsStart, pathCoords = self:FindPass(obsCoords, passStart, passEnd);
	
	if (not pathCoords) then
		print("Failed to find a pass.");
		return nil, 0;
	end
	
	--print("Creating a pass for reachability.");
	local pathSize = 0;
	
	while (not pathCoords:IsEmpty()) do 
		-- Clear obstacles from tiles in the pass.
		
		local coord = pathCoords:Pop();
		local x, y = coord.x, coord.y;
			
		local plotsModded = Stack:Create();
		
		-- Remove the current tile's obstacle.
		self:RemoveObstacle(x, y, plotsModded);
		
		while (not plotsModded:IsEmpty()) do
			-- Add all modded coords to the pass.
			local coordModded = plotsModded:Pop();
			local mX, mY = coordModded.x, coordModded.y;

			local iPass = mY * g_iW + mX + 1; -- add one because Lua arrays start at 1

			self.reachMap[iPass] = self.REACH_NO; -- leave unreached until the next ReachTiles call()
			pathSize = pathSize + 1;
		end		
	end
		
	--print("Pass size: " .. pathSize .. " tile(s).");
	
	-- Use the obstacle start for the next reachability scan, since
	-- we just turned that obstacle into a passable tile.
	return obsStart, pathSize;
end

function PassMaker:EnsureReachability()
	-- Make sure that all passable tiles are reachable 
	-- (that is, connected by at least one path of passable tiles).
	--
	-- Returns 0 if successful, nonzero otherwise.

	local obstaclesName = self.obstaclesName;
	print("PassMaker: Checking for tiles trapped by " .. obstaclesName);	
		
	if (not self.reachMap) then
		print("Failed to analyze plot types.");
		return 1;
	end
	
	local reachStart   = self.reachStart;
	local numReachable = self.numReachable;
	local numUnreached = numReachable;
	local pathSize;
	local obsYesCoords = nil;
	local obsCoords    = nil;
	
	-- Decide whether to search obstacles unreached -> reached or 
	-- vice-versa based on number of unreached tiles on the map.
	local passStart, passEnd;
		
	while (numUnreached > 0) do
		-- Run a reachability scan on the map.
		local numReached, obsYesCoords = self:ReachAroundObstacles(reachStart, obsCoords);
		
		-- Update tile counts based on the reach.
		numUnreached = numUnreached - numReached;
		numReached   = numReachable - numUnreached;
		
		if (numUnreached > 0) then
			-- Not all tiles are reachable.
			print("Detected " .. numUnreached .. " unreachable tile(s) out of " .. 
								 numReachable .. " total.");
			
			if (numReached > (numReachable / 2)) then
				-- We've reached more than half of the reachable tiles.
				-- So it's more efficient to search for obstacles near unreached tiles,
				-- then to reach from an already-reached obstacles.
				--print("Most of the map is within reach.");
				passStart = self.REACH_NO;
				passEnd   = self.REACH_YES;
				
				obsCoords = self:FindRemovableObstaclesNearUnreached();
			else
				-- We haven't reached much of the map yet.
				-- Begin passes near what we've reached so far in order to
				-- expand our reach efficiently.
				--print("Most of the map is out of reach.");
				passStart = self.REACH_YES;
				passEnd   = self.REACH_NO;
				obsCoords = obsYesCoords;
			end
			
			-- Create a pass to make tiles reachable.
			-- Use the start of the pass for the next reachability scan.
			reachStart, pathSize = self:CreatePass(obsCoords, passStart, passEnd);
			
			if (not reachStart) then
				print("Failed to reach around " .. obstaclesName);
				return 2;
			end
			
			-- Update tile counts since we removed ice.
			numReachable = numReachable + pathSize;
			numUnreached = numUnreached + pathSize;
		end
	end
	
	print("All tiles are reachable around or through " .. obstaclesName);
	return 0;
end

function PassMaker:IsMountain(x,y)
	return self:GetPlotType(x,y) == g_PLOT_TYPE_MOUNTAIN;
end

function PassMaker:IsIce(x,y)
	return self:GetFeatureType(x,y) == g_FEATURE_ICE;
end

function PassMaker:Field_GetPlotType(x,y)
	-- Get plot type from a private field.
	local i = y*g_iW + x + 1;
	return self.plotTypes[i];
end

function PassMaker:Field_SetPlotType(x,y,plotType)
	-- Set plot type for a plot in a private field.
	local i = y*g_iW + x + 1;
	self.plotTypes[i] = plotType;
end

function PassMaker:Field_GetFeatureType(x,y)
	-- Features don't exist until after terrain generation.
	return nil;
end

function PassMaker:Field_SetFeatureType(x,y,featureType)
	-- Features don't exist until after terrain generation.
end

function PassMaker:Field_HasCliff(x,y)
	-- Cliffs don't exist until after terrain.
	return false;
end

function PassMaker:Field_RemoveCliffs(x,y)
	-- Cliffs don't exist until after terrain.
end

function PassMaker:Field_IsNaturalWonder(x,y)
	-- Natural wonders don't exist until after terrain generation.
	return false;
end

function PassMaker:Map_GetPlotType(x,y)
	-- Get plot type from the global map.
	return MapGetPlotType(x,y);
end

function PassMaker:Map_SetPlotType(x,y,plotType)
	-- Set plot type (terrain) for a plot in the global map.
	local plot = Map.GetPlot(x,y);
	
	if (plotType == g_PLOT_TYPE_OCEAN) then
		-- Set shallow water.
		TerrainBuilder.SetTerrainType(plot, g_TERRAIN_TYPE_COAST);
	end
	
	-- Map land plot type to a terrain type.
	local terrainOffset = 0; -- default for flat land.
	if (plotType == g_PLOT_TYPE_HILLS) then
		terrainOffset = 1;
		
	elseif (plotType == g_PLOT_TYPE_MOUNTAIN) then
		terrainOffset = 2;
	end
	
	local origTerrain = plot:GetTerrainType();
	local baseTerrain = g_TERRAIN_TYPE_GRASS;
	
	if ((origTerrain == g_TERRAIN_TYPE_PLAINS_HILLS) or 
		(origTerrain == g_TERRAIN_TYPE_PLAINS_MOUNTAIN)) then
		baseTerrain = g_TERRAIN_TYPE_PLAINS;
	end
	
	if ((origTerrain == g_TERRAIN_TYPE_DESERT_HILLS) or 
		(origTerrain == g_TERRAIN_TYPE_DESERT_MOUNTAIN)) then
		baseTerrain = g_TERRAIN_TYPE_DESERT;
	end
	
	if ((origTerrain == g_TERRAIN_TYPE_TUNDRA_HILLS) or 
		(origTerrain == g_TERRAIN_TYPE_TUNDRA_MOUNTAIN)) then
		baseTerrain = g_TERRAIN_TYPE_TUNDRA;
	end
	
	if ((origTerrain == g_TERRAIN_TYPE_SNOW_HILLS) or 
		(origTerrain == g_TERRAIN_TYPE_SNOW_MOUNTAIN)) then
		baseTerrain = g_TERRAIN_TYPE_SNOW;
	end
	
	TerrainBuilder.SetTerrainType(plot, baseTerrain + terrainOffset);
end

function PassMaker:Map_GetFeatureType(x,y)
	-- Get feature type from global Map.
	local plot = Map.GetPlot(x,y);
	return plot:GetFeatureType();
end

function PassMaker:Map_SetFeatureType(x,y,featureType)
	-- Set a plot's feature type.
	TerrainBuilder.SetFeatureType(Map.GetPlot(x,y), featureType)
end

function PassMaker:Map_HasCliff(x,y)
	local plot = Map.GetPlot(x,y);
	-- Ignore cliffs when considering land tiles,
	-- so that we can have tiny islands with cliffs.
	return plot:IsWater() and HasCliff(plot);
end

function PassMaker:Map_RemoveCliffs(x,y)
	local plot = Map.GetPlot(x,y);
	
	-- To ensure passability, just remove cliffs in all directions.
	TerrainBuilder.SetNEOfCliff(plot, false);	
	TerrainBuilder.SetNWOfCliff(plot, false);	
	TerrainBuilder.SetWOfCliff(plot,  false);
	
	local plotNW = Map.GetAdjacentPlot(x, y, DIRECTION_NW);
	local plotW  = Map.GetAdjacentPlot(x, y, DIRECTION_W);
	local plotNE = Map.GetAdjacentPlot(x, y, DIRECTION_NE);

	if (plotNW ~= nil) then
		TerrainBuilder.SetNWOfCliff(plotNW, false);	
	end
	if (plotW ~= nil) then
		TerrainBuilder.SetWOfCliff(plotW, false);	
	end
	if (plotNE ~= nil) then
		TerrainBuilder.SetNEOfCliff(plotNE, false);	
	end
end

function PassMaker:Map_IsNaturalWonder(x,y)
	-- True iff the plot at (x,y) is an impassable natural wonder.
	-- Disregard passable wonders such as Cliffs of Dover
	local plot = Map.GetPlot(x,y);
	return plot:IsNaturalWonder() and plot:IsImpassable();	
end

function PassMaker:Ice_InitReachPlot(x,y)
	-- Calculate reachability val for a plot at (x,y).
	if (self:IsNaturalWonder(x,y) or 
	    self:IsMountain(x,y) or 
		self:HasCliff(x,y)) then
		-- Natural Wonder or mountain/cliff.
		return self.REACH_NONREMOVABLE;
	end
				
	if self:IsIce(x,y) then
		-- Ice
		return self.REACH_REMOVABLE;
	end
	
	return self.REACH_NO; -- Not yet reached.
end

function PassMaker:Ice_RemoveObstacle(x, y, plotsModded)
	-- Remove any obstacle at (x,y), which is expected to have an obstacle.
	-- Push any affected plots to plotsModded.
	
	-- Assert that this plot actually has ice to remove.
	if (self:IsIce(x,y) == false) then
		-- Programmer error....
		print("Whoops, somehow added a non-ice tile to a pass at " .. CoordString(x, y));
		
	elseif self:IsNaturalWonder(x,y) then
		-- Programmer error....
		print("Whoops, somehow added a wonder tile to a pass at " .. CoordString(x, y));	
		
	else
		-- Remove the ice from the plot.
		print("Removing ice at " .. CoordString(x,y));	
		self:SetFeatureType(x,y,g_FEATURE_NONE);
		plotsModded:Push({x=x, y=y});
	end
end

function PassMaker:Mtn_InitReachPlot(x,y)
	-- Calculate reachability val for a plot at (x,y).
	if self:IsNaturalWonder(x,y) then
		-- Natural Wonder.
		return self.REACH_NONREMOVABLE;
	end
	
	if self:IsMountain(x,y) or self:HasCliff(x,y) then
		-- Mountain or cliff.
		return self.REACH_REMOVABLE;
	end
	
	return self.REACH_NO; -- Not yet reached.
end

function PassMaker:Mtn_RemoveObstacle(x,y, plotsModded)
	-- Remove any obstacle at (x,y), which is expected to have an obstacle.
	-- Push any affected plots to plotsModded.

	local iPass = y * g_iW + x + 1; -- add one because Lua arrays start at 1
	
	-- The plot at (x,y) could be a mountain, have cliffs, or both;
	-- so always check for both mountain and cliffs.
	if self:IsMountain(x,y) then
		print("Removing mountain at " .. CoordString(x,y));	
		
		if (ScriptData.preMtnPlotTypes[iPass] == g_PLOT_TYPE_MOUNTAIN) then
			-- This tile somehow used to be a mountain, so just use a hill tile instead.
			self:SetPlotType(x,y,g_PLOT_TYPE_HILLS);
			
		elseif (ScriptData.preMtnPlotTypes[iPass] == g_PLOT_TYPE_LAND) then
			-- This tile used to be flat land before we turned it into a mountain.
			local hillsDice = TerrainBuilder.GetRandomNumber(100, "Mountain Pass Hill");
			local hillsRate = 15;
			
			if (hillsDice < hillsRate) then
				-- Turn the mountain into a hill so that mountain passes
				-- can have random hills.
				self:SetPlotType(x,y,g_PLOT_TYPE_HILLS);
			else
				-- Keep the original flat land.
				self:SetPlotType(x,y,g_PLOT_TYPE_LAND);
			end
		else
			-- Revert to the tile's original plot type.
			local preMtnPlotType = ScriptData.preMtnPlotTypes[iPass];
			
			if ((UserInput.landmass == Landmasses.LAKES) and 
			    (preMtnPlotType     == g_PLOT_TYPE_OCEAN)) then
				-- For lakes landmass, avoid re-adding water
				-- to preserve fresh lakes.
				preMtnPlotType = g_PLOT_TYPE_LAND;
			end
			
			self:SetPlotType(x,y, preMtnPlotType);
		end
		plotsModded:Push({x=x, y=y});
	end
	
	if self:HasCliff(x,y) then
		print("Removing cliff(s) at " .. CoordString(x,y));	
		self:RemoveCliffs(x,y);
		
		if (self:IsMountain() == false) then
			-- Avoid double-counting plots that have both mountain and cliffs.
			plotsModded:Push({x=x, y=y});
		end
	end	
end

function PassMaker:Water_InitReachPlot(x,y)
	-- Calculate reachability val for a plot at (x,y).
	if (self:GetPlotType(x,y) == g_PLOT_TYPE_OCEAN) then
		return self.REACH_REMOVABLE;
	end
	
	return self.REACH_NO; -- Not yet reached.
end

function PassMaker:Water_RemoveObstacle(x,y, plotsModded)
	-- Remove any obstacle at (x,y), which is expected to have an obstacle.
	-- Push any affected plots to plotsModded.
	
	-- Assert that this plot actually has water to remove.
	if (self:GetPlotType(x,y) ~= g_PLOT_TYPE_OCEAN) then
		-- Programmer error....
		print("Whoops, somehow added a non-water tile to a pass at " .. CoordString(x, y));
	else
		-- Turn water into land.
		self:SetPlotType(x,y,g_PLOT_TYPE_LAND);
		plotsModded:Push({x=x, y=y});
	end	
end

function PassMaker:Wonder_InitReachPlot(x,y)
	-- Calculate reachability val for a plot at (x,y).
	if self:IsNaturalWonder(x,y) then
		-- Natural wonder.
		return self.REACH_REMOVABLE;
	end
	
	return self.REACH_NO; -- Not yet reached.
end

function PassMaker:Wonder_RemoveObstacle(x,y, plotsModded)
	-- Remove any obstacle at (x,y), which is expected to have an obstacle.
	-- Push any affected plots to plotsModded.
		
	-- Assert that this plot actually has a wonder to remove.
	if (self:IsNaturalWonder(x,y) == false) then
		-- A previously-removed wonder had multiple plots along the path.
		return;
	end
	
	-- Remove the wonder.
	local featureToRemove = self:GetFeatureType(x, y);
	
	-- Use a visit map to avoid double-counting wonder plots.
	local visited = table.fill(0, g_iW*g_iH);
	local iVisited = y * g_iW + x + 1; -- add one because Lua arrays start at 1
	visited[iVisited] = 1;
	
	-- Use a stack to search neighboring tiles in DFS order until we've
	-- removed the entire natural wonder.
	local stack = Stack:Create();	
	stack:Push({x=x, y=y});
	
	while (not stack:IsEmpty()) do
	
		local coord = stack:Pop();
		local cX, cY = coord.x, coord.y;
		
		-- Remove natural wonder from the plot.
		print("Removing natural wonder at " .. CoordString(cX,cY));
		self:SetFeatureType(cX,cY,g_FEATURE_NONE);
		plotsModded:Push({x=cX, y=cY});
		
		for i, direction in ipairs(HexDirections) do
			local adjacentPlot = Map.GetAdjacentPlot(cX, cY, direction);
			
			if (adjacentPlot ~= nil) then
				-- We have an adjacent plot.
				local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
				local iCouldVisit = aY * g_iW + aX + 1;

				if ((featureToRemove == self:GetFeatureType(aX, aY)) and
					(visited[iCouldVisit] == 0))then
					-- Adjacent matching natural wonder.
					-- Add to the stack of plots to visit.
					stack:Push({x=aX, y=aY});
					visited[iCouldVisit] = 1;
				end
			end
		end
	end
end

function EnsureReachabilityWithPassMaker(args)
	-- Make sure that all passable tiles are reachable 
	-- (that is, connected by at least one path of passable tiles).
	local passMaker = PassMaker.Create(args);
	passMaker:EnsureReachability();
end
------------------------------------------------------------------------------
-- Mountain Patterns for mountain/hill generation.
------------------------------------------------------------------------------
function GetMountainHeightAdjustment(args)
	-- Get the "adjustment" offset for mountain/hill fractal height 
	-- percentages.
	--
	-- Args: mountain args
	-- Return the appropriate adjustment for the world's age.
	
	local args = args or {};
	
	local adjustments = args.adjustments or {5,3,2};
	
	local adjustment = adjustments[UserInput.worldAge];
	--print("Mountain Height adjustment: " .. adjustment);
	
	return adjustment;
end

function GetMountainGrain(baseGrain)
	-- Varying grains for hills/peaks per map size.
	local grain = ScriptData.worldSizeInfo.mtnGrain;
	
	-- Don't make the grain too high or else mountain ranges become dots.
	return math.clamp(baseGrain + grain, 1, 5);
end

function Clustered_GenerateMountainsAndHills(args, mtnPlots, trenchPlots)
	-- Mountains and hills - pattern-specific
	print("Generating mountains and hills (clustered highlands).");
	
	-- Concentrate all of the mountains and hills into a small number of
	-- clusters.
	
	
	
	-- landmass-specific mountain args + fractal config
	local args = args or {};
	
	-- Some island maps need more mountains due to lower
	-- chances of a mountain being on habitable land.
	local extra_mountains  = args.extra_mountains  or 0;
	local terrain_flags    = args.terrain_flags   or g_iFlags;

	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = GetMountainHeightAdjustment(args);
	
	-- Varying grains for hills/peaks per map size.
	local grain = GetMountainGrain(2);
		
	-- Generate fractals to govern mountain ranges
	local rangesFrac = Fractal.Create(
		g_iW, g_iH, 
		grain,
		terrain_flags, 
		g_fracXExp, g_fracYExp);
	
	-- Get height values for mountains and hills
	local mtsAdjust   = 5 + (2*adjustment) + (extra_mountains);
	
	local hillsAdjust = mtsAdjust + 10 + (6 * adjustment);
	
	local trenches = 28;
	
	local iMtnsBottom  = rangesFrac:GetHeight(100 - mtsAdjust);  
	local iHillsBottom = rangesFrac:GetHeight(100 - hillsAdjust); 
	
	local iTrenchesTop = rangesFrac:GetHeight(trenches); 
	
	-- Now the main loop, which will assign the plot types.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			local rangeVal = rangesFrac:GetHeight(x,y);
			
			if (rangeVal >= iMtnsBottom) then
				-- Mountain cluster
				mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;

			elseif (rangeVal >= iHillsBottom) then
				-- Hills surrounding the mountain cluster
				mtnPlots[i] = g_PLOT_TYPE_HILLS;
			
			elseif (rangeVal <= iTrenchesTop) then	
				-- Trenches surrounding the mountain cluster
				trenchPlots[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end	
end

function FRanges_GenerateMountainsAndHills(args, mtnPlots, trenchPlots)
	-- Mountains and hills - pattern-specific
	print("Generating mountains and hills (fractal ranges).");
	
	-- Create a snaky fractal pattern without using tectonic 
	-- rifts. Instead, use the 50% percentile of a fractal's
	-- height map as the focus for mountain ranges.
	
	
	
	-- landmass-specific mountain args + fractal config
	local args = args or {};
	
	-- Some island maps need more mountains due to lower
	-- chances of a mountain being on habitable land.
	local extra_mountains  = args.extra_mountains  or 0;
	local terrain_flags    = args.terrain_flags   or g_iFlags;

	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = GetMountainHeightAdjustment(args);
	
	-- Varying grains for hills/peaks per map size.
	local grain = GetMountainGrain(1);
		
	-- Generate fractals to govern mountain ranges
	local rangesFrac = Fractal.Create(
		g_iW, g_iH, 
		grain,
		terrain_flags, 
		g_fracXExp, g_fracYExp);
	
	-- Apply adjustment to hills and peaks settings.
	-- Base all ranges at the middle (50th percentile) of the fratal heights
	-- so that mountains are as snaky and non-blob-like as possible
	-- The fractal becomes blob-like at extreme values (0 and 100).
	local middle = 50;
	
	-- Get height values for mountains and hills
	local mtsAdjust   = Round(1 + adjustment + (0.5 * extra_mountains));
	
	local hillsAdjust = mtsAdjust + (3 * adjustment);
	
	local trenches     = 15;
	
	local iHillsBottom = rangesFrac:GetHeight(middle - hillsAdjust); 
	local iMtnsBottom  = rangesFrac:GetHeight(middle - mtsAdjust);  
	local iMtnsTop     = rangesFrac:GetHeight(middle + mtsAdjust);
	local iHillsTop    = rangesFrac:GetHeight(middle + hillsAdjust);
	
	local iTrenchesLow  = rangesFrac:GetHeight(trenches);
	local iTrenchesHigh = rangesFrac:GetHeight(100 - trenches);

	
	-- Now the main loop, which will assign the plot types.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			local rangeVal = rangesFrac:GetHeight(x,y);
			
			if (iMtnsBottom <= rangeVal and rangeVal <= iMtnsTop) then
				-- Mountains sandwiched between hills, as far as possible
				-- from fractal extrema.
				mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;

			elseif (iHillsBottom <= rangeVal and rangeVal <= iHillsTop) then
				-- Hills surrounding and running parallel to the mountains.
				mtnPlots[i] = g_PLOT_TYPE_HILLS;
				
			elseif (rangeVal <= iTrenchesLow or rangeVal >= iTrenchesHigh) then
				-- Trench representing water at the map's lowest altitudes.
				trenchPlots[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end
end

function Scattered_GenerateMountainsAndHills(args, mtnPlots, trenchPlots)
	-- Mountains and hills - pattern-specific
	print("Generating mountains and hills (scattered crags).");
	
	-- This pattern combines fractal heights with die rolls in order
	-- to create "scattered" mountains with some areas of the map more
	-- dense than others. Yet no area will become too dense because 
	-- mountains will never be adjacent to each other.
	--
	-- Instead of threshold-based mountains and hills, there are zones 
	-- which determine the chances of crags and clumps. 
	--   -Crags are either mountains or hills such that no two 
	--    mountains are adjacent.
	--   -Clumps are crags that might have foothills.
	--
	-- The zones:
	-- 1) Clump zone: 100% chance of clumps for tiles that are 
	--    not already adjacent to mountains.
	--
	-- 2) Possible clump zone: 100% chance of crags, but not 
	--    all crags are clumps
	--
	-- 3) Isolated crag zone: small chance of crags, none of which
	--    are clumps
	
	
	
	-- landmass-specific mountain args + fractal config
	local args = args or {};
	
	-- Some island maps need more mountains due to lower
	-- chances of a mountain being on habitable land.
	local extra_mountains  = args.extra_mountains  or 0;
	local terrain_flags    = args.terrain_flags   or g_iFlags;

	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = GetMountainHeightAdjustment(args);
	
	-- Varying grains for hills/peaks per map size.
	local grain = GetMountainGrain(1);
		
	-- Generate fractals to govern clumps of mountains
	local clumpsFrac = Fractal.Create(
		g_iW, g_iH, 
		grain,
		terrain_flags, 
		g_fracXExp, g_fracYExp);
	
	-- Get height values for mountain clumps
	local clumpsAdjust       = 5 + (4*adjustment);
	local clumpsChanceAdjust = clumpsAdjust + (6*adjustment);
	
	-- Percentage chance that a tile outside of the clump zones becomes an 
	-- isolated crag. Isolated crags are never clumps.
	local cragRate = 10 + (5*adjustment);
	
	-- Percentage chance that a crag near a mountain range becomes a clump.
	-- A clump is a crag that possibly has foothills.
	local clumpRate = 15 + (3*adjustment);
	
	-- The percentage of crags that are mountains
	local mtnRate  = 5 + (8*adjustment) + (2*extra_mountains);
	
	-- The percentage of tiles surrounding clumps that are hills.
	local foothillRate = 8*adjustment;
	
	-- Trenches: the lowest altitudes of the clumps fractal.
	local trenches      = 65;
	
	local iClumps       = clumpsFrac:GetHeight(clumpsAdjust);	
	local iClumpsChance = clumpsFrac:GetHeight(clumpsChanceAdjust);
	local iTrenches     = clumpsFrac:GetHeight(trenches);
	
	-- If the map is nothing but scattered crags, then allow
	-- mountains on adjacent tiles. Otherwise, ensure that crags
	-- are completely scattered.
	local allowAdjacentMtns = args.adjacent_crags or false;
	if (allowAdjacentMtns == true) then
		print("Allowing adjacent crags.");
	end

	-- Now the main loop, which will assign the plot types.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			local rangeVal = clumpsFrac:GetHeight(x,y);

			local cragDice = 0;
			
			if (mtnPlots[i] == g_PLOT_TYPE_HILLS) then
				-- There is already a foothill here. Skip this tile.
				cragDice = cragRate;
			
			elseif (rangeVal > iClumpsChance) then
				-- Isolated crag zone, so roll dice for isolated crag.
				cragDice = TerrainBuilder.GetRandomNumber(100, "Scattered Crag");
			end
			
			if (cragDice < cragRate) then
				-- Scattered crag
				local nextToMtn = false;
				if (allowAdjacentMtns == false) then
					-- Check for adjacent mountains.
					for i, direction in ipairs(HexDirections) do

						local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
			
						if (adjacentPlot ~= nil) then
							-- We have an adjacent plot.
							-- Let's see if it's unvisited.
							local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
							local iAdjacent = aY * g_iW + aX + 1;
					
							if ((mtnPlots[iAdjacent] == g_PLOT_TYPE_MOUNTAIN) or 
								((mtnPlots[iAdjacent] == g_PLOT_TYPE_HILLS))) then
								nextToMtn = true;
							end
						end
					end
				end
						
				if (nextToMtn == false) then
					-- We're not next to a mountain, so let's make a crag.
					
					-- Decide whether the crag is a mountain or hill..
					local mtnDice = TerrainBuilder.GetRandomNumber(100, "Crag: mountain vs hill");
					
					if (mtnDice < mtnRate) then
						mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;
					else
						mtnPlots[i] = g_PLOT_TYPE_HILLS;
					end
					
					
					local clumpDice = 100;
					
					if (rangeVal <= iClumps) then
						-- Inside the guaranteed clumps zone.
						clumpDice = 0;
					
					elseif (rangeVal <= iClumpsChance) then
						-- Inside the possible clumps zone, so roll dice.
						clumpDice = TerrainBuilder.GetRandomNumber(100, "Scattered Crag Clump");
					end
					
					if (clumpDice < clumpRate) then
						-- Inside the clump zone.
						
						-- Roll dice, add surrounding hills.
						for i, direction in ipairs(HexDirections) do

							local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
				
							if (adjacentPlot ~= nil) then
								-- We have an adjacent plot.
								-- Let's see if it's unvisited.
								local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
								local iAdjacent = aY * g_iW + aX + 1;
						
								local hillDice = TerrainBuilder.GetRandomNumber(100, "Crag Foothill");
						
								if (hillDice < foothillRate) then
									mtnPlots[iAdjacent] = g_PLOT_TYPE_HILLS;
								end
							end
						end
					end
				end
				
			elseif (rangeVal >= iTrenches) then 
				-- Not a scattered crag, but a trench!
				trenchPlots[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end
end

function Tectonic_GenerateMountainsAndHills(args, mtnPlots, trenchPlots)
	-- Mountains and hills - pattern-specific
	print("Generating mountains and hills (tectonic plates).");
	
	-- The "vanilla" mountain pattern.
	
	-- landmass-specific mountain args + fractal config
	local args = args or {};
	
	-- Some island maps need more mountains due to lower
	-- chances of a mountain being on habitable land.
	local extra_mountains = args.extra_mountains  or 0;
	
	-- Larger landmasses like lakes and pangea work better 
	-- with less plates (lower adjustment), while 
	-- the more granular landmasses like archipelago work
	-- better with more plates (higher adjustment).
	--
	-- Adjustment is calibrated so that 1.0 works best for
	-- continents and fractal.
	-- 
	-- This value must be positive.
	local adjust_plates     = args.adjust_plates     or 1.0;
	local terrain_flags     = args.terrain_flags     or g_iFlags;
	
	
	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = GetMountainHeightAdjustment(args);
	
	local plate_multipliers = args.plate_multipliers or {1.5, 1.0, 0.75};
	local plate_multiplier = plate_multipliers[UserInput.worldAge];
	
	adjust_plates = adjust_plates * plate_multiplier;

	local grain = GetMountainGrain(3);
		
	-- Tectonic Plate Counts, for use in generating passes through the mountains.
	local numPlates = ScriptData.worldSizeInfo.numPlates;
	numPlates       = numPlates * adjust_plates;
	
	local blendRidge = args.blendRidge or 5;
	local blendFract = args.blendFract or 5;
	
	
	-- Generate fractals to govern hills and mountains
	local hillsFrac = Fractal.Create(
		g_iW, g_iH, 
		grain, 
		terrain_flags, 
		g_fracXExp, g_fracYExp);
		
	local mountainsFrac = Fractal.Create(
		g_iW, g_iH, 
		grain, 
		terrain_flags, 
		g_fracXExp, g_fracYExp);
	
	-- Add ridges to hills, even for non-tectonic mountain patterns.
	hillsFrac:BuildRidges(numPlates, g_iFlags, blendRidge, blendFract);
	
	-- Tectonic mountains
	mountainsFrac:BuildRidges(numPlates, g_iFlags, blendRidge, blendFract);
	
	-- Apply adjustment to hills and peaks settings.
	local mountains          = 94 - adjustment - extra_mountains;
	local hillsNearMountains = 88 - (2*adjustment) - extra_mountains;
	
	local hillsBottom1       = 28 - adjustment;
	local hillsTop1          = 28 + adjustment;
	local hillsBottom2       = 62 - adjustment;
	local hillsTop2          = 62 + adjustment;

	local trenches           = 30; -- Water unaffected by world age.
	
	-- Get height values for mountains and hills
	local iMountainThreshold  = mountainsFrac:GetHeight(mountains);
	local iHillsNearMountains = mountainsFrac:GetHeight(hillsNearMountains);
	local iTrenchesTop        = mountainsFrac:GetHeight(trenches);
	
	local iPassThreshold      = hillsFrac:GetHeight(hillsNearMountains);
	local iHillsBottom1       = hillsFrac:GetHeight(hillsBottom1);
	local iHillsTop1          = hillsFrac:GetHeight(hillsTop1);
	local iHillsBottom2       = hillsFrac:GetHeight(hillsBottom2);
	local iHillsTop2          = hillsFrac:GetHeight(hillsTop2);
		
	-- Now the main loop, which will assign the plot types.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			local mountainVal = mountainsFrac:GetHeight(x, y);
			local hillVal = hillsFrac:GetHeight(x, y);
			
			if (mountainVal >= iMountainThreshold) then
				-- Possible mountain or mountain pass
				if (hillVal >= iPassThreshold) then
					 -- Mountain Pass though the ridgeline
					mtnPlots[i] = g_PLOT_TYPE_HILLS;
				else 
					-- Mountain
					mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;
				end
				
			elseif (mountainVal >= iHillsNearMountains) then
				mtnPlots[i] = g_PLOT_TYPE_HILLS;
			
			elseif (mountainVal <= iTrenchesTop) then
				trenchPlots[i] = g_PLOT_TYPE_OCEAN;	
				
			elseif ((hillVal >= iHillsBottom1 and hillVal <= iHillsTop1) or 
					(hillVal >= iHillsBottom2 and hillVal <= iHillsTop2)) then
				mtnPlots[i] = g_PLOT_TYPE_HILLS;
			end
		end
	end
end

function Uncanny_GenerateMountainsAndHills(args, mtnPlots, trenchPlots)
	-- Mountains and hills - pattern-specific
	print("Generating mountains and hills (uncanny valleys).");
	
	-- A variant of Fractal Ranges that usually has more valleys.
	
	-- landmass-specific mountain args + fractal config
	local args = args or {};
	
	-- Some island maps need more mountains due to lower
	-- chances of a mountain being on habitable land.
	local extra_mountains  = args.extra_mountains  or 0;
	local terrain_flags    = args.terrain_flags   or g_iFlags;
	

	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = GetMountainHeightAdjustment(args);
	
	-- Varying grains for hills/peaks per map size.
	-- Grain-5 creates too many ranges, even for huge maps.
	local grain = math.clamp(GetMountainGrain(1), 1, 4);
		
	-- Generate fractals to govern mountain ranges
	local rangesFrac = Fractal.Create(
		g_iW, g_iH, 
		grain,
		terrain_flags, 
		g_fracXExp, g_fracYExp);
	
	-- Apply adjustment to hills and peaks settings.
	-- Base all ranges at the middle (50th percentile) of the fratal heights
	-- so that valleys are as snaky and non-blob-like as possible
	-- The fractal becomes blob-like at extreme values (0 and 100).
	local middle = 50;
	
	local innerTrenches = 2.5;
	
	-- Get height values for mountains and hills
	local valleysAdjust = 5;
				
	local mtsAdjust = valleysAdjust + (2 * adjustment);
	
	local hillsAdjust = mtsAdjust + (3*adjustment);
	
	local trenches = 8;
	
	local iTrenchesBottom = rangesFrac:GetHeight(trenches); 
	local iHillsBottom    = rangesFrac:GetHeight(middle - hillsAdjust); 
	local iMtsBottom      = rangesFrac:GetHeight(middle - mtsAdjust); 
	local iValleysBottom  = rangesFrac:GetHeight(middle - valleysAdjust);  
	local iVTrenchBottom  = rangesFrac:GetHeight(middle - innerTrenches);  
	local iVTrenchTop     = rangesFrac:GetHeight(middle + innerTrenches);
	local iValleysTop     = rangesFrac:GetHeight(middle + valleysAdjust);
	local iMtsTop         = rangesFrac:GetHeight(middle + mtsAdjust);	
	local iHillsTop       = rangesFrac:GetHeight(middle + hillsAdjust);
	local iTrenchesTop    = rangesFrac:GetHeight(100 - trenches); 

	
	-- Percent chance that a valley tile becomes a hill.
	local valleyHillsRate = 40;
	
	-- Percent chance that a valley mountain becomes a hill.
	local valleyPassRate = 10;
	
	local footMtnRate  = 10 + extra_mountains;
	local footHillRate = 40;
	
	local scatteredHillRate = 10;
	
	-- Now the main loop, which will assign the plot types.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			local rangeVal = rangesFrac:GetHeight(x,y);
			
			local dice = TerrainBuilder.GetRandomNumber(100, "Uncanney Valleys");
			
			if (iValleysBottom <= rangeVal and rangeVal <= iValleysTop) then
				-- Trench in a valley between mountains.
				trenchPlots[i] = g_PLOT_TYPE_OCEAN;
			
			elseif (iValleysBottom <= rangeVal and rangeVal <= iValleysTop) then
				-- Land sandwiched between mountains, as far as possible
				-- from fractal extrema. These are the uncanny valley tiles.
				if (dice < valleyHillsRate) then
					-- Random valley hill.
					mtnPlots[i] = g_PLOT_TYPE_HILLS;
				end

			elseif (iMtsBottom <= rangeVal and rangeVal <= iMtsTop) then
				-- Mountains surrounding and running parallel to the hills.
				if (dice < valleyPassRate) then
					-- Random mountain pass between valley mountains
					mtnPlots[i] = g_PLOT_TYPE_HILLS;
				else
					mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;
				end
				
			elseif (iHillsBottom <= rangeVal and rangeVal <= iHillsTop) then
				-- Foothills.
				if (dice < footMtnRate) then
					mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;	
				
				elseif (dice < footHillRate) then
					mtnPlots[i] = g_PLOT_TYPE_HILLS;	
				end
				
			elseif (rangeVal <= iTrenchesBottom or rangeVal >= iTrenchesTop) then
				-- Trench representing water at the map's lowest altitudes.
				trenchPlots[i] = g_PLOT_TYPE_OCEAN;
			else
				-- Flat area: add some scattered hills.
				if (dice < scatteredHillRate) then
					-- Random scattered hill
					mtnPlots[i] = g_PLOT_TYPE_HILLS;	
				end
			end
		end
	end
end

function Winding_GenerateMountainsAndHills(args, mtnPlots, trenchPlots)
	-- Mountains and hills - pattern-specific
	print("Generating mountains and hills (winding canyons).");
	
	-- A mountain pattern that uses a fractal to decide where
	-- *not* to put mountains. Relies on a reachabiliy algorithm
	-- to create mountain passes between canyon clumps.
	
	-- landmass-specific mountain args + fractal config
	local args = args or {};
	
	-- Ignores the extra_mountains option.
	local terrain_flags    = args.terrain_flags   or g_iFlags;
	

	-- Set values for hills and mountains according to World Age chosen by user.
	local adjustment = GetMountainHeightAdjustment(args);
	
	-- Fixed mountain grain: windy, but still playable-looking
	local grain = 4;
		
	-- Generate fractals to govern mountain ranges
	local rangesFrac = Fractal.Create(
		g_iW, g_iH, 
		grain,
		terrain_flags, 
		g_fracXExp, g_fracYExp);
	
	local mtsBase = 80;
	
	-- Get height values for mountains and hills
	local mtsAdjust   = mtsBase - 5*adjustment;
	
	local hillsAdjust = mtsAdjust * 0.75;
	
	local trenches = hillsAdjust * 0.5;
	
	local iMtnsBottom  = rangesFrac:GetHeight(mtsAdjust);
	local iHillsBottom = rangesFrac:GetHeight(hillsAdjust); 
	local iTrenchesTop = rangesFrac:GetHeight(trenches); 
	
	-- Now the main loop, which will assign the plot types.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			local rangeVal = rangesFrac:GetHeight(x,y);
			
			if (rangeVal >= iMtnsBottom) then
				-- Mountains forming canyon walls
				mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;
				
			elseif (rangeVal >= iHillsBottom) then
				-- Hills near the canyon walls
				mtnPlots[i] = g_PLOT_TYPE_HILLS;
			
			elseif (rangeVal <= iTrenchesTop) then
				-- Trench in the canyon
				trenchPlots[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end	
end

function GenerateMountainsAndHills(args, plotTypes)
	-- Generate a layer of mountains and hills.
	--
	-- Call this before generating extras.
	--
	-- Args:
	--   args: mountain args from fractal-taylored landmass config
	--   plotTypes: landmass mainland plots, not including extras
	
	local numTiles = g_iW * g_iH;

	-- Make a separate map for mountain pattern output that is
	-- agnostic of actual landmass info.
	local mtnPlots = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);
	
	-- track any pre-existing mountains in the mountain pattern's map.
	for i = 1, numTiles do
		if (plotTypes[i] == g_PLOT_TYPE_MOUNTAIN) then 
			mtnPlots[i] = g_PLOT_TYPE_MOUNTAIN;
		end
	end
	
	local trenchPlots = table.fill(g_PLOT_TYPE_LAND, g_iW * g_iH);
	
	if (UserInput.mountains == MtnPatterns.TECTONIC) then
		-- Tectonic plates
		Tectonic_GenerateMountainsAndHills(args, mtnPlots, trenchPlots);
	
	elseif (UserInput.mountains == MtnPatterns.FRACTAL) then
		-- Fractal ranges
		FRanges_GenerateMountainsAndHills(args, mtnPlots, trenchPlots);
		
	elseif (UserInput.mountains == MtnPatterns.SCATTERED) then
		-- Scattered crags
		Scattered_GenerateMountainsAndHills(args, mtnPlots, trenchPlots);
		
	elseif (UserInput.mountains == MtnPatterns.CLUSTERED) then
		-- Clustered highlands
		Clustered_GenerateMountainsAndHills(args, mtnPlots, trenchPlots);
		
	elseif (UserInput.mountains == MtnPatterns.UNCANNY) then
		-- Uncanny valleys
		Uncanny_GenerateMountainsAndHills(args, mtnPlots, trenchPlots);
		
	elseif (UserInput.mountains == MtnPatterns.WINDING) then
		-- Winding canyons
		Winding_GenerateMountainsAndHills(args, mtnPlots, trenchPlots);
	else
		print("Unexpected mountain pattern: " .. UserInput.mountains);
	end
	
	-- Set global so that we can extend mountain ranges to other parts
	-- of this script.
	ScriptData.mtnPlots    = mtnPlots;
	ScriptData.trenchPlots = trenchPlots;
end

function AddMountainsAndHillsToPlotTypes(mtnArgs, plotTypes)
	-- Apply mountain and hill plots to a world's plot types.
	--
	-- Args: 
	--   mtnArgs: mountain-pattern args for adding lonely mountains
	--   plotTypes: destination map for adding mountains and hills.
	--	
	-- Mountains and hills - fractal ranges
	print("Adding mountains and hills to the world.");
	
	
	local numTiles = g_iW * g_iH;
	
	-- Make a copy of the plot types before adding mountains, so that
	-- we know what to put back if we remove mountains.
	ScriptData.preMtnPlotTypes = table.fill(g_PLOT_TYPE_LAND, numTiles);
	for i = 1, numTiles do
		ScriptData.preMtnPlotTypes[i] = plotTypes[i];
	end
	
	local got_tectonic = (ScriptData.tectonicIslesMap ~= nil);
	
	-- Count # of mts, hills added
	local numMts   = 0;
	local numHills = 0;
	
	-- Apply mountain plot types to the main map.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			
			-- Check mountain pattern for mountains and hills.
			local srcPlotType = ScriptData.mtnPlots[i];
			
			if ((srcPlotType  ~= g_PLOT_TYPE_OCEAN) and
				(plotTypes[i] ~= g_PLOT_TYPE_OCEAN) and
				((got_tectonic == false) or (ScriptData.tectonicIslesMap[i] == g_PLOT_TYPE_OCEAN))) then
				-- Got a plot to add to the world.
				-- Replacing a land tile that is *not* a tectonic island.
				plotTypes[i] = srcPlotType;
				
				if (srcPlotType == g_PLOT_TYPE_MOUNTAIN) then
					numMts = numMts + 1;
				
				elseif (srcPlotType == g_PLOT_TYPE_HILLS) then
					numHills = numHills + 1;
				end
			end
		end
	end
	
	-- Note: coastal mountains may become hills during feature generation.
	print("Added " .. numMts .. " mountains and " .. numHills .. " hills.");
	
	if ((UserInput.mountains == MtnPatterns.SCATTERED) or
	    (UserInput.mountains == MtnPatterns.CLUSTERED)) then
		-- Scattered already has its own version of lonely mountains,
		-- and the whole point of clustered is *not* to have lonely mountains.
		print("Skipping lonely mountains.");
	else
		-- Add some lonely mountains.
		-- Need to do this after populating plot types with mountains
		-- so that lonely-mountain logic can use mountain-ratio.
		print("Adding lonely mountains.");
		local adjustment = GetMountainHeightAdjustment(mtnArgs);
		local mountainRatio = 11 + adjustment * 2;
		AddLonelyMountains(plotTypes, mountainRatio);
	end
end
------------------------------------------------------------------------------
-- Lake Freshening and Clusters
------------------------------------------------------------------------------
function IsWater(plotTypes, plot)
	-- Args:
	--   plotTypes: map of plot types.
	--   plot: plot to check against plot types.
	--
	-- Return true iff the plot is a water tile in the specified map.
	
	if (not plot) then
		-- Beyond the edge of the map: not water.
		return false;
	end
	
	local i = plot:GetY() * g_iW + plot:GetX() + 1;
	
	return plotTypes[i] == g_PLOT_TYPE_OCEAN;
end

function IsStrait(plotTypes, x, y)
	-- Check whether the plot at {x,y} could potentially break
	-- a lake into multiple lakes.
	--
	-- Args:
	--   plotTypes: a map of work-in-progress plot types
	--   x,y location of the plot to check
	--
	-- Returns true iff {x,y} is surrounded by at least two
	-- distinct groups of water tiles which are connected
	-- only by {x,y}.
	
	local prevIsWater = IsWater(plotTypes, Map.GetAdjacentPlot(x, y, HexDirections[6]));
	local numDiffs = 0;
	
	for i, direction in ipairs(HexDirections) do
	
		local curIsWater = IsWater(plotTypes, Map.GetAdjacentPlot(x, y, direction));
		
		if (prevIsWater ~= curIsWater) then
			-- Water/Land changed.
			numDiffs = numDiffs + 1;
			
			if (numDiffs > 2) then
				-- L,W,L,W or W,L,W,L pattern detected.
				-- Two distinct sets of water tiles would
				-- become locally separated if we made {x,y}
				-- a land tile.
				--
				-- Therefore, this tile is a lake chokepoint.
				--print("Lake chokepoint detected at " .. CoordString(x,y));
				return true;
			end
		end
		
		prevIsWater = curIsWater;
	end
	
	return false;
end

function FindWaterEdges(plotTypes, iX, iY)
	-- Traverse a lake to find all water tiles that are adjacent to at least one 
	-- land tile and at least two water tiles.
	--
	-- Args:
	--   plotTypes: map of plot types (search space)
	--   iX, iY: starting location; exclude this tile from results
	--
	-- Returns:
	--   waterEdges: a list of water tiles that are in the lake and adjacent
	--     to at least one land tile.
	--   visited: a map indicating which tiles belong to the lake.
	
	local waterEdges = {};
	
	local visited = table.fill(0, g_iW*g_iH);
	
	-- Use a stack to search neighboring tiles in DFS order until we've
	-- covered all of the water tiles that share the same body of water.
	local stack = Stack:Create();
	
	local iVisited = iY * g_iW + iX + 1; -- add one because Lua arrays start at 1
	
	stack:Push({x=iX, y=iY});
	visited[iVisited] = 1;
	
	while (not stack:IsEmpty()) do
	
		local coord = stack:Pop();
		local x, y = coord.x, coord.y;
		
		if ((x ~= iX) or (y ~= iY)) then
			-- Check neighbors for land. Ignore the starting tile.
			local isWaterEdge = false;
			local numLandNeighbors = 0;
			local numWaterNeighbors = 0;
			for i, direction in ipairs(HexDirections) do
				local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
				
				if (adjacentPlot) then
					-- We have an adjacent plot.
					-- Let's see if it's land.
					local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
					local iCouldVisit = aY * g_iW + aX + 1;
				
					if (plotTypes[iCouldVisit] == g_PLOT_TYPE_OCEAN) then
						-- Got an adjacent water tile.
						numWaterNeighbors = numWaterNeighbors + 1;
					else
						-- Got an adjacent land tile.
						numLandNeighbors = numLandNeighbors + 1;
					end
				else
					-- Treat edges of the map as potential lake edges.
					numLandNeighbors = numLandNeighbors + 1;
				end
			end
			
			if (numLandNeighbors > 0) then 
				-- Next to land or map edge.
				if (((numWaterNeighbors == 2) and IsStrait(plotTypes, x, y)) or 
					(numWaterNeighbors > 2)) then
					-- Also next to at least two water tiles.
					--
					-- If next to only two water tiles, then the water tiles must be separated by land.
					--   Otherwise, it's possible that we're at the corner of a lake.
					--
					-- If next to three more more water tiles, then we can start a land bridge without
					-- simply erasing the lake's coast.
					table.insert(waterEdges, {x=x, y=y});
				end
			end
		end
		
		
		for i, direction in ipairs(HexDirections) do
			local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
			
			if (adjacentPlot ~= nil) then
				-- We have an adjacent plot.
				-- Let's see if it's unvisited water.
				local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
				local iCouldVisit = aY * g_iW + aX + 1;
			
				if ((plotTypes[iCouldVisit] == g_PLOT_TYPE_OCEAN) and 
					(visited[iCouldVisit] == 0)) then
					-- Add to the stack of plots to visit.
					stack:Push({x=aX, y=aY});
					visited[iCouldVisit] = 1;
				end
			end
		end
	end

	return waterEdges, visited;
end

function CreateLandBridge(plotTypes, visited, iX, iY, lakeSizeMax, waterArea)
	-- Create a random path of land (or random land/sea) from one edge of the
	-- lake to another.
	--
	-- Args: 
	--   plotTypes: map of plot types
	--   visited: map indicating which tiles are part of the lake.
	--   iX, iY: where to begin the land bridge
	--   lakeSizeMax: the maximum number of water tiles to allow in the lake
	--     at the specified location; must be >= 9
	--   waterArea: the number of water tiles in the lake.
	--
	-- Returns:
	--   landBridgeSize: the number of tiles turned into land.
	
	--print("Creating a land bridge at " .. CoordString(iX, iY));
	
	if (IsStrait(plotTypes, iX, iY)) then
		--print("Land bridge complete (single-tile bridge).");
		local i = iY * g_iW + iX + 1;
		plotTypes[i] = g_PLOT_TYPE_LAND;
		return 1;
	end
	
	-- Use this value in the visited map to indicate land near the 
	local waterUnvisited = 1;
	local pathVisited    = 2;
	
	-- Scan the starting point's neighbors for adjacent lake-edges
	-- so that we don't simply walk along the lake's edge.
	local prevPlot    = Map.GetAdjacentPlot(iX, iY, HexDirections[6]);
	local prevIsWater = IsWater(plotTypes, prevPlot);
	for i, direction in ipairs(HexDirections) do
		local adjacentPlot = Map.GetAdjacentPlot(iX, iY, direction);
		
		local curIsWater = IsWater(plotTypes, adjacentPlot);
		
		if (curIsWater ~= prevIsWater) then
			-- Water next to land.
			local waterPlot = adjacentPlot;
			if (curIsWater == false) then
				waterPlot = prevPlot;
			end
			
			local aX, aY = waterPlot:GetX(), waterPlot:GetY();
			local iCouldVisit = aY * g_iW + aX + 1;
			
			-- Mark plot as visited so that the path
			-- avoids the lake's edge.
			visited[iCouldVisit] = pathVisited;
			--print("Avoiding plot at " .. CoordString(aX,aY));
		end
		
		prevIsWater = curIsWater;
		prevPlot    = adjacentPlot;
	end
	
	
	local x = iX;
	local y = iY;
	
	local landBridgeSize = 0;
	local prevDirectionIndex = nil;
	
	while (waterArea > lakeSizeMax) do
	
		waterArea = waterArea - 1;
	
		iVisited = y*g_iW + x + 1; -- add one because Lua arrays start at 1
		visited[iVisited]   = pathVisited;
		
		plotTypes[iVisited] = g_PLOT_TYPE_LAND;
		
		landBridgeSize = landBridgeSize + 1;
		
		if (waterArea <= lakeSizeMax) then
			-- Check water area before exploring neighbors.
			--print("Land bridge complete: lake is now fresh.");
			return landBridgeSize;
		end
		
		if (IsStrait(plotTypes, x, y)) then
			-- We successfully crossed the lake.
			--print("Land bridge complete: lake divided.");
			return landBridgeSize;
		end
		
		-- Search for neighbors that could extend the path.
		local adjacentCoords = {};
		for i, direction in ipairs(HexDirections) do
			local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
			
			if (adjacentPlot ~= nil) then
				-- We have an adjacent plot.
				-- Let's see if it's unvisited water.
				local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
				local iCouldVisit = aY * g_iW + aX + 1;
				
				if (visited[iCouldVisit] == waterUnvisited) then
					-- Unvisited water.
					table.insert(adjacentCoords, {iDir=i,x=aX,y=aY});
				end
			end
		end
		
		local numCoords = #adjacentCoords;
		if (numCoords == 0) then
			-- No way to extend the path. Not sure how we didn't notice
			-- that the prev tile was a chokepoint.
			--
			-- Maybe we tried too hard to avoid the start of the bridge.
			--print("Can't extend the path at " .. CoordString(x,y));
			return landBridgeSize;
		end
		
		local adjacentCoordsSimilarD = {};
		if (prevDirectionIndex ~= nil) then
			-- Look for adjacent coords within prevDirection +/- 1 to 
			-- straighten the land-bridge a bit.
			local dirLeft     = (prevDirectionIndex - 1) % 6;
			local dirStraight = prevDirectionIndex;
			local dirRight    = (prevDirectionIndex + 1) % 6;
			
			--print("Searching for adjacent coords in similar direction.");
			for i, coord in ipairs(adjacentCoords) do
				local iDir = coord.iDir;
				
				if ((iDir == dirLeft) or
				    (iDir == dirStraight) or
					(iDir == dirRight)) then
					table.insert(adjacentCoordsSimilarD, coord);
				end
			end
		end
		local numCoordsSimilarD = #adjacentCoordsSimilarD;
		
		if (numCoordsSimilarD > 0) then
			-- Limit selection of adjacent coords to similar direction.
			--print("Got " .. numCoordsSimilarD .. " adjacent coord(s) in similar direction.");
			adjacentCoords = adjacentCoordsSimilarD;
			numCoords      = numCoordsSimilarD;
		end

		-- Choose a random set of adjacent coords.
		local iCoord = 1 + TerrainBuilder.GetRandomNumber(numCoords, "Lake Freshening: extending land bridge");
		local coord = adjacentCoords[iCoord];
		x = coord.x;
		y = coord.y;
		prevDirectionIndex = coord.iDir;

		--print("Extending land bridge to " .. CoordString(x,y));
	end
	
	--print("Land bridge complete: lake is now fresh (exited loop).");
	return landBridgeSize;
end

function FreshenLake(plotTypes, x, y, lakeSizeMax)
	-- Add land until the specified tile is either land or part of a 
	-- body of water that is within the maximum size.
	--
	-- Args:
	--   plotTypes: map of plot types
	--   x, y: location of a salt-water tile.
	--   lakeSizeMax: the maximum number of water tiles to allow in the lake
	--     at the specified location; must be >= 9

	local waterArea = GetWaterArea(plotTypes, x, y, false);
	if (waterArea <= lakeSizeMax) then
		-- Already fresh enough.
		return;
	end
		
	--print("Freshening lake at " .. CoordString(x,y) .. ": water area is " .. waterArea);
	
	local totalLandAdded = 0;
	
	
	while (waterArea > lakeSizeMax) do
	
		-- Detect all water tiles at the edge of the lake (BFS/DFS traversal).
		local waterEdges, visited = FindWaterEdges(plotTypes, x, y);
		if (not waterEdges) then
			print("Failed to compute water edges; replacing water with land at " .. CoordString(x,y));
			-- The Brute-force approach: replace the salt water with land.
			local i = y * g_iW + x + 1;
			self.wholeworldPlotTypes[i] = g_PLOT_TYPE_LAND;
			return;
		end

		local numWaterEdges = #waterEdges;
		if (numWaterEdges == 0) then
			print("No water edges, so replacing water with land at " .. CoordString(x,y));
			-- The Brute-force approach: replace the salt water with land.
			local i = y * g_iW + x + 1;
			self.wholeworldPlotTypes[i] = g_PLOT_TYPE_LAND;
			return;
		end
		--print("Found " .. numWaterEdges .. " water edges at " .. CoordString(x,y));
		
		-- Choose an edge tile at random, but not (x,y)
		local bridgeStart = 
			waterEdges[1 + TerrainBuilder.GetRandomNumber(numWaterEdges,"Lake Freshening: bridge start")];
		
		-- Draw a random path of land or maybe rand mix of land/water from
		-- the edge tile to a non-adjacent edge tile.
		local landBridgeSize = CreateLandBridge(
			plotTypes, visited, bridgeStart.x, bridgeStart.y, lakeSizeMax, waterArea);
		--print("Land bridge size: ", landBridgeSize);
		totalLandAdded = totalLandAdded + landBridgeSize;
		
		-- Repeat until (x,y) is fresh water.
		waterArea = GetWaterArea(plotTypes, x, y, false);
	end
	
	--print("Added " .. totalLandAdded .. " land tile(s) for freshening.");
end

function FreshenLakes(plotTypes, lakeSizeMax, verbose)
	-- Ensure that all bodies of water are within the max lake size.
	--
	-- Args:
	--   plotTypes: map to freshen
	--   lakeSizeMax: target max lake size; some lakes may become smaller
	--     than this target.
	--   verbose: enable/disable debug messages
	
	
	
	if (verbose == true) then
		print("Freshening lakes.");
	end
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local i = y * g_iW + x + 1;
			if (plotTypes[i] == g_PLOT_TYPE_OCEAN) then
				-- Check the freshness of each water tile.
				-- Add land to salt-water lakes until they are fresh.
				FreshenLake(plotTypes, x, y, lakeSizeMax);
			end
		end
	end
	
	-- Do another pass just to make sure we didn't miss any salty lakes.
	if (verbose == true) then
		print("Verifying that lakes are now fresh.");
	end
	local numBruteForceTiles = 0;
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local i = y * g_iW + x + 1;
			if (plotTypes[i] == g_PLOT_TYPE_OCEAN) then
				-- Check the freshness of each water tile.
				-- Use brute-force freshening in this pass.
				if (GetWaterArea(plotTypes, x, y, false) > lakeSizeMax) then
					plotTypes[i] = g_PLOT_TYPE_LAND;
					numBruteForceTiles = numBruteForceTiles + 1;
				end
			end
		end
	end
	if (verbose == true) then
		if (numBruteForceTiles > 0) then
			print("Lakes are now fresh, but had to use brute force on " .. 
				numBruteForceTiles .. " tile(s).");
		else
			print("Lakes are now fresh.");
		end
	end
end

function GenerateLakesFractal(water_level, lake_grain)
	-- Use a lake fractal to add water to a land-heavy map.
	-- For lakes-family landmasses with global world wrap.
	--
	-- Args:
	--   water_level: lake water level
	--   lake_grain:  lake fractal granularity
	-- Returns:
	--   plotTypes: a map of generated lakes
	
	
	
	local plotTypes = table.fill(g_PLOT_TYPE_LAND, g_iW * g_iH);
	
	-- Let lakes touch / wrap around the edges.
	--local fracFlags = {FRAC_WRAP_X = ScriptData.wrapX, FRAC_POLAR = true};
	
	local lakesFrac = Fractal.Create(g_iW, g_iH, lake_grain, g_iFlags, -1, -1)
	local lakesThreshold = lakesFrac:GetHeight(water_level);
	
	for y = 0, g_iH-1 do
		for x = 0, g_iW - 1 do
			local i = y * g_iW + x + 1; -- add one because Lua arrays start at 1
			local lakeVal = lakesFrac:GetHeight(x, y);
			if lakeVal >= lakesThreshold then
				plotTypes[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end
	
	return plotTypes;
end

function GenerateLakesForClusters(inverted)
	-- Use a lake fractal to add water to a land-heavy map.
	-- For generating clusters if lakes (hides lake-levels and lake grains).
	--
	-- Args:
	--   inverted: true iff these lakes are for inverted clusters
	--
	-- Returns:
	--   plotTypes: a map of generated lakes that are not yet clusters
	
	local invertableSeaLevel = UserInput.seaLevel;
	if ((not inverted) or (inverted == false)) then
		-- The above levels are calibrated for islands.
		-- Reverse order for lakes.
		invertableSeaLevel = 4 - invertableSeaLevel;
	end
	
	-- Map sea levels to water percent thresholds.
	local lake_levels = {85,78,70};
	local lake_level  = lake_levels[invertableSeaLevel] or 
			            lake_levels[SEALEVEL_NORMAL]    or 78;
	
	-- Any smaller and lakes might not spread out enough.
	-- Any larger and clusters might be too much like chains.
	local lake_grain = 3;
	
	if (gotIceAgeFamily) then
		-- Use Ice Age sea levels, which are lower than normal sea levels.
		lake_level = lake_level + 5;
	end
	
	if (ExtrasGot.everything == true) then
		-- Got "everything" for extras and an extras landmass.
		-- Use a lower grain to distinguish clusters from other extras.
		lake_grain = 2;
		
	elseif (UserInput.landmass == Landmasses.EXTRAS) then
		-- Extras landmass
		-- Lower lake grain to create fewer, larger clusters
		lake_grain = 1;
		lake_level = lake_level - 16;
	
	elseif (UserInput.landmass == Landmasses.INVEXTRAS) then
		-- Inverted extras
		-- Raise sea level to increase lakes
		lake_level = lake_level - 12;
	
	elseif (UserInput.landmass == Landmasses.LAKES) then
		-- Lakes
		-- Raise sea level to increase lakes
		lake_level = lake_level - 16;
	end
	
	
	
	print("Clusters sea level: ", lake_level);
	print("Clusters grain: ", lake_grain);
	
	-- Use the general-purpose lake fractal with special inputs.
	return GenerateLakesFractal(lake_level, lake_grain);
end
------------------------------------------------------------------------------
-- Extras
------------------------------------------------------------------------------
function GenerateTectonicIslands()
	-- Generate tectonic islands.
	-- Apply a fractal filter to the mountain pattern.
	
	
	local plotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);
	
	local args = ScriptData.landInfo.extraArgs or {};
	
	local tectonic_isles = args.tectonic_isles;
	if (tectonic_isles == nil) then 
		tectonic_isles = true; 
	end
	
	if (tectonic_isles == false) then
		print("No tectonic extras for " .. ScriptData.landInfo.name);
		return plotTypes;
	end
	
	
	print("Generating tectonic extras.");
	
	local flags       = args.tectonicFlags  or {};
	local grain       = args.tectonicGrain  or 2;
	
	-- Add a tectonic fractal layer to govern mountains replacing water tiles.
	local tectonicFrac = Fractal.Create(
		g_iW, g_iH, 
		grain,
		flags, 
		g_fracXExp, g_fracYExp);
	
	local mtnPlots     = ScriptData.mtnPlots;
	local srcPlotMatch = g_PLOT_TYPE_MOUNTAIN;
	
	local seaLevel  = UserInput.seaLevel;
	local invertExtras = args.invertExtras;
	
	if invertExtras then
		-- Invert sea level.
		seaLevel = 4 - seaLevel;
		
		-- Place islands at trenches instead of mountains.
		-- Prevents overwriting mountain ranges with water.
		mtnPlots     = ScriptData.trenchPlots;
		srcPlotMatch = g_PLOT_TYPE_OCEAN;
	end	
			
	-- Higher sea level => higher thresholds => less islands
	local seaLevels = {0, 1, 2}; 
	local seaLevel = seaLevels[seaLevel];
	
	local tecSeaMultiplier = args.tectonicSeaMultiplier  or 1;
	seaLevel = seaLevel * tecSeaMultiplier;
	
	local iMtnDiff      = 20;
	local iHillDiff     = 30;
	local iFlatDiff     = 50;
	
	if (ExtrasGot.everything == true) then
		-- Got "everything" for extras.
		-- Reduce islands to make room for other types of extras.
		iMtnDiff      = 0.6 * iMtnDiff;
		iHillDiff     = 0.6 * iHillDiff;
		iFlatDiff     = 0.6 * iFlatDiff;
		
	elseif (ExtrasGot.tecAndChains == true) then
		-- Got "tectonic and chains" for extras.
		-- Reduce islands to make room for other extras.
		iMtnDiff      = 0.8 * iMtnDiff;
		iHillDiff     = 0.8 * iHillDiff;
		iFlatDiff     = 0.8 * iFlatDiff;
	end
	
	local iMtnBase      = (100 - iMtnDiff);      -- 80;
	local iHillBase     = (100 - iHillDiff);     -- 70;
	local iFlatBase     = (100 - iFlatDiff);     -- 50;
	
	-- Offset thresholds for certain extras-oriented landmasses.
	local landmassOffset  = args.tectonicLandOffset or 0;
	iMtnBase      = iMtnBase      - (0.75 * landmassOffset * iMtnDiff);
	iHillBase     = iHillBase     - (0.75 * landmassOffset * iHillDiff);
	iFlatBase     = iFlatBase     - (1.0  * landmassOffset * iFlatDiff);
 	
	-- Add sea level to thresholds.
	iMtnBase      = iMtnBase      + (0.25  * seaLevel * iMtnDiff);
	iHillBase     = iHillBase     + (0.25  * seaLevel * iHillDiff);
	iFlatBase     = iFlatBase     + (0.2   * seaLevel * iFlatDiff);
	
	iMtnBase      = math.min(95, iMtnBase);
	iHillBase     = math.min(iMtnBase-1, iHillBase);
	iFlatBase     = math.min(iHillBase-1, iFlatBase);
	
	--print("iMtnBase: ", iMtnBase);
	--print("iHillBase: ", iHillBase);
	--print("iFlatBase: ", iFlatBase);
	
	-- Thresholds at which mountains rise above the surface of the ocean.
	local iMtn  = tectonicFrac:GetHeight(iMtnBase);  -- 80
	local iHill = tectonicFrac:GetHeight(iHillBase); -- 75
	local iFlat = tectonicFrac:GetHeight(iFlatBase); -- 70
	
	-- Count # of mts, hills, islands added
	local numMts   = 0;
	local numHills = 0;
	local numFlats = 0;
	
	local gotScatteredTecExtras = 
		(UserInput.mountains == MtnPatterns.SCATTERED) and
		(UserInput.extras    == ExtraIslesLakes.TECTONIC) and
		(UserInput.landmass  == Landmasses.EXTRAS);
		
	if gotScatteredTecExtras then
		print("Including hills in tectonic islands (scattered crags + tectonic extras + extras landmass)");
	end
	
	-- Apply mountain plot types to the main map.
	for x = 0, g_iW - 1, 1 do
		for y = 0, g_iH - 1, 1 do
			local i = y*g_iW + x + 1; -- add one because Lua arrays start at 1
			
			if (mtnPlots[i] == srcPlotMatch) then
				-- Match from the mountain pattern
				local tectonicVal = tectonicFrac:GetHeight(x,y);
			
				-- Use a fractal to decide which mountains rise above the water.
				if (tectonicVal >= iMtn) then
					-- Replace a water tile with a mountain.
					plotTypes[i] = g_PLOT_TYPE_MOUNTAIN;
					numMts = numMts + 1;
					
				elseif (tectonicVal >= iHill) then
					-- Replace a water tile with a hill.
					plotTypes[i] = g_PLOT_TYPE_HILLS;
					numHills = numHills + 1;
				
				elseif (tectonicVal >= iFlat) then
					-- Replace a water tile with flat land
					plotTypes[i] = g_PLOT_TYPE_LAND;
					numFlats = numFlats + 1;
				end
				
			elseif (gotScatteredTecExtras and (mtnPlots[i] == g_PLOT_TYPE_HILLS)) then
				-- Include hills from scattered crags + tectonic extras + extras landmass
				-- so that the world has more than just super-tiny islands.
				local tectonicVal = tectonicFrac:GetHeight(x,y);
			
				-- Use a fractal to decide which hills rise above the water.
				if (tectonicVal >= iMtn) then
					-- Replace a water tile with flat land
					plotTypes[i] = g_PLOT_TYPE_LAND;
					numFlats = numFlats + 1;
				end
			end
		end
	end
	
	-- Note: coastal mountains may become hills during feature generation.
	print("Tectonic tiles added: ");
	print("  Mtns: ", numMts);
	print("  Hills: ", numHills);
	print("  Flats: ", numFlats);
	
	-- Set global so that we can reuse for shallow water.
	ScriptData.tectonicIslesFractal = tectonicFrac;
	
	-- The mountain pattern needs this in order to avoid
	-- clobbering the tectonic islands with mountains and hills.
	ScriptData.tectonicIslesMap = plotTypes;
	
	return plotTypes;
end

function RandomSlice(rDonut)
	-- Args: 
	--    rDonut: the radius of the donut; determines mimimum slice length
	--
	-- Return [length, offset] 
	--   length: the length of the slice in radians
	--   offset: rotate the slice counter-clockwise by this much in radians.
	--     A slice with an offset of 0 begins at 6:00

	-- Enforce a minimum slice length based on donut size to avoid slices 
	-- that don't look anything like a donut.

	-- Default min length.
	local minLength = math.pi;
	
	if (rDonut >= 8) then
		minLength = TAU/3;
	end
	
	local length = RandomRadians(minLength);
	local offset = RandomRadians();
	
	--print("Slice length " .. RadString(length));
	--print("Slice offset " .. RadString(offset));
	
	return length, offset;
end

function IsInSlice(enableSlicing, theta, sliceLength)
	-- Determine whether a certain donut plot is part of the slice to include.
	--
	-- Args:
	--   enableSlicing: if false, then bypass this check and just return true
	--     so that the entire donut is in the slice.
	--   theta: angle to check
	--   sliceLength: the length of the slice in radians
	--
	-- Returns true if doWholeSlice is enabled or theta is in [0,sliceLength).

	if (enableSlicing == false) then
		-- Forcing an unsliced donut.
		return true;
	end
	
	return ((0 <= theta) and (theta < sliceLength));
end

function GenerateMiniDonut(island_plotTypes, centerX, centerY, args)
	-- Generate a single mini-donut.
	--
	-- Args
	--   island_plotTypes: add the mini-donut to this map.
	--   centerX,centerY: target location of the donut's center.
	--   args: donut args from landmass/extras config; ignored if nil
	
	-- Note: some landmasses such as Pangea may shift the mini donut 
	-- as part of landmass generation.
	--print("Generating mini-donut at " .. CoordString(centerX,centerY));
	
	local args = args or {};
	
	
		
	-- If the center is on an odd row then shift x to account for hex grid.
	-- When calculating Euclidian distance.
	local centerXEucl = centerX;
	if ((centerY % 2) == 1) then
		centerXEucl = centerXEucl + 0.5;
	end
	
	-- Donut radius range.
	local rMin = args.rMin or 2;
	local rMax = args.rMax or 3;
	local rDonut = rMin + TerrainBuilder.GetRandomNumber(rMax-rMin+1, "M-Donut radius");
	--print("Donut radius: " .. rDonut);

	
	-- Donut thicknesss range.
	local tMin = args.tMin or 1;
	local tMax = args.tMax or rDonut+1;
	tMin = math.clamp(tMin, 1,    rDonut+1);
	tMax = math.clamp(tMax, tMin, rDonut+1);
	--print("Donut thickness range: " .. CoordString(tMin,tMax));
	local thickness = tMin + TerrainBuilder.GetRandomNumber(tMax-tMin+1, "M-Donut thickness");
	--print("Donut thickness: " .. thickness);
	
	-- Use thickness to determine range for hole radius.
	-- The hole is the circle of water at the donut's center.
	-- A negative radius means no hole (filled donut).
	local rHole = rDonut - thickness;
	--print("Donut hole radius: " .. rHole);
	
	
	-- Donut slice boundaries.
	local sliceLength = TAU;
	local sliceOffset = 0;
	local enableSlicing = args.enableSlicing or false;
	if (enableSlicing == false) then
		--print("No slicing.");
	else
		sliceLength, sliceOffset = RandomSlice(rDonut);
	end
	
	
	-- Limit which plots to visit; improves performance.
	local rDonutBound = 2*rDonut;	
	local westX = centerX - rDonutBound;
	local eastX = centerX + rDonutBound;
	local southY = math.max(0,    centerY - rDonutBound);
	local northY = math.min(g_iH-1, centerY + rDonutBound);

	if (UserInput.worldWrap == WWRAP_REGION) then
		-- Regional map: don't wrap donuts in the x-direction. 
		westX  = math.max(0,    westX);
		eastX  = math.min(g_iW-1, eastX);
	end
	
	--print("Donut boundary: " .. CoordString(westX, southY) .. " to " .. CoordString(eastX, northY)); 
				  

	-- Use a fractal to add watery imperfections to the donut.
	local baysGrain  = 5;
	if (rDonut >= 10) then
		-- Use larger bays for larger donuts.
		baysGrain = 4;
	end
	--print("Bays grain " .. baysGrain);	
	local baysHeight = args.baysHeight or 74;
	local fracFlags  = {FRAC_WRAP_X = true, FRAC_POLAR = true};
	local baysFrac   = Fractal.Create(g_iW, g_iH, baysGrain, fracFlags, -1, -1);
	local iBaysThreshold = baysFrac:GetHeight(baysHeight);
	
	
	-- Loop through plots in the donut's bounding box and add land
	-- to the plots that are part of the donut.
	for x = westX, eastX do
		local plotX = x % g_iW;
		local euclX = x;
		--print("Column " .. plotX);
	
		for y = southY, northY do		
			-- Use Euclidian distance so that donuts look circular despite hex tiles.
			-- If y is an odd row then account for hex-grid tile-shift.
			if ((y % 2) == 1) then
				euclX = x + 0.5;
			else
				euclX = x;
			end
			
			local deltaX = euclX - centerXEucl;
			local deltaY = y - centerY;
			local deltaXSquared = deltaX * deltaX;
			local deltaYSquared = deltaY * deltaY;
			
			-- Use rounded Euclidian distance to improve donut symmetry.
			local d = Round(math.sqrt(deltaXSquared + deltaYSquared));
						
			-- Convert cartesian coords to theta so that we can take a 
			-- slice of the donut.
			--
			-- Offset theta to the range [0,2pi]
			local theta = math.atan2(deltaX, deltaY) + math.pi + sliceOffset;
			if (theta > TAU) then
				theta = theta - TAU;
			end

			-- Check whether to add fractal water to this tile.
			local baysVal = baysFrac:GetHeight(x, y);
			
			if ((rHole < d) and (d <= rDonut) and
			    (IsInSlice(enableSlicing, theta, sliceLength) == true) and
				(baysVal < iBaysThreshold)) then
				-- Donut land.		
				local i = y * g_iW + plotX + 1;
				
				island_plotTypes[i] = g_PLOT_TYPE_LAND;
				--print("Donut at " .. CoordString(deltaX,deltaY) .. " theta " .. theta);
				
			elseif (d <= rHole) then
				-- Donut hole.
				--local i = y * g_iW + plotX + 1;
				--ScriptData.donutFillingMap[i] = FILLING_YES;	
				--print("Filling at " .. CoordString(deltaX,deltaY) .. " theta " .. theta);
			end
		end
	end
end

function GenerateMiniDonuts()
	-- Generate mini-donut islands
	--
	-- Return a map of generated island plot types.
	print("Generating mini-donuts.");
	
	-- Args to pass into the per-donut helper function.
	local donutArgs = {};
	
	

	local landArgs = ScriptData.landInfo.extraArgs or {};
	local invertExtras = landArgs.invertExtras;
	
	-- Some sea-level settings reverse when we have inverted extras.
	local invertableSeaLevel  = UserInput.seaLevel;
	
	if (invertExtras == true) then
		-- Flip the sea level.
		print("Inverting sea level.");
		invertableSeaLevel = 4 - invertableSeaLevel;
	end	
	
	local island_PlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);
	
	local worldSizeArgs = ScriptData.worldSizeInfo.miniDonuts or {};
	
	
	-- Donut integrity.
	-- higher number => more donut
	local baysHeight = landArgs.miniDonutBaysHeight or 74;
	donutArgs.baysHeight = baysHeight;
	--print("Bays Height ", donutArgs.baysHeight);
	
	
	-- Min/Max donut radius based on world size and landmass.
	local rMin = worldSizeArgs.radiusMin or 2;
	local rMax = worldSizeArgs.radiusMax or 3;
	local radiusOffset = landArgs.miniDonutROffset or 0;
	rMax = rMax + radiusOffset;
	if ((UserInput.landmass == Landmasses.INVEXTRAS) and
		(UserInput.extras   == ExtraIslesLakes.DONUTS)) then
		-- Inverted Extras landmass with nothing but mini-donuts.
		-- Increase the size of the donuts.
		rMin = rMin * 2;
		rMax = rMax * 2;
	end
	
	donutArgs.rMin = rMin;
	donutArgs.rMax = rMax;
	print("Radius range: ", CoordString(donutArgs.rMin, donutArgs.rMax));
	
	
	-- Min/Max donut thickness based on sea level.
	local tMins  = {rMax-2, Round(rMax/3),   1};
	local tMaxes = {rMax,   Round(2*rMax/3), 2};
	donutArgs.tMin = tMins[invertableSeaLevel];
	donutArgs.tMax = tMaxes[invertableSeaLevel];
	print("Thickness range: ", CoordString(donutArgs.tMin, donutArgs.tMax));

	
	-- Use a cell system to distribute donuts across the map.
	-- Divide the map into a dozen cells by default.
	local numCellColumns = 4;
	local numCellRows    = 3;
	local numTotalCells  = numCellColumns * numCellRows;
	--print("Number of donut cells: ", numTotalCells);

	-- Choose a random cell to let a donut "escape" to anywhere on the map.
	-- Makes the donut grid less obvious.
	local iEmptyCell = 1 + TerrainBuilder.GetRandomNumber(numTotalCells, "Empty mini-donut cell");
	
	-- Use a shuffled grid of donut types so that we can balance
	-- randomness with consistency between maps.
	local cellNotSliced = 0; -- Not sliced.
	local cellSliced    = 1; -- Randomly sliced.	
	local cellTypes = table.fill(cellNotSliced, numTotalCells);

	-- Percentage of sliced donut instead of unsliced donut.
	local slicedPct = 25;
	local numSlicedCells = math.clamp(0, numTotalCells, Round(numTotalCells*slicedPct/100));
	--print("Number of sliced cells: ", numSlicedCells);
	for i = 1, numSlicedCells do
		cellTypes[i] = cellSliced;
	end
	cellTypes = GetShuffledCopyOfTable(cellTypes);
	
	-- Use floating numbers here; will round when calculating cell boundaries.
	local cellWidth  = g_iW / numCellColumns;
	local cellHeight = g_iH / numCellRows;
	print("Cell size (w,h): ",  CoordString(cellWidth, cellHeight));
	
	-- Add padding to cells to reduce overlap between donuts.
	-- Slightly less than rMax to make cell grid less obvious.
	local cellPaddingX = math.floor(math.clamp(donutArgs.rMax-1, 0, cellWidth/3));
	local cellPaddingY = math.floor(math.clamp(donutArgs.rMax-1, 0, cellHeight/3));
	print("Cell padding: ", CoordString(cellPaddingX, cellPaddingY));

	
	local numDonutsPerCell = worldSizeArgs.numDonutsPerCell or 1;
	numDonutsPerCell = math.max(1,numDonutsPerCell);
	if ((UserInput.landmass == Landmasses.EXTRAS) and
		(UserInput.extras   == ExtraIslesLakes.DONUTS)) then
		-- Extras landmass with nothing but mini-donuts.
		-- Increase the number of donuts.
		numDonutsPerCell = numDonutsPerCell * 2;
	end
	print("Numer of donuts per cell: ", numDonutsPerCell);
	
	-- Boundaries for the current cell.
	local iWestX  = 0;
	local iEastX  = 0;
	local iSouthY = 0;
	local iNorthY = 0;
	local iCell = 1;
	for iRow = 1, numCellRows do
		-- Advance to the next row.
		iEastX  = 0;
		iSouthY = iNorthY;
		iNorthY = math.floor(iRow * cellHeight);
	
		for iCol = 1, numCellColumns do
			-- Advance to the next column.
			iWestX = iEastX;
			iEastX = math.floor(iCol * cellWidth);
			
			for iDonutPerCell = 1, numDonutsPerCell do
			
				local donutX, donutY;
				
				if (iCell == iEmptyCell) then
					-- Empty cell: donut can be anywhere on the map.
					donutX = cellPaddingX + TerrainBuilder.GetRandomNumber(g_iW - 2*cellPaddingX, "Mini-Donut X (empty cell)");
					donutY = cellPaddingY + TerrainBuilder.GetRandomNumber(g_iH - 2*cellPaddingY, "Mini-Donut Y (empty cell)");
				else
					--print("Current cell is " .. CoordString(iWestX, iSouthY) .. " to " .. CoordString(iEastX, iNorthY));
				
					-- Choose a random location for the donut in the cell.
					donutX = iWestX + cellPaddingX + 
						TerrainBuilder.GetRandomNumber(iEastX  - iWestX - 2*cellPaddingX,  "Mini-Donut Cell X");
						
					donutY = iSouthY + cellPaddingY + 
						TerrainBuilder.GetRandomNumber(iNorthY - iSouthY - 2*cellPaddingY, "Mini-Donut Cell Y");
				end
				
				-- Check the cell grid to decide whether to slice this donut.
				donutArgs.enableSlicing = (cellTypes[iCell] == cellSliced);
					
				-- Generate the donut.
				GenerateMiniDonut(island_PlotTypes, donutX, donutY, donutArgs);
			end
			
			-- Advance to the next cell.
			iCell = iCell + 1;
		end
	end
	
		
	-- Place additional, smaller donuts anywhere on the map.
	local numExtraDonuts = worldSizeArgs.numExtraDonuts or 10;
	if (ExtrasGot.everything == true) then
		-- Got "everything" for extras.
		-- Reduce donuts to make room for other types of extras.
		numExtraDonuts = Round(numExtraDonuts/3);
	end
	
	print("Placing " .. numExtraDonuts .. " additional donuts.");
	donutArgs.rMin = 1;
	donutArgs.rMax = math.max(1,Round(donutArgs.rMax/2));
	
	for iDonut = 1, numExtraDonuts do
		local donutX = cellPaddingX + TerrainBuilder.GetRandomNumber(g_iW - 2*cellPaddingX, "Mini-Donut X");
		local donutY = cellPaddingY + TerrainBuilder.GetRandomNumber(g_iH - 2*cellPaddingY, "Mini-Donut Y");
		
		-- Randomly decide whether to slice the extra donut.
		donutArgs.enableSlicing = (TerrainBuilder.GetRandomNumber(100, "Extra donut slicing.") < slicedPct);
		
		GenerateMiniDonut(island_PlotTypes, donutX, donutY, donutArgs);
	end
	
	
	--print("Done generating mini-donuts.");
	
	return island_PlotTypes;
end

function CreateSingleAxisIslandChain(iWidth, iHeight, fTilt)
	-- Copied from Island Maker for the sole purpose of supressing
	-- obnoxious debug messages.
	
	-- This function will create the simplest form of island chain. (fTilt should be in degrees.)
	-- Data returned will be a table of LOCAL plot indices, true for land, false for water.
	local iWH = iWidth * iHeight;
	-- If dimensions are too small to operate, or the angle is not between 0 and 180 degrees, abort.
	if iWidth < 4 or iHeight < 4 or iWH < 24 or fTilt < 0 or fTilt > 180 then
		print("*** ERROR! ***"); print("Cell dimensions or Axis angle invalid!"); print("-");
		return nil
	end
	-- This function is going to execute, so set up a data table for the plots.
	local plot_data = table.fill(false, iWH);
	-- Define the center and the radii.
	local iCenterX, iCenterY = (iWidth - 1) / 2, (iHeight - 1) / 2;
	local fRadiusX, fRadiusY = iWidth / 2, iHeight / 2;
	-- Using triangular method of trigonometry. If angle > 90 degrees, swap sides and convert the angle.
	local fTilt_in_degrees = fTilt;
	if fTilt > 90 then
		fTilt_in_degrees = 180 - fTilt;
	end
	-- Convert angle to radians. (Lua math table functions operate on radians).
	local fTilt_in_radians = math.rad(fTilt_in_degrees);
	-- Determine traits of the angle.
	local is_horizontal, is_vertical, is_steep, leans_left = false, false, false, false;
	if fTilt == 0 or fTilt == 180 then
		is_horizontal = true;
	elseif fTilt == 90 then
		is_vertical = true;
	else
		if fTilt > 45 and fTilt < 135 then
			is_steep = true;
		end
		if fTilt > 90 then
			leans_left = true;
		end
	end
	-- Obtain start and end plots for the axis.
	local start_x, start_y, end_x, end_y;
	if is_horizontal then
		start_x, start_y, end_x, end_y = 0, iCenterY, iWidth - 1, iCenterY;
	elseif is_vertical then
		start_x, start_y, end_x, end_y = iCenterX, 0, iCenterX, iHeight - 1;
	else
		-- Obtain the tangent of this angle.
		local TAN = math.tan(fTilt_in_radians);
		-- Figure out which axis to use as the measurement basis for the triangle.
		-- (The one that won't let the trangle exceed the defined width and height).
		local width_viable, height_viable = false, false;
		--
		-- Check with width as primary axis.
		local width_check = fRadiusY / TAN;
		if width_check <= fRadiusX then
			width_viable = true;
		end
		-- Check with height as primary axis.
		local height_check = fRadiusX * TAN;
		if height_check <= fRadiusY then
			height_viable = true;
		end
		-- Check for error.
		if width_viable == false and height_viable == false then
			-- Neither width nor height checked out as viable? That should not be mathematically possible!
			print("!!!"); print("! ERROR ! - Trigonometric failure !"); print("!!!");
			return nil
		end
		-- Assign the coordinates.
		if width_viable then
			if leans_left == false then
				start_x, start_y, end_x, end_y = iCenterX - width_check, 0, iCenterX + width_check, iHeight - 1;
			else
				start_x, start_y, end_x, end_y = iCenterX - width_check, iHeight - 1, iCenterX + width_check, 0;
			end
		elseif height_viable then
			if leans_left == false then
				start_x, start_y, end_x, end_y = 0, iCenterY - height_check, iWidth - 1, iCenterY + height_check;
			else
				start_x, start_y, end_x, end_y = 0, iCenterY + height_check, iWidth - 1, iCenterY - height_check;
			end
		else
			print("*** ERROR *** Width and Height Checks failed! ***");
		end
	end
	
	-- Debugging
	--print("-"); print("- Traits List - Width:", iWidth, "Height:", iHeight, "Angle:", fTilt);
	--print("-"); print("Degrees:", fTilt_in_degrees, " - Radians:", fTilt_in_radians); print("-");
	--print("Start Coordinates:", start_x, start_y, " - End Coordinates:", end_x, end_y); print("-");

	local startX, startY, endX, endY, bReverseFlag, slope;
	-- Note: Incoming x/y are starting at 0/0, but we need to convert them to Lua array values that start at 1.
	if math.abs(end_y - start_y) < math.abs(end_x - start_x) then -- line is closer to horizontal
		if start_x > end_x then
			startX, startY, endX, endY = end_x + 1, end_y + 1, start_x + 1, start_y + 1; -- swap start and end
			bReverseFlag = true;
			--print("Path reversed, working from the end plot.");
		else -- don't swap
			startX, startY, endX, endY = start_x + 1, start_y + 1, end_x + 1, end_y + 1;
			bReverseFlag = false;
			--print("Path not reversed.");
		end
		local dx = endX - startX;
		local dy = endY - startY;
		if dx == 0 or dy == 0 then
			slope = 0;
		else
			slope = dy/dx;
		end
		--print("Slope: ", slope);
		local y, y_true = nil, startY;
		-- Process the Axis.
		for x = startX, endX do
			local y_int, y_frac = math.modf(y_true);
			if y_frac >= 0.5 then
				y = math.ceil(y_true);
			else
				y = math.floor(y_true);
			end
			-- Determine if this plot is to become land.
			local diceroll = TerrainBuilder.GetRandomNumber(5, "Island Making - Lua");
			if diceroll > 1 then -- Yes, turn this plot to land.
				local i = (y - 1) * iWidth + x;
				plot_data[i] = true;
			end	
			-- Adjust y for the next iteration of the loop.		
			y_true = y_true + slope;
			--print("y plus slope: ", y_true);
		end

	else -- line is closer to vertical
		if start_y > end_y then
			startX, startY, endX, endY = end_x + 1, end_y + 1, start_x + 1, start_y + 1; -- swap start and end
			bReverseFlag = true;
			--print("Path reversed, working from the end plot.");
		else -- don't swap
			startX, startY, endX, endY = start_x + 1, start_y + 1, end_x + 1, end_y + 1;
			bReverseFlag = false;
			--print("Path not reversed.");
		end
		local dx = endX - startX;
		local dy = endY - startY;
		if dx == 0 or dy == 0 then
			slope = 0;
		else
			slope = dx/dy;
		end
		--print("Slope: ", slope);
		local x, x_true = nil, startX;
		-- Process the Axis.
		for y = startY, endY do
			local x_int, x_frac = math.modf(x_true);
			if x_frac >= 0.5 then
				x = math.ceil(x_true);
			else
				x = math.floor(x_true);
			end
			-- Determine if this plot is to become land.
			local diceroll = TerrainBuilder.GetRandomNumber(5, "Island Making - Lua");
			if diceroll > 1 then -- Yes, turn this plot to land.
				local i = (y - 1) * iWidth + x;
				plot_data[i] = true;
			end	
			-- Adjust x for the next iteration of the loop.		
			x_true = x_true + slope;
			--print("x plus slope: ", x_true);
		end
	end
	
	return plot_data
end

function CreateAxisChainWithDots(iWidth, iHeight, fTilt, iInnerWidth, iInnerHeight, iNumDots)
	-- This function adds "dots" (randomly placed single plots of land) to an axis island chain.
	local plot_data = CreateSingleAxisIslandChain(iWidth, iHeight, fTilt)
	-- Error handling.
	if iInnerWidth > iWidth - 2 then
		iInnerWidth = iWidth - 2;
	end
	if iInnerHeight > iHeight - 2 then
		iInnerHeight = iHeight - 2;
	end
	local westX = 1 + math.floor((iWidth - iInnerWidth) / 2);
	local southY = 1 + math.floor((iHeight - iInnerHeight) / 2);
	-- Add dots. ... Note: No collision checks. Any dots placed on existing plots are lost. This is intended.
	for loop = 1, iNumDots do
		local randomX = westX + TerrainBuilder.GetRandomNumber(iInnerWidth, "Random X coord for Dot - Lua");
		local randomY = southY + TerrainBuilder.GetRandomNumber(iInnerHeight, "Random Y coord for Dot - Lua");
		local i = (randomY - 1) * iWidth + randomX;
		plot_data[i] = true;
	end

	return plot_data
end

function CreateAxisChainWithShiftedDots(iWidth, iHeight, fTilt, iInnerWidth, iInnerHeight, iInnerWest, iInnerSouth, iNumDots)
	-- The dots do not necessarily go in the middle, in this function. Use West and South to set the location of the dots.
	local plot_data = CreateSingleAxisIslandChain(iWidth, iHeight, fTilt)
	-- Error handling.
	if iInnerWidth > iWidth - 2 then
		iInnerWidth = iWidth - 2;
	end
	if iInnerHeight > iHeight - 2 then
		iInnerHeight = iHeight - 2;
	end
	-- Add dots. ... Note: No collision checks. Any dots placed on existing plots are lost. This is intended.
	for loop = 1, iNumDots do
		local randomX = iInnerWest + TerrainBuilder.GetRandomNumber(iInnerWidth, "Random X coord for Dot - Lua");
		local randomY = iInnerSouth + TerrainBuilder.GetRandomNumber(iInnerHeight, "Random Y coord for Dot - Lua");
		local i = (randomY - 1) * iWidth + randomX;
		plot_data[i] = true;
	end

	return plot_data
end

function CreateAxisChainWithDoubleDots(iWidth, iHeight, fTilt, iInnerWidth1, iInnerHeight1, iInnerWest1, iInnerSouth1,
                                       iNumDots1, iInnerWidth2, iInnerHeight2, iInnerWest2, iInnerSouth2, iNumDots2)
	-- Two sets of dots, with control over size and location.
	local plot_data = CreateSingleAxisIslandChain(iWidth, iHeight, fTilt)
	-- Error handling.
	if iInnerWidth1 > iWidth - 2 then
		iInnerWidth1 = iWidth - 2;
	end
	if iInnerHeight1 > iHeight - 2 then
		iInnerHeight1 = iHeight - 2;
	end
	if iInnerWidth2 > iWidth - 2 then
		iInnerWidth2 = iWidth - 2;
	end
	if iInnerHeight2 > iHeight - 2 then
		iInnerHeight2 = iHeight - 2;
	end
	-- Dots1
	for loop = 1, iNumDots1 do
		local randomX = iInnerWest1 + TerrainBuilder.GetRandomNumber(iInnerWidth1, "Random X coord for Dot - Lua");
		local randomY = iInnerSouth1 + TerrainBuilder.GetRandomNumber(iInnerHeight1, "Random Y coord for Dot - Lua");
		local i = (randomY - 1) * iWidth + randomX;
		plot_data[i] = true;
	end
	-- Dots2
	for loop = 1, iNumDots2 do
		local randomX = iInnerWest2 + TerrainBuilder.GetRandomNumber(iInnerWidth2, "Random X coord for Dot - Lua");
		local randomY = iInnerSouth2 + TerrainBuilder.GetRandomNumber(iInnerHeight2, "Random Y coord for Dot - Lua");
		local i = (randomY - 1) * iWidth + randomX;
		plot_data[i] = true;
	end

	return plot_data
end

function GenerateChains()
	-- Generate Chains of extra islands
	--
	-- Return a map of generated island chain plot types.
	
	-- Now we begin to add the island chains.
	print("Generating chains of islands/lakes.");

	-- Higher sea level => use less cells.
	local seaLevel  = UserInput.seaLevel;
	
	local landArgs = ScriptData.landInfo.extraArgs or {};
	local invertExtras = landArgs.invertExtras;
	if invertExtras then
		-- Invert sea level.
		seaLevel = 4 - seaLevel;
	end
	
	local cellTargetFracs = {0.88, 0.66, 0.44}; 
	
	local cellTargetFrac  = cellTargetFracs[seaLevel]; -- originally fixed at 0.66
	
	if (ExtrasGot.everything == true) then
		-- Got "everything" for extras.
		-- Reduce islands to make room for other types of extras.
		cellTargetFrac = cellTargetFrac / 3;
	
	elseif (ExtrasGot.tecAndChains == true) then
		-- Got "tectonic and chains" for extras.
		-- Reduce islands to make room for other extras.
		cellTargetFrac = cellTargetFrac / 2;
	end
	
	-- A cell system will be used to combine predefined land chunks with randomly generated island groups.
	-- Define the cell traits. (These need to fit correctly with the map grid width and height.)
	
	local iCellWidth = 10;
	local iCellHeight = 8;
	local iNumCellColumns = math.floor(g_iW / iCellWidth);
	local iNumCellRows = math.floor(g_iH / iCellHeight);
	local iNumTotalCells = iNumCellColumns * iNumCellRows;
	local cell_data = table.fill(false, iNumTotalCells) -- Stores data on map cells in use. All cells begin as empty.
	local iNumCellsInUse = 0;
	local iNumCellTarget = math.floor(iNumTotalCells * cellTargetFrac);
	local island_chain_PlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);

	-- Margins for centering the cell grid on the map.
	local xMargin = math.floor((g_iW - iNumCellColumns * iCellWidth) /2);
	local yMargin = math.floor((g_iH - iNumCellRows    * iCellHeight)/2);
	
	--print("Target number of cells: " .. iNumCellTarget);
	
	-- Add randomly generated island groups
	local iNumGroups = iNumCellTarget; -- Should virtually never use all the groups.
	for group = 1, iNumGroups do
		if (iNumCellsInUse >= iNumCellTarget) then -- Map has reached desired island population.
			--print("Number of Island Groups produced:", group - 1);
			break
		end
		--[[ Formation Chart
		1. Single Cell, Axis Only
		2. Double Cell, Horizontal, Axis Only
		3. Single Cell With Dots
		4. Double Cell, Horizontal, With Shifted Dots
		5. Double Cell, Vertical, Axis Only
		6. Double Cell, Vertical, With Shifted Dots
		7. Triple Cell, Vertical, With Double Dots
		8. Square of Cells 2x2 With Double Dots
		9. Rectangle 3x2 With Double Dots
		10. Rectangle 2x3 With Double Dots ]]--
		--
		-- Choose a formation
		local rate_threshold = {};
		local total_appearance_rate, iNumFormations = 0, 0;
		local appearance_rates = { -- These numbers are relative to one another. No specific target total is necessary.
			7, -- #1
			3, -- #2
			15, --#3
			8, -- #4
			3, -- #5
			6, -- #6
			4, -- #7
			6, -- #8
			4, -- #9
			3, -- #10
		};
		for i, rate in ipairs(appearance_rates) do
			total_appearance_rate = total_appearance_rate + rate;
			iNumFormations = iNumFormations + 1;
		end
		local accumulated_rate = 0;
		for index = 1, iNumFormations do
			local threshold = (appearance_rates[index] + accumulated_rate) * 10000 / total_appearance_rate;
			table.insert(rate_threshold, threshold);
			accumulated_rate = accumulated_rate + appearance_rates[index];
		end
		local formation_type;
		local diceroll = TerrainBuilder.GetRandomNumber(10000, "Choose formation type - Island Making - Lua");
		for index, threshold in ipairs(rate_threshold) do
			if diceroll <= threshold then -- Choose this formation type.
				formation_type = index;
				break
			end
		end
		-- Choose cell(s) not in use;
		local iNumAttemptsToFindOpenCells = 0;
		local found_unoccupied_cell = false;
		local anchor_cell, cell_x, cell_y, foo;
		while found_unoccupied_cell == false do
			if iNumAttemptsToFindOpenCells > 100 then -- Too many attempts on this pass. Might not be any valid locations for this formation.
				--print("-"); print("*-* ALERT! Formation type of", formation_type, "for island group#", group, "unable to find an open space. Switching to single-cell.");
				formation_type = 3; -- Reset formation type.
				iNumAttemptsToFindOpenCells = 0;
			end
			local diceroll = 1 + TerrainBuilder.GetRandomNumber(iNumTotalCells, "Choosing a cell for an island group");
			if cell_data[diceroll] == false then -- Anchor cell is unoccupied.
				-- If formation type is multiple-cell, all secondary cells must also be unoccupied.
				if formation_type == 1 or formation_type == 3 then -- single cell, proceed.
					anchor_cell = diceroll;
					found_unoccupied_cell = true;
				elseif formation_type == 2 or formation_type == 4 then -- double cell, horizontal.
					-- Check to see if anchor cell is in the final column. If so, reject.
					cell_x = math.fmod(diceroll, iNumCellColumns);
					if cell_x ~= 0 then -- Anchor cell is valid, but still have to check adjacent cell.
						if cell_data[diceroll + 1] == false then -- Adjacent cell is unoccupied.
							anchor_cell = diceroll;
							found_unoccupied_cell = true;
						end
					end
				elseif formation_type == 5 or formation_type == 6 then -- double cell, vertical.
					-- Check to see if anchor cell is in the final row. If so, reject.
					cell_y, foo = math.modf(diceroll / iNumCellColumns);
					cell_y = cell_y + 1;
					if cell_y < iNumCellRows then -- Anchor cell is valid, but still have to check cell above it.
						if cell_data[diceroll + iNumCellColumns] == false then -- Adjacent cell is unoccupied.
							anchor_cell = diceroll;
							found_unoccupied_cell = true;
						end
					end
				elseif formation_type == 7 then -- triple cell, vertical.
					-- Check to see if anchor cell is in the northern two rows. If so, reject.
					cell_y, foo = math.modf(diceroll / iNumCellColumns);
					cell_y = cell_y + 1;
					if cell_y < iNumCellRows - 1 then -- Anchor cell is valid, but still have to check cells above it.
						if cell_data[diceroll + iNumCellColumns] == false then -- Cell directly above is unoccupied.
							if cell_data[diceroll + (iNumCellColumns * 2)] == false then -- Cell two rows above is unoccupied.
								anchor_cell = diceroll;
								found_unoccupied_cell = true;
							end
						end
					end
				elseif formation_type == 8 then -- square, 2x2.
					-- Check to see if anchor cell is in the final row or column. If so, reject.
					cell_x = math.fmod(diceroll, iNumCellColumns);
					if cell_x ~= 0 then
						cell_y, foo = math.modf(diceroll / iNumCellColumns);
						cell_y = cell_y + 1;
						if cell_y < iNumCellRows then -- Anchor cell is valid. Still have to check the other three cells.
							if cell_data[diceroll + iNumCellColumns] == false then
								if cell_data[diceroll + 1] == false then
									if cell_data[diceroll + iNumCellColumns + 1] == false then -- All cells are open.
										anchor_cell = diceroll;
										found_unoccupied_cell = true;
									end
								end
							end
						end
					end
				elseif formation_type == 9 then -- horizontal, 3x2.
					-- Check to see if anchor cell is too near to an edge. If so, reject.
					cell_x = math.fmod(diceroll, iNumCellColumns);
					if cell_x ~= 0 and cell_x ~= iNumCellColumns - 1 then
						cell_y, foo = math.modf(diceroll / iNumCellColumns);
						cell_y = cell_y + 1;
						if cell_y < iNumCellRows then -- Anchor cell is valid. Still have to check the other cells.
							if cell_data[diceroll + iNumCellColumns] == false then
								if cell_data[diceroll + 1] == false and cell_data[diceroll + 2] == false then
									if cell_data[diceroll + iNumCellColumns + 1] == false then
										if cell_data[diceroll + iNumCellColumns + 2] == false then -- All cells are open.
											anchor_cell = diceroll;
											found_unoccupied_cell = true;
										end
									end
								end
							end
						end
					end
				elseif formation_type == 10 then -- vertical, 2x3.
					-- Check to see if anchor cell is too near to an edge. If so, reject.
					cell_x = math.fmod(diceroll, iNumCellColumns);
					if cell_x ~= 0 then
						cell_y, foo = math.modf(diceroll / iNumCellColumns);
						cell_y = cell_y + 1;
						if cell_y < iNumCellRows - 1 then -- Anchor cell is valid. Still have to check the other cells.
							if cell_data[diceroll + iNumCellColumns] == false then
								if cell_data[diceroll + 1] == false then
									if cell_data[diceroll + iNumCellColumns + 1] == false then
										if cell_data[diceroll + (iNumCellColumns * 2)] == false then
											if cell_data[diceroll + (iNumCellColumns * 2) + 1] == false then -- All cells are open.
												anchor_cell = diceroll;
												found_unoccupied_cell = true;
											end
										end
									end
								end
							end
						end
					end
				end
			end
			iNumAttemptsToFindOpenCells = iNumAttemptsToFindOpenCells + 1;
		end
		-- Find Cell X and Y
		cell_x = math.fmod(anchor_cell, iNumCellColumns);
		if cell_x == 0 then
			cell_x = iNumCellColumns;
		end
		cell_y, foo = math.modf(anchor_cell / iNumCellColumns);
		cell_y = cell_y + 1;
	
		-- Create this island group.
		local iWidth, iHeight, fTilt; -- Scope the variables needed for island group creation.
		local plot_data = {};
		local x_shift, y_shift;
		if formation_type == 1 then -- single cell
			local x_shrinkage = TerrainBuilder.GetRandomNumber(4, "Cell Width adjustment - Lua");
			if x_shrinkage > 2 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Height adjustment - Lua");
			if y_shrinkage > 2 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth - x_shrinkage;
			iHeight = iCellHeight - y_shrinkage;
			fTilt = TerrainBuilder.GetRandomNumber(181, "Angle for island chain axis - LUA");
			plot_data = CreateSingleAxisIslandChain(iWidth, iHeight, fTilt)

		elseif formation_type == 2 then -- two cells, horizontal
			local x_shrinkage = TerrainBuilder.GetRandomNumber(8, "Cell Width adjustment - Lua");
			if x_shrinkage > 5 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Height adjustment - Lua");
			if y_shrinkage > 2 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth * 2 - x_shrinkage;
			iHeight = iCellHeight - y_shrinkage;
			-- Limit angles to mostly horizontal ones.
			fTilt = 145 + TerrainBuilder.GetRandomNumber(90, "Angle for island chain axis - LUA");
			if fTilt > 180 then
				fTilt = fTilt - 180;
			end
			plot_data = CreateSingleAxisIslandChain(iWidth, iHeight, fTilt)
		
		elseif formation_type == 3 then -- single cell, with dots
			local x_shrinkage = TerrainBuilder.GetRandomNumber(4, "Cell Width adjustment - Lua");
			if x_shrinkage > 2 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Height adjustment - Lua");
			if y_shrinkage > 2 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth - x_shrinkage;
			iHeight = iCellHeight - y_shrinkage;
			fTilt = TerrainBuilder.GetRandomNumber(181, "Angle for island chain axis - LUA");
			-- Determine "dot box"
			local iInnerWidth, iInnerHeight = iWidth - 2, iHeight - 2;
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(8, "Diceroll - Lua");
			local iNumDots = 4;
			if die_1 + die_2 > 1 then
				iNumDots = iNumDots + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots = iNumDots + die_1 + die_2;
			end
			plot_data = CreateAxisChainWithDots(iWidth, iHeight, fTilt, iInnerWidth, iInnerHeight, iNumDots)

		elseif formation_type == 4 then -- two cells, horizontal, with dots
			local x_shrinkage = TerrainBuilder.GetRandomNumber(8, "Cell Width adjustment - Lua");
			if x_shrinkage > 5 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Height adjustment - Lua");
			if y_shrinkage > 2 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth * 2 - x_shrinkage;
			iHeight = iCellHeight - y_shrinkage;
			-- Limit angles to mostly horizontal ones.
			fTilt = 145 + TerrainBuilder.GetRandomNumber(90, "Angle for island chain axis - LUA");
			if fTilt > 180 then
				fTilt = fTilt - 180;
			end
			-- Determine "dot box"
			local iInnerWidth = math.floor(iWidth / 2);
			local iInnerHeight = iHeight - 2;
			local iInnerWest = 2 + TerrainBuilder.GetRandomNumber((iWidth - 1) - iInnerWidth, "Shift for sub island group - Lua");
			local iInnerSouth = 2;
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(10, "Diceroll - Lua");
			local iNumDots = 5;
			if die_1 + die_2 > 1 then
				iNumDots = iNumDots + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots = iNumDots + die_1 + die_2;
			end
			plot_data = CreateAxisChainWithShiftedDots(iWidth, iHeight, fTilt, iInnerWidth, iInnerHeight, iInnerWest, iInnerSouth, iNumDots)
		
		elseif formation_type == 5 then -- Double Cell, Vertical, Axis Only
			local x_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Width adjustment - Lua");
			if x_shrinkage > 2 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(7, "Cell Height adjustment - Lua");
			if y_shrinkage > 4 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth - x_shrinkage;
			iHeight = iCellHeight * 2 - y_shrinkage;
			-- Limit angles to mostly vertical ones.
			fTilt = 55 + TerrainBuilder.GetRandomNumber(71, "Angle for island chain axis - LUA");
			plot_data = CreateSingleAxisIslandChain(iWidth, iHeight, fTilt)
	
		elseif formation_type == 6 then -- Double Cell, Vertical With Dots
			local x_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Width adjustment - Lua");
			if x_shrinkage > 2 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(7, "Cell Height adjustment - Lua");
			if y_shrinkage > 4 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth - x_shrinkage;
			iHeight = iCellHeight * 2 - y_shrinkage;
			-- Limit angles to mostly vertical ones.
			fTilt = 55 + TerrainBuilder.GetRandomNumber(71, "Angle for island chain axis - LUA");
			-- Determine "dot box"
			local iInnerWidth = iWidth - 2;
			local iInnerHeight = math.floor(iHeight / 2);
			local iInnerWest = 2;
			local iInnerSouth = 2 + TerrainBuilder.GetRandomNumber((iHeight - 1) - iInnerHeight, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(10, "Diceroll - Lua");
			local iNumDots = 5;
			if die_1 + die_2 > 1 then
				iNumDots = iNumDots + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots = iNumDots + die_1 + die_2;
			end
			plot_data = CreateAxisChainWithShiftedDots(iWidth, iHeight, fTilt, iInnerWidth, iInnerHeight, iInnerWest, iInnerSouth, iNumDots)
	
		elseif formation_type == 7 then -- Triple Cell, Vertical With Double Dots
			local x_shrinkage = TerrainBuilder.GetRandomNumber(4, "Cell Width adjustment - Lua");
			if x_shrinkage > 1 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(9, "Cell Height adjustment - Lua");
			if y_shrinkage > 5 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth - x_shrinkage;
			iHeight = iCellHeight * 3 - y_shrinkage;
			-- Limit angles to steep ones.
			fTilt = 70 + TerrainBuilder.GetRandomNumber(41, "Angle for island chain axis - LUA");
			-- Handle Dots Group 1.
			local iInnerWidth1 = iWidth - 3;
			local iInnerHeight1 = iCellHeight - 1;
			local iInnerWest1 = 2 + TerrainBuilder.GetRandomNumber(2, "Shift for sub island group - Lua");
			local iInnerSouth1 = 2 + TerrainBuilder.GetRandomNumber(iCellHeight - 3, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(8, "Diceroll - Lua");
			local iNumDots1 = 4;
			if die_1 + die_2 > 1 then
				iNumDots1 = iNumDots1 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots1 = iNumDots1 + die_1 + die_2;
			end
			-- Handle Dots Group 2.
			local iInnerWidth2 = iWidth - 3;
			local iInnerHeight2 = iCellHeight - 1;
			local iInnerWest2 = 2 + TerrainBuilder.GetRandomNumber(2, "Shift for sub island group - Lua");
			local iInnerSouth2 = iCellHeight + 2 + TerrainBuilder.GetRandomNumber(iCellHeight - 3, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(8, "Diceroll - Lua");
			local iNumDots2 = 4;
			if die_1 + die_2 > 1 then
				iNumDots2 = iNumDots2 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots2 = iNumDots2 + die_1 + die_2;
			end
			plot_data = CreateAxisChainWithDoubleDots(iWidth, iHeight, fTilt, iInnerWidth1, iInnerHeight1, iInnerWest1, iInnerSouth1,
													  iNumDots1, iInnerWidth2, iInnerHeight2, iInnerWest2, iInnerSouth2, iNumDots2)
		elseif formation_type == 8 then -- Square Block 2x2 With Double Dots
			local x_shrinkage = TerrainBuilder.GetRandomNumber(6, "Cell Width adjustment - Lua");
			if x_shrinkage > 4 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Height adjustment - Lua");
			if y_shrinkage > 3 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth * 2 - x_shrinkage;
			iHeight = iCellHeight * 2 - y_shrinkage;
			-- Full range of angles
			fTilt = TerrainBuilder.GetRandomNumber(181, "Angle for island chain axis - LUA");
			-- Handle Dots Group 1.
			local iInnerWidth1 = iCellWidth - 2;
			local iInnerHeight1 = iCellHeight - 2;
			local iInnerWest1 = 3 + TerrainBuilder.GetRandomNumber(iCellWidth - 2, "Shift for sub island group - Lua");
			local iInnerSouth1 = 3 + TerrainBuilder.GetRandomNumber(iCellHeight - 2, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(6, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(10, "Diceroll - Lua");
			local iNumDots1 = 5;
			if die_1 + die_2 > 1 then
				iNumDots1 = iNumDots1 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots1 = iNumDots1 + die_1 + die_2;
			end
			-- Handle Dots Group 2.
			local iInnerWidth2 = iCellWidth - 2;
			local iInnerHeight2 = iCellHeight - 2;
			local iInnerWest2 = 3 + TerrainBuilder.GetRandomNumber(iCellWidth - 2, "Shift for sub island group - Lua");
			local iInnerSouth2 = 3 + TerrainBuilder.GetRandomNumber(iCellHeight - 2, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(8, "Diceroll - Lua");
			local iNumDots2 = 5;
			if die_1 + die_2 > 1 then
				iNumDots2 = iNumDots2 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots2 = iNumDots2 + die_1 + die_2;
			end
			plot_data = CreateAxisChainWithDoubleDots(iWidth, iHeight, fTilt, iInnerWidth1, iInnerHeight1, iInnerWest1, iInnerSouth1,
													  iNumDots1, iInnerWidth2, iInnerHeight2, iInnerWest2, iInnerSouth2, iNumDots2)

		elseif formation_type == 9 then -- Horizontal Block 3x2 With Double Dots
			local x_shrinkage = TerrainBuilder.GetRandomNumber(8, "Cell Width adjustment - Lua");
			if x_shrinkage > 5 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(5, "Cell Height adjustment - Lua");
			if y_shrinkage > 3 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth * 3 - x_shrinkage;
			iHeight = iCellHeight * 2 - y_shrinkage;
			-- Limit angles to mostly horizontal ones.
			fTilt = 145 + TerrainBuilder.GetRandomNumber(90, "Angle for island chain axis - LUA");
			if fTilt > 180 then
				fTilt = fTilt - 180;
			end
			-- Handle Dots Group 1.
			local iInnerWidth1 = iCellWidth;
			local iInnerHeight1 = iCellHeight - 2;
			local iInnerWest1 = 4 + TerrainBuilder.GetRandomNumber(iCellWidth + 2, "Shift for sub island group - Lua");
			local iInnerSouth1 = 3 + TerrainBuilder.GetRandomNumber(iCellHeight - 2, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(8, "Diceroll - Lua");
			local iNumDots1 = 9;
			if die_1 + die_2 > 1 then
				iNumDots1 = iNumDots1 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots1 = iNumDots1 + die_1 + die_2;
			end
			-- Handle Dots Group 2.
			local iInnerWidth2 = iCellWidth;
			local iInnerHeight2 = iCellHeight - 2;
			local iInnerWest2 = 4 + TerrainBuilder.GetRandomNumber(iCellWidth + 2, "Shift for sub island group - Lua");
			local iInnerSouth2 = 3 + TerrainBuilder.GetRandomNumber(iCellHeight - 2, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(5, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(7, "Diceroll - Lua");
			local iNumDots2 = 8;
			if die_1 + die_2 > 1 then
				iNumDots2 = iNumDots2 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots2 = iNumDots2 + die_1 + die_2;
			end
			plot_data = CreateAxisChainWithDoubleDots(iWidth, iHeight, fTilt, iInnerWidth1, iInnerHeight1, iInnerWest1, iInnerSouth1,
													  iNumDots1, iInnerWidth2, iInnerHeight2, iInnerWest2, iInnerSouth2, iNumDots2)

		elseif formation_type == 10 then -- Vertical Block 2x3 With Double Dots
			local x_shrinkage = TerrainBuilder.GetRandomNumber(6, "Cell Width adjustment - Lua");
			if x_shrinkage > 4 then
				x_shrinkage = 0;
			end
			local y_shrinkage = TerrainBuilder.GetRandomNumber(8, "Cell Height adjustment - Lua");
			if y_shrinkage > 5 then
				y_shrinkage = 0;
			end
			x_shift, y_shift = 0, 0;
			if x_shrinkage > 0 then
				x_shift = TerrainBuilder.GetRandomNumber(x_shrinkage, "Cell Width offset - Lua");
			end
			if y_shrinkage > 0 then
				y_shift = TerrainBuilder.GetRandomNumber(y_shrinkage, "Cell Height offset - Lua");
			end
			iWidth = iCellWidth * 2 - x_shrinkage;
			iHeight = iCellHeight * 3 - y_shrinkage;
			-- Mostly vertical
			fTilt = 55 + TerrainBuilder.GetRandomNumber(71, "Angle for island chain axis - LUA");
			-- Handle Dots Group 1.
			local iInnerWidth1 = iCellWidth - 2;
			local iInnerHeight1 = iCellHeight;
			local iInnerWest1 = 3 + TerrainBuilder.GetRandomNumber(iCellWidth - 2, "Shift for sub island group - Lua");
			local iInnerSouth1 = 4 + TerrainBuilder.GetRandomNumber(iCellHeight + 2, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(10, "Diceroll - Lua");
			local iNumDots1 = 8;
			if die_1 + die_2 > 1 then
				iNumDots1 = iNumDots1 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots1 = iNumDots1 + die_1 + die_2;
			end
			-- Handle Dots Group 2.
			local iInnerWidth2 = iCellWidth - 2;
			local iInnerHeight2 = iCellHeight;
			local iInnerWest2 = 3 + TerrainBuilder.GetRandomNumber(iCellWidth - 2, "Shift for sub island group - Lua");
			local iInnerSouth2 = 4 + TerrainBuilder.GetRandomNumber(iCellHeight + 2, "Shift for sub island group - Lua");
			-- Determine number of dots
			local die_1 = TerrainBuilder.GetRandomNumber(4, "Diceroll - Lua");
			local die_2 = TerrainBuilder.GetRandomNumber(8, "Diceroll - Lua");
			local iNumDots2 = 7;
			if die_1 + die_2 > 1 then
				iNumDots2 = iNumDots2 + TerrainBuilder.GetRandomNumber(die_1 + die_2, "Number of dots to add to island chain - Lua");
			else
				iNumDots2 = iNumDots2 + die_1 + die_2;
			end
			plot_data = CreateAxisChainWithDoubleDots(iWidth, iHeight, fTilt, iInnerWidth1, iInnerHeight1, iInnerWest1, iInnerSouth1,
													  iNumDots1, iInnerWidth2, iInnerHeight2, iInnerWest2, iInnerSouth2, iNumDots2)
		end

		-- Obtain land plots from the plot data
		local x_adjust = xMargin + (cell_x - 1) * iCellWidth  + x_shift;
		local y_adjust = yMargin + (cell_y - 1) * iCellHeight + y_shift;
		for y = 1, iHeight do
			for x = 1, iWidth do
				local data_index = (y - 1) * iWidth + x;
				if plot_data[data_index] == true then -- This plot is land.
					local real_x = x + x_adjust - 1;
					local real_y = y + y_adjust - 1;
					local plot_index = real_y * g_iW + real_x + 1;
					island_chain_PlotTypes[plot_index] = g_PLOT_TYPE_LAND;
				end
			end
		end
	
		-- Record cells in use
		if formation_type == 1 then -- single cell
			cell_data[anchor_cell] = true;
			iNumCellsInUse = iNumCellsInUse + 1;
		elseif formation_type == 2 then
			cell_data[anchor_cell], cell_data[anchor_cell + 1] = true, true;
			iNumCellsInUse = iNumCellsInUse + 2;
		elseif formation_type == 3 then
			cell_data[anchor_cell] = true;
			iNumCellsInUse = iNumCellsInUse + 1;
		elseif formation_type == 4 then
			cell_data[anchor_cell], cell_data[anchor_cell + 1] = true, true;
			iNumCellsInUse = iNumCellsInUse + 2;
		elseif formation_type == 5 then
			cell_data[anchor_cell], cell_data[anchor_cell + iNumCellColumns] = true, true;
			iNumCellsInUse = iNumCellsInUse + 2;
		elseif formation_type == 6 then
			cell_data[anchor_cell], cell_data[anchor_cell + iNumCellColumns] = true, true;
			iNumCellsInUse = iNumCellsInUse + 2;
		elseif formation_type == 7 then
			cell_data[anchor_cell], cell_data[anchor_cell + iNumCellColumns] = true, true;
			cell_data[anchor_cell + (iNumCellColumns * 2)] = true;
			iNumCellsInUse = iNumCellsInUse + 3;
		elseif formation_type == 8 then
			cell_data[anchor_cell], cell_data[anchor_cell + 1] = true, true;
			cell_data[anchor_cell + iNumCellColumns], cell_data[anchor_cell + iNumCellColumns + 1] = true, true;
			iNumCellsInUse = iNumCellsInUse + 4;
		elseif formation_type == 9 then
			cell_data[anchor_cell], cell_data[anchor_cell + 1] = true, true;
			cell_data[anchor_cell + iNumCellColumns], cell_data[anchor_cell + iNumCellColumns + 1] = true, true;
			cell_data[anchor_cell + 2], cell_data[anchor_cell + iNumCellColumns + 2] = true, true;
			iNumCellsInUse = iNumCellsInUse + 6;
		elseif formation_type == 10 then
			cell_data[anchor_cell], cell_data[anchor_cell + 1] = true, true;
			cell_data[anchor_cell + iNumCellColumns], cell_data[anchor_cell + iNumCellColumns + 1] = true, true;
			cell_data[anchor_cell + (iNumCellColumns * 2)], cell_data[anchor_cell + (iNumCellColumns * 2) + 1] = true, true;
			iNumCellsInUse = iNumCellsInUse + 6;
		end
	end
	
	return island_chain_PlotTypes;
end

function GenerateClusters()
	-- Generate clusters of islands
	--
	-- Return a map of generated island plot types.
	print("Generating clusters.");

	local landArgs = ScriptData.landInfo.extraArgs or {};
	local invertExtras = landArgs.invertExtras;
		
	local worldSizeArgs = ScriptData.worldSizeInfo.clusters or {};
	
	-- Target size for islands/lakes in a cluster.
	local targetSize = LAKE_MAXNUMTILES;
	
	if ((gotExtrasLandmass == true) and 
	    (UserInput.extras == ExtraIslesLakes.CLUSTERS)) then
		-- 
		local targetSizeOffset = worldSizeArgs.targetSizeOffset or 0;
		targetSize = targetSize + targetSizeOffset + 1;
	end
	
	print("Target island/lake size: ", targetSize);
	
	-- Generate a lakes fractal that doesn't have clusters yet.
	local plotTypes = GenerateLakesForClusters(invertExtras);
	
	-- Freshen the lakes.
	FreshenLakes(plotTypes, targetSize, false);
	
	-- We want islands, not lakes.
	-- But we started with lakes, so invert the lakes into islands.
	for y = 0, g_iH-1 do
		for x = 0, g_iW - 1 do
			local i = y * g_iW + x + 1; -- add one because Lua arrays start at 1
			if (plotTypes[i] == g_PLOT_TYPE_OCEAN) then
				plotTypes[i] = g_PLOT_TYPE_LAND;
			else
				plotTypes[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end
	
	return plotTypes;
end

function MergeChains(destPlotTypes, srcPlotTypes)
	-- Merge chains onto a map, allowing land and possibly adjacent water 
	-- from the source to overwrite plots from the destination.
	--
	-- Do this before adding mountains and hills.
	-- 
	-- args:
	--   destPlotTypes: destination for merged extras
	--   srcPlotTypes: extras to merge
	
	--print("Merging chains.");
	local sea_thresholds = {10, 20, 30};
	local adj_thresholds = {75, 80, 85, 90, 95, 100};
	
	local sea_level = UserInput.seaLevel;
	
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local iPlot = y * g_iW + x + 1;
				
			local srcPlotType = srcPlotTypes[iPlot];
			
			if (srcPlotType ~= g_PLOT_TYPE_OCEAN) then
				-- Something to merge.
				local dice = TerrainBuilder.GetRandomNumber(100,"Chains isle exclusion");
				if (dice < sea_thresholds[sea_level]) then
					-- Occasionally add land from island chains for inversion.
					destPlotTypes[iPlot] = srcPlotType;
				end
			else
				-- Check for adjacent land.
				local numAdjacentLands = 0;
				for i, direction in ipairs(HexDirections) do
					local adjacentPlot = Map.GetAdjacentPlot(x, y, direction);
					
					if (adjacentPlot ~= nil) then
						-- We have an adjacent plot.
						-- Let's see if it's unvisited.
						local aX, aY = adjacentPlot:GetX(), adjacentPlot:GetY();
						local iCouldVisit = aY * g_iW + aX + 1;
					
						if (srcPlotTypes[iCouldVisit] ~= g_PLOT_TYPE_OCEAN) then
							-- Adjacent to land.
							numAdjacentLands = numAdjacentLands + 1;
						end
					end
				end
				
				if (numAdjacentLands > 0) then
					-- This tile is adjacent to an island chain.
					local dice = TerrainBuilder.GetRandomNumber(100,"Chains expansion");		
					if (dice < adj_thresholds[numAdjacentLands]) then
						-- Pretend that coastal water is part of the island chain.
						-- This adds water to inverted chains.
						destPlotTypes[iPlot] = g_PLOT_TYPE_LAND;
					end
				end
			end
		end
	end
	
	--print("Done merging lake chains.");
end

function MergeExtras(mergeLand, destPlotTypes, srcPlotTypes)
	-- Merge extras onto a map, allowing all land from the source to 
	-- overwrite plots from the destination.
	--
	-- Do this before adding mountains and hills.
	-- 
	-- args:
	--   mergeLand: whether to merge as land (true) or as water (false)
	--   destPlotTypes: destination for merged extras
	--   srcPlotTypes: extras to merge
	
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local iPlot = y * g_iW + x + 1;
			local srcPlotType = srcPlotTypes[iPlot];
			
			if (srcPlotType ~= g_PLOT_TYPE_OCEAN) then
				-- Something to merge.
				if (mergeLand == true) then	
					-- Merge the extra land as-is.
					destPlotTypes[iPlot] = srcPlotType;
				else
					-- Turn the extra land into a body of water.
					destPlotTypes[iPlot] = g_PLOT_TYPE_OCEAN;
				end				
			end
		end
	end
	
	--print("Done merging extras.");
end

function MergeExtrasPreserveSaltLakes(mergeLand, destPlotTypes, srcPlotTypes)
	-- Add extras without freshening lakes.
	--
	-- args:
	--   mergeLand: whether to merge land (true) or water (false)
	--   destPlotTypes: destination for merged extras
	--   srcPlotTypes: extras to merge

	print("Merging extras, but preserving salty lakes.");
	
	if (mergeLand == false) then
		-- We're adding water, so lakes will stay salty.
		MergeExtras(mergeLand, destPlotTypes, srcPlotTypes);
		return;
	end
	
	local gotIceCaps = ScriptData.landInfo.gotIceCaps;
	if (gotIceCaps == nil) then
		gotIceCaps = false;
	end
	
	
	-- The max size of a lake that could become three fresh water lakes
	-- if we added an island.
	local waterAreaThresh = 3*LAKE_MAXNUMTILES + 1;


	-- Apply island data to the main plot data.
	local numTilesAdded = 0;

	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local i = y * g_iW + x + 1;
			if srcPlotTypes[i] ~= g_PLOT_TYPE_OCEAN then
				if destPlotTypes[i] == g_PLOT_TYPE_OCEAN then
					local waterArea = GetWaterArea(destPlotTypes, x, y, gotIceCaps);
				
					-- Only add islands to sufficiently-large lakes.
					if (waterArea > waterAreaThresh) then
						destPlotTypes[i] = srcPlotTypes[i];
						numTilesAdded = numTilesAdded+1;
						--print("Water area is " .. waterArea .. " at (" .. x .. "," .. y .. "); adding island.");
					else
						--print("Water area is " .. waterArea .. " at (" .. x .. "," .. y .. "); discarding island.");
					end
				end
			end
		end
	end	
	
	print("Added " .. numTilesAdded .. " land tile(s) for extras.");
end

function GenerateExtras(plotTypes)
	-- Generate extra islands or lakes, depending on landmass.
	--
	-- args:
	--   plotTypes: 
	--      apply extra-island data to this set of plot types; 
	--      ignored if nil
	--
	-- Return a map of extra islands for customized merging with landmass.
	
	print("Adding extras to the world.");
	
	-- Check whether we actually wanted extra *lakes* instead of
	-- extra islands.
	local args = ScriptData.landInfo.extraArgs or {};
	local invertExtras = args.invertExtras;
		
	-- Start with a blank map for extras.
	local extras_PlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);
		
	if (ExtrasGot.chains == true) then
		-- Add island chains to the extras map.
		MergeChains(extras_PlotTypes, GenerateChains());
		
		if (UserInput.landmass == Landmasses.EXTRAS) then
			-- Extras landmass: add more chains!
			MergeChains(extras_PlotTypes, GenerateChains());
		end
	end	
	
	if ((ExtrasGot.clusters == true) and 
	    (UserInput.landmass ~= Landmasses.LAKES)) then
		-- Add clusters to the extras map.
		-- The lakes landmass has its own logic for adding clusters.
		MergeExtras(true, extras_PlotTypes, GenerateClusters());
	end
	
	if (ExtrasGot.miniDonuts == true) then
		-- Add mini-donuts to the extras map.
		MergeExtras(true, extras_PlotTypes, GenerateMiniDonuts());
	end

	if (ExtrasGot.tectonic == true) then
		-- Add tectonic islands to the extras map.
		MergeExtras(true, extras_PlotTypes, GenerateTectonicIslands());
	end
		
	-- Finished generating extras (1st pass).

	print("Merging extras to the landmass.");	
	
	-- Merge extras onto the mainland.
	local MergeToMainland = args.MergeToMainland or MergeExtras;
	
		-- By default, use extras to generate islands (land).
	local mergeLand  = true;
	if invertExtras then
		-- Use extras to generate water instead of land.
		mergeLand  = false;
	end
	
	MergeToMainland(mergeLand, plotTypes, extras_PlotTypes);
		
	print("Done adding extras.");
end

function GenerateMtnsHillsAndExtras(mtnArgs, plotTypes)
	-- Add mountains, hills, and extras to the landmass.
	--
	-- Args:
	--   mtnArgs: args to pass to the mountain pattern
	--   plotTypes: add to this map of plot types.

	-- Generate mountain pattern.
	GenerateMountainsAndHills(mtnArgs, plotTypes);	
	
	-- Add extras, which depends on mountain pattern.
	GenerateExtras(plotTypes);
	
	-- Add mountains and hills after adding all land tiles.
	AddMountainsAndHillsToPlotTypes(mtnArgs, plotTypes);
end
------------------------------------------------------------------------------
-- Plot Type overrides and helper methods
-- Need to declare these before declaring land info.
------------------------------------------------------------------------------
GotLakes_Fractal = {}

function GotLakes_Fractal:AddLandAndWater(water_percent)
	-- Use the continents fractal to decide land vs water for each plot.
	--
	-- Returns the total number of land tiles on the map
	
	local iWaterThreshold = self.continentsFrac:GetHeight(water_percent);
		
	iNumTotalLandTiles = 0;
	for x = 0, g_iW - 1 do
		for y = 0, g_iH - 1 do
			local i = y * g_iW + x;
			local val = self.continentsFrac:GetHeight(x, y);
			if(val <= iWaterThreshold) then
				self.plotTypes[i] = g_PLOT_TYPE_OCEAN;
			else
				self.plotTypes[i] = g_PLOT_TYPE_LAND;
				iNumTotalLandTiles = iNumTotalLandTiles + 1;
			end
		end
	end

	return iNumTotalLandTiles;
end

function GotLakes_Fractal:AddMtsHillsExtras()
	-- Add mountains, hills, and extra islands to the plot-types map.
	
	-- Get landmass-specific args.
	local mtnArgs = ScriptData.landInfo.mountainArgs;
	
	GenerateMtnsHillsAndExtras(mtnArgs, self.plotTypes);
end

function GotLakes_Fractal:Continents_IceAge_init()
    -- Fractal init for ice age continents

    -- Ice Age: lower sea levels and use wide continents.
    local rift_dice   = TerrainBuilder.GetRandomNumber(3, "Rift Grain roll - Ice Age");
    if rift_dice == 0 then
        rift_dice = -1;
    end

    local center_dice = TerrainBuilder.GetRandomNumber(7, "Continental Grain roll - Ice Age");
    
    local args = 
    {
        continent_grain = 3,
        rift_grain      = rift_dice,
        has_center_rift = (center_dice < 3)
    };
    
    self.continentsFrac = nil;
    self:InitFractal(args);
end

function GotLakes_Fractal:Continents_Normal_init()
    -- Fractal init for continents normal
    -- Continents plot type generator may call this repeatedly.

    local grain_dice = TerrainBuilder.GetRandomNumber(7, "Continental Grain roll");
    
    if grain_dice < 4 then
        grain_dice = 2;
    else
        grain_dice = 1;
    end
    
    local rift_dice = TerrainBuilder.GetRandomNumber(3, "Rift Grain roll");
    if rift_dice < 1 then
        rift_dice = -1;
    end
    
    local args = 
    {
        continent_grain = grain_dice, 
        rift_grain = rift_dice
    };
    
    self.continentsFrac = nil;
    self:InitFractal(args);
end

function GotLakes_Fractal:Continents_GeneratePlotTypes(args)
	-- Continents landmass
  	local args = args or {};
    
    local water_percent = GetSeaLevel();
        
    local shift_plot_types = args.shift_plot_types;
	if (shift_plot_types == nil) then
		shift_plot_types = true;
	end

    local has_center_rift = args.has_center_rift;
	if (has_center_rift == nil) then
		has_center_rift = true;
	end
    
    -- Reject the continental fractal layer until the largest 
    -- landmass occupies this percent or less of the total land.
    local biggest_area_max_pct = 64;
    
    if (gotIceAgeFamily) then
        print("Applying plot type overrides (continents + ice age)");
    
        -- Change fractal initialization for ice age.
        self.continents_init =
            GotLakes_Fractal.Continents_IceAge_init;
            
        -- Remove the restriction on largest landmass size.
        biggest_area_max_pct = 100;
        
        -- Remove the center rift to allow for snaky continents.
        has_center_rift = false;
    else
        self.continents_init = 
            GotLakes_Fractal.Continents_Normal_init;
    end
    
    local biggest_area_thresh = biggest_area_max_pct * 0.01;
    
        
    -- Generate continental fractal layer and examine the largest landmass
    -- until we meet our critera.
	local done = false;
	local iAttempts = 0;
	local iWaterThreshold, biggest_area, iNumTotalLandTiles, iNumBiggestAreaTiles, iBiggestID;
	while done == false do
		local grain_dice = TerrainBuilder.GetRandomNumber(7, "Continental Grain roll - LUA Continents");
		if grain_dice < 4 then
			grain_dice = 2;
		else
			grain_dice = 1;
		end
		local rift_dice = TerrainBuilder.GetRandomNumber(3, "Rift Grain roll - LUA Continents");
		if rift_dice < 1 then
			rift_dice = -1;
		end
		
		self:InitFractal{continent_grain = grain_dice, rift_grain = rift_dice};
		iWaterThreshold = self.continentsFrac:GetHeight(water_percent);
		local iBuffer = math.floor(g_iH/13.0);

		iNumTotalLandTiles = 0;
		for x = 0, g_iW - 1 do
			for y = iBuffer, g_iH - iBuffer - 1 do
				local i = y * g_iW + x;
				local val = self.continentsFrac:GetHeight(x, y);
				local pPlot = Map.GetPlotByIndex(i);
				if(val >= iWaterThreshold) then
					self.plotTypes[i] = g_PLOT_TYPE_LAND;
					-- temporary setting so can calculate areas
					TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_DESERT);  
					iNumTotalLandTiles = iNumTotalLandTiles + 1;
				else
					self.plotTypes[i] = g_PLOT_TYPE_OCEAN;
					-- temporary setting so can calculate areas
					TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_OCEAN);  
				end
			end
		end

		ShiftPlotTypes(self.plotTypes);
		self:GenerateCenterRift(self.plotTypes);

		AreaBuilder.Recalculate();
		local biggest_area = Areas.FindBiggestArea(false);
		iNumBiggestAreaTiles = biggest_area:GetPlotCount();
		
		-- Now test the biggest landmass to see if it is large enough.
		if iNumBiggestAreaTiles <= iNumTotalLandTiles * biggest_area_thresh then
			done = true;
			iBiggestID = biggest_area:GetID();
		end
		iAttempts = iAttempts + 1;
	end
	
	-- Add hills, mountains, and extras which are mostly landmass-agnostic.
	self:AddMtsHillsExtras();
	
	return self.plotTypes;
end

function GotLakes_Fractal:Fractal_GeneratePlotTypes(args)
	-- Fractal landmass
	local args = args or {};
	
    local shift_plot_types = args.shift_plot_types;
	if (shift_plot_types == nil) then
		shift_plot_types = true;
	end
	
	local has_center_rift  = args.has_center_rift or false;
	
 	self:AddLandAndWater(GetSeaLevel());

	-- Add hills, mountains, and extras which are mostly landmass-agnostic.
	self:AddMtsHillsExtras();
	
	if(shift_plot_types == true) then
		ShiftPlotTypes(self.plotTypes);
		-- Center Rift warrants plot shifting to guarantee centered landmasses.
		if(has_center_rift == true) then
			self:GenerateCenterRift()
		end
	end

	return self.plotTypes;
end

function GotLakes_Fractal:Pangea_GeneratePlotTypes(args)
	-- Pangea landmass
	local args = args or {};
	
	-- The following fractal-gen args are fixed for pangea:
	--    shift_plot_types (true)
	--    has_center_rift  (false)
	
	local water_percent = GetSeaLevel();
		
	if (ExtrasGot.chains == true) then
		-- Pangea with chains: slightly higher sea level
		water_percent = math.min(100, water_percent + 3);
	end

	-- Generate continental fractal layer and examine the largest landmass. Reject
	-- the result until the largest landmass occupies 84% or more of the total land.
	local done = false;
	local iAttempts = 0;
	local iWaterThreshold, biggest_area, iNumBiggestAreaTiles, iBiggestID;
	while done == false do
		local grain_dice = TerrainBuilder.GetRandomNumber(7, "Continental Grain roll - LUA Pangea");
		if grain_dice < 4 then
			grain_dice = 1;
		else
			grain_dice = 2;
		end
		local rift_dice = TerrainBuilder.GetRandomNumber(3, "Rift Grain roll - LUA Pangea");
		if rift_dice < 1 then
			rift_dice = -1;
		end
		
		self.continentsFrac = nil;
		self:InitFractal{continent_grain = grain_dice, rift_grain = rift_dice};
		iWaterThreshold = self.continentsFrac:GetHeight(water_percent);
		
		local iNumTotalLandTiles = 0;
		for x = 0, g_iW - 1 do
			for y = 0, g_iH - 1 do
				local i = y * g_iW + x;
				local val = self.continentsFrac:GetHeight(x, y);
				local pPlot = Map.GetPlotByIndex(i);
				if(val <= iWaterThreshold) then
					self.plotTypes[i] = g_PLOT_TYPE_OCEAN;
					-- temporary setting so can calculate areas
					TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_OCEAN);  
				else
					self.plotTypes[i] = g_PLOT_TYPE_LAND;
					-- temporary setting so can calculate areas
					TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_DESERT);  
					iNumTotalLandTiles = iNumTotalLandTiles + 1;
				end
			end
		end
		
		AreaBuilder.Recalculate();
		local biggest_area = Areas.FindBiggestArea(false);
		iNumBiggestAreaTiles = biggest_area:GetPlotCount();
		
		-- Now test the biggest landmass to see if it is large enough.
		if iNumBiggestAreaTiles >= iNumTotalLandTiles * 0.84 then
			done = true;
			iBiggestID = biggest_area:GetID();
		end
		iAttempts = iAttempts + 1;
	end
		
	-- Because we haven't yet shifted the plot types, we will not be able to take advantage 
	-- of having water and flatland plots already set. We still have to generate all data
	-- for hills and mountains, too, then shift everything, then set plots one more time.
	self:AddMtsHillsExtras();

	ShiftPlotTypes(self.plotTypes);
	
	-- Now shift everything toward one of the poles, 
	-- to reduce how much jungles tend to dominate this script.
	local shift_dice = TerrainBuilder.GetRandomNumber(2, "Shift direction - LUA Pangaea");
	if ((UserInput.worldWrap == WWRAP_GLOBE) and 
	    (UserInput.climate == Climates.TILTEDFULL)) then
		-- Always shift towards the sun for global tilted axis full.
		print("Attempting to shift pangea to the south (global tilted-axis full)");
		shift_dice = PANGEA_SHIFT_SOUTH;
	end
	
	local iStartRow, iNumRowsToShift;
	local bFoundPangaea, bDoShift = false, false;
	if shift_dice == PANGEA_SHIFT_NORTH then
		-- Shift North
		for y = g_iH - 2, 1, -1 do
			for x = 0, g_iW - 1 do
				local i = y * g_iW + x;
				if self.plotTypes[i] == g_PLOT_TYPE_HILLS or self.plotTypes[i] == g_PLOT_TYPE_LAND then
					local plot = Map.GetPlot(x, y);
					local iAreaID = plot:GetArea():GetID();
					if iAreaID == iBiggestID then
						bFoundPangaea = true;
						iStartRow = y + 1;
						if iStartRow < g_iH - 4 then -- Enough rows of water space to do a shift.
							bDoShift = true;
						end
						break
					end
				end
			end
			-- Check to see if we've found the Pangaea.
			if bFoundPangaea == true then
				break
			end
		end
	else
		-- Shift South
		for y = 1, g_iH - 2 do
			for x = 0, g_iW- 1 do
				local i = y * g_iW + x;
				if self.plotTypes[i] == g_PLOT_TYPE_HILLS or self.plotTypes[i] == g_PLOT_TYPE_LAND then
					local plot = Map.GetPlot(x, y);
					local iAreaID = plot:GetArea():GetID();
					if iAreaID == iBiggestID then
						bFoundPangaea = true;
						iStartRow = y - 1;
						if iStartRow > 3 then -- Enough rows of water space to do a shift.
							bDoShift = true;
						end
						break
					end
				end
			end
			-- Check to see if we've found the Pangaea.
			if bFoundPangaea == true then
				break
			end
		end
	end
	if bDoShift == true then
		if shift_dice == PANGEA_SHIFT_NORTH then -- Shift North
			local iRowsDifference = g_iH - iStartRow - 2;
			local iRowsInPlay = math.floor(iRowsDifference * 0.7);
			local iRowsBase = math.ceil(iRowsDifference * 0.3);
			local rows_dice = TerrainBuilder.GetRandomNumber(iRowsInPlay, "Number of Rows to Shift - LUA Pangaea");
			local iNumRows = math.min(iRowsDifference - 1, iRowsBase + rows_dice);
			local iNumEvenRows = 2 * math.floor(iNumRows / 2); -- MUST be an even number or we risk breaking a 1-tile isthmus and splitting the Pangaea.
			local iNumRowsToShift = math.max(2, iNumEvenRows);
			--print("-"); print("Shifting lands northward by this many plots: ", iNumRowsToShift); print("-");
			-- Process from top down.
			for y = (g_iH - 1) - iNumRowsToShift, 0, -1 do
				for x = 0, g_iW - 1 do
					local sourcePlotIndex = y * g_iW + x + 1;
					local destPlotIndex = (y + iNumRowsToShift) * g_iW + x + 1;
					self.plotTypes[destPlotIndex] = self.plotTypes[sourcePlotIndex]
				end
			end
			for y = 0, iNumRowsToShift - 1 do
				for x = 0, g_iW - 1 do
					local i = y * g_iW + x + 1;
					self.plotTypes[i] = g_PLOT_TYPE_OCEAN;
				end
			end
		else -- Shift South
			local iRowsDifference = iStartRow - 1;
			local iRowsInPlay = math.floor(iRowsDifference * 0.7);
			local iRowsBase = math.ceil(iRowsDifference * 0.3);
			local rows_dice = TerrainBuilder.GetRandomNumber(iRowsInPlay, "Number of Rows to Shift - LUA Pangaea");
			local iNumRows = math.min(iRowsDifference - 1, iRowsBase + rows_dice);
			local iNumEvenRows = 2 * math.floor(iNumRows / 2); -- MUST be an even number or we risk breaking a 1-tile isthmus and splitting the Pangaea.
			local iNumRowsToShift = math.max(2, iNumEvenRows);
			--print("-"); print("Shifting lands southward by this many plots: ", iNumRowsToShift); print("-");
			-- Process from bottom up.
			for y = 0, (g_iH - 1) - iNumRowsToShift do
				for x = 0, g_iW - 1 do
					local sourcePlotIndex = (y + iNumRowsToShift) * g_iW + x + 1;
					local destPlotIndex = y * g_iW + x + 1;
					self.plotTypes[destPlotIndex] = self.plotTypes[sourcePlotIndex]
				end
			end
			for y = g_iH - iNumRowsToShift, g_iH - 1 do
				for x = 0, g_iW - 1 do
					local i = y * g_iW + x + 1;
					self.plotTypes[i] = g_PLOT_TYPE_OCEAN;
				end
			end
		end
	end

	return self.plotTypes;
end
-------------------------------------------------------------------------------------------
FractalWorld = {};

function FractalWorld.Create()
	
	local data = {
		InitFractal = FractalWorld.InitFractal,
		GenerateCenterRift = FractalWorld.GenerateCenterRift,
		GeneratePlotTypes = FractalWorld.GeneratePlotTypes,
		
		plotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH)
	};
	
	return data;
end	

function FractalWorld:InitFractal(args)
	if(args == nil) then args = {}; end
	
	local continent_grain = args.continent_grain or 2;
	local rift_grain = args.rift_grain or -1; -- Default no rifts. Set grain to between 1 and 3 to add rifts. - Bob
	local invert_heights = args.invert_heights or false;
	local polar = args.polar or true;
	local ridge_flags = args.ridge_flags or g_iFlags;
	
	local fracFlags = {};
	
	if(invert_heights) then
		fracFlags.FRAC_INVERT_HEIGHTS = true;
	end
	
	if(polar) then
		fracFlags.FRAC_POLAR = true;
	end
	
	if(rift_grain > 0 and rift_grain < 4) then
		self.riftsFrac = Fractal.Create(g_iW, g_iH, rift_grain, {}, g_fracXExp, g_fracYExp);
		self.continentsFrac = Fractal.CreateRifts(g_iW, g_iH, continent_grain, fracFlags, self.riftsFrac, g_fracXExp, g_fracYExp);
	else
		self.continentsFrac = Fractal.Create(g_iW, g_iH, continent_grain, fracFlags, g_fracXExp, g_fracYExp);	
	end

	-- Use Brian's tectonics method to weave ridgelines in to the continental fractal.
	-- Without fractal variation, the tectonics come out too regular.
	--
	--[[ "The principle of the RidgeBuilder code is a modified Voronoi diagram. I 
	added some minor randomness and the slope might be a little tricky. It was 
	intended as a 'whole world' modifier to the fractal class. You can modify 
	the number of plates, but that is about it." ]]-- Brian Wade - May 23, 2009
	--
	local numPlates = ScriptData.worldSizeInfo.numRPlates; -- Ridge Plates
	-- Blend a bit of ridge into the fractal.
	-- This will do things like roughen the coastlines and build inland seas. - Brian
	self.continentsFrac:BuildRidges(numPlates, ridge_flags, 1, 2);
end

function FractalWorld:GenerateCenterRift()
	-- Causes a rift to break apart and separate any landmasses overlaying the map center.
	-- Rift runs south to north ala the Atlantic Ocean.
	-- Any land plots in the first or last map columns will be lost, overwritten.
	-- This rift function is hex-dependent. It would have to be adapted to work with squares tiles.
	-- Center rift not recommended for non-oceanic worlds or with continent grains higher than 2.
	-- 
	-- First determine the rift "lean". 0 = Starts west, leans east. 1 = Starts east, leans west.
	local riftLean = TerrainBuilder.GetRandomNumber(2, "FractalWorld Center Rift Lean - Lua");
	
	-- Set up tables recording the rift line and the edge plots to each side of the rift line.
	local riftLine = {};
	local westOfRift = {};
	local eastOfRift = {};
	-- Determine minimum and maximum length of line segments for each possible direction.
	local primaryMaxLength = math.max(1, math.floor(g_iH / 8));
	local secondaryMaxLength = math.max(1, math.floor(g_iH / 11));
	local tertiaryMaxLength = math.max(1, math.floor(g_iH / 14));
	
	-- Set rift line starting plot and direction.
	local startDistanceFromCenterColumn = math.floor(g_iH / 8);
	if riftLean == 0 then
		startDistanceFromCenterColumn = -(startDistanceFromCenterColumn);
	end
	local startX = math.floor(g_iW / 2) + startDistanceFromCenterColumn;
	local startY = 0;
	local startingDirection = DirectionTypes.DIRECTION_NORTHWEST;
	if riftLean == 0 then
		startingDirection = DirectionTypes.DIRECTION_NORTHEAST;
	end
	-- Set rift X boundary.
	local riftXBoundary = math.floor(g_iW / 2) - startDistanceFromCenterColumn;
	
	-- Rift line is defined by a series of line segments traveling in one of three directions.
	-- East-leaning lines move NE primarily, NW secondarily, and E tertiary.
	-- West-leaning lines move NW primarily, NE secondarily, and W tertiary.
	-- Any E or W segments cause a wider gap on that row, requiring independent storage of data regarding west or east of rift.
	--
	-- Key variables need to be defined here so they persist outside of the various loops that follow.
	-- This requires that the starting plot be processed outside of those loops.
	local currentDirection = startingDirection;
	local currentX = startX;
	local currentY = startY;
	table.insert(riftLine, {currentX, currentY});
	-- Record west and east of the rift for this row.
	local rowIndex = currentY + 1;
	westOfRift[rowIndex] = currentX - 1;
	eastOfRift[rowIndex] = currentX + 1;
	-- Set this rift plot as type Ocean.
	local plotIndex = currentX + 1; -- Lua arrays starting at 1 sure makes for a lot of extra work and chances for bugs.
	self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN; -- Tiles crossed by the rift all turn in to water.
	
	-- Generate the rift line.
	if riftLean == 0 then -- Leans east
		while currentY < g_iH - 1 do
			-- Generate a line segment
			local nextDirection = 0;

			if currentDirection == DirectionTypes.DIRECTION_EAST then
				local segmentLength = TerrainBuilder.GetRandomNumber(tertiaryMaxLength + 1, "FractalWorld Center Rift Segment Length - Lua");
				-- Choose next direction
				if currentX >= riftXBoundary then -- Gone as far east as allowed, must turn back west.
					nextDirection = DirectionTypes.DIRECTION_NORTHWEST;
				else
					local dice = TerrainBuilder.GetRandomNumber(3, "FractalWorld Center Rift Direction - Lua");
					if dice == 1 then
						nextDirection = DirectionTypes.DIRECTION_NORTHWEST;
					else
						nextDirection = DirectionTypes.DIRECTION_NORTHEAST;
					end
				end
				-- Process the line segment
				local plotsToDo = segmentLength;
				while plotsToDo > 0 do
					currentX = currentX + 1; -- Moving east, no change to Y.
					rowIndex = currentY;
					-- westOfRift[rowIndex] does not change.
					eastOfRift[rowIndex] = currentX + 1;
					plotIndex = currentY * g_iW + currentX + 1;
					self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
					plotsToDo = plotsToDo - 1;
				end

			elseif currentDirection == DirectionTypes.DIRECTION_NORTHWEST then
				local segmentLength = TerrainBuilder.GetRandomNumber(secondaryMaxLength + 1, "FractalWorld Center Rift Segment Length - Lua");
				-- Choose next direction
				if currentX >= riftXBoundary then -- Gone as far east as allowed, must turn back west.
					nextDirection = DirectionTypes.DIRECTION_NORTHWEST;
				else
					local dice = TerrainBuilder.GetRandomNumber(4, "FractalWorld Center Rift Direction - Lua");
					if dice == 2 then
						nextDirection = DirectionTypes.DIRECTION_EAST;
					else
						nextDirection = DirectionTypes.DIRECTION_NORTHEAST;
					end
				end
				-- Process the line segment
				local plotsToDo = segmentLength;
				while plotsToDo > 0 and currentY < g_iH - 1 do
					-- Identifying hex plots other than to east or west is tricky.
					-- The X coord could be one of two possibilities.
					-- Call on Map.GetAdjacentPlot to safely handle this task.
					local nextPlot = Map.GetAdjacentPlot(currentX, currentY, currentDirection);
					currentX = nextPlot:GetX();
					currentY = currentY + 1;
					rowIndex = currentY;
					westOfRift[rowIndex] = currentX - 1;
					eastOfRift[rowIndex] = currentX + 1;
					plotIndex = currentY * g_iW + currentX + 1;
					self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
					plotsToDo = plotsToDo - 1;
				end
				
			else -- NORTHEAST
				local segmentLength = TerrainBuilder.GetRandomNumber(primaryMaxLength + 1, "FractalWorld Center Rift Segment Length - Lua");
				-- Choose next direction
				if currentX >= riftXBoundary then -- Gone as far east as allowed, must turn back west.
					nextDirection = DirectionTypes.DIRECTION_NORTHWEST;
				else
					local dice = TerrainBuilder.GetRandomNumber(2, "FractalWorld Center Rift Direction - Lua");
					if dice == 1 and currentY > g_iH * 0.28 then
						nextDirection = DirectionTypes.DIRECTION_EAST;
					else
						nextDirection = DirectionTypes.DIRECTION_NORTHWEST;
					end
				end
				-- Process the line segment
				local plotsToDo = segmentLength;
				while plotsToDo > 0 and currentY < g_iH - 1 do
					-- Identifying hex plots other than to east or west is tricky.
					-- The X coord could be one of two possibilities.
					-- Call on Map.GetAdjacentPlot to safely handle this task.
					local nextPlot = Map.GetAdjacentPlot(currentX, currentY, currentDirection);
					currentX = nextPlot:GetX();
					currentY = currentY + 1;
					rowIndex = currentY;
					westOfRift[rowIndex] = currentX - 1;
					eastOfRift[rowIndex] = currentX + 1;
					plotIndex = currentY * g_iW + currentX + 1;
					self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
					plotsToDo = plotsToDo - 1;
				end
			end
			
			-- Line segment is done, set next direction.
			currentDirection = nextDirection;
		end

	else -- Leans west
		while currentY < g_iH - 1 do
			-- Generate a line segment
			local nextDirection = 0;

			if currentDirection == DirectionTypes.DIRECTION_WEST then
				local segmentLength = TerrainBuilder.GetRandomNumber(tertiaryMaxLength + 1, "FractalWorld Center Rift Segment Length - Lua");
				-- Choose next direction
				if currentX <= riftXBoundary then -- Gone as far west as allowed, must turn back east.
					nextDirection = DirectionTypes.DIRECTION_NORTHEAST;
				else
					local dice = TerrainBuilder.GetRandomNumber(3, "FractalWorld Center Rift Direction - Lua");
					if dice == 1 then
						nextDirection = DirectionTypes.DIRECTION_NORTHEAST;
					else
						nextDirection = DirectionTypes.DIRECTION_NORTHWEST;
					end
				end
				-- Process the line segment
				local plotsToDo = segmentLength;
				while plotsToDo > 0 do
					currentX = currentX - 1; -- Moving west, no change to Y.
					rowIndex = currentY;
					westOfRift[rowIndex] = currentX - 1;
					-- eastOfRift[rowIndex] does not change.
					plotIndex = currentY * g_iW + currentX + 1;
					self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
					plotsToDo = plotsToDo - 1;
				end

			elseif currentDirection == DirectionTypes.DIRECTION_NORTHEAST then
				local segmentLength = TerrainBuilder.GetRandomNumber(secondaryMaxLength + 1, "FractalWorld Center Rift Segment Length - Lua");
				-- Choose next direction
				if currentX <= riftXBoundary then -- Gone as far west as allowed, must turn back east.
					nextDirection = DirectionTypes.DIRECTION_NORTHEAST;
				else
					local dice = TerrainBuilder.GetRandomNumber(4, "FractalWorld Center Rift Direction - Lua");
					if dice == 2 then
						nextDirection = DirectionTypes.DIRECTION_WEST;
					else
						nextDirection = DirectionTypes.DIRECTION_NORTHWEST;
					end
				end
				-- Process the line segment
				local plotsToDo = segmentLength;
				while plotsToDo > 0 and currentY < g_iH - 1 do
					-- Identifying hex plots other than to east or west is tricky.
					-- The X coord could be one of two possibilities.
					-- Call on Map.GetAdjacentPlot to safely handle this task.
					local nextPlot = Map.GetAdjacentPlot(currentX, currentY, currentDirection);
					currentX = nextPlot:GetX();
					currentY = currentY + 1;
					rowIndex = currentY;
					westOfRift[rowIndex] = currentX - 1;
					eastOfRift[rowIndex] = currentX + 1;
					plotIndex = currentY * g_iW + currentX + 1;
					self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
					plotsToDo = plotsToDo - 1;
				end
				
			else -- NORTHWEST
				local segmentLength = TerrainBuilder.GetRandomNumber(primaryMaxLength + 1, "FractalWorld Center Rift Segment Length - Lua");
				-- Choose next direction
				if currentX <= riftXBoundary then -- Gone as far west as allowed, must turn back east.
					nextDirection = DirectionTypes.DIRECTION_NORTHEAST;
				else
					local dice = TerrainBuilder.GetRandomNumber(2, "FractalWorld Center Rift Direction - Lua");
					if dice == 1 and currentY > g_iH * 0.28 then
						nextDirection = DirectionTypes.DIRECTION_WEST;
					else
						nextDirection = DirectionTypes.DIRECTION_NORTHEAST;
					end
				end
				-- Process the line segment
				local plotsToDo = segmentLength;
				while plotsToDo > 0 and currentY < g_iH - 1 do
					-- Identifying hex plots other than to east or west is tricky.
					-- The X coord could be one of two possibilities.
					-- Call on Map.GetAdjacentPlot to safely handle this task.
					local nextPlot = Map.GetAdjacentPlot(currentX, currentY, currentDirection);
					currentX = nextPlot:GetX();
					currentY = currentY + 1;
					rowIndex = currentY;
					westOfRift[rowIndex] = currentX - 1;
					eastOfRift[rowIndex] = currentX + 1;
					plotIndex = currentY * g_iW + currentX + 1;
					self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
					plotsToDo = plotsToDo - 1;
				end
			end
			
			-- Line segment is done, set next direction.
			currentDirection = nextDirection;
		end
	end
	-- Process the final plot in the rift.
	westOfRift[g_iH] = currentX - 1;
	eastOfRift[g_iH] = currentX + 1;
	plotIndex = (g_iH - 1) * g_iW + currentX + 1;
	self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;

	-- Now force the rift to widen, causing land on either side of the rift to drift apart.
	local horizontalDrift = 3;
	local verticalDrift = 2;
	--
	if riftLean == 0 then
		-- Process Western side from top down.
		for y = g_iH - 1 - verticalDrift, 0, -1 do
			local thisRowX = westOfRift[y+1];
			for x = horizontalDrift, thisRowX do
				local sourcePlotIndex = y * g_iW + x + 1;
				local destPlotIndex = (y + verticalDrift) * g_iW + (x - horizontalDrift) + 1;
				self.plotTypes[destPlotIndex] = self.plotTypes[sourcePlotIndex]
			end
		end
		-- Process Eastern side from bottom up.
		for y = verticalDrift, g_iH - 1 do
			local thisRowX = eastOfRift[y+1];
			for x = thisRowX, g_iW - horizontalDrift - 1 do
				local sourcePlotIndex = y * g_iW + x + 1;
				local destPlotIndex = (y - verticalDrift) * g_iW + (x + horizontalDrift) + 1;
				self.plotTypes[destPlotIndex] = self.plotTypes[sourcePlotIndex]
			end
		end
		-- Clean up remainder of tiles (by turning them all to Ocean).
		-- Clean up bottom left.
		for y = 0, verticalDrift - 1 do
			local thisRowX = westOfRift[y+1];
			for x = 0, thisRowX do
				local plotIndex = y * g_iW + x + 1;
				self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
			end
		end
		-- Clean up top right.
		for y = g_iH - verticalDrift, g_iH - 1 do
			local thisRowX = eastOfRift[y+1];
			for x = thisRowX, g_iW - 1 do
				local plotIndex = y * g_iW + x + 1;
				self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
			end
		end
		-- Clean up the rift.
		for y = verticalDrift, g_iH - 1 - verticalDrift do
			local westX = westOfRift[y-verticalDrift+1] - horizontalDrift + 1;
			local eastX = eastOfRift[y+verticalDrift+1] + horizontalDrift - 1;
			for x = westX, eastX do
				local plotIndex = y * g_iW + x + 1;
				self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
			end
		end

	else -- riftLean = 1
		-- Process Western side from bottom up.
		for y = verticalDrift, g_iH - 1 do
			local thisRowX = westOfRift[y+1];
			for x = horizontalDrift, thisRowX do
				local sourcePlotIndex = y * g_iW + x + 1;
				local destPlotIndex = (y - verticalDrift) * g_iW + (x - horizontalDrift) + 1;
				self.plotTypes[destPlotIndex] = self.plotTypes[sourcePlotIndex]
			end
		end
		-- Process Eastern side from top down.
		for y = g_iH - 1 - verticalDrift, 0, -1 do
			local thisRowX = eastOfRift[y+1];
			for x = thisRowX, g_iW - horizontalDrift - 1 do
				local sourcePlotIndex = y * g_iW + x + 1;
				local destPlotIndex = (y + verticalDrift) * g_iW + (x + horizontalDrift) + 1;
				self.plotTypes[destPlotIndex] = self.plotTypes[sourcePlotIndex]
			end
		end
		-- Clean up remainder of tiles (by turning them all to Ocean).
		-- Clean up top left.
		for y = g_iH - verticalDrift, g_iH - 1 do
			local thisRowX = westOfRift[y+1];
			for x = 0, thisRowX do
				local plotIndex = y * g_iW + x + 1;
				self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
			end
		end
		-- Clean up bottom right.
		for y = 0, verticalDrift - 1 do
			local thisRowX = eastOfRift[y+1];
			for x = thisRowX, g_iW - 1 do
				local plotIndex = y * g_iW + x + 1;
				self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
			end
		end
		-- Clean up the rift.
		for y = verticalDrift, g_iH - 1 - verticalDrift do
			local westX = westOfRift[y+verticalDrift+1] - horizontalDrift + 1;
			local eastX = eastOfRift[y-verticalDrift+1] + horizontalDrift - 1;
			for x = westX, eastX do
				local plotIndex = y * g_iW + x + 1;
				self.plotTypes[plotIndex] = g_PLOT_TYPE_OCEAN;
			end
		end
	end
end
-------------------------------------------------------------------------------------------
MultilayeredFractal = {};

function MultilayeredFractal.Create()
	-- Set up flags for the fractal generator. These are no longer bit values.
	-- For Lua, we had to convert to a table of booleans to simulate bits.
	local iNoFlags = {};
	local iTerrainFlags = g_iFlags; -- Matches wrap to that of the map.
	iTerrainFlags.FRAC_POLAR = false;
	local iRoundFlags = {};
	local iHorzFlags = {
		FRAC_WRAP_X = true,
	};
	local iVertFlags = {
		FRAC_WRAP_Y = true,
	};
	local iXYFlags = {
		FRAC_WRAP_X = true,
		FRAC_WRAP_Y = true,
	};
	
	-- Current region, reinitiated in GeneratePlotsByRegion for each layer.
	local plotTypes = {};

	-- Sum of all layered regions
	local wholeworldPlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);

	-- Create data.
	local data = {
	
		-- member variables
		g_iW				= g_iW,
		g_iH				= g_iH,
		iTerrainFlags	= iTerrainFlags,
		iNoFlags		= iNoFlags,
		iRoundFlags		= iRoundFlags,
		iHorzFlags		= iHorzFlags,
		iVertFlags		= iVertFlags,
		iXYFlags		= iXYFlags,
		
		-- plot arrays
		plotTypes		= plotTypes,
		wholeworldPlotTypes = wholeworldPlotTypes,		
	}
	
	setmetatable(data, {__index = MultilayeredFractal});

	return data;
end

function MultilayeredFractal:ShiftRegionPlots(iRegionWidth, iRegionHeight)
	-- Minimizes land plots along the region's edges by shifting the coordinates.
	local shift_x = self:DetermineXShift(iRegionWidth, iRegionHeight);
	local shift_y = self:DetermineYShift(iRegionWidth, iRegionHeight);
	self:ShiftRegionPlotsBy(shift_x, shift_y, iRegionWidth, iRegionHeight);
end

function MultilayeredFractal:ShiftRegionPlotsBy(xshift, yshift, iRegionWidth, iRegionHeight)
	if (xshift > 0 or yshift > 0) then
		local iWH = iRegionWidth * iRegionHeight;
		local buf = {};
		for i = 1, iWH + 1, 1 do
			buf[i] = self.plotTypes[i];
		end

		for iDestY = 0, iRegionHeight, 1 do
			for iDestX = 0, iRegionWidth, 1 do
				local iDestI = iRegionWidth * iDestY + iDestX;
				local iSourceX = (iDestX + xshift) % iRegionWidth;
				local iSourceY = (iDestY + yshift) % iRegionHeight;
				local iSourceI = iRegionWidth * iSourceY + iSourceX;
				self.plotTypes[iDestI] = buf[iSourceI]
			end
		end
	end
end

function MultilayeredFractal:DetermineXShift(iRegionWidth, iRegionHeight)
	--[[ This function will align the most water-heavy vertical portion of the region with
	its vertical edge. This is a form of centering the landmasses, but it emphasizes the
	edge not the middle. If there are columns completely empty of land, these will tend to
	be chosen as the new edge. The operation looks at a group of columns not just a single
	column, then picks the center of the most water heavy group of columns to be the new
	vertical reguin edge. ]]--

	-- First loop through the map columns and record land plots in each column.
	local land_totals = {};
	for x = 0, iRegionWidth - 1 do
		local current_column = 0;
		for y = 0, iRegionHeight - 1 do
			local i = y * iRegionWidth + x + 1;
			if (self.plotTypes[i] ~= g_PLOT_TYPE_OCEAN) then
				current_column = current_column + 1;
			end
		end
		table.insert(land_totals, current_column);
	end
	
	-- Now evaluate column groups, each record applying to the center column of the group.
	local column_groups = {};
	-- Determine the group size in relation to map width.
	local group_radius = math.max(1, math.floor(iRegionWidth / 10));
	-- Measure the groups.
	for column_index = 1, iRegionWidth do
		local current_group_total = 0;
		for current_column = column_index - group_radius, column_index + group_radius do
			local current_index = current_column % iRegionWidth;
			if current_index == 0 then -- Modulo of the last column will be zero; this repairs the issue.
				current_index = iRegionWidth;
			end
			current_group_total = current_group_total + land_totals[current_index];
		end
		table.insert(column_groups, current_group_total);
	end
	
	-- Identify the group with the least amount of land in it.
	local best_value = iRegionHeight * (2 * group_radius + 1); -- Set initial value to max possible.
	local best_group = 1; -- Set initial best group as current map edge.
	for column_index, group_land_plots in ipairs(column_groups) do
		if group_land_plots < best_value then
			best_value = group_land_plots;
			best_group = column_index;
		end
	end
	
	-- Determine X Shift
	local x_shift = best_group - 1;
	return x_shift;
end

function MultilayeredFractal:DetermineYShift(iRegionWidth, iRegionHeight)
	-- Counterpart to DetermineXShift()

	-- First loop through the map rows and record land plots in each row.
	local land_totals = {};
	for y = 0, iRegionHeight - 1 do
		local current_row = 0;
		for x = 0, iRegionWidth - 1 do
			local i = y * iRegionWidth + x + 1;
			if (self.plotTypes[i] ~= g_PLOT_TYPE_OCEAN) then
				current_row = current_row + 1;
			end
		end
		table.insert(land_totals, current_row);
	end
	
	-- Now evaluate row groups, each record applying to the center row of the group.
	local row_groups = {};
	-- Determine the group size in relation to map height.
	local group_radius = math.max(1, math.floor(iRegionHeight / 10));
	-- Measure the groups.
	for row_index = 1, iRegionHeight do
		local current_group_total = 0;
		for current_row = row_index - group_radius, row_index + group_radius do
			local current_index = current_row % iRegionHeight;
			if current_index == 0 then -- Modulo of the last row will be zero; this repairs the issue.
				current_index = iRegionHeight;
			end
			current_group_total = current_group_total + land_totals[current_index];
		end
		table.insert(row_groups, current_group_total);
	end
	
	-- Identify the group with the least amount of land in it.
	local best_value = iRegionWidth * (2 * group_radius + 1); -- Set initial value to max possible.
	local best_group = 1; -- Set initial best group as current map edge.
	for row_index, group_land_plots in ipairs(row_groups) do
		if group_land_plots < best_value then
			best_value = group_land_plots;
			best_group = row_index;
		end
	end
	
	-- Determine Y Shift
	local y_shift = best_group - 1;
	return y_shift;
end

function MultilayeredFractal:GenerateFractalLayerWithoutHills(args)
	--[[ This function is intended to be paired with ApplyTectonics. If all the hills and
	mountains plots are going to be overwritten by the tectonics results, then why waste
	calculations generating them? ]]--
	local args = args or {};
	
	-- Handle args or assign defaults.
	local iWaterPercent = args.iWaterPercent or 55;
	local iRegionWidth = args.iRegionWidth; -- Mandatory Parameter, no default
	local iRegionHeight = args.iRegionHeight; -- Mandatory Parameter, no default
	local iRegionWestX = args.iRegionWestX; -- Mandatory Parameter, no default
	local iRegionSouthY = args.iRegionSouthY; -- Mandatory Parameter, no default
	local iRegionGrain = args.iRegionGrain or 1;
	local iRegionPlotFlags = args.iRegionPlotFlags or self.iRoundFlags;
	local iRegionTerrainFlags = self.iTerrainFlags; -- Removed from args list.
	local iRegionFracXExp = args.iRegionFracXExp or 6;
	local iRegionFracYExp = args.iRegionFracYExp or 5;
	local iRiftGrain = args.iRiftGrain or -1;
	local bShift = args.bShift or true;

	--print("Received Region Data");
	--print(iRegionWidth, iRegionHeight, iRegionWestX, iRegionSouthY, iRegionGrain);
	--print("- - -");
	
	-- Init the plot types array for this region's plot data. Redone for each new layer.
	-- Compare to self.wholeworldPlotTypes, which contains the sum of all layers.
	self.plotTypes = {};
	table.fill(self.plotTypes, g_PLOT_TYPE_OCEAN, iRegionWidth * iRegionHeight);
	
	--print("Filled regional table.");

	-- Init the land/water fractal
	local regionContinentsFrac;
	if (iRiftGrain > 0) and (iRiftGrain < 4) then
		local riftsFrac = Fractal.Create(iRegionWidth, iRegionHeight, iRiftGrain, {}, iRegionFracXExp, iRegionFracYExp);
		regionContinentsFrac = Fractal.CreateRifts(iRegionWidth, iRegionHeight, iRegionGrain, iRegionPlotFlags, riftsFrac, iRegionFracXExp, iRegionFracYExp);
	else
		regionContinentsFrac = Fractal.Create(iRegionWidth, iRegionHeight, iRegionGrain, iRegionPlotFlags, iRegionFracXExp, iRegionFracYExp);	
	end
	--print("Initialized main fractal");
	
	-- temp workaround, remove when the chunk above is activated.
	-- local regionContinentsFrac = Fractal.Create(iRegionWidth, iRegionHeight, iRegionGrain, iRegionPlotFlags, iRegionFracXExp, iRegionFracYExp);	
	
	-- Using the fractal matrices we just created, determine fractal-height values for sea level.
	local iWaterThreshold = regionContinentsFrac:GetHeight(iWaterPercent);

	-- Loop through the region's plots
	for x = 0, iRegionWidth - 1, 1 do
		for y = 0, iRegionHeight - 1, 1 do
			local i = y * iRegionWidth + x + 1; -- Lua arrays start at 1.
			local val = regionContinentsFrac:GetHeight(x,y);
			if val <= iWaterThreshold then
				--do nothing
			else
				self.plotTypes[i] = g_PLOT_TYPE_LAND
			end
		end
	end
	
	--print("Generated Plot Types");
	
	if bShift then -- Shift plots to obtain a more natural shape.
		self:ShiftRegionPlots(iRegionWidth, iRegionHeight);
	end

	--print("Shifted Plots");

	-- Once the plot types for the region have been generated, they must be
	-- applied to the global plot array.
	--
	-- Default approach is to ignore water and layer the lands over one another.
	-- Land of any type in each new layer overwrites whatever had been there.
	-- If you want to layer the water, too, or some other method, then
	-- you need to replace this function with a custom version, in your script.
	--
	-- Apply the region's plots to the global plot array.
	for x = 0, iRegionWidth - 1, 1 do
		local wholeworldX = x + iRegionWestX;
		for y = 0, iRegionHeight - 1, 1 do
			local i = y * iRegionWidth + x + 1
			
			-- Check for land to exclude both ocean and nil.
			if self.plotTypes[i] == g_PLOT_TYPE_LAND then
				local wholeworldY = y + iRegionSouthY;
				local iWorld = wholeworldY * self.g_iW + wholeworldX + 1
				self.wholeworldPlotTypes[iWorld] = self.plotTypes[i]
			end
		end
	end
	
	--print("Applied region's plots to global plot array.");

	-- This region is done.
	return
end

function MultilayeredFractal:GenerateWaterLayer(args)
	-- This function is intended to allow adding seas to specific areas of large continents.
	--
	-- Usage: Because of how the Shift operation works, we have to treat the "water" we are
	-- trying to place in this layer as "land", until the shift (if any) has occurred. Once
	-- the shift is done, the actual layering of the plots is handled by turning any "land"
	-- plots from this layer in to water plots when applied to the whole world plot table.
	-- This is all under the hood. The only thing is that the WaterPercent in the args table
	-- does NOT apply to the water you are trying to place, but rather the dead space where
	-- no changes will occur. Put another way, you should treat this layer of water as if
	-- it were an island or continent you are trying to add, except that in actual placement
	-- it will come out as water in the end, letting you place a sea inside a larger piece 
	-- of land, or a bay that eats in to a continent from the sea. -- Sirian
	local args = args or {};
	
	-- Handle args or assign defaults.
	local iWaterPercent = args.iWaterPercent or 55;
	local iRegionWidth = args.iRegionWidth; -- Mandatory Parameter, no default
	local iRegionHeight = args.iRegionHeight; -- Mandatory Parameter, no default
	local iRegionWestX = args.iRegionWestX; -- Mandatory Parameter, no default
	local iRegionSouthY = args.iRegionSouthY; -- Mandatory Parameter, no default
	local iRegionGrain = args.iRegionGrain or 1;
	local iRegionPlotFlags = args.iRegionPlotFlags or self.iRoundFlags;
	local iRegionFracXExp = args.iRegionFracXExp or 6;
	local iRegionFracYExp = args.iRegionFracYExp or 5;
	local iRiftGrain = args.iRiftGrain or -1;
	local bShift = args.bShift or true;

	-- Init the plot types array for this region's plot data. Redone for each new layer.
	-- Compare to self.wholeworldPlotTypes, which contains the sum of all layers.
	self.plotTypes = {};
	table.fill(self.plotTypes, g_PLOT_TYPE_OCEAN, iRegionWidth * iRegionHeight);
	
	-- Init the land/water fractal
	local regionContinentsFrac;
	if (iRiftGrain > 0) and (iRiftGrain < 4) then
		local riftsFrac = Fractal.Create(iRegionWidth, iRegionHeight, iRiftGrain, {}, iRegionFracXExp, iRegionFracYExp);
		regionContinentsFrac = Fractal.CreateRifts(iRegionWidth, iRegionHeight, iRegionGrain, iRegionPlotFlags, riftsFrac, iRegionFracXExp, iRegionFracYExp);
	else
		regionContinentsFrac = Fractal.Create(iRegionWidth, iRegionHeight, iRegionGrain, iRegionPlotFlags, iRegionFracXExp, iRegionFracYExp);	
	end
	
	-- Using the fractal matrices we just created, determine fractal-height values for sea level.
	local iWaterThreshold = regionContinentsFrac:GetHeight(iWaterPercent);

	-- Loop through the region's plots
	for x = 0, iRegionWidth - 1, 1 do
		for y = 0, iRegionHeight - 1, 1 do
			local i = y * iRegionWidth + x + 1; -- Lua arrays start at 1.
			local val = regionContinentsFrac:GetHeight(x,y);
			if val <= iWaterThreshold then
				--do nothing
			else
				self.plotTypes[i] = g_PLOT_TYPE_LAND
			end
		end
	end
	
	if bShift then -- Shift plots to obtain a more natural shape.
		self:ShiftRegionPlots(iRegionWidth, iRegionHeight);
	end

	-- Apply the region's plots to the global plot array.
	for x = 0, iRegionWidth - 1, 1 do
		local wholeworldX = x + iRegionWestX;
		for y = 0, iRegionHeight - 1, 1 do
			local i = y * iRegionWidth + x + 1
			
			-- Check for land to exclude both ocean and nil.
			if self.plotTypes[i] == g_PLOT_TYPE_LAND then
				local wholeworldY = y + iRegionSouthY;
				local iWorld = wholeworldY * self.g_iW + wholeworldX + 1
				self.wholeworldPlotTypes[iWorld] = g_PLOT_TYPE_OCEAN
			end
		end
	end
	
	-- This region is done.
	return
end

function MultilayeredFractal:GetMountainArgs()
	-- Return args for generating mountains and hills for MLF landmasses.

	-- Get landmass-specific args.
	return ScriptData.landInfo.mountainArgs;
end

function MultilayeredFractal:AddMtsHillsExtras()
	-- Add mountains, hills, and extra islands to the plot-types map.
	--
	-- This method replaces MLF:ApplyTectonics.
	
	-- Get landmass-specific args.
	local mtnArgs = self:GetMountainArgs();
	
	GenerateMtnsHillsAndExtras(mtnArgs, self.wholeworldPlotTypes);
end

function MultilayeredFractal:DoLandHeavyIceCaps()
	-- Add ice caps to a land-heavy map, if applicable.
	--
	-- No-op for any of the following conditions
	--    Regional world wrap
	--    Tilted Axis climate
	--    0 rows of polar ice
	--
	-- Return true iff we made ice caps.
	local polarIceNumRows = ScriptData.landInfo.polarIceNumRows;
	
	local makeIceCaps = 
		(UserInput.worldWrap == WWRAP_GLOBE) and 
		(gotTiltedFamily == false) and 
		(polarIceNumRows > 0);
	
	if (makeIceCaps == false) then
		-- We don't need ice caps.
		return false;
	end
	
	-- Ensure top two / bottom rows are ocean (to give 
	-- rivers a place to end and to create polar ice).
	for x = 0, g_iW - 1 do
		for y = 0, polarIceNumRows - 1 do
			self.wholeworldPlotTypes[y*g_iW + x + 1] = g_PLOT_TYPE_OCEAN;
			self.wholeworldPlotTypes[g_iW * (g_iH - y) - x] = g_PLOT_TYPE_OCEAN;
		end
	end
	
	-- Augment land info with knowledge of ice caps.
	ScriptData.landInfo.gotIceCaps = true;
	
	return true;
end

function MultilayeredFractal:Bagels_GeneratePlotsByRegion()
	-- Bagels	
	
	-- Fill all rows with water plots.
	self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);

	-- Args to pass into the per-bagel-slice helper function.
	-- Each bagel slice is basically a large mini-donut.
	local bagelArgs = {};
	
	local worldSizeArgs = ScriptData.worldSizeInfo.bagels or {};
	
	-- Get user inputs.
	local sea_level = UserInput.seaLevel;
	
	-- Bagel integrity.
	-- Controls the frequency of bays/lakes
	local baysHeight = 80;
	
	if (gotIceAgeFamily) then
		print("Lowering sea levels (ice age)");
		baysHeight    = baysHeight    + 8;
	end
	
	-- higher number => more land for bagels
	bagelArgs.baysHeight = baysHeight;
	print("Bays Height ", bagelArgs.baysHeight);
	
	
	-- Use a cell system to distribute donuts across the map.
	-- Divide the map into half a dozen cells by default.
	local numCellColumns = 3;
	local numCellRows    = 2;
	local numEmptyCells  = 2;
	
	if (UserInput.worldWrap == WWRAP_GLOBE) then
		-- Globe map--check for cropped climates.

		if (UserInput.climate == Climates.TILTEDCROP) then
			print("Using 3x1 bagel grid (globe + tilted axis cropped)");
			numCellRows   = 1;
			numEmptyCells = 1;
		   
		elseif (UserInput.climate == Climates.ICEAGECROP) then
			print("Using 4x2 bagel grid (globe + ice age cropped)");
			numCellColumns = 4;
			numEmptyCells  = 3;
		end
	end

	local numCells       = numCellColumns * numCellRows;
	
	-- Use floating numbers here; will round when calculating cell boundaries.
	local cellWidth  = g_iW / numCellColumns;
	local cellHeight = g_iH / numCellRows;
	local cellLength = math.min(cellWidth, cellHeight);
	print("Cell size (w,h): ",  CoordString(cellWidth, cellHeight));
	
	
	-- Min/Max donut radius based on cell size.
	-- A bagel radius of cellLength/2 would force the bagel to be at the
	-- center of the cell, so use something smaller for wiggle-room.
	local bagelSize = math.max(Round(cellLength*0.4), 1);
	local bagelOffsets = {1, 0, -1};
	bagelSize = bagelSize + bagelOffsets[sea_level];
	bagelArgs.rMin  = bagelSize;
	bagelArgs.rMax  = bagelSize;
	local     rMax  = bagelArgs.rMax;
	print("Radius: ", bagelSize);
	
	
	-- Min/Max donut thickness based on sea level.
	bagelArgs.tMin = Round(  rMax/2);
	bagelArgs.tMax = Round(3*rMax/4);
	print("Thickness range: ", CoordString(bagelArgs.tMin, bagelArgs.tMax));

	
	-- Choose random cells to be empty so that there is ocean somewhere.
	local cellEmpty = 0;
	local cellBagel = 1;
	local cellTypes = table.fill(cellBagel, numCells);
	for i = 1, numEmptyCells do
		cellTypes[i] = cellEmpty;
	end
	cellTypes = GetShuffledCopyOfTable(cellTypes);
	
	
	-- Add padding to cells to reduce overlap between bagels.
	-- Slightly less than rMax to make cell grid less obvious.
	local cellPaddingX = math.floor(math.clamp(bagelArgs.rMax-1, 0, cellWidth/3));
	local cellPaddingY = math.floor(math.clamp(bagelArgs.rMax-1, 0, cellHeight/3));
	print("Cell padding: ", CoordString(cellPaddingX, cellPaddingY));

		
	-- Boundaries for the current cell.
	local iWestX  = 0;
	local iEastX  = 0;
	local iSouthY = 0;
	local iNorthY = 0;
	local iCell   = 1;
	for iRow = 1, numCellRows do
		-- Advance to the next row.
		iEastX  = 0;
		iSouthY = iNorthY;
		iNorthY = math.floor(iRow * cellHeight);
	
		for iCol = 1, numCellColumns do
			-- Advance to the next column.
			iWestX = iEastX;
			iEastX = math.floor(iCol * cellWidth);
			
			local bagelX, bagelY;
			
			if (cellTypes[iCell] == cellBagel) then
				print("Current cell is " .. CoordString(iWestX, iSouthY) .. " to " .. CoordString(iEastX, iNorthY));
			
				bagelArgs.enableSlicing = false;
			
				-- Choose a random location for the bagel in the cell.
				bagelX = iWestX + cellPaddingX + 
					TerrainBuilder.GetRandomNumber(iEastX  - iWestX - 2*cellPaddingX,  "Bagel Cell X");
					
				bagelY = iSouthY + cellPaddingY + 
					TerrainBuilder.GetRandomNumber(iNorthY - iSouthY - 2*cellPaddingY, "Bagel Cell Y");

			else
				bagelArgs.enableSlicing = true;
			
				-- Empty cell: a bagel piece could be anywhere on the map.
				bagelX = cellPaddingX + TerrainBuilder.GetRandomNumber(g_iW - 2*cellPaddingX, "Bagel Cell X");
				bagelY = cellPaddingY + TerrainBuilder.GetRandomNumber(g_iH - 2*cellPaddingY, "Bagel Cell Y");
			end
			
			-- Generate the bagel slice using the mini-donut helper function.
			GenerateMiniDonut(self.wholeworldPlotTypes, bagelX, bagelY, bagelArgs);	
			
			-- Advance to the next cell.
			iCell = iCell + 1;
		end
	end
	

	-- Land and water are set. Now apply mountains, hills, and extra islands.
	self:AddMtsHillsExtras();

	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:Donut_GeneratePlotsByRegion()
	-- Donut
						
	-- Note: fractal wrap flags do not have to match world shape.
	-- Let wrap-x be true here even if the world is flat.
	local fracFlags = {FRAC_WRAP_X = true, FRAC_POLAR = true};

	-- Get user inputs.
	
	-- Special donut filling behavior for sea level:
	-- low    -> land/water mix a.k.a. "standard"
	-- medium -> ocean
	-- high   -> ocean (overrides all other changes, including extra islands)
	local sea_level = UserInput.seaLevel;
	
	-- Special donut filling behavior for world age:
	-- young  -> water becomes mountains
	local world_age = UserInput.worldAge;
	
	-- Special donut filling behavior for mountain pattern:
	-- clustered highlands -> land becomes hills
	local mtn_pattern = UserInput.mountains;
	
	-- Controls the initial height and width of the donut (before adding bays/lakes)
	--
	-- Don't make this too large (> 95) or else the the landmass will intersect
	-- itself and the poles.
	local axis_multiplier = 0.55;
	
	-- Tiles within this distance to the center are part of the donut's filling.
	local filling_distance = 0.25; -- Original donut seems to be 0.25
	
	-- Tiles between this and the filling distance become mountains
	-- when the world age is young.
	local mtn_filling_distance = 0.2;
	
	-- Controls the minimum distance of bays/lakes from the inner/outer edges 
	-- of the donut
	cohesion_diff = 0.06;
	
	-- Controls the frequency of bays/lakes
	local baysHeight = 74;
	
	if (gotIceAgeFamily) then
		print("Lowering sea levels (ice age)");
		baysHeight    = baysHeight    + 8;
	end
	
	local inner_cohesion_multiplier = filling_distance + 2*cohesion_diff;
	local outer_cohesion_multiplier = axis_multiplier - cohesion_diff;
				
	-- Fill all rows with water plots.
	self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);
	ScriptData.donutFillingMap = table.fill(FILLING_NO, g_iW * g_iH);

	
	local centerX = g_iW / 2;
	local centerY = g_iH / 2;   
	local donutWidth = math.max(g_iW, g_iH);
	
	-- Add the donut as land plots.
	local centerS = donutWidth/2;
	local axis = centerS * axis_multiplier;
	local axisSquared = axis * axis;
	for x = 0, g_iW - 1 do
		for y = 0, g_iH - 1 do
			local deltaX = x - centerX;
			local deltaY = y - centerY;
			local deltaXSquared = deltaX * deltaX;
			local deltaYSquared = deltaY * deltaY;
			local d = deltaXSquared/axisSquared + deltaYSquared/axisSquared;
			
			if ((filling_distance <= d) and (d <= 1)) then
				-- Donut land.
				local i = y * g_iW + x + 1;
				self.wholeworldPlotTypes[i] = g_PLOT_TYPE_LAND;
				
			elseif (d < filling_distance) then
				-- Donut filling.
				local i = y * g_iW + x + 1;
				ScriptData.donutFillingMap[i] = FILLING_YES;
				
				if (sea_level == SEALEVEL_LOW) then
					-- Low sea level: turn donut filling into land, but
					-- allow some land to become water.
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_LAND;
				end
			end
		end
	end
	
	-- Now add bays, fjords, inland seas, etc, but not inside the cohesion area.
	local baysGrain = 4;
	local baysFrac = Fractal.Create(g_iW, g_iH, baysGrain, fracFlags, -1, -1);
	local iBaysThreshold = baysFrac:GetHeight(baysHeight);
		
	local axis = centerS * outer_cohesion_multiplier;
	local axisSquared = axis * axis;
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local deltaX = x - centerX;
			local deltaY = y - centerY;
			local deltaXSquared = deltaX * deltaX;
			local deltaYSquared = deltaY * deltaY;
			local d = deltaXSquared/axisSquared + deltaYSquared/axisSquared;
			if d > 1 then
				local i = y * g_iW + x + 1;
				local baysVal = baysFrac:GetHeight(x, y);
				if baysVal >= iBaysThreshold then
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
				end
			end
		end
	end
	
	local axis = centerS * inner_cohesion_multiplier;
	local axisSquared = axis * axis;
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local deltaX = x - centerX;
			local deltaY = y - centerY;
			local deltaXSquared = deltaX * deltaX;
			local deltaYSquared = deltaY * deltaY;
			local d = deltaXSquared/axisSquared + deltaYSquared/axisSquared;
			if d <= 1 then
				local i = y * g_iW + x + 1;
				local baysVal = baysFrac:GetHeight(x, y);
				if baysVal >= iBaysThreshold then
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
					-- Count inner cohesion water as part of the donut filling.
					ScriptData.donutFillingMap[i] = FILLING_YES;
				end
			end
		end
	end
	
	-- Land and water are set. Now apply mountains, hills, and extra islands.
	self:AddMtsHillsExtras();
	
	print("Donut: adding donut filling.");
	
	-- Override the center of the donut with sweet, sweet filling logic.
	
	local landFilling  = g_PLOT_TYPE_LAND;
	local waterFilling = g_PLOT_TYPE_OCEAN;
	
	if (world_age == WORLDAGE_YOUNG) then
		-- Young world: turn water into mountains
		waterFilling = g_PLOT_TYPE_MOUNTAIN;
	end
	
	if (UserInput.mountains == MtnPatterns.CLUSTERED) then
		-- Clustered highlands: turn land into hills
		landFilling = g_PLOT_TYPE_HILLS;
	end
	
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local i = y * g_iW + x + 1;
			
			if (ScriptData.donutFillingMap[i] == FILLING_YES) then
				-- Got donut filling.
				--
				if (sea_level == SEALEVEL_HIGH) then
					-- High sea level: all ocean
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
				
				elseif (self.wholeworldPlotTypes[i] == g_PLOT_TYPE_OCEAN) then
					self.wholeworldPlotTypes[i] = waterFilling;
					
				elseif (self.wholeworldPlotTypes[i] == g_PLOT_TYPE_LAND) then
					self.wholeworldPlotTypes[i] = landFilling;
				end
				--]]
			end
		end
	end

	print("Donut: done!");
	
	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:Extras_GeneratePlotsByRegion()
	-- Extras, Inverted Extras
			
	local landArgs = ScriptData.landInfo.extraArgs or {};
	local invertExtras = landArgs.invertExtras;
	
	local bgPlotType = g_PLOT_TYPE_OCEAN;

	if invertExtras then
		-- Start with all land and add water for extras.
		bgPlotType = g_PLOT_TYPE_LAND;
	end
	
	if ((UserInput.landmass == Landmasses.EXTRAS) and 
	    ((UserInput.extras   == ExtraIslesLakes.TECTONIC) or 
		 (UserInput.extras   == ExtraIslesLakes.TCHAINS))) then
		-- Tectonic islands or tectonic islands and chains.
		-- Use special mountain args to amplify mountain ranges.
		ScriptData.landInfo.mountainArgs = 
		{
			adjustments     = {5,4,3},
			adjust_plates   = 2.5, -- 2.5
			adjacent_crags  = true,
			extra_mountains = 25, -- 15
		};
		
		-- Reduce the effects of sea level
		-- so that more mountains rise above water.
		ScriptData.landInfo.extraArgs.tectonicSeaMultiplier = 1.0; 
	end
	
	-- Fill all rows with background plots.
	self.wholeworldPlotTypes = table.fill(bgPlotType, g_iW * g_iH);
	
	if invertExtras then
		-- Add ice caps, if applicable, to 
		self:DoLandHeavyIceCaps();
	end
	
	-- Apply mountains, hills, and extra islands.
	self:AddMtsHillsExtras();

	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:InlandSea_GeneratePlotsByRegion()
	-- Inland Sea
		
	-- Use sea level input for lake size.
	-- Lake density is inversely proportional to sea level
	local lakes      = GetSeaLevel();
	local lake_grain = ScriptData.landInfo.lake_grain;
	
	
	if (UserInput.worldWrap == WWRAP_REGION) then
		-- Raise water levels to compensate for margins near map edges.
		lakes = lakes - 20;
	end
	
	if (ExtrasGot.anything == true) then
		-- Raise water levels to compensate for extra islands.
		lakes = lakes - 6;
	end
	
	
	if (UserInput.worldWrap == WWRAP_GLOBE) then
		-- Use a globe-specific lake fractal.
		self.wholeworldPlotTypes = GenerateLakesFractal(lakes, lake_grain);
		
		-- Add ice caps if applicable
		self:DoLandHeavyIceCaps();
		
	else
		-- Fill all rows with land plots.
		self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_LAND, g_iW * g_iH);
	
		-- Avoid wrapping lakes around the edges.
		local fracFlags = {FRAC_POLAR = true};
		
		-- Add margins so that edges are relatively dry.
		local marginX = math.max(1,math.floor(0.05*g_iW));
		local marginY = math.max(1,math.floor(0.05*g_iH));
		
		local iWestX = marginX;
		local iEastX = g_iW - marginX;
		local iWidth = iEastX - iWestX;
		
		local iSouthY = marginY;
		local iNorthY = g_iH - marginY;
		local iHeight = iNorthY - iSouthY;
		
		local lakesFrac = Fractal.Create(iWidth, iHeight, lake_grain, fracFlags, -1, -1)
		local lakesThreshold = lakesFrac:GetHeight(lakes);
		
		for y = iSouthY, iNorthY do
			for x = iWestX, iEastX do
				local val = lakesFrac:GetHeight(x, y);
				if val >= lakesThreshold then
					local i = y * g_iW + x + 1; -- add one because Lua arrays start at 1
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
				end
			end
		end
	
		-- Second, oval layer to increase the chance of one main body of water.
		local centerX = (g_iW / 2) - 1;
		local centerY = (g_iH / 2) - 1;
		local xAxis = centerX * 0.75;
		local yAxis = centerY * 0.5;
		local xAxisSquared = xAxis * xAxis;
		local yAxisSquared = yAxis * yAxis;
		for x = 0, g_iW - 1 do
			for y = 0, g_iH - 1 do
				local i = y * g_iW + x + 1;
				local deltaX = x - centerX;
				local deltaY = y - centerY;
				local deltaXSquared = deltaX * deltaX;
				local deltaYSquared = deltaY * deltaY;
				local oval_value = deltaXSquared / xAxisSquared + deltaYSquared / yAxisSquared;
				if oval_value <= 1 then
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
				end
			end
		end
	end

	-- Land and water are set.
	-- Now apply hills, mountains, etc.
	self:AddMtsHillsExtras();
	
	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:IslandPlates_GeneratePlotsByRegion()
	-- Island Plates

	-- Get user inputs.
	local fracFlags = {};
	local diffBase  = 4;
	local seaDiffs  = {-diffBase, 0, diffBase};
	local seaDiff   = seaDiffs[UserInput.seaLevel];
	
	if (gotIceAgeFamily) then
		print("Lowering sea levels (ice age)");
		seaDiff = seaDiff - diffBase;
	end

	-- Fill all rows with water plots.
	self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);
		
	-- Large islands layer.
	local args = {};
	args.iWaterPercent = seaDiff + 80; -- 65;
	args.iRegionWidth = math.ceil(g_iW);
	args.iRegionHeight = math.ceil(g_iH);
	args.iRegionWestX = math.floor(0);
	args.iRegionSouthY = math.floor(0);
	args.iRegionGrain = 3;
	args.iRegionHillsGrain = 4;
	args.iRegionPlotFlags = fracFlags;
	args.iRegionFracXExp = 7;
	args.iRegionFracYExp = 6;
	self:GenerateFractalLayerWithoutHills(args);
	
	-- Bays layer
	-- Do this before adding small islands so that we don't fracture
	-- small islands or extras into tiny islands.
	local args = {};
	args.iWaterPercent = 80 - seaDiff; -- seaDiff + 90
	args.iRegionWidth = math.ceil(g_iW);
	args.iRegionHeight = math.ceil(g_iH);
	args.iRegionWestX = math.floor(0);
	args.iRegionSouthY = math.floor(0);
	args.iRegionGrain = 4;
	args.iRegionHillsGrain = 4;
	args.iRegionPlotFlags = fracFlags;
	args.iRegionFracXExp = 7;
	args.iRegionFracYExp = 6;
	args.iRiftGrain = -1;
	self:GenerateWaterLayer(args);
	
	if (ExtrasGot.anything == false) then
		-- Small islands layer for when there are no extras.
		local args = {};
		args.iWaterPercent = seaDiff + 88; -- 95;
		args.iRegionWidth = math.ceil(g_iW);
		args.iRegionHeight = math.ceil(g_iH);
		args.iRegionWestX = math.floor(0);
		args.iRegionSouthY = math.floor(0);
		args.iRegionGrain = 4;
		args.iRegionHillsGrain = 4;
		args.iRegionPlotFlags = fracFlags;
		args.iRegionFracXExp = 7;
		args.iRegionFracYExp = 6;
		self:GenerateFractalLayerWithoutHills(args);
	end
	
	-- Land and water are set. Now apply mountains, hills, etc.
	self:AddMtsHillsExtras();

	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:Lakes_GeneratePlotsByRegion()
	-- Lakes landmass
	
	-- Note: Lakes does not have polar ice caps because ice caps
	-- are not compatible with lake freshening.
	
	
	-- Use sea level input for lake size.
	-- Lake density is inversely proportional to sea level
	local lakes      = GetSeaLevel();
	local lake_grain = ScriptData.landInfo.lake_grain;
	
	local useFractal = true;
	
	if (ExtrasGot.anything == true) then
		-- Special logic for extras.
		-- Extras provide enough water for us to skip the normal fractal.
		useFractal = false;
		
		if (UserInput.extras == ExtraIslesLakes.CLUSTERS) then
			-- Special logic for clusters.
			
			-- Island Clusters = inverted low-grain fresh lakes,
			-- so inverted clusters = low-grain fresh lakes.
			-- Just create the lakes here to avoid double-inversion.
			self.wholeworldPlotTypes = GenerateLakesForClusters(true);
			
		else
			-- Fill all rows with land plots.
			self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_LAND, g_iW * g_iH);
			
			-- Let GenerateExtras() add the lakes.
		end
	end
	
	
	if (useFractal == true) then
		-- Generate fractal lakes.
		if (UserInput.worldWrap == WWRAP_GLOBE) then
			-- Use the general-purpose lake fractal for global maps.
			self.wholeworldPlotTypes = GenerateLakesFractal(lakes, lake_grain);
		else
			-- Lakes + Regional world wrap
		
			-- Fill all rows with land plots.
			self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_LAND, g_iW * g_iH);
		
			-- Avoid wrapping lakes around the edges.
			local fracFlags = {FRAC_POLAR = true};
			
			-- Use a lakes fractal with no margins.
			local lakesFrac = Fractal.Create(g_iW, g_iH, lake_grain, fracFlags, -1, -1)
			local lakesThreshold = lakesFrac:GetHeight(lakes);
			
			for y = 0, g_iH - 1 do
				for x = 0, g_iW - 1 do
					local val = lakesFrac:GetHeight(x, y);
					if val >= lakesThreshold then
						local i = y * g_iW + x + 1; -- add one because Lua arrays start at 1
						self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
					end
				end
			end	
		end
	end
	
	-- Generate mountain pattern.
	GenerateMountainsAndHills(self:GetMountainArgs(), self.wholeworldPlotTypes);
	
	-- Add extra lakes if applicable.
	GenerateExtras(self.wholeworldPlotTypes);
	
	-- Freshen lakes
	FreshenLakes(self.wholeworldPlotTypes, LAKE_MAXNUMTILES, true);
	
	-- Land is set and lakes are fresh. Now apply hills and mountains.	
	AddMountainsAndHillsToPlotTypes(self:GetMountainArgs(), self.wholeworldPlotTypes);
	
	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:LargeIslands_GeneratePlotsByRegion()
	-- Large Islands.
	
	-- Get user inputs.
	local sea_level = UserInput.seaLevel;

	-- Get the number of civs based on map size.
	local iNumCivs = ScriptData.worldSizeInfo.numCivs or 8;
	
	-- Use Sea Level to change the target number of large islands.
	local island_multipliers = {2, 1.65, 1.3};
	local island_multiplier  = island_multipliers[sea_level]; -- used to be 1.65 always
	
	-- Use Sea Level to change the target number of occupied cells.
	--
	-- We don't need to lower this for high sea level since we'll be making less
	-- large islands anyway.
	--
	-- Raise the multiplier for low sea level so that we have a higher chance of 
	-- fitting more large islands into the map.
	local cell_multipliers   = {0.8, 0.7, 0.7};
	local cell_multiplier    = cell_multipliers[sea_level]; -- used to be 0.7 always
	
	if (gotIceAgeFamily) then
		print("Lowering water levels (ice age).");
		island_multiplier = island_multiplier + 0.35;
		cell_multiplier   = cell_multiplier + 0.1;
	end
	
	-- A cell system will be used to generate the large islands.
	-- Define the cell traits. (These need to fit correctly with the map grid width and height.)
	local iCellWidth = 5;
	local iCellHeight = 4;
	local iNumCellColumns = math.floor(g_iW / iCellWidth);
	local iNumCellRows = math.floor(g_iH / iCellHeight);
	local iNumTotalCells = iNumCellColumns * iNumCellRows;
	local cell_data = table.fill(false, iNumTotalCells) -- Stores data on map cells in use. All cells begin as empty.
	local iNumCellsInUse = 0;
	local iNumCellTarget = math.floor(iNumTotalCells * cell_multiplier);

	-- Add randomly generated island groups
	local iNumGroups = math.floor(iNumCivs * island_multiplier);
	print("-"); print("- Target number of Large Islands: ", iNumGroups); print("-");
	local cannot_find_valid_cell = false;
	for group = 1, iNumGroups do
		if iNumCellsInUse >= iNumCellTarget or cannot_find_valid_cell == true then -- Map has reeached desired island population.
			print("-"); print("** Cannot place all Large Islands. Actual number placed: ", group - 1); print("-");
			break
		end
		
		--print("Group ", group);
		--CheckForNilPlotTypes(self.wholeworldPlotTypes);
		
		--[[ Formation Chart
		1. Rectangle 4x2
		2. Rectangle 2x4
		3. Rectangle 3x3 ]]--
		--
		-- Choose a formation
		local rate_threshold = {};
		local total_appearance_rate, iNumFormations = 0, 0;
		local appearance_rates = { -- These numbers are relative to one another. No specific target total is necessary.
			5, -- #1
			5, -- #2
			6, -- #3
		};
		for i, rate in ipairs(appearance_rates) do
			total_appearance_rate = total_appearance_rate + rate;
			iNumFormations = iNumFormations + 1;
		end
		local accumulated_rate = 0;
		for index = 1, iNumFormations do
			local threshold = (appearance_rates[index] + accumulated_rate) * 10000 / total_appearance_rate;
			table.insert(rate_threshold, threshold);
			accumulated_rate = accumulated_rate + appearance_rates[index];
		end
		local formation_type;
		local diceroll = TerrainBuilder.GetRandomNumber(10000, "Choose formation type - Island Making - Lua");
		for index, threshold in ipairs(rate_threshold) do
			if diceroll <= threshold then -- Choose this formation type.
				formation_type = index;
				break
			end
		end
		-- Choose cell(s) not in use;
		local iNumAttemptsToFindOpenCells = 0;
		local found_unoccupied_cell = false;
		local anchor_cell, cell_x, cell_y, foo;
		while found_unoccupied_cell == false and cannot_find_valid_cell == false do
			if iNumAttemptsToFindOpenCells > 100 then -- Too many attempts on this pass. Might not be any valid locations.
				cannot_find_valid_cell = true;
			else
				local diceroll = 1 + TerrainBuilder.GetRandomNumber(iNumTotalCells, "Choosing a cell for an island group");
				if cell_data[diceroll] == false then -- Anchor cell is unoccupied.
					-- If formation type is multiple-cell, all secondary cells must also be unoccupied.
					if formation_type == 1 then -- horizontal, 4x2.
						-- Check to see if anchor cell is too near to an edge. If so, reject.
						cell_x = math.fmod(diceroll, iNumCellColumns);
						if cell_x ~= 0 and cell_x < iNumCellColumns - 2 then
							cell_y, foo = math.modf(diceroll / iNumCellColumns);
							cell_y = cell_y + 1;
							if cell_y < iNumCellRows then 
								-- Anchor cell is valid. Still have to check the other cells.
								if cell_data[diceroll + iNumCellColumns] == false then
									if cell_data[diceroll + 1] == false and cell_data[diceroll + 2] == false then
										if cell_data[diceroll + 3] == false then
											if cell_data[diceroll + iNumCellColumns + 1] == false then
												if cell_data[diceroll + iNumCellColumns + 2] == false then
													if cell_data[diceroll + iNumCellColumns + 3] == false then -- All cells are open.
														anchor_cell = diceroll;
														found_unoccupied_cell = true;
													end
												end
											end
										end
									end
								end
							end
						end
					elseif formation_type == 2 then -- vertical, 2x4.
						-- Check to see if anchor cell is too near to an edge. If so, reject.
						cell_x = math.fmod(diceroll, iNumCellColumns);
						if cell_x ~= 0 then
							cell_y, foo = math.modf(diceroll / iNumCellColumns);
							cell_y = cell_y + 1;
							if cell_y < iNumCellRows - 2 then
								-- Anchor cell is valid. Still have to check the other cells.
								if cell_data[diceroll + iNumCellColumns] == false then
									if cell_data[diceroll + 1] == false then
										if cell_data[diceroll + iNumCellColumns + 1] == false then
											if cell_data[diceroll + (iNumCellColumns * 2)] == false then
												if cell_data[diceroll + (iNumCellColumns * 2) + 1] == false then
													if cell_data[diceroll + (iNumCellColumns * 3)] == false then
														if cell_data[diceroll + (iNumCellColumns * 3) + 1] == false then -- All cells are open.
															anchor_cell = diceroll;
															found_unoccupied_cell = true;
														end
													end
												end
											end
										end
									end
								end
							end
						end
					elseif formation_type == 3 then -- square, 3x3.
						-- Check to see if anchor cell is too near to an edge. If so, reject.
						cell_x = math.fmod(diceroll, iNumCellColumns);
						if cell_x ~= 0 and cell_x < iNumCellColumns - 1 then
							cell_y, foo = math.modf(diceroll / iNumCellColumns);
							cell_y = cell_y + 1;
							if cell_y < iNumCellRows - 1 then
								-- Anchor cell is valid. Still have to check the other cells.
								if cell_data[diceroll + iNumCellColumns] == false then
									if cell_data[diceroll + 1] == false and cell_data[diceroll + 2] == false then
										if cell_data[diceroll + iNumCellColumns + 1] == false then
											if cell_data[diceroll + (iNumCellColumns * 2)] == false then
												if cell_data[diceroll + (iNumCellColumns * 2) + 1] == false then
													if cell_data[diceroll + iNumCellColumns + 2] == false then
														if cell_data[diceroll + (iNumCellColumns * 2) + 2] == false then -- All cells are open.
															anchor_cell = diceroll;
															found_unoccupied_cell = true;
														end
													end
												end
											end
										end
									end
								end
							end
						end
					end
				end
				iNumAttemptsToFindOpenCells = iNumAttemptsToFindOpenCells + 1;
			end
		end
	
		-- Execute remainder of process on this loop only if a valid cell was found.
		if cannot_find_valid_cell == false then

			-- Find Cell X and Y
			cell_x = math.fmod(anchor_cell, iNumCellColumns);
			if cell_x == 0 then
				cell_x = iNumCellColumns;
			end
			cell_y, foo = math.modf(anchor_cell / iNumCellColumns);
			cell_y = cell_y + 1;

			-- Debug
			--print("-"); print("-"); print("* Group# " .. group, "Formation Type: " .. formation_type, "Cell X, Y: " .. cell_x .. ", " .. cell_y);

			-- Create this large island.
			local iWidth, iHeight; -- Scope the variables needed for island group creation.
			local iWestX = iCellWidth * (cell_x - 1) + 1;
			local iSouthY = iCellHeight * (cell_y - 1) + 1;
			local iCohesionW, iCohesionH;
			local iCoWestX, iCoSouthY;
			local iWater, iCohesionWater = 55, 60;
			local FracXExp, FracYExp = 5, 5;
			if formation_type == 1 then -- four cells wide, two cells tall.
				iWidth = iCellWidth * 4 - 2;
				iHeight = iCellHeight * 2 - 2;
				FracXExp = 6;
				iCohesionW = math.floor(iWidth/2);
				iCohesionH = math.floor(iHeight/2);
				iCoWestX = iWestX + math.floor(iWidth/4);
				iCoSouthY = iSouthY + math.floor(iHeight/4);
			elseif formation_type == 2 then -- two cells wide, four cells tall.
				iWidth = iCellWidth * 2 - 2;
				iHeight = iCellHeight * 4 - 2;
				FracYExp = 6;
				iCohesionW = math.floor(iWidth/2);
				iCohesionH = math.floor(iHeight/2);
				iCoWestX = iWestX + math.floor(iWidth/4);
				iCoSouthY = iSouthY + math.floor(iHeight/4);
			elseif formation_type == 3 then -- three cells wide, three cells tall.
				iWidth = iCellWidth * 3 - 2;
				iHeight = iCellHeight * 3 - 2;
				iWater = 60;
				iCohesionW = math.floor(iWidth/2);
				iCohesionH = math.floor(iHeight/2);
				iCoWestX = iWestX + math.floor(iWidth/4);
				iCoSouthY = iSouthY + math.floor(iHeight/4);
			end

			-- Execute fractal layers for this Large Island.		
			local args = {};
			args.iWaterPercent = iWater;
			args.iRegionWidth = iWidth;
			args.iRegionHeight = iHeight;
			args.iRegionWestX = iWestX;
			args.iRegionSouthY = iSouthY;
			--args.iRegionGrain
			--args.iRegionHillsGrain
			args.iRegionPlotFlags = self.iRoundFlags;
			args.iRegionFracXExp = FracXExp;
			args.iRegionFracYExp = FracYExp;
			args.iRiftGrain = -1;
			--args.bShift

			self:GenerateFractalLayerWithoutHills(args)

			--print("After 1st fractal layer.");
			--CheckForNilPlotTypes(self.wholeworldPlotTypes);
			
			local args = {};
			args.iWaterPercent = iCohesionWater;
			args.iRegionWidth = iCohesionW;
			args.iRegionHeight = iCohesionH;
			args.iRegionWestX = iCoWestX;
			args.iRegionSouthY = iCoSouthY;
			--args.iRegionGrain
			--args.iRegionHillsGrain
			args.iRegionPlotFlags = self.iRoundFlags;
			args.iRegionFracXExp = FracXExp;
			args.iRegionFracYExp = FracYExp;
			args.iRiftGrain = -1;
			--args.bShift

			self:GenerateFractalLayerWithoutHills(args)
	
			-- Record cells in use
			if formation_type == 1 then
				cell_data[anchor_cell], cell_data[anchor_cell + 1] = true, true;
				cell_data[anchor_cell + iNumCellColumns], cell_data[anchor_cell + iNumCellColumns + 1] = true, true;
				cell_data[anchor_cell + 2], cell_data[anchor_cell + 3] = true, true;
				cell_data[anchor_cell + iNumCellColumns + 2], cell_data[anchor_cell + iNumCellColumns + 3] = true, true;
				iNumCellsInUse = iNumCellsInUse + 8;
			elseif formation_type == 2 then
				cell_data[anchor_cell], cell_data[anchor_cell + 1] = true, true;
				cell_data[anchor_cell + iNumCellColumns], cell_data[anchor_cell + iNumCellColumns + 1] = true, true;
				cell_data[anchor_cell + (iNumCellColumns * 2)], cell_data[anchor_cell + (iNumCellColumns * 2) + 1] = true, true;
				cell_data[anchor_cell + (iNumCellColumns * 3)], cell_data[anchor_cell + (iNumCellColumns * 3) + 1] = true, true;
				iNumCellsInUse = iNumCellsInUse + 8;
			elseif formation_type == 3 then
				cell_data[anchor_cell], cell_data[anchor_cell + 1], cell_data[anchor_cell + 2] = true, true, true;
				cell_data[anchor_cell + iNumCellColumns], cell_data[anchor_cell + iNumCellColumns + 1], cell_data[anchor_cell + iNumCellColumns + 2] = true, true, true;
				cell_data[anchor_cell + (iNumCellColumns * 2)], cell_data[anchor_cell + (iNumCellColumns * 2) + 1], cell_data[anchor_cell + (iNumCellColumns * 2) + 2] = true, true, true;
				iNumCellsInUse = iNumCellsInUse + 9;
			end
		end
	end

	--print("Before adding mountains, hills, extras.");
	--CheckForNilPlotTypes(self.wholeworldPlotTypes);
	
	-- Land and water are set. Now apply mountains, etc.
	self:AddMtsHillsExtras();

	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:LargeLakes_GeneratePlotsByRegion()
	-- Salt Lakes, Large Lakes
	
	-- Use sea level input for lake size.
	-- Lake density is inversely proportional to sea level
	local lakes      = GetSeaLevel();
	local lake_grain = ScriptData.landInfo.lake_grain;
	
		
	if (ExtrasGot.anything == true) then
		-- Lower water levels to compensate for extra islands.
		lakes = lakes - 6;
	end
	
	if (UserInput.worldWrap == WWRAP_GLOBE) then
		-- Use a globe-specific lake fractal.
		self.wholeworldPlotTypes = GenerateLakesFractal(lakes, lake_grain);
		
		-- Add ice caps if applicable
		self:DoLandHeavyIceCaps();
		
	else
		-- Fill all rows with land plots.
		self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_LAND, g_iW * g_iH);
	
		-- Avoid wrapping lakes around the edges.
		local fracFlags = {FRAC_POLAR = true};
		
		local marginX = 0;
		local marginY = 0;
		
		local iWestX = marginX;
		local iEastX = g_iW - marginX;
		local iWidth = iEastX - iWestX;
		
		local iSouthY = marginY;
		local iNorthY = g_iH - marginY;
		local iHeight = iNorthY - iSouthY;
		
		local lakesFrac = Fractal.Create(iWidth, iHeight, lake_grain, fracFlags, -1, -1)
		local lakesThreshold = lakesFrac:GetHeight(lakes);
		
		for y = iSouthY, iNorthY do
			for x = iWestX, iEastX do
				local val = lakesFrac:GetHeight(x, y);
				if val >= lakesThreshold then
					local i = y * g_iW + x + 1; -- add one because Lua arrays start at 1
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
				end
			end
		end
	end

	-- Land and water are set.
	-- Now apply hills, mountains, etc.
	self:AddMtsHillsExtras();
	
	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:Maze_GeneratePlotsByRegion()
	-- Maze

	-- Get user inputs.
	local sea_level = GetSeaLevel();

	-- Fill all rows with water plots.
	self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);
	
	-- Use a fractal to create isolated clumps of land.
	local landGrains = {5,4,3};
	local landGrain  = landGrains[UserInput.seaLevel];
	local landFlags  = {FRAC_WRAP_X = true, FRAC_POLAR = true};
	local landFrac   = Fractal.Create(g_iW, g_iH, landGrain, landFlags, -1, -1)
	local landThresh = landFrac:GetHeight(sea_level);
	
	-- Use a pangea-shaped fractal to decide where to allow isolated clumps.
	local panGrain  = 1;
	local panFlags  = {FRAC_WRAP_X = true, FRAC_POLAR = true};
	local panFrac   = Fractal.Create(g_iW, g_iH, panGrain, panFlags, -1, -1)
	local panThresh = panFrac:GetHeight(sea_level-10);
	
	for y = 0, g_iH do
		for x = 0, g_iW do
			local panVal = panFrac:GetHeight(x,y);
			if (panVal >= panThresh) then		
				local landVal = landFrac:GetHeight(x,y);
				if landVal >= landThresh then
					local i = y * g_iW + x + 1; -- add one because Lua arrays start at 1
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_LAND;
				end
			end
		end
	end
	
	-- Use PassMaker to connect the islands with land bridges.
	local bridgeArgs = 
	{
		obstaclesName 		= "water",
		plotTypes           = self.wholeworldPlotTypes,
		InitReachPlot 		= PassMaker.Water_InitReachPlot,
		RemoveObstacle      = PassMaker.Water_RemoveObstacle,
	};
	EnsureReachabilityWithPassMaker(bridgeArgs);
		
	--[
	-- Use a bay fractal to carve bays into the land.
	local bayGrain  = 3;
	local bayFlags  = {FRAC_WRAP_X = true, FRAC_POLAR = true};
	local bayFrac   = Fractal.Create(g_iW, g_iH, panGrain, panFlags, -1, -1)
	local bayThresh = panFrac:GetHeight(72);
	
	for y = 0, g_iH do
		for x = 0, g_iW do
			local bayVal = panFrac:GetHeight(x,y);
			if (bayVal >= bayThresh) then		
				local i = y * g_iW + x + 1; -- add one because Lua arrays start at 1
				self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
			end
		end
	end
	
	-- Reunite the lands with another pass of PassMaker.
	EnsureReachabilityWithPassMaker(bridgeArgs);
	--]]
	
	-- Land and water are set. Now apply mountains, hills, etc.
	self:AddMtsHillsExtras();

	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end

function MultilayeredFractal:Oval_GeneratePlotsByRegion()
	-- Oval
	
	-- Note: fractal wrap flags do not have to match world shape.
	-- Let wrap-x be true here even if the world is flat.
	local fracFlags = {FRAC_WRAP_X = true, FRAC_POLAR = true};

	-- Get user inputs.
	local sea_level = UserInput.seaLevel;

	-- Controls the initial height and width of the oval (before adding bays/lakes)
	--
	-- Don't make this too large (> 95) or else the the landmass will intersect
	-- itself and the poles.
	local axis_list = {0.85, 0.79, 0.73};           -- Used to be {0.87, 0.81, 0.75}
	local axis_multiplier = axis_list[sea_level];
	
	-- Controls the minimum distance of bays/lakes from the oval's center
	--
	-- At high sea level, only the very center of the oval is guaranteed to be dry.
	local cohesion_list = {0.5, 0.3, 0.1};       -- Used to be {0.41, 0.38, 0.35}
	local cohesion_multiplier = cohesion_list[sea_level];

	-- Controls the frequency of bays/lakes
	local baysHeight_list = {86, 80, 74};  -- Used to be hard-coded as 82.
	local baysHeight = baysHeight_list[sea_level];
	local baysGrain  = 3;
	
	if (gotIceAgeFamily) then
		print("Lowering sea levels (ice age)");
		axis_multiplier     = axis_multiplier     + 0.06;
		baysHeight          = baysHeight          + 6;
	end
		
	-- Fill all rows with water plots.
	self.wholeworldPlotTypes = table.fill(g_PLOT_TYPE_OCEAN, g_iW * g_iH);

	local centerX = g_iW / 2;
	local centerY = g_iH / 2;
	local width     = centerX;
	local height    = centerY;

	if (UserInput.worldWrap == WWRAP_GLOBE) then
		-- Global map. Check for cropped climate.
		if (UserInput.climate == Climates.ICEAGECROP) then
			-- Ice age cropped.
			-- Enlarge to compensate for height-crop.
			print("Cropping the oval (globe + ice age cropped)");
			height = height * 1.5;
			
			-- Recalibrate water levels to compensate for larger oval.
			axis_multiplier     = axis_list[sea_level];
			cohesion_multiplier = 0.001;
			baysHeight          = baysHeight - 6;
			
			local baysGrains    = {3,2,1};
			baysGrain           = baysGrains[sea_level];
			
			-- Adjust civ-start strategy and AI hint to be more like a continents map.
			ScriptData.landInfo.aspDivMethod = DIVMETHOD_CONTINENTAL;
			ScriptData.landInfo.expHint   = MAPHINT_OFFSHORE;
			
		elseif (UserInput.climate == Climates.TILTEDCROP) then
			-- Tilted axis cropped.
			-- Enlarge and shift north.
			print("Cropping the oval (globe + tilted axis cropped)");
			height  = g_iH;
			centerY = g_iH;
		end
	end
	
	-- Add the main oval as land plots.
	local majorAxis = width  * axis_multiplier;
	local minorAxis = height * axis_multiplier;
	local majorAxisSquared = majorAxis * majorAxis;
	local minorAxisSquared = minorAxis * minorAxis;	
	for x = 0, g_iW - 1 do
		for y = 0, g_iH - 1 do
			local deltaX = x - centerX;
			local deltaY = y - centerY;
			local deltaXSquared = deltaX * deltaX;
			local deltaYSquared = deltaY * deltaY;
			local d = deltaXSquared/majorAxisSquared + deltaYSquared/minorAxisSquared;
			if d <= 1 then
				local i = y * g_iW + x + 1;
				self.wholeworldPlotTypes[i] = g_PLOT_TYPE_LAND;
			end
		end
	end

	-- Now add bays, fjords, inland seas, etc, but not inside the cohesion area.
	local baysFrac = Fractal.Create(g_iW, g_iH, baysGrain, fracFlags, -1, -1);
	local iBaysThreshold = baysFrac:GetHeight(baysHeight);
	local majorAxis = width * cohesion_multiplier;
	local minorAxis = height * cohesion_multiplier;
	local majorAxisSquared = majorAxis * majorAxis;
	local minorAxisSquared = minorAxis * minorAxis;
	for y = 0, g_iH - 1 do
		for x = 0, g_iW - 1 do
			local deltaX = x - centerX;
			local deltaY = y - centerY;
			local deltaXSquared = deltaX * deltaX;
			local deltaYSquared = deltaY * deltaY;
			local d = deltaXSquared/majorAxisSquared + deltaYSquared/minorAxisSquared;
			if d > 1 then
				local i = y * g_iW + x + 1;
				local baysVal = baysFrac:GetHeight(x, y);
				if baysVal >= iBaysThreshold then
					self.wholeworldPlotTypes[i] = g_PLOT_TYPE_OCEAN;
				end
			end
		end
	end
	
	-- Land and water are set. Now apply mountains, hills, etc.
	self:AddMtsHillsExtras();

	-- Plot Type generation completed. Return global plot array.
	return self.wholeworldPlotTypes;
end
----------------------------------------------------------------------------------
-- River overrides and helper methods
-- Need to declare these before declaring climate info.
----------------------------------------------------------------------------------
function InlandSea_GetRiverValueAtPlot(plot)
	-- Custom method to force rivers to flow toward the map center.
	
    -- Prevent running back into the sea
	if (plot:IsCoastalLand()) then
		return 1000;
	end

	if(plot:IsNWOfCliff() or plot:IsWOfCliff() or plot:IsNEOfCliff()) then
		return 1000;
	elseif(plot:IsNaturalWonder() or AdjacentToNaturalWonder(plot)) then
		return 1000;
	end
	
	local x = plot:GetX()
	local y = plot:GetY()
	local random_factor = TerrainBuilder.GetRandomNumber(3, "River Rand");
	local direction_influence_value = (math.abs(x - (g_iW / 2)) + math.abs(y - (g_iH / 2))) * random_factor;
	local sum = GetPlotElevation(plot) * 20 + direction_influence_value;

	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		local adjacentPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), direction);
		if (adjacentPlot) then
			sum = sum + GetPlotElevation(adjacentPlot);
		else
			sum = 1000;  -- Prevent running off the map
		end
		
	end

	sum = sum + TerrainBuilder.GetRandomNumber(10, "River Rand");

	return sum;
end

function Normal_GetRiverValuesAtPlot(plot)
	-- Default behavior.
	if(plot:IsNWOfCliff() or plot:IsWOfCliff() or plot:IsNEOfCliff()) then
		return -1;
	elseif(plot:IsNaturalWonder() or AdjacentToNaturalWonder(plot)) then
		return -1;
	end


	local sum = GetPlotElevation(plot) * 20;

	local numDirections = DirectionTypes.NUM_DIRECTION_TYPES;
	for direction = 0, numDirections - 1, 1 do

		local adjacentPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), direction);

		if (adjacentPlot ~= nil) then
			sum = sum + GetPlotElevation(adjacentPlot);

			if(g_TERRAIN_TYPE_DESERT == adjacentPlot:GetTerrainType()) then
				sum = sum + 4;
			end		
		else
			sum = sum + 40;
		end
		
	end

	sum = sum + TerrainBuilder.GetRandomNumber(10, "River Rand");

	return sum;
end

function GetRiverValueAtPlot(plot)

	if (gotRegionalInlandSea) then
		-- Special logic for Inland Sea + Region world shape
		return InlandSea_GetRiverValueAtPlot(plot);
	end

	-- Default behavior.
	return Normal_GetRiverValuesAtPlot(plot);
end

function DoRiverReverse(startPlot, thisFlowDirection, originalFlowDirection, minLength, curLength, iRiverID)	

	--print ("Creating River at: " .. tostring(startPlot:GetX()) .. ", ".. tostring(startPlot:GetY()));

	thisFlowDirection = thisFlowDirection or FlowDirectionTypes.NO_FLOWDIRECTION;
	originalFlowDirection = originalFlowDirection or FlowDirectionTypes.NO_FLOWDIRECTION;

	--print ("thisFlowDirection: " .. tostring(thisFlowDirection));
	--print ("originalFlowDirection: " .. tostring(originalFlowDirection));
	--print ("minLength: " .. tostring(minLength));
	--print ("curLength: " .. tostring(curLength));
	--print ("iRiverID: " .. tostring(iRiverID));

	-- pStartPlot = the plot at whose SE corner the river is starting
	local riverPlot;
		
	local bestFlowDirection = FlowDirectionTypes.NO_FLOWDIRECTION;
	if (thisFlowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTH) then

		riverPlot = Map.GetAdjacentPlot(startPlot:GetX(), startPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		if (riverPlot == nil) then
			return;
		end
			
		local adjacentPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		if (adjacentPlot == nil or riverPlot:IsWOfRiver() or riverPlot:IsWater() or adjacentPlot:IsWater()) then
			if (riverPlot:IsWater()) then
				TerrainBuilder.SetWOfRiver(riverPlot, true, thisFlowDirection);	    
			end
			return;
		end
			
		TerrainBuilder.SetWOfRiver(riverPlot, true, thisFlowDirection);
		-- riverPlot does not change
		--print ("At (x,y) SetWOfRiver, flowDirection = ", riverPlot:GetX(), riverPlot:GetY(), thisFlowDirection);
			
	elseif (thisFlowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST) then
		
		riverPlot = startPlot;
		local adjacentPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		if (adjacentPlot == nil or riverPlot:IsNWOfRiver() or riverPlot:IsWater() or adjacentPlot:IsWater()) then
			if (riverPlot:IsWater()) then
				TerrainBuilder.SetNWOfRiver(riverPlot, true, thisFlowDirection);	    
			end
			return;
		end
			
		TerrainBuilder.SetNWOfRiver(riverPlot, true, thisFlowDirection);
		-- riverPlot does not change
		--print ("At (x,y) SetNWOfRiver, flowDirection = ", riverPlot:GetX(), riverPlot:GetY(), thisFlowDirection);
		
	elseif (thisFlowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST) then
		
		riverPlot = startPlot;
		local adjacentPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		if (adjacentPlot == nil or riverPlot:IsNEOfRiver() or riverPlot:IsWater() or adjacentPlot:IsWater()) then
			if (riverPlot:IsWater()) then
					TerrainBuilder.SetNEOfRiver(riverPlot, true, thisFlowDirection);	    
			end
			return;
		end

		TerrainBuilder.SetNEOfRiver(riverPlot, true, thisFlowDirection);
		riverPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_WEST);
		if (riverPlot == nil) then
			return;
		end
		--print ("At (x,y) SetNEOfRiver, flowDirection = ", riverPlot:GetX(), riverPlot:GetY(), thisFlowDirection);
		
	elseif (thisFlowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTH) then
		
		riverPlot = startPlot;
		local adjacentPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		if (adjacentPlot == nil or riverPlot:IsWOfRiver() or riverPlot:IsWater() or adjacentPlot:IsWater() ) then
			if (riverPlot:IsWater()) then
				TerrainBuilder.SetWOfRiver(riverPlot, true, thisFlowDirection);	    
			end
			return;
		end

		TerrainBuilder.SetWOfRiver(riverPlot, true, thisFlowDirection);
		riverPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
		if (riverPlot == nil) then
			return;
		end
		--print ("At (x,y) SetWOfRiver, flowDirection = ", riverPlot:GetX(), riverPlot:GetY(), thisFlowDirection);
		
	elseif (thisFlowDirection == FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST) then

		riverPlot = startPlot;
		local adjacentPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		if (adjacentPlot == nil or riverPlot:IsNWOfRiver() or riverPlot:IsWater() or adjacentPlot:IsWater() ) then
			if (riverPlot:IsWater()) then
				TerrainBuilder.SetNWOfRiver(riverPlot, true, thisFlowDirection);	    
			end
				return;
		end

		TerrainBuilder.SetNWOfRiver(riverPlot, true, thisFlowDirection);
		-- riverPlot does not change
		--print ("At (x,y) SetNWOfRiver, flowDirection = ", riverPlot:GetX(), riverPlot:GetY(), thisFlowDirection);

	elseif (thisFlowDirection == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST) then
			
		riverPlot = Map.GetAdjacentPlot(startPlot:GetX(), startPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		if (riverPlot == nil) then
			return;
		end
			
		local adjacentPlot = Map.GetAdjacentPlot(riverPlot:GetX(), riverPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		if (adjacentPlot == nil or riverPlot:IsNEOfRiver() or riverPlot:IsWater() or adjacentPlot:IsWater()) then
			if (riverPlot:IsWater()) then
				TerrainBuilder.SetNEOfRiver(riverPlot, true, thisFlowDirection);	    
			end
			return;
		end

		TerrainBuilder.SetNEOfRiver(riverPlot, true, thisFlowDirection);
		-- riverPlot does not change
		--print ("At (x,y) SetNEOfRiver, flowDirection = ", riverPlot:GetX(), riverPlot:GetY(), thisFlowDirection);
		
	else
		-- River is starting here, set the direction in the next step
		riverPlot = startPlot;		
	end

	if (IsAdjacentRiver(riverPlot, iRiverID) 
		    or (IsAdjacentMountain(riverPlot) and curLength > minLength) 
		    or (curLength > minLength * 1.5)) then
			
		-- The river has flowed off into a lake or another river, next to a mountain (having met minimum distance), or equalled min length +50%.  We are done.
		--print ("DoRiverReverse() success");
		return; 
	end

	-- Storing X,Y positions as locals to prevent redundant function calls.
	local riverPlotX = riverPlot:GetX();
	local riverPlotY = riverPlot:GetY();
		
	--print ("River Plot now (x, y): ", riverPlotX, riverPlotY);
		
	-- Mark this plot as having a river so we don't come here again
	local iMapIndex = riverPlotY * g_iW + riverPlotX;
	g_riverPlots[iMapIndex] = iRiverID;
		
	-- Table of methods used to determine the adjacent plot.
	local adjacentPlotFunctions = {
		[FlowDirectionTypes.FLOWDIRECTION_NORTH] = function() 
			return Map.GetAdjacentPlot(riverPlotX, riverPlotY, DirectionTypes.DIRECTION_NORTHWEST); 
		end,
		
		[FlowDirectionTypes.FLOWDIRECTION_NORTHEAST] = function() 
			return Map.GetAdjacentPlot(riverPlotX, riverPlotY, DirectionTypes.DIRECTION_NORTHEAST);
		end,
		
		[FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST] = function() 
			return Map.GetAdjacentPlot(riverPlotX, riverPlotY, DirectionTypes.DIRECTION_EAST);
		end,
		
		[FlowDirectionTypes.FLOWDIRECTION_SOUTH] = function() 
			return Map.GetAdjacentPlot(riverPlotX, riverPlotY, DirectionTypes.DIRECTION_SOUTHWEST);
		end,
		
		[FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST] = function() 
			return Map.GetAdjacentPlot(riverPlotX, riverPlotY, DirectionTypes.DIRECTION_WEST);
		end,
		
		[FlowDirectionTypes.FLOWDIRECTION_NORTHWEST] = function() 
			return Map.GetAdjacentPlot(riverPlotX, riverPlotY, DirectionTypes.DIRECTION_NORTHWEST);
		end	
	}
		
	if(bestFlowDirection == FlowDirectionTypes.NO_FLOWDIRECTION) then

		-- Attempt to calculate the best flow direction.
		local bestValue = math.huge;
		for flowDirection, getAdjacentPlot in pairs(adjacentPlotFunctions) do
				
			if (GetOppositeFlowDirection(flowDirection) ~= originalFlowDirection) then
					
				if (thisFlowDirection == FlowDirectionTypes.NO_FLOWDIRECTION or
					flowDirection == TurnRightFlowDirections[thisFlowDirection] or 
					flowDirection == TurnLeftFlowDirections[thisFlowDirection]) then
					
					local adjacentPlot = getAdjacentPlot();

					if (adjacentPlot ~= nil) then

						local value = GetRiverValueAtPlot(adjacentPlot);
						if (flowDirection == originalFlowDirection) then
							value = value / 4;
						end
							
						if (value < bestValue) then
							bestValue = value;
							bestFlowDirection = flowDirection;
						end
					end
				end
			end
		end
			
		-- Try a second pass allowing the river to "flow backwards".
		if(bestFlowDirection == FlowDirectionTypes.NO_FLOWDIRECTION) then
			
			local bestValue = math.huge;
			for flowDirection, getAdjacentPlot in pairs(adjacentPlotFunctions) do
				
				if (thisFlowDirection == FlowDirectionTypes.NO_FLOWDIRECTION or
					flowDirection == TurnRightFlowDirections[thisFlowDirection] or 
					flowDirection == TurnLeftFlowDirections[thisFlowDirection]) then
					
					local adjacentPlot = getAdjacentPlot();			
					if (adjacentPlot ~= nil) then	
						local value = GetRiverValueAtPlot(adjacentPlot);
						if (value < bestValue) then
							bestValue = value;
							bestFlowDirection = flowDirection;
						end
					end	
				end
			end
		end		
	end
		
		--Recursively generate river.
	if (bestFlowDirection ~= FlowDirectionTypes.NO_FLOWDIRECTION) then
		DoRiverReverse(riverPlot, bestFlowDirection, originalFlowDirection, minLength, curLength +1, iRiverID);
	end
end

function IsAdjacentRiver(pPlot, iRiverID)			
	local adjacentPlot;	
	local numDirections = DirectionTypes.NUM_DIRECTION_TYPES;
	for direction = 0, numDirections - 1, 1 do
		adjacentPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), direction);
		if (adjacentPlot ~= nil) then
	   		local i = adjacentPlot:GetY() * g_iW + adjacentPlot:GetX();
			if (g_riverPlots[i] > 0 and iRiverID ~= g_riverPlots[i]) then
				return true;
			end
		end
	end
	return false;
end

function IsAdjacentMountain(pPlot)			
	local adjacentPlot;	
	local numDirections = DirectionTypes.NUM_DIRECTION_TYPES;
	for direction = 0, numDirections - 1, 1 do
		adjacentPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), direction);
		if ((adjacentPlot ~= nil) and adjacentPlot:IsMountain()) then
			return true;
		end
	end
	return false;
end

function NotCloseToAnotherRiver(pStartPlot)
	
	local iNotAllowedAsCloseAs = 4;

	local iPlotIndex = pStartPlot:GetIndex();
	for i, riverPlotIndex in ipairs(g_riverStartPlots) do
		if (Map.GetPlotDistance(iPlotIndex, riverPlotIndex) <= iNotAllowedAsCloseAs) then
		    return false;
		end
	end

	return true;
end

function TryStartRiver(pStartPlot, directionIntoSea)

	-- Check N/S flow direction for match
	if (pStartPlot:GetX() < g_iW / 2) then
		-- Should flow south
		if (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST or directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTH or directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST) then
			return;
		end
	else
		-- Should flow north
		if (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST or directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTH or directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST) then
			return;
		end
	end
	if (pStartPlot:GetY() < g_iH / 2) then
		-- Should flow west
		if (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST or directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST) then
			return;
		end
	else
		-- Should flow east
		if (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST or directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST) then
			return;
		end
	end

	if NotCloseToAnotherRiver(pStartPlot) and not IsAdjacentMountain(pStartPlot) and not IsAdjacentRiver(pStartPlot, -1) then
		table.insert(g_riverStartPlots, pStartPlot:GetIndex());

		local current_direction;
		local iRand = TerrainBuilder.GetRandomNumber (2, "River First Turn Rand");

		if     (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTHEAST) then
			TerrainBuilder.SetNWOfRiver(pStartPlot, true, directionIntoSea);
			current_direction = FlowDirectionTypes.FLOWDIRECTION_NORTH;
			if (iRand == 1) then 
				current_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST;
			end
			
		elseif (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST) then
			TerrainBuilder.SetNEOfRiver(pStartPlot, true, directionIntoSea);
			local pWPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_WEST);
			pStartPlot = pWPlot;
			current_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTH; 
			if (iRand == 1) then 
				current_direction = FlowDirectionTypes.FLOWDIRECTION_NORTHEAST; 
			end

		elseif (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTH) then
			local pWPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_WEST);
			TerrainBuilder.SetWOfRiver(pWPlot, true, directionIntoSea);
			local pNWPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
			pStartPlot = pNWPlot;
			current_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST;
			if (iRand == 1) then 
				current_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST; 
			end

		elseif (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST) then
			local pNWPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
			TerrainBuilder.SetNWOfRiver(pNWPlot, true, directionIntoSea);
			pStartPlot = pNWPlot;
			current_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTH;
			if (iRand == 1) then 
				current_direction = FlowDirectionTypes.FLOWDIRECTION_NORTHWEST;
			end

		elseif (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTHWEST) then
			local pNEPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
			TerrainBuilder.SetNEOfRiver(pNEPlot, true, directionIntoSea);
			pStartPlot = pNEPlot;
			current_direction = FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST;
			if (iRand == 1) then 
				current_direction = FlowDirectionTypes.FLOWDIRECTION_NORTH;
			end

		elseif (directionIntoSea == FlowDirectionTypes.FLOWDIRECTION_NORTH) then
			TerrainBuilder.SetWOfRiver(pStartPlot, true, directionIntoSea);
			current_direction = FlowDirectionTypes.FLOWDIRECTION_NORTHWEST;
			if (iRand == 1) then 
				current_direction = FlowDirectionTypes.FLOWDIRECTION_NORTHEAST; 
			end
		end

		DoRiverReverse(pStartPlot, current_direction, directionIntoSea, 8 + (Map.GetMapSize() * 3), 1, g_iRiverID);
		
		g_iRiverID = g_iRiverID + 1;
	end
end

function InlandSea_AddRivers()
	-- Customization for Inland Sea, to keep river starts away from 
	-- map edges and set river "original flow direction".
	print("Adding Rivers (Inland Sea)");

	local orig_direction, current_direction, pStartPlot;
	
	g_riverPlots = table.fill(0, g_iW * g_iH);

	for i = 0, (g_iW * g_iH) - 1, 1 do
		plot = Map.GetPlotByIndex(i);
		if (plot:IsCoastalLand()) then
			if (plot:IsNaturalWonder() == false and AdjacentToNaturalWonder(plot) == false) then
				local pNWPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
				local pNEPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
				local pEPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_EAST);
				local pSEPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
				local pSWPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
				local pWPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), DirectionTypes.DIRECTION_WEST);

				-- Don't start any rivers really near the map edge 
				if (pNWPlot ~= nil and pNEPlot ~= nil and pEPlot ~= nil and pSEPlot ~= nil and pSWPlot ~= nil and pWPlot ~= nil) then

				    -- ... or near another river
					if (not pNWPlot:IsRiver() and not pNEPlot:IsRiver() and not pEPlot:IsRiver() and not pSEPlot:IsRiver() and not pSWPlot:IsRiver() and not pWPlot:IsRiver()) then

						if     (pEPlot:IsWater()  and not pSEPlot:IsWater() and not pSWPlot:IsWater() and not pWPlot:IsWater())  then
							TryStartRiver(plot, FlowDirectionTypes.FLOWDIRECTION_NORTHEAST);
						elseif (pSEPlot:IsWater() and not pSWPlot:IsWater() and not pWPlot:IsWater()  and not pNWPlot:IsWater()) then
							TryStartRiver(plot, FlowDirectionTypes.FLOWDIRECTION_SOUTHEAST);
						elseif (pSWPlot:IsWater() and not pWPlot:IsWater()  and not pNWPlot:IsWater() and not pNEPlot:IsWater()) then
							TryStartRiver(plot, FlowDirectionTypes.FLOWDIRECTION_SOUTH);
						elseif (pWPlot:IsWater()  and not pNWPlot:IsWater() and not pNEPlot:IsWater() and not pEPlot:IsWater())  then
							TryStartRiver(plot, FlowDirectionTypes.FLOWDIRECTION_SOUTHWEST);
						elseif (pNWPlot:IsWater() and not pNEPlot:IsWater() and not pEPlot:IsWater()  and not pSEPlot:IsWater()) then
							TryStartRiver(plot, FlowDirectionTypes.FLOWDIRECTION_NORTHWEST);
						elseif (pNEPlot:IsWater() and not pEPlot:IsWater()  and not pSEPlot:IsWater() and not pSWPlot:IsWater()) then
							TryStartRiver(plot, FlowDirectionTypes.FLOWDIRECTION_NORTH);
						end
					end
				end
			end			
		end
	end		
end

function Normal_AddRivers()
	-- Default behavior.
	local riverSourceRangeDefault = GlobalParameters.RIVER_SOURCE_RANGE_DEFAULT or 4;
	local seaWaterRangeDefault = GlobalParameters.RIVER_SEA_WATER_RANGE_DEFAULT or 2;
	local plotsPerRiverEdge = GlobalParameters.RIVER_PLOTS_PER_RIVER_EDGE_DEFAULT or 12;
	
	print("Adding Rivers (normal)");
	
	local passConditions = {
		function(plot)
			return (plot:IsHills() or plot:IsMountain());
		end,
		
		function(plot)
			return (not plot:IsCoastalLand()) and (TerrainBuilder.GetRandomNumber(8, "MapGenerator AddRivers") == 0);
		end,
		
		function(plot)
			local area = plot:GetArea();
			return (plot:IsHills() or plot:IsMountain()) and (area:GetRiverEdgeCount() <	((area:GetPlotCount() / plotsPerRiverEdge) + 1));
		end,
		
		function(plot)
			local area = plot:GetArea();
			return (area:GetRiverEdgeCount() < (area:GetPlotCount() / plotsPerRiverEdge) + 1);
		end
	}
	
	for iPass, passCondition in ipairs(passConditions) do
					
		if (iPass <= 2) then
			riverSourceRange = riverSourceRangeDefault;
			seaWaterRange = seaWaterRangeDefault;
		else
			riverSourceRange = (riverSourceRangeDefault / 2);
			seaWaterRange = (seaWaterRangeDefault / 2);
		end
			
		for i = 0, (g_iW * g_iH) - 1, 1 do
			local plot = Map.GetPlotByIndex(i);
			if(not plot:IsWater()) then
				if(passCondition(plot) and plot:IsNaturalWonder() == false and AdjacentToNaturalWonder(plot) == false) then
					if (not Map.FindWater(plot, riverSourceRange, true)) then
						if (not Map.FindWater(plot, seaWaterRange, false)) then
							local inlandCorner = TerrainBuilder.GetInlandCorner(plot);
							if(inlandCorner and plot:IsNaturalWonder() == false and AdjacentToNaturalWonder(plot) == false) then
								DoRiver(inlandCorner);
							end
						end
					end
				end			
			end
		end
	end		
end
----------------------------------------------------------------------------------
-- Terrain overrides and helper methods
-- Need to declare these before declaring climate info.
----------------------------------------------------------------------------------
GotLakes_TerrainGenerator = {};

function GotLakes_TerrainGenerator.Default_Create(args)
	-- Misc override to support global climate info.
	-- This is otherwise normal behavior.
	--
	-- Sea Level and World Age map options affect only plot generation.
	-- Temperature map options affect only terrain generation.
	-- Rainfall map options affect only feature generation.
	--
	local args = args or {};
	local temperature = UserInput.temperature or 2; -- Default setting is Temperate.
	local grain_amount = args.grain_amount or 4;
		
	-- These settings offer a limited ability for map scripts to modify terrain.
	-- Where these are inadequate, replace the TerrainGenerator with a custom method.
	local temperature_shift = args.temperature_shift or 0.1;
	ScriptData.temperatureShift = temperature_shift;
	local desert_shift = args.desert_shift or 12;

	-- Special fractal for the poles of a regional tilted axis.
	-- Because solid tundra/snow/sand is not a fun place to start a game.
	local iHotPolePercent  = args.iHotPolePercent  or 30;
	local iFulldarkPercent = args.iFulldarkPercent or 30;
	local iSemidarkPercent = args.iSemidarkPercent or 45;
	local polar_shift      = args.polar_shift      or 8;
	
	-- Set terrain bands.
	local iDesertPercent = args.iDesertPercent or 28;
	local iPlainsPercent = args.iPlainsPercent or 50; -- Deserts are processed first, so Plains will take this percentage of whatever remains. - Bob
	local fSnowLatitude  = args.fSnowLatitude  or 0.75;
	local fTundraLatitude = args.fTundraLatitude or 0.6;
	local fDesertTopLatitude = args.fDesertTopLatitude or 0.5;
	local fDesertBottomLatitude = args.fDesertBottomLatitude or 0.2;
	local fGrassLatitude = args.fGrassLatitude or 0.1; -- Above this is actually the latitude where it stops being all grass. - Bob
	-- Adjust terrain bands according to user's Temperature selection. (Which must be passed in by the map script.)
	if temperature == TEMPERATURE_COOL then 
		-- Reduce World Temperature.
		iDesertPercent     = iDesertPercent     - desert_shift;
		fSnowLatitude      = fSnowLatitude      - (temperature_shift * 0.5);
		fTundraLatitude    = fTundraLatitude    - (temperature_shift * 1.5);
		fDesertTopLatitude = fDesertTopLatitude - temperature_shift;
		fGrassLatitude     = fGrassLatitude     - (temperature_shift * 0.5);
		
		iHotPolePercent    = iHotPolePercent    - polar_shift;
		iFulldarkPercent   = iFulldarkPercent   + (polar_shift * 0.5);
		iSemidarkPercent   = iSemidarkPercent   + polar_shift;
		
	elseif temperature == TEMPERATURE_HOT then 
		-- Increase World Temperature.
		iDesertPercent     = iDesertPercent     + desert_shift;
		fSnowLatitude      = fSnowLatitude      + (temperature_shift * 0.5);
		fTundraLatitude    = fTundraLatitude    + temperature_shift;
		fDesertTopLatitude = fDesertTopLatitude + temperature_shift;
		fGrassLatitude     = fGrassLatitude     + (temperature_shift * 0.5);
		
		iHotPolePercent    = iHotPolePercent    + (polar_shift * 2);
		iFulldarkPercent   = iFulldarkPercent   - polar_shift;
		iSemidarkPercent   = iSemidarkPercent   - polar_shift;
	end

	local data = 
	{
		-- member methods
		InitFractals			= ScriptData.climateInfo.TInitFractals,
		InitRegion              = ScriptData.climateInfo.TInitRegion,
		GenerateTerrain			= ScriptData.climateInfo.GenerateTerrain,
		GenerateTerrainAtPlot	= ScriptData.climateInfo.GlobeGenerateTerrainAtPlot,
	
		-- member variables
		grain_amount	= grain_amount,
		
		iDesertPercent	        = iDesertPercent,
		iDesertTopPercent		= 100,
		iDesertBottomPercent	= math.max(0, math.floor(100-iDesertPercent)),
		
		iPlainsPercent	        = iPlainsPercent,
		iPlainsTopPercent		= 100,
		iPlainsBottomPercent	= math.max(0, math.floor(100-iPlainsPercent)),
		
		iHotPolePercent         = iHotPolePercent,
		iHotPoleBottomPercent   = math.max(0, math.floor(100-iHotPolePercent)),
		
		iFulldarkPercent        = iFulldarkPercent,
		iFulldarkBottomPercent  = math.max(0, math.floor(100-iFulldarkPercent)),
		
		iSemidarkPercent        = iSemidarkPercent,
		iSemidarkBottomPercent	= math.max(0, math.floor(100-iSemidarkPercent)),
		
		fSnowLatitude			= fSnowLatitude,
		fTundraLatitude			= fTundraLatitude,
		fGrassLatitude			= fGrassLatitude,
		fDesertBottomLatitude	= fDesertBottomLatitude,
		fDesertTopLatitude		= fDesertTopLatitude,
		
		temperatureShift = temperature_shift
	};
	
	data:InitFractals();
	
	if (UserInput.worldWrap == WWRAP_REGION) then
		-- Additional step for climates that need region init.
		data:InitRegion();
	end
	
	return data;
end

function GotLakes_TerrainGenerator:Default_InitFractals()
	self.deserts = Fractal.Create(	g_iW, g_iH, 
									self.grain_amount, g_iFlags, 
									g_fracXExp, g_fracYExp);
									
	self.iDesertTop = self.deserts:GetHeight(self.iDesertTopPercent);
	self.iDesertBottom = self.deserts:GetHeight(self.iDesertBottomPercent);

	self.plains = Fractal.Create(	g_iW, g_iH, 
									self.grain_amount, g_iFlags, 
									g_fracXExp, g_fracYExp);
									
	self.iPlainsTop = self.plains:GetHeight(self.iPlainsTopPercent);
	self.iPlainsBottom = self.plains:GetHeight(self.iPlainsBottomPercent);
end

function GotLakes_TerrainGenerator.Default_InitRegion()
	-- No-op by default. Custom hook for Terrain Generator.
end

function GotLakes_TerrainGenerator:Default_GenerateTerrain()		

	for x = 0, g_iW - 1 do
		for y = 0, g_iH - 1 do 
			local i = y * g_iW + x;
			
			local plot = Map.GetPlotByIndex(i);
						
			if (plot:IsWater() == false) then
				-- Generate terrain for non-water tiles only.
				-- A previous map-gen step already handled water terrain.
				local terrainType = self:GenerateTerrainAtPlot(x, y);
				if terrainType then
					-- Offset terrain type if plot type is hills or mountains.
					-- This assumes that terrain type enums are sorted by 
					-- base terrain, then plot type: 
					--   grass,  grass_hills,  grass_mountains,
					--   plains, plains_hills, plains_mountains,
					--   ...etc.
					if (plot:IsHills()) then
						terrainType = terrainType + 1;
						
					elseif (plot:IsMountain()) then
						terrainType = terrainType + 2;
					end
				
					TerrainBuilder.SetTerrainType(plot, terrainType);
				end
			end
		end
	end
end

function GotLakes_TerrainGenerator:Default_GenerateTerrainAtPlot(iX,iY)
	-- Generate terrain for a non-water tile.
	local lat = GetLatitudeAtPlot(ScriptData.latVariationFrac,iX,iY);
	
	local terrainType = g_TERRAIN_TYPE_GRASS;
			
	if(lat >= self.fSnowLatitude) then
		terrainType = g_TERRAIN_TYPE_SNOW;
		
	elseif(lat >= self.fTundraLatitude) then
		terrainType = g_TERRAIN_TYPE_TUNDRA;
		
	elseif (lat < self.fGrassLatitude) then
		terrainType = g_TERRAIN_TYPE_GRASS;
	else
		local desertVal = self.deserts:GetHeight(iX, iY);
		local plainsVal = self.plains:GetHeight(iX, iY);
		
		if ((desertVal >= self.iDesertBottom)   and (desertVal <= self.iDesertTop) and 
			(lat >= self.fDesertBottomLatitude) and (lat < self.fDesertTopLatitude)) then
			terrainType = g_TERRAIN_TYPE_DESERT;
			
		elseif ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
			terrainType = g_TERRAIN_TYPE_PLAINS;
		end
	end

	return terrainType;
end

function GotLakes_TerrainGenerator.IceAge_Create(args)
	-- Ice Age overrides almost every terrain arg 
	-- and also has asymmetrical temperature shifts,
	-- hence the custom constructor.
	local args = args or {};
	local temperature = UserInput.temperature or 2; -- Default setting is Temperate.
	local grain_amount = 3;
	
	if (UserInput.worldWrap == WWRAP_REGION) then
		-- Use a higher grain for regional ice age.
		-- Improves distribution of good starting spots for civs.
		
		if (UserInput.climate == Climates.ICEAGEFULL) then
			-- Max grain for full ice age.
			grain_amount = 5;
		else
			-- Slightly higher grain for cropped ice age.
			grain_amount = 4;
		end
	end
	
	-- Set terrain bands.
	local iDesertPercent = args.iDesertPercent or 15; -- used to be 20
	local iPlainsPercent = args.iPlainsPercent or 75; -- Deserts are processed first, so Plains will take this percentage of whatever remains.
	local fSnowLatitude  = args.fSnowLatitude  or 0.45;
	local fTundraLatitude = args.fTundraLatitude or 0.275; -- 0.3
	local fDesertTopLatitude = args.fDesertTopLatitude or 0.2;
	local fGrassLatitude = args.fGrassLatitude or 0.0; -- Above this is actually the latitude where it stops being all grass.
	local fDesertBottomLatitude = args.fDesertBottomLatitude or 0.1;
	-- Adjust terrain bands according to user's Temperature selection.
	if temperature == TEMPERATURE_COOL then -- World Temperature is Cool.
		iDesertPercent = 10;
		fSnowLatitude = 0.4; -- Used to be 0.45
		fTundraLatitude = 0.15;
		fDesertTopLatitude = 0.15;
		
	elseif temperature == TEMPERATURE_HOT then -- World Temperature is Hot.
		iDesertPercent = 20;
		fSnowLatitude  = 0.5;
		fTundraLatitude = 0.35; -- 0.4
		fDesertTopLatitude = 0.25;
	else -- Normal Temperature.
	end

	local data = {
	
		-- member methods
		InitFractals			= ScriptData.climateInfo.TInitFractals,
		GenerateTerrain			= ScriptData.climateInfo.GenerateTerrain,
		GenerateTerrainAtPlot	= ScriptData.climateInfo.GlobeGenerateTerrainAtPlot,
	
		-- member variables
		grain_amount	= grain_amount,
		iWidth			= g_iW,
		iHeight			= g_iH,
		
		iDesertPercent	= iDesertPercent,
		iPlainsPercent	= iPlainsPercent,

		iDesertTopPercent		= 100,
		iDesertBottomPercent	= math.max(0, math.floor(100-iDesertPercent)),
		iPlainsTopPercent		= 100,
		iPlainsBottomPercent	= math.max(0, math.floor(100-iPlainsPercent)),
		
		fSnowLatitude			= fSnowLatitude,
		fTundraLatitude			= fTundraLatitude,
		fGrassLatitude			= fGrassLatitude,
		fDesertBottomLatitude	= fDesertBottomLatitude,
		fDesertTopLatitude		= fDesertTopLatitude,
	}

	data:InitFractals();
	
	return data;
end

function GotLakes_TerrainGenerator:Sandstorm_InitFractals()								
	-- Sandstorm
	self.deserts = Fractal.Create(	g_iW, g_iH, 
									self.grain_amount + 1, g_iFlags, 
									g_fracXExp, g_fracYExp);
	
	self.iDesertTop = self.deserts:GetHeight(self.iDesertTopPercent); -- 100
	self.iDesertBottom = self.deserts:GetHeight(self.iDesertBottomPercent); -- 30
	self.iEquatorDesertTop = self.deserts:GetHeight(self.iDesertTopPercent); -- 100
	self.iEquatorDesertBottom = 
		self.deserts:GetHeight(math.clamp(self.iDesertBottomPercent+20,0,100)); -- 50

	self.plains = Fractal.Create(	g_iW, g_iH, 
									self.grain_amount + 2, g_iFlags, 
									g_fracXExp, g_fracYExp);
									
	self.iPlainsTop = self.plains:GetHeight(math.clamp(self.iPlainsTopPercent-20,0,100)); -- 80
	self.iPlainsBottom = self.plains:GetHeight(self.iPlainsBottomPercent); -- 10
	self.iEquatorPlainsTop = self.plains:GetHeight(self.iPlainsTopPercent); -- 100
	self.iEquatorPlainsBottom = 
		self.plains:GetHeight(math.clamp(self.iPlainsBottomPercent+40,0,100)); -- 50
end

function GotLakes_TerrainGenerator:Sandstorm_GenerateTerrainAtPlot(iX,iY)
	local lat = GetLatitudeAtPlot(ScriptData.latVariationFrac,iX,iY);
	
	local terrainVal = g_TERRAIN_TYPE_GRASS;

	-- Special logic for Sandstorm
	if(lat >= self.fSnowLatitude) then
		terrainVal = g_TERRAIN_TYPE_SNOW;
	elseif(lat >= self.fTundraLatitude) then
		terrainVal = g_TERRAIN_TYPE_TUNDRA;
	elseif (lat < self.fGrassLatitude) then
		local desertVal = self.deserts:GetHeight(iX, iY);
		local plainsVal = self.plains:GetHeight(iX, iY);
		if ((desertVal >= self.iEquatorDesertBottom) and (desertVal <= self.iEquatorDesertTop)) then
			terrainVal = g_TERRAIN_TYPE_DESERT;
		elseif ((plainsVal >= self.iEquatorPlainsBottom) and (plainsVal <= self.iEquatorPlainsTop)) then
			terrainVal = g_TERRAIN_TYPE_PLAINS;
		end
	else
		local desertVal = self.deserts:GetHeight(iX, iY);
		local plainsVal = self.plains:GetHeight(iX, iY);
		if ((desertVal >= self.iDesertBottom) and (desertVal <= self.iDesertTop)) then
			terrainVal = g_TERRAIN_TYPE_DESERT;
		elseif ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
			terrainVal = g_TERRAIN_TYPE_PLAINS;
		end
	end

	return terrainVal;
end

function GotLakes_TerrainGenerator:TiltedAxis_InitFractals()
	self.deserts = Fractal.Create(	g_iW, g_iH, 
									self.grain_amount, g_iFlags, 
									g_fracXExp, g_fracYExp);
									
	self.iDesertTop = self.deserts:GetHeight(self.iDesertTopPercent);
	self.iDesertBottom = self.deserts:GetHeight(self.iDesertBottomPercent);

	self.plains = Fractal.Create(	g_iW, g_iH, 
									self.grain_amount, g_iFlags, 
									g_fracXExp, g_fracYExp);
									
	self.iPlainsTop = self.plains:GetHeight(self.iPlainsTopPercent);
	self.iPlainsBottom = self.plains:GetHeight(self.iPlainsBottomPercent);
									
	self.polar = Fractal.Create(	g_iW, g_iH, 
									self.grain_amount, g_iFlags, 
									g_fracXExp, g_fracYExp);
					
	self.iHotPoleBottom  = self.polar:GetHeight(self.iHotPoleBottomPercent);
	self.iFulldarkBottom = self.polar:GetHeight(self.iFulldarkBottomPercent);
	self.iSemidarkBottom = self.polar:GetHeight(self.iSemidarkBottomPercent);
end

function GotLakes_TerrainGenerator:TiltedAxis_InitRegion()
	-- Tilted Axis regional init.
	
	print("Initializing Tilted Axis region.");
	
	-- The Euclidian (x,y) location of the center of the map.
	local mapCenterX = math.floor(g_iW/2);
	local mapCenterY = math.floor(g_iH/2);
	if ((mapCenterY % 2) == 1) then
		mapCenterX = mapCenterX + 0.5;
	end
	
	print("Map Center (x,y): " .. CoordString(mapCenterX, mapCenterY));
	
	-- The location of the hot pole that faces the sun, 
	-- relative to the center of the map.
	--
	-- Start with polar coords so that we can randomize by distance and angle.
	
	-- Scale the maximum hot pole distance using whichever dimension is longer.
	local distRange = math.max(g_iW, g_iH) * 5;
	print("Hot Pole Max R: ", distRange);
	
	local hotPoleRDist = 0; -- Hexes
	local hotPoleTheta = 0; -- Radians
	
	if (UserInput.landmass == Landmasses.DONUT) then
		-- For Donut, fix the hot pole at the center of the map.
		-- This effectively turns the donut filling into desert.
		print("Fixing the hot pole at the donut's filling.");
	else
		-- In hexes
		hotPoleRDist = TerrainBuilder.GetRandomNumber(distRange,"Tilted Axis Region - hot pole distance"); 
		
		-- In Radians
		hotPoleTheta = RandomRadians(0, TAU, 1000);
		print("Hot Pole R: ", hotPoleRDist);
	end	

	
	-- The Location of the pole in cartesion coords, so that we can calculate
	-- distance to other (x,y) locations
	local hotPoleXDiff = hotPoleRDist * math.cos(hotPoleTheta);
	local hotPoleYDiff = hotPoleRDist * math.sin(hotPoleTheta);
	
	-- Convert cartesian (x,y) back to a corresponding map-grid theta.
	-- Forces a one-to-one mapping of tiles to thetas for each possible
	-- location of the pole.
	local hotPoleTileTheta = math.pi/2 - math.atan2(Round(hotPoleXDiff), Round(hotPoleYDiff));
	print("Hot Pole Theta: ", hotPoleTileTheta);

	-- This location may be outside of the hex grid--that is OK!
	local hotPoleX = mapCenterX + hotPoleXDiff;
	local hotPoleY = mapCenterY + hotPoleYDiff;
	print("Hot Pole (x,y): " .. CoordString(hotPoleX, hotPoleY));
	
	-- This prevents the pole from ever covering the entire map, and allows
	-- us to scale the pole's latitude bands based on distance from the 
	-- map's center when the pole is farther away.
	local polarH = g_iW  * math.abs(math.cos(hotPoleTileTheta)) + 
				   g_iH * math.abs(math.sin(hotPoleTileTheta));	
	print("Polar H: ", polarH);
	
	-- Base desert radius
	-- Makes more sense when the pole is far away, since this 
	-- ensures that at the desert intersects with the map regardless
	-- of pole location.
	local desertR = hotPoleRDist - (polarH/4);
	
	-- Min Desert Radius
	-- The minimum radius for the desert portion of the pole, 
	-- in case the pole's location is close to the center of the map.
	local minDesertR = g_iH/5;
	print("Desert Min R: ", minDesertR);
	
	if (desertR < minDesertR) then
		print("Desert R: " .. desertR .. " (too small, using min instead)");
		desertR = minDesertR;
	else
		print("Desert R: ", desertR);
	end
	
	
	-- Desert radius offset based on temperature.
	local temperature = UserInput.temperature;
	local desertROffsets = {-1, 0, 1};
	local desertROffset = desertROffsets[temperature];
	desertROffset = desertROffset*polarH/24;
	
	
	-- The pole's desert radius, based on distance from the center of the map.
	minDesertR = minDesertR + desertROffset;
	desertR    = desertR    + desertROffset;	
	print("Desert R with temperature: ", desertR);
	
	
	-- The "fertile band" starts at the desert radius.
	-- Set the width of this band of terrain.
	local fertileWidth = 1;
	if (Round(desertR) == Round(minDesertR)) then
		-- Hot pole is well inside the map, so the fertile
		-- band is probably visible all around the hot pole.
		--
		-- Use a fixed, small fertile width that grows slightly
		-- with distance from center.
		local widthFracts = {0.3, 0.4, 0.5};
		local widthFract  = widthFracts[temperature]; 
				
		fertileWidth = widthFract*g_iH + 0.6*hotPoleRDist;
				
		print("Close to the center of the map.");
	else
		local hFracts = {0.45, 0.5, 0.55};
		local hFract  = hFracts[temperature];
	
		-- Hot pole far enough from the center of the map to use
		-- a function of both width and height.
		--
		-- Will be widest when the hot pole is on to the
		-- horizontal axis, and thinnest when the hot pole is
		-- on the vertical axis. This keeps all bands visible
		-- on a rectangular map where H < W.
		fertileWidth = hFract*polarH;		
	end
	-- Must be a positive integer.
	fertileWidth = math.max(1, fertileWidth);
	print("fertile band width: " .. fertileWidth);
	
	
	-- The radius beyond which the world faces away from the sun.
	darkSideR = desertR + fertileWidth;
	print("Dark side radius: " .. darkSideR);
	
	-- The latitude to associate with the dark side of the planet.
	-- Also the max latitude of the fertile band.
	-- Must be within (0,1].
	print("Dark side lat: " .. ScriptData.climateInfo.darkSideLat);
	
	-- Set global terrain info so that both terrain and feature gen can
	-- share the same hot pole for latitude calculations.
	ScriptData.terrainInfo =
	{
		hotPoleX     = hotPoleX,
		hotPoleY     = hotPoleY,
		fertileWidth = fertileWidth,
		desertR      = desertR,
		darkSideR    = darkSideR,
	};	
end

function GotLakes_TerrainGenerator:TiltedAxisGlobe_GenerateTerrainAtPlot(iX,iY)
	-- Special logic for tilted axis
	-- Softens the hot pole for improved civ starts and playability.
	
	local lat = GetLatitudeAtPlot(ScriptData.latVariationFrac,iX,iY);
	
	local terrainVal = g_TERRAIN_TYPE_GRASS;
	
	-- Start from dark side inwards for >=
	if ((UserInput.climate == Climates.TILTEDFULL) and
	    (lat >= TILTEDAXIS_SNOWLAT)) then
		-- Snow Zone (dark side for full region and globe)
		terrainVal = g_TERRAIN_TYPE_SNOW;
	elseif (lat >= TILTEDAXIS_TUNDRALAT) then
		-- Tundra Zone (dark side for cropped region)
		terrainVal = g_TERRAIN_TYPE_TUNDRA;
		
	-- Start from hot pole outwards for < and <=
	elseif (lat < TILTEDAXIS_DESERTLAT) then
		-- Sand Zone
		-- Fractal mix of sand and grass.
		local polarVal = self.polar:GetHeight(iX, iY);
		if (polarVal > self.iHotPoleBottom) then
			-- Sand in the sand zone.
			terrainVal = g_TERRAIN_TYPE_DESERT;
		end
	elseif (lat <= TILTEDAXIS_GRASSLAT) then
		-- Fertile band, grass
		local desertVal = self.deserts:GetHeight(iX, iY);
		if ((desertVal >= self.iDesertBottom) and (desertVal <= self.iDesertTop)) then
			terrainVal = g_TERRAIN_TYPE_DESERT;
		end
	else
		-- Fertile band, non-grass
		-- Global fertile band is too thin for desert.
		local plainsVal = self.plains:GetHeight(iX, iY);
		if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
			terrainVal = g_TERRAIN_TYPE_PLAINS;
		end
	end

	return terrainVal;
end

function GotLakes_TerrainGenerator:TiltedAxisRegion_GenerateTerrainAtPlot(iX,iY)
	-- Special logic for regional tilted axis
	-- Softens the poles for improved civ starts and playability.
	-- Dark side is the classic pure wasteland--ignored when placing civ starts.
	
	local lat = GetLatitudeAtPlot(ScriptData.latVariationFrac,iX,iY);
	
	-- Start with grass.
	local terrainVal = g_TERRAIN_TYPE_GRASS;
	
	-- Start from dark side inwards for >=
	if (lat >= TILTEDAXIS_SNOWLAT) then
		-- Snow zone (dark side of the planet)
		if (UserInput.landmass == Landmasses.DONUT) then
			-- Override to force snow
			terrainVal = g_TERRAIN_TYPE_SNOW;
		else
			local polarVal = self.polar:GetHeight(iX, iY);
			
			if (polarVal > self.iFulldarkBottom) then
				-- The full power of the dark side...
				terrainVal = g_TERRAIN_TYPE_SNOW;
				
			elseif (polarVal > self.iSemidarkBottom) then
				-- Only moderately dark.
				terrainVal = g_TERRAIN_TYPE_TUNDRA;
			else
				-- Fertile band, non-grass, non-desert.
				local plainsVal = self.plains:GetHeight(iX, iY);
				if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
					terrainVal = g_TERRAIN_TYPE_PLAINS;
				end
			end
		end
		
	elseif (lat >= TILTEDAXIS_TUNDRALAT) then
		-- Tundra zone (dark side cropped / semi-dark)
		if (UserInput.landmass == Landmasses.DONUT) then
			-- Override to force tundra
			terrainVal = g_TERRAIN_TYPE_TUNDRA;
		else
		
			local polarVal = self.polar:GetHeight(iX, iY);
			
			if (polarVal > self.iSemidarkBottom) then
				-- We're in the tundra zone, so treat "full darkness" as tundra.
				terrainVal = g_TERRAIN_TYPE_TUNDRA;
			else
				-- Fertile band, non-grass, non-desert.
				local plainsVal = self.plains:GetHeight(iX, iY);
				if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
					terrainVal = g_TERRAIN_TYPE_PLAINS;
				end
			end
		end
		
	-- Start from hot pole outwards for < and <=
	elseif (lat < TILTEDAXIS_DESERTLAT) then
		-- Sand Zone (hot pole)
		if (UserInput.landmass == Landmasses.DONUT) then
			-- Override to force all hot-pole tiles to be sand.
			terrainVal = g_TERRAIN_TYPE_DESERT;
		else
			-- Fractal mix of sand and grass.
			local polarVal = self.polar:GetHeight(iX, iY);
			if (polarVal > self.iHotPoleBottom) then
				-- Sand in the sand zone.
				terrainVal = g_TERRAIN_TYPE_DESERT;
			else
				terrainVal = g_TERRAIN_TYPE_GRASS;
			end
		end
			
	elseif (lat <= TILTEDAXIS_GRASSLAT) then
		-- Fertile band, grass
		local desertVal = self.deserts:GetHeight(iX, iY);
		if ((desertVal >= self.iDesertBottom) and (desertVal <= self.iDesertTop)) then
			terrainVal = g_TERRAIN_TYPE_DESERT;
		else
			terrainVal = g_TERRAIN_TYPE_GRASS;
		end
	else
		-- Fertile band
		local desertVal = self.deserts:GetHeight(iX, iY);
		local plainsVal = self.plains:GetHeight(iX, iY);
		if ((lat <= ((TILTEDAXIS_GRASSLAT+TILTEDAXIS_TUNDRALAT)/2)) and
			(desertVal >= self.iDesertBottom) and 
		    (desertVal <= self.iDesertTop)) then
			-- Desert in the sun-facing half of fertile band.
			terrainVal = g_TERRAIN_TYPE_DESERT;
		elseif ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
			terrainVal = g_TERRAIN_TYPE_PLAINS;
		end
	end

	return terrainVal;
end

function GotLakes_TerrainGenerator:TwoSuns_InitRegion()
	-- Two Suns regional init.
	
	print("Initializing Two Suns region.");
	
	-- The Euclidian (x,y) location of the center of the map.
	local mapCenterX = math.floor(g_iW/2);
	local mapCenterY = math.floor(g_iH/2);
	if ((mapCenterY % 2) == 1) then
		mapCenterX = mapCenterX + 0.5;
	end
	
	print("Map Center (x,y): " .. CoordString(mapCenterX, mapCenterY));
	
	-- The location of the poles relative to the center of the map.
	-- Start with polar coords so that we can randomize by angle.
	--
	-- Scale the maximum pole distance using whichever dimension is longer.
	local poleRDist = math.max(g_iW, g_iH) * 5;    -- Hexes
	local southPoleTheta = RandomRadians(0, TAU, 1000); -- Radians

	local northPoleTheta = southPoleTheta + math.pi;
	if (northPoleTheta >= TAU) then
		northPoleTheta = northPoleTheta - TAU;
	end
	
	-- The Location of the poles in cartesion coords, so that we can calculate
	-- distance to other (x,y) locations
	local southPoleXDiff = poleRDist * math.cos(southPoleTheta);
	local southPoleYDiff = poleRDist * math.sin(southPoleTheta);
	
	local northPoleXDiff = poleRDist * math.cos(northPoleTheta);
	local northPoleYDiff = poleRDist * math.sin(northPoleTheta);
	
	-- Convert cartesian (x,y) back to a corresponding map-grid theta.
	-- Forces a one-to-one mapping of tiles to thetas for each possible
	-- location of the poles.
	local southPoleTileTheta = math.pi/2 - 
		math.atan2(Round(southPoleXDiff), Round(southPoleYDiff));
	print("South Pole Theta: ", southPoleTileTheta);
	
	-- These locations may be outside of the hex grid--that is OK!
	local southPoleX = mapCenterX + southPoleXDiff;
	local southPoleY = mapCenterY + southPoleYDiff;
	print("South Pole (x,y): " .. CoordString(southPoleX, southPoleY));
	
	local northPoleX = mapCenterX + northPoleXDiff;
	local northPoleY = mapCenterY + northPoleYDiff;
	print("North Pole (x,y): " .. CoordString(northPoleX, northPoleY));
	
	-- This prevents the pole from ever covering the entire map, and allows
	-- us to scale the pole's latitude bands based on distance from the 
	-- map's center when the pole is farther away.
	local polarH = g_iW * math.abs(math.cos(southPoleTileTheta)) + 
				   g_iH * math.abs(math.sin(southPoleTileTheta));	
	print("Polar H: ", polarH);
	
	-- Base desert radius
	-- Makes more sense when the pole is far away, since this 
	-- ensures that at the desert intersects with the map regardless
	-- of pole location.
	local desertR = poleRDist - (polarH/3);
	
	-- Min Desert Radius
	-- The minimum radius for the desert portion of the pole, 
	-- in case the pole's location is close to the center of the map.
	local minDesertR = g_iH/5;
	print("Desert Min R: ", minDesertR);
	
	if (desertR < minDesertR) then
		print("Desert R: " .. desertR .. " (too small, using min instead)");
		desertR = minDesertR;
	else
		print("Desert R: ", desertR);
	end
	
	
	-- Desert radius offset based on temperature.
	local temperature = UserInput.temperature;
	local desertROffsets = {0, 0.25, 0.5};
	local desertROffset = desertROffsets[temperature];
	desertROffset = desertROffset*polarH/10; -- 24
	
	
	-- The pole's desert radius, based on distance from the center of the map.
	minDesertR = minDesertR + desertROffset;
	desertR    = desertR    + desertROffset;	
	print("Desert R with temperature: ", desertR);
	
	
	-- The "fertile band" starts at the desert radius.
	-- Set the width of this band of terrain.
	local hFracts = {0.45, 0.5, 0.55};
	local hFract  = hFracts[temperature];

	-- Hot pole far enough from the center of the map to use
	-- a function of both width and height.
	--
	-- Will be widest when the hot pole is on to the
	-- horizontal axis, and thinnest when the hot pole is
	-- on the vertical axis. This keeps all bands visible
	-- on a rectangular map where H < W.
	local fertileWidth = hFract*polarH;
	
	-- Cut fertile width in half since there are two fertile bands.
	fertileWidth = fertileWidth/2;
	
	-- Must be a positive integer.
	fertileWidth = math.max(1, fertileWidth);
	print("fertile band width: " .. fertileWidth);
	
	
	-- The radius beyond which the world faces away from the sun.
	darkSideR = desertR + fertileWidth;
	print("Dark side radius: " .. darkSideR);
	
	-- The latitude to associate with the dark side of the planet.
	-- Also the max latitude of the fertile band.
	-- Must be within (0,1].
	print("Dark side lat: " .. ScriptData.climateInfo.darkSideLat);
	
	-- Set global terrain info so that both terrain and feature gen can
	-- share the same hot pole for latitude calculations.
	ScriptData.terrainInfo =
	{
		southPoleX   = southPoleX,
		southPoleY   = southPoleY,
		northPoleX   = northPoleX,
		northPoleY   = northPoleY,
		fertileWidth = fertileWidth,
		desertR      = desertR,
		darkSideR    = darkSideR,
	};	
end

function GotLakes_TerrainGenerator:TwoSunsGlobe_GenerateTerrainAtPlot(iX,iY)
	-- Special logic for two suns
	local lat = GetLatitudeAtPlot(ScriptData.latVariationFrac,iX,iY);
	
	local terrainVal = g_TERRAIN_TYPE_GRASS;
	
	-- Start from dark side inwards for >=
	if (lat >= TILTEDAXIS_SNOWLAT) then
		-- Snow Zone (dark side for full region and globe)		
		local polarVal = self.polar:GetHeight(iX, iY);
			
		if (polarVal > self.iFulldarkBottom) then
			-- The full power of the dark side...
			terrainVal = g_TERRAIN_TYPE_SNOW;
			
		elseif (polarVal > self.iSemidarkBottom) then
			-- Only moderately dark.
			terrainVal = g_TERRAIN_TYPE_TUNDRA;
		else
			-- Fertile band, non-grass, non-desert.
			local plainsVal = self.plains:GetHeight(iX, iY);
			if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
				terrainVal = g_TERRAIN_TYPE_PLAINS;
			end
		end
		
	elseif (lat >= TILTEDAXIS_TUNDRALAT) then
		-- Tundra Zone (dark side for cropped region)
		local polarVal = self.polar:GetHeight(iX, iY);
			
		if (polarVal > self.iSemidarkBottom) then
			-- We're in the tundra zone, so treat "full darkness" as tundra.
			terrainVal = g_TERRAIN_TYPE_TUNDRA;
		else
			-- Fertile band, non-grass, non-desert.
			local plainsVal = self.plains:GetHeight(iX, iY);
			if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
				terrainVal = g_TERRAIN_TYPE_PLAINS;
			end
		end
		
	-- Start from hot pole outwards for < and <=
	elseif (lat < TILTEDAXIS_DESERTLAT) then
		-- Sand Zone
		-- Fractal mix of sand and grass.
		local polarVal = self.polar:GetHeight(iX, iY);
		if (polarVal > self.iHotPoleBottom) then
			-- Sand in the sand zone.
			terrainVal = g_TERRAIN_TYPE_DESERT;
		end
	elseif (lat <= TILTEDAXIS_GRASSLAT) then
		-- Fertile band, grass
		local desertVal = self.deserts:GetHeight(iX, iY);
		if ((desertVal >= self.iDesertBottom) and (desertVal <= self.iDesertTop)) then
			terrainVal = g_TERRAIN_TYPE_DESERT;
		end
	else
		-- Fertile band, non-grass
		-- Global fertile band is too thin for desert.
		local plainsVal = self.plains:GetHeight(iX, iY);
		if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
			terrainVal = g_TERRAIN_TYPE_PLAINS;
		end
	end

	return terrainVal;
end

function GotLakes_TerrainGenerator:TwoSunsRegion_GenerateTerrainAtPlot(iX,iY)
	-- Special logic for regional two suns
	-- Softens the poles for improved civ starts and playability.
	-- Dark side is the classic pure wasteland--ignored when placing civ starts.
	local lat = GetLatitudeAtPlot(ScriptData.latVariationFrac,iX,iY);
	
	-- Start with grass.
	local terrainVal = g_TERRAIN_TYPE_GRASS;
	
	-- Start from dark side inwards for >=
	if (lat >= TILTEDAXIS_SNOWLAT) then
		-- Snow zone (dark side of the planet)
		local polarVal = self.polar:GetHeight(iX, iY);
		
		if (polarVal > self.iFulldarkBottom) then
			-- The full power of the dark side...
			terrainVal = g_TERRAIN_TYPE_SNOW;
			
		elseif (polarVal > self.iSemidarkBottom) then
			-- Only moderately dark.
			terrainVal = g_TERRAIN_TYPE_TUNDRA;
		else
			-- Fertile band, non-grass, non-desert.
			local plainsVal = self.plains:GetHeight(iX, iY);
			if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
				terrainVal = g_TERRAIN_TYPE_PLAINS;
			end
		end
		
	elseif (lat >= TILTEDAXIS_TUNDRALAT) then
		-- Tundra zone (dark side cropped / semi-dark)
		local polarVal = self.polar:GetHeight(iX, iY);
		
		if (polarVal > self.iSemidarkBottom) then
			-- We're in the tundra zone, so treat "full darkness" as tundra.
			terrainVal = g_TERRAIN_TYPE_TUNDRA;
		else
			-- Fertile band, non-grass, non-desert.
			local plainsVal = self.plains:GetHeight(iX, iY);
			if ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
				terrainVal = g_TERRAIN_TYPE_PLAINS;
			end
		end
		
	-- Start from hot pole outwards for < and <=
	elseif (lat < TILTEDAXIS_DESERTLAT) then
		-- Sand Zone (hot pole)
		-- Fractal mix of sand and grass.
		local polarVal = self.polar:GetHeight(iX, iY);
		if (polarVal > self.iHotPoleBottom) then
			-- Sand in the sand zone.
			terrainVal = g_TERRAIN_TYPE_DESERT;
		else
			terrainVal = g_TERRAIN_TYPE_GRASS;
		end
			
	elseif (lat <= TILTEDAXIS_GRASSLAT) then
		-- Fertile band, grass
		local desertVal = self.deserts:GetHeight(iX, iY);
		if ((desertVal >= self.iDesertBottom) and (desertVal <= self.iDesertTop)) then
			terrainVal = g_TERRAIN_TYPE_DESERT;
		else
			terrainVal = g_TERRAIN_TYPE_GRASS;
		end
	else
		-- Fertile band
		local desertVal = self.deserts:GetHeight(iX, iY);
		local plainsVal = self.plains:GetHeight(iX, iY);
		if ((lat <= ((TILTEDAXIS_GRASSLAT+TILTEDAXIS_TUNDRALAT)/2)) and
			(desertVal >= self.iDesertBottom) and 
		    (desertVal <= self.iDesertTop)) then
			-- Desert in the sun-facing half of fertile band.
			terrainVal = g_TERRAIN_TYPE_DESERT;
		elseif ((plainsVal >= self.iPlainsBottom) and (plainsVal <= self.iPlainsTop)) then
			terrainVal = g_TERRAIN_TYPE_PLAINS;
		end
	end

	return terrainVal;
end
------------------------------------------------------------------------------
-- Latitude Lookup
-- Need to declare these before declaring climate info.
------------------------------------------------------------------------------
function GetLatitudeVariation(variationFrac, iX, iY, scale)
	local scale = scale or 0.1;

	local fracLevel = variationFrac:GetHeight(iX, iY);
		
	-- Map fractal height value to the range [0,1].
	local variation = 
		(fracLevel - ScriptData.latVariationMinHeight) / ScriptData.latVariationHeightDiff;
		
	-- Shift variation to [-0.5, 0.5].
	variation = variation - 0.5;
		
	-- Transform variation to [-scale/2, scale/2]
	variation = variation * scale;
	
	return variation;
end

function DefaultGlobe_GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Default latitude for globe maps.
	-- Returns a latitude value between 0.0 (tropical) and 1.0 (polar).
	local lat = math.abs((g_iH / 2) - iY) / (g_iH / 2);
	
	-- Adjust latitude using variation fractal, to roughen the border between bands:
	lat = lat + GetLatitudeVariation(variationFrac, iX, iY);
	
	-- Limit to the range [0, 1]:
	lat = math.clamp(lat, 0, 1);
	return lat;
end

function DefaultRegion_GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Default latitude for region maps.
	-- Variation adds +/- regLatVariation latitude.
	local variation = GetLatitudeVariation(
		variationFrac, iX, iY, ScriptData.climateInfo.regLatVariation);
		
	lat = ScriptData.climateInfo.regionLat + variation;
	lat = math.clamp(lat, 0, 1);
	return lat;
end

function IceAgeGlobe_GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Custom lat for Ice Age.
	local lat;

	if (UserInput.climate == Climates.ICEAGEFULL) then
		-- Full ice age: expand polar ice caps
		if ((iY <= g_iH/6) or (iY >= 5*g_iH/6)) then
			lat = 1.0;
		else
			-- Squish the latitude function into the 2/3 non-polar region
			lat = math.abs((g_iH / 2) - iY) / (g_iH / 3);
		end
	else
		-- Cropped ice age: crop-out the frozen wasteland
		lat = math.abs((g_iH / 2) - iY) / (g_iH / 2);
	end

	-- Adjust latitude using variationFrac fractal, to roughen the border between bands:
	lat = lat + GetLatitudeVariation(variationFrac, iX, iY);
	
	-- In order to increase the coverage of tundra latitude, had to increase the "power" of each 0.1 of latitudinal effect.
	lat = lat * 0.6;
	
	-- Limit to the range [0, 1]:
	lat = math.clamp(lat, 0, 1);
	return lat
end

function TiltedAxisGlobe_GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Custom lat for Tilted Axis Globe.
	local lat;

	if (UserInput.climate == Climates.TILTEDFULL) then
		-- full: northern hemisphere is frozen wasteland 
		lat = 2 * iY / g_iH;
	else
		-- cropped: crop-out the frozen wasteland
		lat = iY / g_iH;
	end
	
	if (UserInput.temperature == TEMPERATURE_COOL) then
		lat = lat + lat * ScriptData.temperatureShift;
	elseif (UserInput.temperature == TEMPERATURE_HOT) then
		lat = lat - lat * ScriptData.temperatureShift;
	end

	-- Adjust latitude using variationFrac fractal, to roughen the border between bands:
	lat = lat + GetLatitudeVariation(variationFrac, iX, iY);
	-- Limit to the range [0, 1]:
	lat = math.clamp(lat, 0, 1);
	return lat;
end

function TiltedAxisRegion_GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Custom lat for Tilted Axis Region.
	local lat = 0;
	
	-- Distance from this point to the pole which faces the sun.
	-- Use Euclidian distance so that the pole is circular-ish (give or take
	-- a column). This was good enough for donut, so why not use it here too!
	--
	-- Note: I tried Map.PlotDistance(), and not only was the hex shape super-
	-- ugly, but the distance calculations had some weird anomalies near the 
	-- corners of the map.
	local tiltedLat = Distance(iX, iY, 
		ScriptData.terrainInfo.hotPoleX, ScriptData.terrainInfo.hotPoleY);
	
	if (tiltedLat <= ScriptData.terrainInfo.desertR) then
		-- Location: the "sand zone" of the hot pole!
		return 0;
	end
	
	if (tiltedLat > ScriptData.terrainInfo.darkSideR) then
		-- Location: the dark side of the planet!
		return ScriptData.climateInfo.darkSideLat;
	end
	
	-- Location: the fertile band.

	-- Interpolate distance onto a [0,1], where 0 represents desert radius
	-- and 1 represents the dark side of the planet.
	lat = math.clamp((tiltedLat - ScriptData.terrainInfo.desertR) / 
		ScriptData.terrainInfo.fertileWidth, 0.0, 1.0);
		
	-- Adjust latitude using variationFrac fractal, to roughen the border between bands:
	-- Variation adds +/- regLatVariation latitude.
	local variation = GetLatitudeVariation(
		variationFrac, iX, iY, ScriptData.climateInfo.regLatVariation);
		
	lat = math.clamp(lat + variation, 0, 1);
	
	-- Scale and Translate the lat onto [desertLat, darkSideLat].
	return TILTEDAXIS_DESERTLAT + (lat * 
		(ScriptData.climateInfo.darkSideLat - TILTEDAXIS_DESERTLAT));
end

function TwoSunsGlobe_GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Custom lat for Two Suns Globe.
	-- Flip the scale so that the equator is cold.
	local lat = 1 - math.abs((g_iH / 2) - iY) / (g_iH / 2);
	
	if (UserInput.temperature == TEMPERATURE_COOL) then
		lat = lat + lat * ScriptData.temperatureShift;
	elseif (UserInput.temperature == TEMPERATURE_HOT) then
		lat = lat - lat * ScriptData.temperatureShift;
	end

	-- Adjust latitude using variationFrac fractal, to roughen the border between bands:
	lat = lat + GetLatitudeVariation(variationFrac, iX, iY, 0.03);
		
	-- Limit to the range [0, 1]
	return math.clamp(lat, 0, 1);
end

function TwoSunsRegion_GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Custom lat for Two Suns Region.
	local lat = 0;
	
	-- Distance from this point to the nearest sun-facing pole.
	-- Use Euclidian distance so that the pole is circular-ish (give or take
	-- a column). This was good enough for donut, so why not use it here too!
	--
	-- Note: I tried Map.PlotDistance(), and not only was the hex shape super-
	-- ugly, but the distance calculations had some weird anomalies near the 
	-- corners of the map.
	local tiltedLat = math.min(
		Distance(iX, iY, 
			ScriptData.terrainInfo.southPoleX, 
			ScriptData.terrainInfo.southPoleY),
		Distance(iX, iY, 
			ScriptData.terrainInfo.northPoleX, 
			ScriptData.terrainInfo.northPoleY));
			
	if (tiltedLat <= ScriptData.terrainInfo.desertR) then
		-- Location: the "sand zone" of the hot pole!
		return 0;
	end
	
	if (tiltedLat > ScriptData.terrainInfo.darkSideR) then
		-- Location: the dark side of the planet!
		return ScriptData.climateInfo.darkSideLat;
	end
	
	-- Location: the fertile band.

	-- Interpolate distance onto a [0,1], where 0 represents desert radius
	-- and 1 represents the dark side of the planet.
	lat = math.clamp((tiltedLat - ScriptData.terrainInfo.desertR) / 
		ScriptData.terrainInfo.fertileWidth, 0.0, 1.0);
		
	-- Adjust latitude using variationFrac fractal, to roughen the border between bands:
	-- Variation adds +/- 0.1 latitude.
	local variation = GetLatitudeVariation(
		variationFrac, iX, iY, ScriptData.climateInfo.regLatVariation);
		
	lat = ScriptData.climateInfo.regionLat + variation;
	lat = math.clamp(lat, 0, 1);
		
	-- Scale and Translate the lat onto [desertLat, darkSideLat].
	return TILTEDAXIS_DESERTLAT + (lat * 
		(ScriptData.climateInfo.darkSideLat - TILTEDAXIS_DESERTLAT));
end
------------------------------------------------------------------------------
-- Natural Wonder Generator overrides
-- Improved edge-case handling
------------------------------------------------------------------------------
function CustomGetMultiTileFeaturePlotList(pPlot, eFeatureType, aPlots)
	-- Override to improve stability.
	
	-- First check this plot itself
	if ((pPlot == nil) or (not TerrainBuilder.CanHaveFeature(pPlot, eFeatureType, true))) then
		return false;
	else
		table.insert(aPlots, pPlot:GetIndex());
	end

	-- Which type of custom placement is it?
	local customPlacement = GameInfo.Features[eFeatureType].CustomPlacement;

	-- 2 tiles inland, east-west facing camera
	if (customPlacement == "PLACEMENT_TORRES_DEL_PAINE" or
	    customPlacement == "PLACEMENT_YOSEMITE") then

		-- Assume first tile is the western one, check the one to the east
		local pAdjacentPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		if (pAdjacentPlot ~= nil and TerrainBuilder.CanHaveFeature(pAdjacentPlot, eFeatureType, true) == true) then
			table.insert(aPlots, pAdjacentPlot:GetIndex());
			return true;
		end

	-- 2 tiles on coast, roughly facing camera
	elseif (customPlacement == "PLACEMENT_CLIFFS_DOVER") then
		local pNEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
		local pWPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_WEST);
		local pSWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		local pSEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		local pEPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);

		-- W and SW are water, see if SE works
		local pSecondPlot;
		if (pWPlot ~= nil and pSWPlot ~= nil and pWPlot:IsWater() and pWPlot:IsLake() == false and pSWPlot:IsWater() and pWPlot:IsLake() == false) then
			pSecondPlot = pSEPlot;

		-- SW and SE are water, see if E works
		elseif (pSWPlot ~= nil and pSEPlot ~= nil and pSWPlot:IsWater() and pSWPlot:IsLake() == false and pSEPlot:IsWater() and pSEPlot:IsLake() == false) then
			pSecondPlot = pEPlot;

		-- SE and E are water, see if NE works
		elseif (pSWPlot ~= nil and pEPlot ~= nil  and pSEPlot:IsWater() and pSEPlot:IsLake() == false and pEPlot:IsWater() and pEPlot:IsLake() == false) then
			pSecondPlot = pNEPlot;
		else
			return false;
		end
		
		if (pSecondPlot == nil) then
			return false;
		end

		if (TerrainBuilder.CanHaveFeature(pSecondPlot, eFeatureType, true)) then
			table.insert(aPlots, pSecondPlot:GetIndex());
			return true;
		end

	-- 2 tiles, one on coastal land and one in water, try to face camera if possible
	elseif (customPlacement == "PLACEMENT_GIANTS_CAUSEWAY") then

		-- Assume first tile a land tile without hills, check around it in a preferred order for water
		if (pPlot:IsWater() or pPlot:IsHills()) then
			return false;
		end

		local pSWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		if (pSWPlot ~= nil and pSWPlot:IsWater() and pSWPlot:IsLake() == false) then
			table.insert(aPlots, pSWPlot:GetIndex());
			return true;
		end

		local pSEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		if (pSEPlot ~= nil and pSEPlot:IsWater() and pSEPlot:IsLake() == false) then
			table.insert(aPlots, pSEPlot:GetIndex());
			return true;
		end

		local pWPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_WEST);
		if (pWPlot ~= nil and pWPlot:IsWater() and pWPlot:IsLake() == false) then
			table.insert(aPlots, pWPlot:GetIndex());
			return true;
		end

		local pEPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		if (pEPlot ~= nil and pEPlot:IsWater() and pEPlot:IsLake() == false) then
			table.insert(aPlots, pEPlot:GetIndex());
			return true;
		end

		local pNWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
		if (pNWPlot ~= nil and pNWPlot:IsWater() and pNWPlot:IsLake() == false) then
			table.insert(aPlots, pNWPlot:GetIndex());
			return true;
		end

		local pNEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
		if (pNEPlot ~= nil and pNEPlot:IsWater() and pNEPlot:IsLake() == false) then
			table.insert(aPlots, pNEPlot:GetIndex());
			return true;
		end

	-- 3 tiles in triangle coast on front edge, land behind (with any rotation)
	elseif (customPlacement == "PLACEMENT_PIOPIOTAHI") then

		local pWPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_WEST);
		local pNWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
		local pNEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
		local pEPlot  = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_EAST);
		local pSEPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
		local pSWPlot = Map.GetAdjacentPlot(pPlot:GetX(), pPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
		
		-- all 6 hexes around must be land

		if (pNWPlot == nil or pNEPlot == nil or pWPlot== nil or pSWPlot== nil or pSEPlot== nil or pEPlot== nil or pNWPlot:IsWater() or pNEPlot:IsWater() or pWPlot:IsWater() or pSWPlot:IsWater() or pSEPlot:IsWater() or pEPlot:IsWater()) then
			return false;
		else
			-- find two adjacent plots that can both serve for this NW
			local bWValid  = TerrainBuilder.CanHaveFeature(pWPlot, eFeatureType, true);
			local bNWValid = TerrainBuilder.CanHaveFeature(pNWPlot, eFeatureType, true);
			local bNEValid = TerrainBuilder.CanHaveFeature(pNEPlot, eFeatureType, true);
			local bEValid  = TerrainBuilder.CanHaveFeature(pEPlot, eFeatureType, true);
			local bSEValid = TerrainBuilder.CanHaveFeature(pSEPlot, eFeatureType, true);
			local bSWValid = TerrainBuilder.CanHaveFeature(pSWPlot, eFeatureType, true);

			if (bSEValid ~= nil and bSWValid ~= nil and bSEValid == true and bSWValid == true ) then
				pWaterCheck1 = Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
				pWaterCheck2 = Map.GetAdjacentPlot(pSWPlot:GetX(), pSWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);
				pWaterCheck3 = Map.GetAdjacentPlot(pSWPlot:GetX(), pSWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
				--- mod <<<
				if ((pWaterCheck1 == nil) or (pWaterCheck2 == nil) or (pWaterCheck3 == nil)) then
					-- Off the edge of the map. This prevents NPE.
					return false;
				end
				--- mod >>>
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pSEPlot:GetIndex());
					table.insert(aPlots, pSWPlot:GetIndex());
					return true;
				end
			end
			
			if (bEValid  ~= nil and bSEValid ~= nil and bEValid  == true and bSEValid == true ) then
				pWaterCheck1 = Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_EAST);
				pWaterCheck2 = Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_EAST);
				pWaterCheck3 = Map.GetAdjacentPlot(pSEPlot:GetX(), pSEPlot:GetY(), DirectionTypes.DIRECTION_SOUTHEAST);

				if ((pWaterCheck1 == nil) or (pWaterCheck2 == nil) or (pWaterCheck3 == nil)) then
					-- Off the edge of the map. This prevents NPE.
					return false;
				end
				
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pEPlot:GetIndex());
					table.insert(aPlots, pSEPlot:GetIndex());
					return true;
				end
			end
			
			if (bSWValid  ~= nil and bWValid ~= nil and bSWValid  == true and bWValid == true ) then
				pWaterCheck1 = Map.GetAdjacentPlot(pSWPlot:GetX(), pSWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
				pWaterCheck2 = Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_SOUTHWEST);
				pWaterCheck3 = Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_WEST);
				
				if ((pWaterCheck1 == nil) or (pWaterCheck2 == nil) or (pWaterCheck3 == nil)) then
					-- Off the edge of the map. This prevents NPE.
					return false;
				end
				
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pSWPlot:GetIndex());
					table.insert(aPlots, pWPlot:GetIndex());
					return true;
				end
			end

			if (bWValid  ~= nil and bNWValid ~= nil and bWValid  == true and bNWValid == true ) then	
				pWaterCheck1 = Map.GetAdjacentPlot(pWPlot:GetX(), pWPlot:GetY(), DirectionTypes.DIRECTION_WEST);
				pWaterCheck2 = Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_WEST);
				pWaterCheck3 = Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
				
				if ((pWaterCheck1 == nil) or (pWaterCheck2 == nil) or (pWaterCheck3 == nil)) then
					-- Off the edge of the map. This prevents NPE.
					return false;
				end
				
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pWPlot:GetIndex());
					table.insert(aPlots, pNWPlot:GetIndex());
					return true;
				end
			end
			
			if (bNEValid  ~= nil and bEValid ~= nil and bNEValid  == true and bEValid == true ) then	
				pWaterCheck1 = Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
				pWaterCheck2 = Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
				pWaterCheck3 = Map.GetAdjacentPlot(pEPlot:GetX(), pEPlot:GetY(), DirectionTypes.DIRECTION_EAST);
				
				if ((pWaterCheck1 == nil) or (pWaterCheck2 == nil) or (pWaterCheck3 == nil)) then
					-- Off the edge of the map. This prevents NPE.
					return false;
				end
				
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pNEPlot:GetIndex());
					table.insert(aPlots, pEPlot:GetIndex());
					return true;
				end
			end

			if (bNWValid  ~= nil and bNEValid ~= nil and bNWValid  == true and bNEValid == true ) then
				pWaterCheck1 = Map.GetAdjacentPlot(pNWPlot:GetX(), pNWPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
				pWaterCheck2 = Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHWEST);
				pWaterCheck3 = Map.GetAdjacentPlot(pNEPlot:GetX(), pNEPlot:GetY(), DirectionTypes.DIRECTION_NORTHEAST);
				
				if ((pWaterCheck1 == nil) or (pWaterCheck2 == nil) or (pWaterCheck3 == nil)) then
					-- Off the edge of the map. This prevents NPE.
					return false;
				end
				
				if (pWaterCheck1:IsWater() == false or pWaterCheck2:IsWater() == false or pWaterCheck3:IsWater() == false) then
					return false;
				else
					table.insert(aPlots, pNWPlot:GetIndex());
					table.insert(aPlots, pNEPlot:GetIndex());
					return true;
				end
			end
		end
	end

	return false;
end

function ResetTerrain(iPlot)
	-- No-op - don't let NW generator mess with terrain.
end

function ResetNaturalWonderTerrain(pPlot)
	-- Call this after running PassMaker on natural wonders.
	local iTerrain = pPlot:GetTerrainType();
	if(iTerrain == g_TERRAIN_TYPE_SNOW_HILLS  or iTerrain == g_TERRAIN_TYPE_SNOW_MOUNTAIN) then
		TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_SNOW);
	elseif(iTerrain == g_TERRAIN_TYPE_DESERT_HILLS  or iTerrain == g_TERRAIN_TYPE_DESERT_MOUNTAIN) then
		TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_DESERT);
	elseif(iTerrain == g_TERRAIN_TYPE_PLAINS_HILLS  or iTerrain == g_TERRAIN_TYPE_PLAINS_MOUNTAIN) then
		TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_PLAINS);
	elseif(iTerrain == g_TERRAIN_TYPE_GRASS_HILLS  or iTerrain == g_TERRAIN_TYPE_GRASS_MOUNTAIN) then
		TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_GRASS);
	elseif(iTerrain == g_TERRAIN_TYPE_TUNDRA_HILLS  or iTerrain == g_TERRAIN_TYPE_TUNDRA_MOUNTAIN) then
		TerrainBuilder.SetTerrainType(pPlot, g_TERRAIN_TYPE_TUNDRA);
	end
end
------------------------------------------------------------------------------
-- Feature-Gen overrides and helper methods
-- Need to declare these before declaring climate info.
------------------------------------------------------------------------------
GotLakes_FeatureGenerator = {};

function GotLakes_FeatureGenerator.Default_Create(args)
	
	-- Get temperature setting for ice variation.
	local temperature = UserInput.temperature or TEMPERATURE_NORMAL;
	local rainfall    = UserInput.rainfall    or RAINFALL_NORMAL;
	
	local args = args or {};
	local jungle_grain = args.jungle_grain or 5;
	local forest_grain = args.forest_grain or 5;
	local clump_grain = args.clump_grain or 3;
	
	local iJungleChange = args.iJungleChange or 20;
	local iForestChange = args.iForestChange or 15;
	local iClumpChange = args.iClumpChange or 5;
	local fMarshChange = args.fMarshChange or 1.5;
	local fOasisChange = args.fOasisChange or 1.5;
		
	-- Latitude boundary for jungle; 0 = hot equator, 1 = frozen pole.
	local jungleMaxLat = args.jungleMaxLat or 0.34; -- 0.32 was slightly too low
	local jungleMaxLatChange = args.jungleMaxLatChange or 0.1;
	
	-- Set feature traits.
	local iJunglePercent = args.iJunglePercent or 65;
	local iForestPercent = args.iForestPercent or 31;
	local iClumpHeight = args.iClumpHeight or 90;
	local fMarshPercent = args.fMarshPercent or 8;
	local iOasisPercent = args.iOasisPercent or 6;
	
	
	-- Custom field for fractal ice.
	local iIcePercent = args.iIcePercent or 25;
	local iIceChange  = args.iIceChange or 10;
	local ice_grain   = args.ice_grain or 4;
	
	-- Custom fields for jungle terrain variation
	local jungle_plains_grain  = args.jungle_plains_grain or 4;
	local IJunglePlainsPercent = args.IJunglePlainsPercent or 50;
	local IJunglePlainsChange  = args.IJunglePlainsChange or 25;
	
	-- Custom field for evergreen trees.
	-- Controls a range of reduced change of trees in snow.
	local fEvergreenBottom = args.fEvergreenBottom or 0.6;
	local fEvergreenChange = args.fEvergreenChange or 0.1;
	
	if (temperature == TEMPERATURE_COOL) then
		-- Cool climate => more ice
		iIcePercent = iIcePercent + iIceChange;
		
		-- Cooler terrain under jungle (more plains)
		IJunglePlainsPercent = IJunglePlainsPercent + IJunglePlainsChange;
		
		-- Trees farther from the poles.
		fEvergreenBottom = fEvergreenBottom - fEvergreenChange;
		
		-- Jungles further from poles.
		jungleMaxLat = jungleMaxLat - jungleMaxLatChange;
		
	elseif (temperature == TEMPERATURE_HOT) then
		-- Hot climate => less ice
		iIcePercent = iIcePercent - iIceChange;
		
		-- Warmer terrain under jungle (more grassland)
		IJunglePlainsPercent = IJunglePlainsPercent - IJunglePlainsChange;
		
		-- Trees closer to the poles.
		fEvergreenBottom = fEvergreenBottom + fEvergreenChange;
		
		-- Jungles further from equator
		jungleMaxLat = jungleMaxLat + jungleMaxLatChange;
	end
	
	
	-- Adjust foliage amounts according to user's Rainfall selection. (Which must be passed in by the map script.)
	if rainfall == RAINFALL_DRY then 
		-- Rainfall is sparse, climate is Arid.
		iJunglePercent = iJunglePercent - iJungleChange;
		iForestPercent = iForestPercent - iForestChange;
		iClumpHeight   = iClumpHeight - iClumpChange;
		fMarshPercent  = fMarshPercent / fMarshChange;
		iOasisPercent  = iOasisPercent / fOasisChange;
		
	elseif rainfall == RAINFALL_WET then 
		-- Rainfall is abundant, climate is Wet.
		iJunglePercent = iJunglePercent + iJungleChange;
		iForestPercent = iForestPercent + iForestChange;
		iClumpHeight   = iClumpHeight + iClumpChange;
		fMarshPercent  = fMarshPercent * fMarshChange;
		iOasisPercent  = iOasisPercent * fOasisChange;
	end

	-- create instance data
	local instance = 
	{
		-- methods
		__initFractals		= GotLakes_FeatureGenerator.__Default_initFractals,
		AddFeatures			= GotLakes_FeatureGenerator.Default_AddFeatures,
		AddOasisAtPlot		= GotLakes_FeatureGenerator.Default_AddOasisAtPlot,
		AddIceAtPlot		= ScriptData.climateInfo.GlobeAddIceAtPlot,
		AddMarshAtPlot		= GotLakes_FeatureGenerator.Default_AddMarshAtPlot,
		AddJunglesAtPlot	= ScriptData.climateInfo.AddJunglesAtPlot,
		AddForestsAtPlot	= ScriptData.climateInfo.AddForestsAtPlot,
		AdjustJungleAtPlot  = ScriptData.climateInfo.AdjustJungleAtPlot,
		AdjustRiverTerrain	= GotLakes_FeatureGenerator.Default_AdjustRiverTerrain,
		
		-- members
		iJunglePercent = iJunglePercent,
		iForestPercent = iForestPercent,
		iClumpHeight = iClumpHeight,
		fMarshPercent = fMarshPercent,
		iOasisPercent = iOasisPercent,
	
		iIcePercent = iIcePercent,
		ice_grain = ice_grain,
	
		jungleMaxLat = jungleMaxLat,
	
		IJunglePlainsPercent = IJunglePlainsPercent,
		jungle_plains_grain = jungle_plains_grain,
	
		fEvergreenBottom = fEvergreenBottom,
		fEvergreenTop    = fEvergreenTop,
	
		jungle_grain = jungle_grain,
		forest_grain = forest_grain,
		marsh_grain  = 4,
		clump_grain = clump_grain,		
	};

	-- initialize instance data
	instance:__initFractals()
	
	return instance;
end

function GotLakes_FeatureGenerator:__Default_initFractals()
	-- Initialize fractals
	self.jungles		= Fractal.Create(g_iW, g_iH, self.jungle_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.forests		= Fractal.Create(g_iW, g_iH, self.forest_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.forestclumps	= Fractal.Create(g_iW, g_iH, self.clump_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.marsh			= Fractal.Create(g_iW, g_iH, self.marsh_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.ice			= Fractal.Create(g_iW, g_iH, self.ice_grain, g_iFlags, g_fracXExp, g_fracYExp);
	
	-- For deciding whether jungle should have plains underneath.
	self.junglePlains   = Fractal.Create(g_iW, g_iH, self.jungle_plains_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.iJunglePTop    = self.junglePlains:GetHeight(100 - self.IJunglePlainsPercent);
	
	-- Get heights
	self.iJungleBottom	= self.jungles:GetHeight(50 - Round(self.iJunglePercent/2));
	self.iJungleTop		= self.jungles:GetHeight(50 + Round(self.iJunglePercent/2));
	self.iForestLevel	= self.forests:GetHeight(100 - self.iForestPercent)
	self.iClumpLevel	= self.forestclumps:GetHeight(self.iClumpHeight)
	self.iMarshLevel	= self.marsh:GetHeight(100 - self.fMarshPercent)
	self.iIceLevel		= self.ice:GetHeight(100 - self.iIcePercent)
end

function GotLakes_FeatureGenerator:Default_AddFeatures()
	-- Default behavior for adding features.
	self.iNumLandPlots = 0;
		
	local latMin = 1.0;
	local latMax = 0.0;
		
	-- Main loop, adds features to all plots as appropriate
	for y = 0, g_iH - 1, 1 do
		for x = 0, g_iW - 1, 1 do
			local i = y * g_iW + x;
			local plot = Map.GetPlotByIndex(i);
			
			if (plot ~= nil) then
				local featureType = plot:GetFeatureType();
						
				local lat = GetLatitudeAtPlot(ScriptData.latVariationFrac, x, y);

				if ((UserInput.worldWrap == WORLDWRAP_GLOBE) and 
				    (gotIceAgeFamily == true)) then
					-- Global ice age climate -- squish latitude bands.
					lat = lat * 0.6;
				end
				
				latMin = math.min(latMin, lat);
				latMax = math.max(latMax, lat);
				
				if (plot:IsImpassable()    or 
				    plot:IsNaturalWonder() or 
					(featureType ~= g_FEATURE_NONE)) then
					-- Impassable terrain / natural wonder / already has a feature
					-- Don't add a feature.
				elseif plot:IsWater() then
					-- Water feature
					
					-- TerrainBuilder.CanHaveFeature() doesn't allow ice in lakes, 
					-- so skip that check because we're OK with frozen lakes.
					if (IsAdjacentToRiver(x, y) == false) then
						-- Add ice, but avoid ice next to rivers.
						if (self:AddIceAtPlot(plot, x, y, lat) == true) then
							-- Added ice.
						end
					end
				else
					-- Land feature
					self.iNumLandPlots = self.iNumLandPlots + 1;
					if (TerrainBuilder.CanHaveFeature(plot, g_FEATURE_FLOODPLAINS) == true) then
						-- All desert plots along river are set to flood plains.
						TerrainBuilder.SetFeatureType(plot, g_FEATURE_FLOODPLAINS);
					else
						if(self:AddOasisAtPlot(plot, x, y, lat) == true) then
							-- Added an oasis.
						end
					end

					local bMarsh = false;
					local bJungle = false;
					--None of these are guarenteed
					if(featureType == g_FEATURE_NONE) then
						--First check to add Marsh
						bMarsh = self:AddMarshAtPlot(plot, x, y, lat);

						if (bMarsh == true) then
							-- Added marsh
						end
						
						if(featureType == g_FEATURE_NONE and  bMarsh == false) then
							--check to add Jungle
							bJungle = self:AddJunglesAtPlot(plot, x, y, lat);
							
							if (bJungle == true) then
								local terrainType = 
									self:AdjustJungleAtPlot(plot, x, y, plot:GetTerrainType());
								TerrainBuilder.SetTerrainType(plot, terrainType);
							end
						end
						
						if(featureType == g_FEATURE_NONE and bMarsh== false and bJungle == false) then 
							--check to add Forest
							if (self:AddForestsAtPlot(plot, x, y, lat) == true) then
								-- Added forest.
							end
						end
					end
				end
			end
		end
	end
	
	latMin = Round(latMin*100)/100;
	latMax = Round(latMax*100)/100;
	print("Feature latitude range: " .. CoordString(latMin, latMax));
	
	self:AdjustRiverTerrain(); -- Adjust terrain near rivers
end

function GotLakes_FeatureGenerator:Default_AddOasisAtPlot(plot, iX, iY, lat)
	if(TerrainBuilder.CanHaveFeature(plot, g_FEATURE_OASIS)) then
		if TerrainBuilder.GetRandomNumber(100, "Add Oasis Lua") <= self.iOasisPercent then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_OASIS);
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:Default_AddMarshAtPlot(plot, iX, iY, lat)
	local marsh_height = self.marsh:GetHeight(iX, iY)
	if(marsh_height >= self.iMarshLevel) then
		if(TerrainBuilder.CanHaveFeature(plot, g_FEATURE_MARSH)) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_MARSH)
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:Default_AddJunglesAtPlot(plot, iX, iY, lat)
	if (lat > self.jungleMaxLat) then
		-- This tile is too cold / far from the equator for jungles.
		return false;
	end
	
	local jungle_height = self.jungles:GetHeight(iX, iY);
	
	if jungle_height >= self.iJungleBottom and jungle_height <= self.iJungleTop then
		local terrain_type = plot:GetTerrainType();
		if (((terrain_type == g_TERRAIN_TYPE_GRASS) or
			 (terrain_type == g_TERRAIN_TYPE_GRASS_HILLS)) and
		    (TerrainBuilder.CanHaveFeature(plot, g_FEATURE_JUNGLE)) ) then
			
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_JUNGLE);
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:Default_AddForestsAtPlot(plot, iX, iY, lat)
	if (self.forests:GetHeight(iX, iY) >= self.iForestLevel) or (self.forestclumps:GetHeight(iX, iY) >= self.iClumpLevel) then
		if TerrainBuilder.CanHaveFeature(plot, g_FEATURE_FOREST) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_FOREST);
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:Default_AdjustJungleAtPlot(plot, iX, iY, terrainType)
	-- Use a fractal to decide whether to adjust jungle.
	if (self.junglePlains:GetHeight(iX, iY) >= self.iJunglePTop) then
		-- Jungle plains
		if ((terrainType == g_TERRAIN_TYPE_PLAINS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_GRASS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_DESERT_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_TUNDRA_HILLS)) then
			return g_TERRAIN_TYPE_PLAINS_HILLS;
			
		elseif ((terrainType == g_TERRAIN_TYPE_PLAINS) or 
				(terrainType == g_TERRAIN_TYPE_GRASS) or 
				(terrainType == g_TERRAIN_TYPE_DESERT) or
				(terrainType == g_TERRAIN_TYPE_TUNDRA)) then
			return g_TERRAIN_TYPE_PLAINS;
		end
	else
		-- Jungle grassland
		if ((terrainType == g_TERRAIN_TYPE_PLAINS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_GRASS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_DESERT_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_TUNDRA_HILLS)) then
			return g_TERRAIN_TYPE_GRASS_HILLS;
			
		elseif ((terrainType == g_TERRAIN_TYPE_PLAINS) or 
				(terrainType == g_TERRAIN_TYPE_GRASS) or 
				(terrainType == g_TERRAIN_TYPE_DESERT) or
				(terrainType == g_TERRAIN_TYPE_TUNDRA)) then
			return g_TERRAIN_TYPE_GRASS;
		end
	end
	return terrainType;
end

function GotLakes_FeatureGenerator:Default_AdjustRiverTerrain()
	-- Custom function that does not soften icy rivers.
	for y = 0, g_iH - 1 do 
		for x = 0, g_iW - 1 do
			local i = y * g_iW + x;
			local plot = Map.GetPlotByIndex(i);
			
			-- These flags are for recalc of areas and rebuild of graphics. 
			-- No need to recalc from any of these changes.
			if (plot:IsRiver()) then
				local terrainType = plot:GetTerrainType();
				local newTerrainType = terrainType;
				
				if (terrainType == g_TERRAIN_TYPE_TUNDRA_MOUNTAIN) then
					newTerrainType = g_TERRAIN_TYPE_PLAINS_MOUNTAIN;
				
				elseif (terrainType == g_TERRAIN_TYPE_TUNDRA_HILLS) then
					newTerrainType = g_TERRAIN_TYPE_PLAINS_HILLS;
					
				elseif (terrainType == g_TERRAIN_TYPE_TUNDRA) then
					newTerrainType = g_TERRAIN_TYPE_PLAINS;
				end

				TerrainBuilder.SetTerrainType(plot, newTerrainType);
			end
		end
	end
end

function GotLakes_FeatureGenerator:Evergreen_AddForestsAtPlot(plot, iX, iY, lat)
	
	if (plot:IsMountain()) then
		-- Don't allow forests on mountains, because that looks wrong.
		return false;
	end
	
	if (self.forests:GetHeight(iX, iY) >= self.iForestLevel) or 
	   (self.forestclumps:GetHeight(iX, iY) >= self.iClumpLevel) then
	   
	    -- Allow forests on snow.
		if (TerrainBuilder.CanHaveFeature(plot, g_FEATURE_FOREST) or
			(plot:GetTerrainType() == g_TERRAIN_TYPE_SNOW) or 
			(plot:GetTerrainType() == g_TERRAIN_TYPE_SNOW_HILLS)) then
			
			if (lat <= self.fEvergreenBottom) then
				TerrainBuilder.SetFeatureType(plot, g_FEATURE_FOREST);
				return true;
			end
			
			-- Roll dice, check latitude.
			local rand = TerrainBuilder.GetRandomNumber(100, "Add Forest Lua")/100.0;
			if (rand > 2.5*(lat - self.fEvergreenBottom)) then
				TerrainBuilder.SetFeatureType(plot, g_FEATURE_FOREST);
				return true;
			end	
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:Globe_AddIceAtPlot(plot, iX, iY, lat)
	-- Overridden for Lakes and WB wrap-x awareness
	
	-- Overriding default feature ice to ensure that the ice fully covers
	-- a landmass-specific number of rows, at top and bottom
	local iceNumRows = ScriptData.landInfo.polarIceNumRows;
	
	if ((iY < iceNumRows) or (iY >= g_iH - iceNumRows)) then
		-- Frozen wasteland.
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE)
		return true;
	end
	
	-- Assuming we only have at most two rows at each vertical edge
	-- of the map, just use white noise for scarce ice tiles.
	local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua")/100.0;
	local iceRollPassed = false;
	
	if(rand < 8 * (lat - 0.875)) then
		iceRollPassed = true;
	elseif(rand < 4 * (lat - 0.75)) then
		iceRollPassed = true;
	end
	
	if (iceRollPassed == true) then
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
		return true;
	end
	return false;
end

function GotLakes_FeatureGenerator.IceAge_Create(args)
	-- Customizations for Ice Age include adding much more forest in tundra 
	-- regions, but reducing foliage elsewhere, and adding more random feature ice.
	local args = args or {};
	local jungle_grain = args.jungle_grain or 7;
	local forest_grain = args.forest_grain or 6;
	local clump_grain = args.clump_grain or 3;
	local iFoliageChange = args.iFoliageChange or 10;
	local iClumpChange = args.iClumpChange or 5;
	local fMarshChange = args.fMarshChange or 1.5;
	local fOasisChange = args.fOasisChange or 1.5;
	
	-- Get temperature setting for ice variation.
	local temperature = UserInput.temperature or TEMPERATURE_NORMAL;
	
	-- Get Rainfall setting.
	local rain = UserInput.rainfall or RAINFALL_NORMAL;
	
	-- Latitude boundary for jungle; 0 = hot equator, 1 = frozen pole.
	local jungleMaxLat = args.jungleMaxLat or 0.1;
	local jungleMaxLatChange = args.jungleMaxLatChange or 0.05;	
	
	-- Set feature traits.
	local iJunglePercent = args.iJunglePercent or 30;
	local iForestPercent = args.iForestPercent or 28;
	local iTundraPercent = args.iTundraPercent or 60; -- originally 70
	--
	local iClumpHeight  = args.iClumpHeight or 93;
	local fMarshPercent = args.fMarshPercent or 20;
	local iOasisPercent = args.iOasisPercent or 6;
	--
	-- Custom field for fractal ice.
	local iIcePercent = args.iIcePercent or 30;
	local iIceChange  = args.iIceChange or 15;
	local ice_grain   = args.ice_grain or 4;
	
	-- Custom fields for jungle terrain variation
	local jungle_plains_grain  = args.jungle_plains_grain or 4;
	local IJunglePlainsPercent = args.IJunglePlainsPercent or 50;
	local IJunglePlainsChange  = args.IJunglePlainsChange or 25;
	
	
	if (temperature == TEMPERATURE_COOL) then
		-- Cool climate => more ice
		iIcePercent = iIcePercent + iIceChange;
		
		-- Cooler terrain under jungle (more plains)
		IJunglePlainsPercent = IJunglePlainsPercent + IJunglePlainsChange;
		
		-- Jungles further from poles.
		jungleMaxLat = jungleMaxLat - jungleMaxLatChange;
		
	elseif (temperature == TEMPERATURE_HOT) then
		-- Hot climate => less ice
		iIcePercent = iIcePercent - iIceChange;
		
		-- Warmer terrain under jungle (more grassland)
		IJunglePlainsPercent = IJunglePlainsPercent - IJunglePlainsChange;
		
		-- Jungles further from equator
		jungleMaxLat = jungleMaxLat + jungleMaxLatChange;
	end

	
	-- Adjust terrain bands according to user's Temperature selection.
	if rain == RAINFALL_DRY then 
		-- Rainfall is sparse, climate is Arid.
		iJunglePercent = iJunglePercent - iFoliageChange;
		iForestPercent = iForestPercent - iFoliageChange;
		--
		iTundraPercent = iTundraPercent - iFoliageChange;
		--
		iClumpHeight = iClumpHeight - iClumpChange;
		fMarshPercent = fMarshPercent / fMarshChange;
		iOasisPercent = iOasisPercent / fOasisChange;
		
	elseif rain == RAINFALL_WET then 
		-- Rainfall is abundant, climate is Wet.
		iJunglePercent = iJunglePercent + iFoliageChange;
		iForestPercent = iForestPercent + iFoliageChange;
		--
		iTundraPercent = iTundraPercent + iFoliageChange;
		--
		iClumpHeight = iClumpHeight + iClumpChange;
		fMarshPercent = fMarshPercent * fMarshChange;
		iOasisPercent = iOasisPercent * fOasisChange;
	end

	-- create instance data
	local instance = {
	
		-- methods
		__initFractals		= GotLakes_FeatureGenerator.__IceAge_initFractals,
		AddFeatures			= GotLakes_FeatureGenerator.Default_AddFeatures,
		AddOasisAtPlot		= GotLakes_FeatureGenerator.Default_AddOasisAtPlot,
		AddIceAtPlot		= GotLakes_FeatureGenerator.IceAgeGlobe_AddIceAtPlot,
		AddMarshAtPlot		= GotLakes_FeatureGenerator.Default_AddMarshAtPlot,
		AddJunglesAtPlot	= GotLakes_FeatureGenerator.Default_AddJunglesAtPlot,
		AddForestsAtPlot	= GotLakes_FeatureGenerator.IceAge_AddForestsAtPlot,
		AdjustJungleAtPlot  = GotLakes_FeatureGenerator.Default_AdjustJungleAtPlot,
		AdjustRiverTerrain  = GotLakes_FeatureGenerator.Default_AdjustRiverTerrain,		

		-- members		
		iJunglePercent = iJunglePercent,
		iForestPercent = iForestPercent,
		--
		iTundraPercent = iTundraPercent,
		--
		iIcePercent = iIcePercent,
		ice_grain = ice_grain,
		--
		
		jungleMaxLat = jungleMaxLat,
	
		IJunglePlainsPercent = IJunglePlainsPercent,
		jungle_plains_grain = jungle_plains_grain,
		
		iClumpHeight = iClumpHeight,
		fMarshPercent = fMarshPercent,
		iOasisPercent = iOasisPercent,
	
		jungle_grain = jungle_grain,
		forest_grain = forest_grain,
		marsh_grain  = 4,
		clump_grain  = clump_grain,
	};

	-- initialize instance data
	instance:__initFractals()
	
	return instance;
end

function GotLakes_FeatureGenerator:__IceAge_initFractals()
	-- Create fractals
	self.jungles		= Fractal.Create(g_iW, g_iH, self.jungle_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.forests		= Fractal.Create(g_iW, g_iH, self.forest_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.forestclumps	= Fractal.Create(g_iW, g_iH, self.clump_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.marsh			= Fractal.Create(g_iW, g_iH, self.marsh_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.ice			= Fractal.Create(g_iW, g_iH, self.ice_grain, g_iFlags, g_fracXExp, g_fracYExp);
	
	-- For deciding whether jungle should have plains underneath.
	self.junglePlains   = Fractal.Create(g_iW, g_iH, self.jungle_plains_grain, g_iFlags, g_fracXExp, g_fracYExp);
	self.iJunglePTop    = self.junglePlains:GetHeight(100 - self.IJunglePlainsPercent);
	
	-- Get heights
	self.iJungleBottom	= self.jungles:GetHeight(50 - Round(self.iJunglePercent/2));
	self.iJungleTop		= self.jungles:GetHeight(50 + Round(self.iJunglePercent/2));
	self.iTundraLevel	= self.forests:GetHeight(100 - self.iTundraPercent)
	self.iForestLevel	= self.forests:GetHeight(100 - self.iForestPercent)
	self.iClumpLevel	= self.forestclumps:GetHeight(self.iClumpHeight)
	self.iMarshLevel	= self.marsh:GetHeight(100 - self.fMarshPercent)
	self.iIceLevel		= self.ice:GetHeight(100 - self.iIcePercent)

end

function GotLakes_FeatureGenerator:IceAgeGlobe_AddIceAtPlot(plot, iX, iY, lat)
	-- Ice Age globe.
		
	local iceNumRows = ScriptData.landInfo.polarIceNumRows;
	
	if ((iY < iceNumRows) or (iY >= g_iH - iceNumRows)) then
		-- Top or bottom N rows of the map ==> frozen
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
		return true;
	end
	
	
	if (UserInput.climate == Climates.ICEAGEFULL) then
		-- Full ice age: extend ice by 1/6 of the map's height
		iceNumRows = iceNumRows + (g_iH / 6);
	end
	
	if ((iY < iceNumRows) or (iY >= g_iH - iceNumRows)) then
		-- Check against extended row count.
		-- This could be a lot of ice, so use a fractal pattern
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
		return false;
	end
	
	-- Roll dice, check latitude.
	local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua")/100;
	local iceRollPassed = false;
	
	if (rand < 8 * (lat - 0.875)) then
		-- Chance of ice ranging from 0 to 25%, starting at latitude 0.875
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			iceRollPassed = true;
		end
	elseif (rand < 4 * (lat - 0.75)) then
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			iceRollPassed = true;
		end
	end	

	if lat > 0.47 then
		local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua") / 100;
		if rand < 8 * (lat - 0.50) then
			iceRollPassed = true;
		elseif rand < 4 * (lat - 0.46) then
			iceRollPassed = true;
		end
		
	-- Add encroaching icebergs reaching out beyond normal range
	elseif lat > 0.39 then
		local rand = TerrainBuilder.GetRandomNumber(100, "Add Encroaching Ice - Sirian's Ice Age - Lua") / 100;
		if rand < 0.06 then
			iceRollPassed = true;
		end
	elseif lat > 0.32 then
		local rand = TerrainBuilder.GetRandomNumber(100, "Add Encroaching Ice - Sirian's Ice Age - Lua") / 100;
		if rand < 0.04 then
			iceRollPassed = true;
		end
	elseif lat > 0.27 then
		local rand = TerrainBuilder.GetRandomNumber(100, "Add Encroaching Ice - Sirian's Ice Age - Lua") / 100;
		if rand < 0.02 then
			iceRollPassed = true;
		end
	end
	
	-- Intersect the white noise with a fractal pattern so that the encroaching ice
	-- blends nicely with the polar ice.
	if ((iceRollPassed == true) and 
	    (self.ice:GetHeight(iX, iY) >= self.iIceLevel)) then
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
		return true;
	end
	return false;
end

function GotLakes_FeatureGenerator:IceAgeRegion_AddIceAtPlot(plot, iX, iY, lat)
	-- Ice Age region.	
	if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
		return true;
	end
	return false;
end

function GotLakes_FeatureGenerator:IceAge_AddForestsAtPlot(plot, iX, iY, lat)
	local terrainType = plot:GetTerrainType()
	if ((terrainType == g_TERRAIN_TYPE_TUNDRA) or
		(terrainType == g_TERRAIN_TYPE_TUNDRA_HILLS)) then
		if (self.forests:GetHeight(iX, iY) >= self.iTundraLevel) or 
		   (self.forestclumps:GetHeight(iX, iY) >= self.iClumpLevel) then
			if TerrainBuilder.CanHaveFeature(plot, g_FEATURE_FOREST) then
				TerrainBuilder.SetFeatureType(plot, g_FEATURE_FOREST)
				return true;
			end
		end
	else
		if (self.forests:GetHeight(iX, iY) >= self.iForestLevel) or 
		   (self.forestclumps:GetHeight(iX, iY) >= self.iClumpLevel) then
			if TerrainBuilder.CanHaveFeature(plot, g_FEATURE_FOREST) then
				TerrainBuilder.SetFeatureType(plot, g_FEATURE_FOREST)
				return true;
			end
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:Rainforest_AddJunglesAtPlot(plot, iX, iY, lat)
	-- Rainforest override
	-- Ignore latitude.
	local jungle_height = self.jungles:GetHeight(iX, iY);
	
	if jungle_height >= self.iJungleBottom and jungle_height <= self.iJungleTop then
		local terrain_type = plot:GetTerrainType()
		if (((terrain_type == g_TERRAIN_TYPE_GRASS) or
			 (terrain_type == g_TERRAIN_TYPE_GRASS_HILLS)) and
		    (TerrainBuilder.CanHaveFeature(plot, g_FEATURE_JUNGLE)) ) then
			
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_JUNGLE);
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:Region_AddIceAtPlot(plot, iX, iY, lat)
	-- Overridden for regions
	
	-- White noise is good enough for maps that don't have much ice.
	local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua")/100.0;
	
	if(rand < 8 * (lat - 0.875)) then
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
		return true;
	elseif(rand < 4 * (lat - 0.75)) then
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
		return true;
	end
	return false;
end

function GotLakes_FeatureGenerator:SandyJungle_AddJunglesAtPlot(plot, iX, iY, lat)
	-- Sandy Jungle override
	
	if (plot:IsMountain()) then
		-- Don't allow jungles on mountains, because that looks wrong.
		return false;
	end
	
	-- Ignore latitude.
	local jungle_height = self.jungles:GetHeight(iX, iY);
	
	if jungle_height >= self.iJungleBottom and jungle_height <= self.iJungleTop then
		local terrainType = plot:GetTerrainType();
		
		if ((terrainType == g_TERRAIN_TYPE_DESERT) or 
			(terrainType == g_TERRAIN_TYPE_DESERT_HILLS) or 
			TerrainBuilder.CanHaveFeature(plot, g_FEATURE_JUNGLE)) then
			-- Allow jungles on desert.
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_JUNGLE);
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:SandyJungle_AdjustJungleAtPlot(plot, iX, iY, terrainType)
	-- Allow desert under jungle.
	-- For other terrain types, use a fractal to decide whether to adjust jungle.
	if (self.junglePlains:GetHeight(iX, iY) >= self.iJunglePTop) then
		-- Jungle plains
		if ((terrainType == g_TERRAIN_TYPE_PLAINS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_GRASS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_TUNDRA_HILLS)) then
			return g_TERRAIN_TYPE_PLAINS_HILLS;
			
		elseif ((terrainType == g_TERRAIN_TYPE_PLAINS) or 
				(terrainType == g_TERRAIN_TYPE_GRASS) or 
				(terrainType == g_TERRAIN_TYPE_TUNDRA)) then
			return g_TERRAIN_TYPE_PLAINS;
		end
	else
		-- Jungle grassland
		if ((terrainType == g_TERRAIN_TYPE_PLAINS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_GRASS_HILLS) or 
			(terrainType == g_TERRAIN_TYPE_TUNDRA_HILLS)) then
			return g_TERRAIN_TYPE_GRASS_HILLS;
			
		elseif ((terrainType == g_TERRAIN_TYPE_PLAINS) or 
				(terrainType == g_TERRAIN_TYPE_GRASS) or 
				(terrainType == g_TERRAIN_TYPE_TUNDRA)) then
			return g_TERRAIN_TYPE_GRASS;
		end
	end
	return terrainType;
end

function GotLakes_FeatureGenerator:TiltedAxisGlobeCropped_AddIceAtPlot(plot, iX, iY, lat)
	-- Tilted Axis Cropped Globe
		
	local iceNumRows = ScriptData.landInfo.polarIceNumRows;
	
	if (iY >= g_iH - iceNumRows) then
		-- Top rows for cropped climate => frozen
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
		return true;
	end
	-- Don't bother with ice on the bottom rows,
	-- since the south pole is hot.
	
	-- Not much ice on this map, so just use white noise.
	-- Roll dice, check latitude.
	local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua")/100.0;
	
	-- Use fractal here because tilted axis can have a lot of ice, 
	-- even when cropped.
	if (rand < 8 * (lat - 0.875)) then
		-- Chance of ice ranging from 0 to 25%, starting at latitude 0.875
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
	elseif (rand < 4 * (lat - 0.75)) then
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:TiltedAxisGlobeFull_AddIceAtPlot(plot, iX, iY, lat)
	-- Tilted Axis Full globe
	--print("Possibly adding ice at " .. CoordString(iX, iY));
	local iceNumRows = ScriptData.landInfo.polarIceNumRows;
	
	local iceHeight = self.ice:GetHeight(iX, iY);
	--print("Ice Height vs. Ice Threshold: " .. CoordString(iceHeight, self.iIceLevel));
	
	if (iY >= g_iH - iceNumRows) then
		-- Frozen wasteland at the edge of the map.
		TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE)
		--print("Top-row ice.");
		return true;
	end
		
	if (iY >= g_iH / 2) then
		-- Northern hemisphere for full climate => frozen
		-- This is usually a ton of ice, so use fractal pattern.
		if (iceHeight >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			--print("Dark Side ice.");
			return true;
		end
		return false;
	end	
	-- Don't bother with ice on the bottom rows,
	-- since the south pole is hot.
	
	-- Roll dice, check latitude.
	local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua")/100;
	
	if (rand < 8 * (lat - 0.875)) then
		-- Chance of ice ranging from 0 to 25%, starting at latitude 0.875
		if (iceHeight >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			--print("Creeping ice 1.");
			return true;
		end
	elseif (rand < 4 * (lat - 0.75)) then
		if (iceHeight >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			--print("Creeping ice 2.");
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:TiltedAxisRegion_AddIceAtPlot(plot, iX, iY, lat)
	-- Tilted Axis region	
	if (lat >= TILTEDAXIS_SNOWLAT) then
		-- Location: snowy dark side of the planet!
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
		return false;
	end
	
	-- Roll dice, check latitude.
	local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua")/100.0;
	
	-- Use fractal here because tilted axis region can have a lot of ice.
	if (rand < 8 * (lat - 0.875)) then
		-- Chance of ice ranging from 0 to 25%, starting at latitude 0.875
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
	elseif (rand < 4 * (lat - 0.75)) then
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
	end
	return false;
end

function GotLakes_FeatureGenerator:TwoSuns_AddIceAtPlot(plot, iX, iY, lat)
	-- Two Suns
	if (lat >= TILTEDAXIS_TUNDRALAT) then
		-- Location: snowy dark side of the planet!
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
		return false;
	end
	
	-- Roll dice, check latitude.
	--[[
	local rand = TerrainBuilder.GetRandomNumber(100, "Add Ice Lua")/100.0;
	
	-- Use fractal here because tilted axis region can have a lot of ice.
	if (rand < 8 * (lat - 0.875)) then
		-- Chance of ice ranging from 0 to 25%, starting at latitude 0.875
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
	elseif (rand < 4 * (lat - 0.75)) then
		if (self.ice:GetHeight(iX, iY) >= self.iIceLevel) then
			TerrainBuilder.SetFeatureType(plot, g_FEATURE_ICE);
			return true;
		end
	end
	--]]
	return false;
end
------------------------------------------------------------------------------
-- Continent Stamper overrides
-- Need to declare these before declaring land info.
------------------------------------------------------------------------------
function Default_DetermineContinents()
	-- Just use the default stamper.
	print("Using default stamper for continent art.");
		
	-- The default behavior operates in C++..
	TerrainBuilder.StampContinents();
end

function Uniform_DetermineContinents()
	-- Choose an art set randomly and use that one art set for the entire map.
	local artNum   = 1 + TerrainBuilder.GetRandomNumber(4, "Random uniform continent art.");
	local artNames = {"American", "Asian", "African", "European"};
	
	print ("Using " .. artNames[artNum] .. " art for the entire map.");
	
	for i, plot in Plots() do
		if plot:IsWater() then
			plot:SetContinentArtType(0); -- ocean art
		else
			plot:SetContinentArtType(artNum);
		end
	end
end
------------------------------------------------------------------------------
-- Resource Generator overrides
-- Need to declare these before declaring land and climate info.
------------------------------------------------------------------------------
GotLakes_ResourceGenerator = {}

function ResourceGenerator.Create(args)
	-- We have no choice but to override the constructor in order to change
	-- behavior, because the constructor does EVERYTHING in addition to 
	-- construction.

	-- create instance data
	local instance = 
	{		
		-- methods
		__InitResourceData				= ResourceGenerator.__InitResourceData,
		__FindValidLocs					= ResourceGenerator.__FindValidLocs,
		__GetLuxuryResources			= ResourceGenerator.__GetLuxuryResources,
		__ValidLuxuryPlots				= ScriptData.climateInfo.ValidLuxuryPlots,
		__PlaceLuxuryResources			= ResourceGenerator.__PlaceLuxuryResources,
		__ScoreLuxuryPlots				= ScriptData.climateInfo.ScoreLuxuryPlots,
		__GetWaterLuxuryResources		= ResourceGenerator.__GetWaterLuxuryResources,
		__SetWaterLuxury				= ResourceGenerator.__SetWaterLuxury,
		__PlaceWaterLuxury				= ResourceGenerator.__PlaceWaterLuxury,
		__GetStrategicResources			= ResourceGenerator.__GetStrategicResources,
		__ValidStrategicPlots			= ResourceGenerator.__ValidStrategicPlots,
		__ScoreStrategicPlots			= ResourceGenerator.__ScoreStrategicPlots,
		__PlaceStrategicResources		= ResourceGenerator.__PlaceStrategicResources,
		__GetWaterStrategicResources	= ResourceGenerator.__GetWaterStrategicResources,
		__SetWaterStrategic				= ResourceGenerator.__SetWaterStrategic,
		__PlaceWaterStrategic			= ResourceGenerator.__PlaceWaterStrategic,
		__GetOtherResources				= ResourceGenerator.__GetOtherResources,
		__PlaceOtherResources			= ResourceGenerator.__PlaceOtherResources,
		__GetWaterOtherResources		= ResourceGenerator.__GetWaterOtherResources,
		__PlaceWaterOtherResources		= ResourceGenerator.__PlaceWaterOtherResources,
		__RemoveOtherDuplicateResources	= ResourceGenerator.__RemoveOtherDuplicateResources,
		__RemoveDuplicateResources		= ResourceGenerator.__RemoveDuplicateResources,
		__ScorePlots					= ResourceGenerator.__ScorePlots,
		__ScoreWaterPlots				= ResourceGenerator.__ScoreWaterPlots,

		-- data
		iWaterLux = args.iWaterLux or 3;
		iWaterBonus = args.iWaterBonus or 1.25;
		iLuxuriesPerRegion = args.LuxuriesPerRegion or 4;
		resources = args.resources;

		iResourcesInDB      = 0;
		iNumContinents		= 0;
		iTotalValidPlots    = 0;
		iWaterPlots = 0;
		iFrequencyTotal     = 0;
		iFrequencyTotalWater     = 0;
		iFrequencyStrategicTotal     = 0;
		iFrequencyStrategicTotalWater    = 0;
		iTargetPercentage   = 28;
		iStandardPercentage = 28;
		iLuxuryPercentage   = 20;
		iStrategicPercentage   = 18;
		iOccurencesPerFrequency = 0;
		iNumWaterLuxuries = 0;
		iNumWaterStrategics = 0;
		bOdd = false;
		eResourceType		= {},
		eResourceClassType	= {},
		iFrequency          = {},
		iSeaFrequency          = {},
		aLuxuryType		= {},
		aLuxuryTypeCoast		= {},
		aStrategicType		= {},
		aOtherType		= {},
		aOtherTypeWater		= {},
		aStrategicTypeCoast = {},
		aIndex = {},
		aaPossibleLuxLocs		= {},
		aaPossibleStratLocs		= {},
		aaPossibleLocs		= {},
		aaPossibleWaterLocs		= {},
		aResourcePlacementOrderStrategic = {},
		aResourcePlacementOrder = {},
		aWaterResourcePlacementOrder = {},
		aPeakEra = {},
	};
	
	-- initialize instance data
	instance:__InitResourceData()
	
	-- Chooses and then places the land luxury resources
	instance:__GetLuxuryResources()

	-- Chooses and then places the water luxury resources
	instance:__GetWaterLuxuryResources()

	-- Chooses and then places the land strategic resources
	instance:__GetStrategicResources()

	-- Chooses and then places the water strategic resources
	instance:__GetWaterStrategicResources()

	-- Chooses and then places the other resources [other is now only bonus, but later could be resource types added through mods]
	instance:__GetOtherResources()

	-- Chooses and then places the water other resources [other is now only bonus, but later could be resource types added through mods]
	instance:__GetWaterOtherResources()

	-- Removes too many adjacent other resources.
	instance:__RemoveOtherDuplicateResources()

	return instance;
end

function PlotCanHaveResource(plot)
	-- A simple test for land resource placement.
	-- Much easier than overriding terrain-resource mappings.

	if (plot:IsWater() or plot:IsMountain()) then
		-- Avoid certain plots.
		return false;
	end
	
	local terrainType = plot:GetTerrainType();
	
	if ((terrainType == g_TERRAIN_TYPE_SNOW) or 
		(terrainType == g_TERRAIN_TYPE_SNOW_HILLS)) then
		-- Avoid snow.
		return false;
	end
	
	local featureType = plot:GetFeatureType();
	
	if ((featureType == g_FEATURE_NONE) or
	    (featureType == g_FEATURE_FLOODPLAINS) or
	    (featureType == g_FEATURE_JUNGLE) or
	    (featureType == g_FEATURE_FOREST) or
	    (featureType == g_FEATURE_MARSH)) then
		-- Allow only these features.
		-- Avoid ice and oasis.
		-- Avoid natural wonders, which are special feature types.
		return true;
	end

	return false;
end

function GotLakes_ResourceGenerator:__Custom_ValidLuxuryPlots(eContinent)
	-- Custom logic to allow more luxuries on non-normal climates.
	-- go through each plot on the continent and put the luxuries	
	local iSize = #self.aLuxuryType;
	local iBaseScore = 1;
	self.iTotalValidPlots = 0;
	
	local avoidIce = ScriptData.climateInfo.resourcesAvoidIce;

	plots = Map.GetContinentPlots(eContinent);
	for i, plot in ipairs(plots) do
	
		local bCanHaveSomeResource = false;
		local pPlot = Map.GetPlotByIndex(plot);

		if(pPlot~=nil and pPlot:IsWater() == false) then
		
			-- See which resources can appear here
			for iI = 1, iSize do
				local bIce = false;

				if(IsAdjacentToIce(pPlot:GetX(), pPlot:GetY()) == true) then
					bIce = avoidIce;
				end
				
				-- Relax constraints on resource placement.
				if (PlotCanHaveResource(pPlot) and (bIce == false)) then
					row = {};
					row.MapIndex = plot;
					row.Score = iBaseScore;

					table.insert (self.aaPossibleLuxLocs[self.aLuxuryType[iI]], row);
					bCanHaveSomeResource = true;
				end
			end


			if (bCanHaveSomeResource == true) then
				self.iTotalValidPlots = self.iTotalValidPlots + 1;
			end
		end
		-- Compute how many of each resource to place
	end

	self.iOccurencesPerFrequency = self.iTargetPercentage / 100 * #plots * self.iLuxuryPercentage / 100 / self.iLuxuriesPerRegion;
end

function GotLakes_ResourceGenerator:__Custom_ScoreLuxuryPlots(iResourceIndex, eContinent)
	-- Custom logic to allow more luxuries on non-normal climates.
	-- Clear all earlier entries (some might not be valid if resources have been placed
	for k, v in pairs(self.aaPossibleLuxLocs[iResourceIndex]) do
		self.aaPossibleLuxLocs[iResourceIndex][k] = nil;
	end
	
	local avoidIce = ScriptData.climateInfo.resourcesAvoidIce;

	plots = Map.GetContinentPlots(eContinent);
	for i, plot in ipairs(plots) do
		local pPlot = Map.GetPlotByIndex(plot);
		local bIce = false;
		
		if(IsAdjacentToIce(pPlot:GetX(), pPlot:GetY()) == true) then
			bIce = avoidIce;
		end

		-- Relax constraints on resource placement.
		if (PlotCanHaveResource(pPlot) and bIce == false) then
			row = {};
			row.MapIndex = plot;
			row.Score = 500;
			row.Score = row.Score / ((ResourceBuilder.GetAdjacentResourceCount(pPlot) + 1) * 3.5);
			row.Score = row.Score + TerrainBuilder.GetRandomNumber(100, "Resource Placement Score Adjust");
			
			if(ResourceBuilder.GetAdjacentResourceCount(pPlot) <= 1 or #self.aaPossibleLuxLocs == 0) then
					table.insert (self.aaPossibleLuxLocs[iResourceIndex], row);
			end
		end
	end
end
------------------------------------------------------------------------------
-- Custom start positioner to handle civ starts on a variety of maps.
--
-- Based on AssignStartingPlots from Civ 5.
------------------------------------------------------------------------------
GotLakes_ASP = {}

function GotLakes_ASP.Create()
	-- There are three methods of dividing the map in to regions.
	-- OneLandmass, Continents, Oceanic. Default method is Continents.
	--
	-- Standard start plot finding uses a regional division method, then
	-- assigns one civ per region. Regions with lowest average fertility
	-- get their assignment first, to avoid the poor getting poorer.
	--
	-- Default methods for civ and city state placement both rely on having
	-- regional division data. If the desired process for a given map script
	-- would not define regions of this type, replace the start finder
	-- with your custom method.
	--
	-- Note that this operation relies on inclusion of the Mapmaker Utilities.

	-- Main data table ("self dot" table).
	--	
	local resourceSetting = UserInput.resources;
	local aspDivMethod    = ScriptData.landInfo.aspDivMethod;
	
	local findStarts = 
	{
		-- Core Process member methods
		__Init = GotLakes_ASP.__Init,
		ApplyHexAdjustment = GotLakes_ASP.ApplyHexAdjustment,
		GenerateRegions = GotLakes_ASP.GenerateRegions,
		ChooseLocations = GotLakes_ASP.ChooseLocations,
		BalanceAndAssign = GotLakes_ASP.BalanceAndAssign,
		PlaceNaturalWonders = GotLakes_ASP.PlaceNaturalWonders,
		PlaceResourcesAndCityStates = GotLakes_ASP.PlaceResourcesAndCityStates,
		
		-- Generate Regions member methods
		MeasureStartPlacementFertilityOfPlot = GotLakes_ASP.MeasureStartPlacementFertilityOfPlot,
		MeasureStartPlacementFertilityInRectangle = GotLakes_ASP.MeasureStartPlacementFertilityInRectangle,
		MeasureStartPlacementFertilityOfLandmass = GotLakes_ASP.MeasureStartPlacementFertilityOfLandmass,
		RemoveDeadRows = GotLakes_ASP.RemoveDeadRows,
		DivideIntoRegions = GotLakes_ASP.DivideIntoRegions,
		ChopIntoThreeRegions = GotLakes_ASP.ChopIntoThreeRegions,
		ChopIntoTwoRegions = GotLakes_ASP.ChopIntoTwoRegions,
		CustomOverride = GotLakes_ASP.CustomOverride,

		-- Choose Locations member methods
		MeasureTerrainInRegions = GotLakes_ASP.MeasureTerrainInRegions,
		DetermineRegionTypes = GotLakes_ASP.DetermineRegionTypes,
		PlaceImpactAndRipples = GotLakes_ASP.PlaceImpactAndRipples,
		MeasureSinglePlot = GotLakes_ASP.MeasureSinglePlot,
		EvaluateCandidatePlot = GotLakes_ASP.EvaluateCandidatePlot,
		IterateThroughCandidatePlotList = GotLakes_ASP.IterateThroughCandidatePlotList,
		FindStart = GotLakes_ASP.FindStart,
		FindCoastalStart = GotLakes_ASP.FindCoastalStart,
		FindStartWithoutRegardToAreaID = GotLakes_ASP.FindStartWithoutRegardToAreaID,
		
		-- Balance and Assign member methods
		AttemptToPlaceBonusResourceAtPlot = GotLakes_ASP.AttemptToPlaceBonusResourceAtPlot,
		AttemptToPlaceHillsAtPlot = GotLakes_ASP.AttemptToPlaceHillsAtPlot,
		AttemptToPlaceSmallStrategicAtPlot = GotLakes_ASP.AttemptToPlaceSmallStrategicAtPlot,
		FindFallbackForUnmatchedRegionPriority = GotLakes_ASP.FindFallbackForUnmatchedRegionPriority,
		AddStrategicBalanceResources = GotLakes_ASP.AddStrategicBalanceResources,
		AttemptToPlaceStoneAtGrassPlot = GotLakes_ASP.AttemptToPlaceStoneAtGrassPlot,
		NormalizeStartLocation = GotLakes_ASP.NormalizeStartLocation,
		NormalizeTeamLocations = GotLakes_ASP.NormalizeTeamLocations,

		-- Civ start position variables
		startingPlots = {},				-- Stores x and y coordinates (and "score") of starting plots for civs, indexed by region number
		resource_setting = resourceSetting,
		method = aspDivMethod,			-- Method of regional division, default is continental
		iNumCivs = 0,					-- Number of civs at game start
		plotDataIsCoastal = {},			-- Stores table of NextToSaltWater plots to reduce redundant calculations
		plotDataIsNextToCoast = {},		-- Stores table of TwoAwayFromSaltWater plots to reduce redundant calculations
		regionData = {},				-- Stores data returned from regional division algorithm
		regionTerrainCounts = {},		-- Stores counts of terrain elements for all regions
		regionTypes = {},				-- Stores region types
		distanceData = table.fill(0, g_iW * g_iH), -- Stores "impact and ripple" data of start points as each is placed
		startLocationConditions = {},   -- Stores info regarding conditions at each start location
				
		-- Rectangular Division, dimensions within which all regions reside. (Unused by the other methods)
		inhabited_WestX,
		inhabited_SouthY,
		inhabited_Width,
		inhabited_Height,
		
		-- Positioner defaults. These are the controls for the "Center Bias" placement method for civ starts in regions.
		centerBias = 34, -- % of radius from region center to examine first
		middleBias = 67, -- % of radius from region center to check second
		minFoodInner = 1,
		minProdInner = 0,
		minGoodInner = 3,
		minFoodMiddle = 4,
		minProdMiddle = 0,
		minGoodMiddle = 6,
		minFoodOuter = 4,
		minProdOuter = 2,
		minGoodOuter = 8,
		maxJunk = 1000000, -- 9

		-- Hex Adjustment tables. These tables direct plot by plot scans in a radius 
		-- around a center hex, starting to Northeast, moving clockwise.
		firstRingYIsEven = {{0, 1}, {1, 0}, {0, -1}, {-1, -1}, {-1, 0}, {-1, 1}},
		secondRingYIsEven = {
		{1, 2}, {1, 1}, {2, 0}, {1, -1}, {1, -2}, {0, -2},
		{-1, -2}, {-2, -1}, {-2, 0}, {-2, 1}, {-1, 2}, {0, 2}
		},
		thirdRingYIsEven = {
		{1, 3}, {2, 2}, {2, 1}, {3, 0}, {2, -1}, {2, -2},
		{1, -3}, {0, -3}, {-1, -3}, {-2, -3}, {-2, -2}, {-3, -1},
		{-3, 0}, {-3, 1}, {-2, 2}, {-2, 3}, {-1, 3}, {0, 3}
		},
		firstRingYIsOdd = {{1, 1}, {1, 0}, {1, -1}, {0, -1}, {-1, 0}, {0, 1}},
		secondRingYIsOdd = {		
		{1, 2}, {2, 1}, {2, 0}, {2, -1}, {1, -2}, {0, -2},
		{-1, -2}, {-1, -1}, {-2, 0}, {-1, 1}, {-1, 2}, {0, 2}
		},
		thirdRingYIsOdd = {		
		{2, 3}, {2, 2}, {3, 1}, {3, 0}, {3, -1}, {2, -2},
		{2, -3}, {1, -3}, {0, -3}, {-1, -3}, {-2, -2}, {-2, -1},
		{-3, 0}, {-2, 1}, {-2, 2}, {-1, 3}, {0, 3}, {1, 3}
		},
		-- Direction types table, another method of handling hex adjustments, in combination with Map.PlotDirection()
		direction_types = 
		{
			DirectionTypes.DIRECTION_NORTHEAST,
			DirectionTypes.DIRECTION_EAST,
			DirectionTypes.DIRECTION_SOUTHEAST,
			DirectionTypes.DIRECTION_SOUTHWEST,
			DirectionTypes.DIRECTION_WEST,
			DirectionTypes.DIRECTION_NORTHWEST
		},
	}
	
	findStarts:__Init()
	
	return findStarts
end

function GotLakes_ASP:__Init()
	-- Set up data tables that record whether a plot is coastal land and whether a plot is adjacent to coastal land.
	self.plotDataIsCoastal, self.plotDataIsNextToCoast = GenerateNextToCoastalLandDataTables()
end

function GotLakes_ASP:ApplyHexAdjustment(x, y, plot_adjustments)
	-- Used this bit of code so many times, I had to make it a function.
	local adjusted_x, adjusted_y;
	if ScriptData.wrapX == true then
		adjusted_x = (x + plot_adjustments[1]) % g_iW;
	else
		adjusted_x = x + plot_adjustments[1];
	end
	if ScriptData.wrapY == true then
		adjusted_y = (y + plot_adjustments[2]) % g_iH;
	else
		adjusted_y = y + plot_adjustments[2];
	end
	return adjusted_x, adjusted_y;
end
------------------------------------------------------------------------------
-- Start of functions tied to GenerateRegions()
------------------------------------------------------------------------------
function GotLakes_ASP:MeasureStartPlacementFertilityOfPlot(x, y, checkForCoastalLand)
	-- Fertility of plots is used to divide continents or areas in to Regions.
	-- Regions are used to assign starting plots and place some resources.
	-- Usage: x, y are plot coords, with 0,0 in SW. The check is a boolean.
	--
	-- Mountain, Oasis, FloodPlain tiles = -2, 5, 6 and do not count anything else.
	--   Rest of the tiles add up values of tile traits.
	--   Terrain: Grass 4, Plains 3, Tundra 2, Coast 2, Desert 1, Snow -1
	--   Features: Hill 1, Forest 1, FreshWater 1, River 1, Jungle -1, Marsh 1, Ice -1
	--   We want players who start in Grass to have the least room to expand.

	local plot = Map.GetPlot(x, y);
	local plotFertility = 0;
	local terrainType = plot:GetTerrainType();
	local featureType = plot:GetFeatureType();
	-- Measure Fertility -- Any cases absent from the process have a 0 value.
	if plot:IsMountain() then -- Note, mountains cannot belong to a landmass AreaID, so they usually go unmeasured.
		plotFertility = -2;
	elseif IsBaseTerrainType(g_TERRAIN_TYPE_SNOW, terrainType) then
		plotFertility = -1;
	elseif (featureType == g_FEATURE_OASIS) then
		plotFertility = 4; -- Reducing Oasis value slightly. -1/26/2011 BT
	elseif (featureType == g_FEATURE_FLOOD_PLAINS) then
		plotFertility = 5; -- Reducing Flood Plains value slightly. -1/26/2011 BT
	else
		if IsBaseTerrainType(g_TERRAIN_TYPE_GRASS, terrainType) then -- Reversing values for Grass and Plains. -1/26/2011 BT
			plotFertility = 3;
		elseif IsBaseTerrainType(g_TERRAIN_TYPE_PLAINS, terrainType) then
			plotFertility = 4;
		elseif IsBaseTerrainType(g_TERRAIN_TYPE_TUNDRA, terrainType) then
			plotFertility = 2;
		elseif (terrainType == g_TERRAIN_TYPE_COAST) then
			plotFertility = 2;
		elseif IsBaseTerrainType(g_TERRAIN_TYPE_DESERT, terrainType) then
			plotFertility = 1;
		end
		if plot:IsHills() then
			plotFertility = plotFertility + 1;
		end
		if featureType == g_FEATURE_FOREST then
			plotFertility = plotFertility + 0; -- Removing forest bonus as a balance tweak. -1/26/2011 BT
		elseif featureType == g_FEATURE_JUNGLE then
			plotFertility = plotFertility - 1;
		elseif featureType == g_FEATURE_MARSH then
			plotFertility = plotFertility + 1; -- Marsh plots add food in Civ 6
		elseif featureType == g_FEATURE_ICE then
			plotFertility = plotFertility - 1;
		end
		if plot:IsRiver() then
			plotFertility = plotFertility + 1;
		end
		if plot:IsWater() and plot:IsLake() then
			-- Fresh water
			plotFertility = plotFertility + 1;
		end
		if checkForCoastalLand == true then 
			-- When measuring only one AreaID, this shortcut helps account for coastal plots not measured.
			if plot:IsCoastalLand() then
				plotFertility = plotFertility + 2;
			end
		end
	end

	return plotFertility
end

function GotLakes_ASP:MeasureStartPlacementFertilityInRectangle(iWestX, iSouthY, iWidth, iHeight)
	-- This function is designed to provide initial data for regional division recursion.
	-- Loop through plots in this rectangle and measure Fertility Rating.
	-- Results will include a data table of all measured plots.
	local areaFertilityTable = {};
	local areaFertilityCount = 0;
	local plotCount = iWidth * iHeight;
	-- When generating a plot data table incrementally, process Y first so that plots go row by row.
	for y = iSouthY, iSouthY + iHeight - 1 do 
		for x = iWestX, iWestX + iWidth - 1 do
			local plotFertility = self:MeasureStartPlacementFertilityOfPlot(x, y, false); -- Check for coastal land is disabled.
			table.insert(areaFertilityTable, plotFertility);
			areaFertilityCount = areaFertilityCount + plotFertility;
		end
	end

	-- Returns table, integer, integer.
	return areaFertilityTable, areaFertilityCount, plotCount
end

function GotLakes_ASP:MeasureStartPlacementFertilityOfLandmass(iAreaID, iWestX, iEastX, iSouthY, iNorthY, wrapsX, wrapsY)
	-- This function is designed to provide initial data for regional division recursion.
	-- Loop through plots in this landmass and measure Fertility Rating.
	-- Results will include a data table of all plots within the rectangle that includes the entirety of this landmass.
	--
	-- This function will account for any wrapping around the world this landmass may do.
	local xEnd, yEnd; --[[ These coordinates will be used in case of wrapping landmass, 
	                       extending the landmass "off the map", in to imaginary space 
	                       to process it. Modulo math will correct the coordinates for 
	                       accessing the plot data array. ]]--
	if wrapsX then
		xEnd = iEastX + g_iW;
	else
		xEnd = iEastX;
	end
	if wrapsY then
		yEnd = iNorthY + g_iH;
	else
		yEnd = iNorthY;
	end
	--
	local areaFertilityTable = {};
	local areaFertilityCount = 0;
	local plotCount = 0;
	for yLoop = iSouthY, yEnd do -- When generating a plot data table incrementally, process Y first so that plots go row by row.
		for xLoop = iWestX, xEnd do
			plotCount = plotCount + 1;
			local x = xLoop % g_iW;
			local y = yLoop % g_iH;
			local plot = Map.GetPlot(x, y);
			local thisPlotsArea =plot:GetArea():GetID();
			if thisPlotsArea ~= iAreaID then -- This plot is not a member of the landmass, set value to 0
				table.insert(areaFertilityTable, 0);
			else -- This plot is a member, process it.
				local plotFertility = self:MeasureStartPlacementFertilityOfPlot(x, y, true); -- Check for coastal land is enabled.
				table.insert(areaFertilityTable, plotFertility);
				areaFertilityCount = areaFertilityCount + plotFertility;
			end
		end
	end
	
	-- Note: The table accounts for world wrap, so make sure to translate its index correctly.
	-- Plots in the table run from the southwest corner along the bottom row, then upward row by row, per normal plot data indexing.
	return areaFertilityTable, areaFertilityCount, plotCount
end

function GotLakes_ASP:RemoveDeadRows(fertility_table, iWestX, iSouthY, iWidth, iHeight)
	-- Any outside rows in the fertility table of a just-divided region that 
	-- contains all zeroes can be safely removed.
	-- This will improve the accuracy of operations involving any applicable region.
	local adjusted_table = {};
	local adjusted_WestX;
	local adjusted_SouthY
	local adjusted_Width
	local adjusted_Height;
	
	-- Check for rows to remove on the bottom.
	local adjustSouth = 0;
	for y = 0, iHeight - 1 do
		local bKeepThisRow = false;
		for x = 0, iWidth - 1 do
			local i = y * iWidth + x + 1;
			if fertility_table[i] ~= 0 then
				bKeepThisRow = true;
				break
			end
		end
		if bKeepThisRow == true then
			break
		else
			adjustSouth = adjustSouth + 1;
		end
	end

	-- Check for rows to remove on the top.
	local adjustNorth = 0;
	for y = iHeight - 1, 0, -1 do
		local bKeepThisRow = false;
		for x = 0, iWidth - 1 do
			local i = y * iWidth + x + 1;
			if fertility_table[i] ~= 0 then
				bKeepThisRow = true;
				break
			end
		end
		if bKeepThisRow == true then
			break
		else
			adjustNorth = adjustNorth + 1;
		end
	end

	-- Check for columns to remove on the left.
	local adjustWest = 0;
	for x = 0, iWidth - 1 do
		local bKeepThisColumn = false;
		for y = 0, iHeight - 1 do
			local i = y * iWidth + x + 1;
			if fertility_table[i] ~= 0 then
				bKeepThisColumn = true;
				break
			end
		end
		if bKeepThisColumn == true then
			break
		else
			adjustWest = adjustWest + 1;
		end
	end

	-- Check for columns to remove on the right.
	local adjustEast = 0;
	for x = iWidth - 1, 0, -1 do
		local bKeepThisColumn = false;
		for y = 0, iHeight - 1 do
			local i = y * iWidth + x + 1;
			if fertility_table[i] ~= 0 then
				bKeepThisColumn = true;
				break
			end
		end
		if bKeepThisColumn == true then
			break
		else
			adjustEast = adjustEast + 1;
		end
	end

	if adjustSouth > 0 or adjustNorth > 0 or adjustWest > 0 or adjustEast > 0 then
		-- Truncate this region to remove dead rows.
		adjusted_WestX = (iWestX + adjustWest) % g_iW;
		adjusted_SouthY = (iSouthY + adjustSouth) % g_iH;
		adjusted_Width = (iWidth - adjustWest) - adjustEast;
		adjusted_Height = (iHeight - adjustSouth) - adjustNorth;
		-- Reconstruct fertility table. This must be done row by row, so process Y coord first.
		for y = 0, adjusted_Height - 1 do
			for x = 0, adjusted_Width - 1 do
				local i = (y + adjustSouth) * iWidth + (x + adjustWest) + 1;
				local plotFert = fertility_table[i];
				table.insert(adjusted_table, plotFert);
			end
		end
		--[[
		print("-");
		print("Removed Dead Rows, West: ", adjustWest, " East: ", adjustEast);
		print("Removed Dead Rows, South: ", adjustSouth, " North: ", adjustNorth);
		print("-");
		print("Incoming values: ", iWestX, iSouthY, iWidth, iHeight);
		print("Outgoing values: ", adjusted_WestX, adjusted_SouthY, adjusted_Width, adjusted_Height);
		print("-");
		local incoming_index = table.maxn(fertility_table);
		local outgoing_index = table.maxn(adjusted_table);
		print("Size of incoming fertility table: ", incoming_index);
		print("Size of outgoing fertility table: ", outgoing_index);
		--]]
		return adjusted_table, adjusted_WestX, adjusted_SouthY, adjusted_Width, adjusted_Height;
	
	else -- Region not adjusted, return original values unaltered.
		return fertility_table, iWestX, iSouthY, iWidth, iHeight;
	end
end

function GotLakes_ASP:DivideIntoRegions(iNumDivisions, fertility_table, rectangle_data_table)
	-- This is a recursive algorithm. (Original concept and implementation by Ed Beach).
	--
	-- Fertility table is a plot data array including data for all plots to be processed here.
	-- The fertility table is obtained as part of the MeasureFertility functions, or via division during the recursion.
	--
	-- Rectangle table includes seven data fields:
	-- westX, southY, width, height, AreaID, fertilityCount, plotCount
	--
	-- If AreaID is -1, it means the rectangle contains fertility data from all plots regardless of their AreaIDs.
	-- The plotCount is an absolute count of plots within the rectangle, without regard to AreaID membership.
	-- This is going to purposely reduce average fertility per plot for Order-of-Assignment priority.
	-- Rectangles with a lot of non-member plots will tend to be misshapen and need to be on the favorable side of minDistance elements.
	-- print("-"); print("DivideIntoRegions called.");

	--[[ Log dump of incoming table data. Activate for debug only.
	print("Data tables passed to DivideIntoRegions.");
	PrintContentsOfTable(fertility_table)
	PrintContentsOfTable(rectangle_data_table)
	print("End of this instance, DivideIntoRegions tables.");
	]]--
	
	local iNumDivides = 0;
	local iSubdivisions = 0;
	local bPrimeGreaterThanThree = false;
	local firstSubdivisions = 0;
	local laterSubdivisions = 0;

	-- If this rectangle is not to be divided, break recursion and record the data.
	if (iNumDivisions == 1) then -- This area is to be defined as a Region.
		-- Expand rectangle table to include an eighth field for average fertility per plot.
		local fAverageFertility = rectangle_data_table[6] / rectangle_data_table[7]; -- fertilityCount/plotCount
		table.insert(rectangle_data_table, fAverageFertility);
		-- Insert this record in to the instance data for start placement regions for this game.
		-- (This is the crux of the entire regional definition process, determining an actual region.)
		table.insert(self.regionData, rectangle_data_table);
		--[[
		local iNumberOfThisRegion = table.maxn(self.regionData);
		print("-");
		print("---------------------------------------------");
		print("Defined location of Start Region #", iNumberOfThisRegion);
		print("---------------------------------------------");
		print("-");
		--]]
		return

	--[[ Divide this rectangle into iNumDivisions worth of subdivisions, then send each
	     subdivision back through this function in a recursive loop. ]]--
	elseif (iNumDivisions > 1) then
		-- See if region is taller or wider.
		local iWidth  = Round(rectangle_data_table[3]);
		local iHeight = Round(rectangle_data_table[4]);
		local bTaller = false;
		if iHeight > iWidth then
			bTaller = true;
		end

		-- If the number of divisions is 2 or 3, no further subdivision is required.
		if (iNumDivisions == 2) then
			iNumDivides = 2;
			iSubdivisions = 1;
		elseif (iNumDivisions == 3) then
			iNumDivides = 3;
			iSubdivisions = 1;
		
		-- If the number of divisions is greater than 3 and a prime number,
		-- divide all of these cases in to an odd plus an even number, then subdivide.
		--
		--[[ Ed's original algorithm skipped this step and produced "extra" divisions,
		     which I would have had to account for. I decided it was far far easier to
		     improve the algorithm and remove all extra divisions than it was to have
		     to write large chunks of code trying to process empty regions. Not to 
		     mention the added precision of using all land on the continent or map to
		     determine where to place major civilizations.  - Bob Thomas, April 2010 ]]--
		elseif (iNumDivisions == 5) then
			bPrimeGreaterThanThree = true;
			chopPercent = 59.2; -- These chopPercents are all set to undershoot slightly, averaging out the actual result closer to target.
			firstSubdivisions = 3; -- This is because if you aim for the exact target, there is never undershoot and almost always overshoot.
			laterSubdivisions = 2; -- So a well calibrated target ends up compensating for that overshoot factor, to improve total fairness.
		elseif (iNumDivisions == 7) then
			bPrimeGreaterThanThree = true;
			chopPercent = 42.2;
			firstSubdivisions = 3;
			laterSubdivisions = 4;
		elseif (iNumDivisions == 11) then
			bPrimeGreaterThanThree = true;
			chopPercent = 27;
			firstSubdivisions = 3;
			laterSubdivisions = 8;
		elseif (iNumDivisions == 13) then
			bPrimeGreaterThanThree = true;
			chopPercent = 38.1;
			firstSubdivisions = 5;
			laterSubdivisions = 8;
		elseif (iNumDivisions == 17) then
			bPrimeGreaterThanThree = true;
			chopPercent = 52.8;
			firstSubdivisions = 9;
			laterSubdivisions = 8;
		elseif (iNumDivisions == 19) then
			bPrimeGreaterThanThree = true;
			chopPercent = 36.7;
			firstSubdivisions = 7;
			laterSubdivisions = 12;

		-- If the number of divisions is greater than 3 and not a prime number,
		-- then chop this rectangle in to 2 or 3 parts and subdivide those.
		elseif (iNumDivisions == 4) then
			iNumDivides = 2;
			iSubdivisions = 2;
		elseif (iNumDivisions == 6) then
			iNumDivides = 3;
			iSubdivisions = 2;
		elseif (iNumDivisions == 8) then
			iNumDivides = 2;
			iSubdivisions = 4;
		elseif (iNumDivisions == 9) then
			iNumDivides = 3;
			iSubdivisions = 3;
		elseif (iNumDivisions == 10) then
			iNumDivides = 2;
			iSubdivisions = 5;
		elseif (iNumDivisions == 12) then
			iNumDivides = 3;
			iSubdivisions = 4;
		elseif (iNumDivisions == 14) then
			iNumDivides = 2;
			iSubdivisions = 7;
		elseif (iNumDivisions == 15) then
			iNumDivides = 3;
			iSubdivisions = 5;
		elseif (iNumDivisions == 16) then
			iNumDivides = 2;
			iSubdivisions = 8;
		elseif (iNumDivisions == 18) then
			iNumDivides = 3;
			iSubdivisions = 6;
		elseif (iNumDivisions == 20) then
			iNumDivides = 2;
			iSubdivisions = 10;
		elseif (iNumDivisions == 21) then
			iNumDivides = 3;
			iSubdivisions = 7;
		elseif (iNumDivisions == 22) then
			iNumDivides = 2;
			iSubdivisions = 11;
		else
			print("Erroneous number of regional divisions : ", iNumDivisions);
		end

		-- Now process the division via one of the three methods.
		-- All methods involve recursion, to obtain the best manner of subdividing each rectangle involved.
		if bPrimeGreaterThanThree then
			--print("DivideIntoRegions: Uneven Division for handling prime numbers selected.");
			local results = self:ChopIntoTwoRegions(fertility_table, rectangle_data_table, bTaller, chopPercent);
			local first_section_fertility_table = results[1];
			local first_section_data_table = results[2];
			local second_section_fertility_table = results[3];
			local second_section_data_table = results[4];
			--
			self:DivideIntoRegions(firstSubdivisions, first_section_fertility_table, first_section_data_table)
			self:DivideIntoRegions(laterSubdivisions, second_section_fertility_table, second_section_data_table)

		else
			if (iNumDivides == 2) then
				--print("DivideIntoRegions: Divide in to Halves selected.");
				local results = self:ChopIntoTwoRegions(fertility_table, rectangle_data_table, bTaller, 49.5); -- Undershoot by design, to compensate for inevitable overshoot. Gets the actual result closer to target.
				local first_section_fertility_table = results[1];
				local first_section_data_table = results[2];
				local second_section_fertility_table = results[3];
				local second_section_data_table = results[4];
				--
				self:DivideIntoRegions(iSubdivisions, first_section_fertility_table, first_section_data_table)
				self:DivideIntoRegions(iSubdivisions, second_section_fertility_table, second_section_data_table)

			elseif (iNumDivides == 3) then
				--print("DivideIntoRegions: Divide in to Thirds selected.");
				local results = self:ChopIntoThreeRegions(fertility_table, rectangle_data_table, bTaller);
				local first_section_fertility_table = results[1];
				local first_section_data_table = results[2];
				local second_section_fertility_table = results[3];
				local second_section_data_table = results[4];
				local third_section_fertility_table = results[5];
				local third_section_data_table = results[6];
				--
				self:DivideIntoRegions(iSubdivisions, first_section_fertility_table, first_section_data_table)
				self:DivideIntoRegions(iSubdivisions, second_section_fertility_table, second_section_data_table)
				self:DivideIntoRegions(iSubdivisions, third_section_fertility_table, third_section_data_table)

			else
				print("Invalid iNumDivides value (from DivideIntoRegions): must be 2 or 3.");
			end
		end
	end
end

function GotLakes_ASP:ChopIntoThreeRegions(fertility_table, rectangle_data_table, bTaller, chopPercent)
	--print("-"); print("ChopIntoThree called.");
	-- Performs the mechanics of dividing a region into three roughly equal fertility subregions.
	local results = {};

	-- Chop off the first third.
	local initial_results = self:ChopIntoTwoRegions(fertility_table, rectangle_data_table, bTaller, 33); -- Undershoot by a bit, tends to make the actual result closer to accurate.
	-- add first subdivision to results
	local temptable = initial_results[1];
	table.insert(results, temptable); 

	--[[ Activate table printouts for debug purposes only, then deactivate when done. ]]--
	--print("Data returned to ChopIntoThree from ChopIntoTwo.");
	--PrintContentsOfTable(temptable)

	local temptable = initial_results[2];
	table.insert(results, temptable);

	--PrintContentsOfTable(temptable)

	-- Prepare the remainder for further processing.
	local second_section_fertility_table = initial_results[3]; 

	--PrintContentsOfTable(second_section_fertility_table)

	local second_section_data_table = initial_results[4];

	--PrintContentsOfTable(second_section_data_table)
	--print("End of this instance, ChopIntoThree tables.");

	-- See if this piece is taller or wider. (Ed's original implementation skipped this step).
	local bTallerForRemainder = false;
	local width = second_section_data_table[3];
	local height = second_section_data_table[4];
	if height > width then
		bTallerForRemainder = true;
	end

	-- Chop the bigger piece in half.		
	local interim_results = self:ChopIntoTwoRegions(second_section_fertility_table, second_section_data_table, bTallerForRemainder, 48.5); -- Undershoot just a little.
	table.insert(results, interim_results[1]); 
	table.insert(results, interim_results[2]); 
	table.insert(results, interim_results[3]); 
	table.insert(results, interim_results[4]); 

	--[[ Returns a table of six entries, each of which is a nested table.
	1: fertility_table of first subdivision
	2: rectangle_data_table of first subdivision.
	3: fertility_table of second subdivision
	4: rectangle_data_table of second subdivision.
	5: fertility_table of third subdivision
	6: rectangle_data_table of third subdivision.  ]]--
	return results
end

function GotLakes_ASP:ChopIntoTwoRegions(fertility_table, rectangle_data_table, bTaller, chopPercent)
	-- Performs the mechanics of dividing a region into two subregions.
	--
	-- Fertility table is a plot data array including data for all plots to be processed here.
	-- This data already factors any need for processing AreaID.
	--
	-- Rectangle table includes seven data fields:
	-- westX, southY, width, height, AreaID, fertilityCount, plotCount
	--print("-"); print("ChopIntoTwo called.");

	--[[ Log dump of incoming table data. Activate for debug only.
	print("Data tables passed to ChopIntoTwoRegions.");
	PrintContentsOfTable(fertility_table)
	PrintContentsOfTable(rectangle_data_table)
	print("End of this instance, ChopIntoTwoRegions tables.");
	--]]--

	-- Read the incoming data table.
	local iWestX      = Round(rectangle_data_table[1]);
	local iSouthY     = Round(rectangle_data_table[2]);
	local iRectWidth  = Round(rectangle_data_table[3]);
	local iRectHeight = Round(rectangle_data_table[4]);
	local iAreaID = rectangle_data_table[5];
	local iTargetFertility = rectangle_data_table[6] * chopPercent / 100;
	
	-- Now divide the region.
	--
	-- West and South edges remain the same for first region.
	local firstRegionWestX = iWestX;
	local firstRegionSouthY = iSouthY;
	-- scope variables that get decided conditionally.
	local firstRegionWidth, firstRegionHeight;
	local secondRegionWestX, secondRegionSouthY, secondRegionWidth, secondRegionHeight;
	local iFirstRegionFertility = 0;
	local iSecondRegionFertility = 0;
	local region_one_fertility = {};
	local region_two_fertility = {};

	if (bTaller) then -- We will divide horizontally, resulting in first region on bottom, second on top.
		--
		-- Width for both will remain the same as the parent rectangle.
		firstRegionWidth = iRectWidth;
		secondRegionWestX = iWestX;
		secondRegionWidth = iRectWidth;

		-- Measure one row at a time, moving up from bottom, until we have exceeded the target fertility.
		local reachedTargetRow = false;
		local rectY = 0;
		while reachedTargetRow == false do
			-- Process the next row in line.
			for rectX = 0, iRectWidth - 1 do
				local fertIndex = rectY * iRectWidth + rectX + 1;
				local plotFertility = fertility_table[fertIndex];
				-- Add this plot's fertility to the region total so far.
				iFirstRegionFertility = iFirstRegionFertility + plotFertility;
				-- Record this plot in a new fertility table. (Needed for further subdivisions).
				-- Note, building this plot data table incrementally, so it must go row by row.
				table.insert(region_one_fertility, plotFertility);
			end
			if iFirstRegionFertility >= iTargetFertility then
				-- This row has completed the region.
				firstRegionHeight = rectY + 1;
				secondRegionSouthY = (iSouthY + rectY + 1) % g_iH;
				secondRegionHeight = iRectHeight - firstRegionHeight;
				reachedTargetRow = true;
				break
			else
				rectY = rectY + 1;
			end
		end
		
		-- Debug printout of division location.
		--print("Dividing along horizontal line between rows: ", secondRegionSouthY - 1, "-", secondRegionSouthY);
		
		-- Create the fertility table for the second region, the one on top.
		-- Data must be added row by row, to keep the table index behavior consistent.
		for rectY = firstRegionHeight, iRectHeight - 1 do
			for rectX = 0, iRectWidth - 1 do
				local fertIndex = rectY * iRectWidth + rectX + 1;
				local plotFertility = fertility_table[fertIndex];
				-- Add this plot's fertility to the region total so far.
				iSecondRegionFertility = iSecondRegionFertility + plotFertility;
				-- Record this plot in a new fertility table. (Needed for further subdivisions).
				-- Note, building this plot data table incrementally, so it must go row by row.
				table.insert(region_two_fertility, plotFertility);
			end
		end
				
	else -- We will divide vertically, resulting in first region on left, second on right.
		--
		-- Height for both will remain the same as the parent rectangle.
		firstRegionHeight = iRectHeight;
		secondRegionSouthY = iSouthY;
		secondRegionHeight = iRectHeight;
		
		--[[ First region's new fertility table will be a little tricky. We don't know how many 
		     table entries it will need beforehand, and we cannot add the entries sequentially
		     when the data is being generated column by column, yet the table index needs to 
		     proceed row by row. So we will have to make a second pass.  ]]--

		-- Measure one column at a time, moving left to right, until we have exceeded the target fertility.
		local reachedTargetColumn = false;
		local rectX = 0;
		while reachedTargetColumn == false do
			-- Process the next column in line.
			for rectY = 0, iRectHeight - 1 do
				local fertIndex = rectY * iRectWidth + rectX + 1;
				local plotFertility = fertility_table[fertIndex];
				-- Add this plot's fertility to the region total so far.
				iFirstRegionFertility = iFirstRegionFertility + plotFertility;
				-- No table record here, handle later row by row.
			end
			if iFirstRegionFertility >= iTargetFertility then
				-- This column has completed the region.
				firstRegionWidth = rectX + 1;
				secondRegionWestX = (iWestX + rectX + 1) % g_iW;
				secondRegionWidth = iRectWidth - firstRegionWidth;
				reachedTargetColumn = true;
				break
			else
				rectX = rectX + 1;
			end
		end

		-- Debug printout of division location.
		--print("Dividing along vertical line between columns: ", secondRegionWestX - 1, "-", secondRegionWestX);

		-- Create the fertility table for the second region, the one on the right.
		-- Data must be added row by row, to keep the table index behavior consistent.
		for rectY = 0, iRectHeight - 1 do
			for rectX = firstRegionWidth, iRectWidth - 1 do
				local fertIndex = rectY * iRectWidth + rectX + 1;
				local plotFertility = fertility_table[fertIndex];
				-- Add this plot's fertility to the region total so far.
				iSecondRegionFertility = iSecondRegionFertility + plotFertility;
				-- Record this plot in a new fertility table. (Needed for further subdivisions).
				-- Note, building this plot data table incrementally, so it must go row by row.
				table.insert(region_two_fertility, plotFertility);
			end
		end
		-- Now create the fertility table for the first region.
		for rectY = 0, iRectHeight - 1 do
			for rectX = 0, firstRegionWidth - 1 do
				local fertIndex = rectY * iRectWidth + rectX + 1;
				local plotFertility = fertility_table[fertIndex];
				table.insert(region_one_fertility, plotFertility);
			end
		end
	end
	
	-- Now check the newly divided regions for dead rows (all zero values) along
	-- the edges and remove any found.
	--
	-- First region
	local FRFertT, FRWX, FRSY, FRWid, FRHei;
	FRFertT, FRWX, FRSY, FRWid, FRHei = self:RemoveDeadRows(region_one_fertility,
		firstRegionWestX, firstRegionSouthY, firstRegionWidth, firstRegionHeight);
	--
	-- Second region
	local SRFertT, SRWX, SRSY, SRWid, SRHei;
	SRFertT, SRWX, SRSY, SRWid, SRHei = self:RemoveDeadRows(region_two_fertility,
		secondRegionWestX, secondRegionSouthY, secondRegionWidth, secondRegionHeight);
	--
	
	-- Generate the data tables that record the location of the new subdivisions.
	local firstPlots = FRWid * FRHei;
	local secondPlots = SRWid * SRHei;
	local region_one_data = {FRWX, FRSY, FRWid, FRHei, iAreaID, iFirstRegionFertility, firstPlots};
	local region_two_data = {SRWX, SRSY, SRWid, SRHei, iAreaID, iSecondRegionFertility, secondPlots};
	-- Generate the final data.
	local outcome = {FRFertT, region_one_data, SRFertT, region_two_data};
	return outcome
end

function GotLakes_ASP:GenerateRegions()
	print("GotLakes ASP: Dividing the map into Regions");
	-- This function stores its data in the instance (self) data table.
	--
	-- The "Three Methods" of regional division:
	-- 1. Biggest Landmass: All civs start on the biggest landmass.
	-- 2. Continental: Civs are assigned to continents. Any continents with more than one civ are divided.
	-- 3. Rectangular: Civs start within a given rectangle that spans the whole map, without regard to landmass sizes.
	--                 This method is primarily applied to Archipelago and other maps with lots of tiny islands.
	-- 4. Rectangular: Civs start within a given rectangle defined by arguments passed in on the function call.
	--                 Arguments required for this method: iWestX, iSouthY, iWidth, iHeight

	-- Determine number of civilizations present in this game.
	self.iNumCivs = PlayerManager.GetAliveMajorsCount();
	print("Civs: ", self.iNumCivs);

	if self.method == 1 then -- Biggest Landmass
		print("Division Method: biggest landmass");
		-- Identify the biggest landmass.
		local biggest_area = Areas.FindBiggestArea(False);
		local iAreaID = biggest_area:GetID();
		-- We'll need all eight data fields returned in the results table from the boundary finder:
		local landmass_data = ObtainLandmassBoundaries(iAreaID);
		local iWestX = landmass_data[1];
		local iSouthY = landmass_data[2];
		local iEastX = landmass_data[3];
		local iNorthY = landmass_data[4];
		local iWidth = landmass_data[5];
		local iHeight = landmass_data[6];
		local wrapsX = landmass_data[7];
		local wrapsY = landmass_data[8];
		
		-- Obtain "Start Placement Fertility" of the landmass. (This measurement is customized for start placement).
		-- This call returns a table recording fertility of all plots within a rectangle that contains the landmass,
		-- with a zero value for any plots not part of the landmass -- plus a fertility sum and plot count.
		local fert_table, fertCount, plotCount = self:MeasureStartPlacementFertilityOfLandmass(iAreaID, 
		                                         iWestX, iEastX, iSouthY, iNorthY, wrapsX, wrapsY);
		-- Now divide this landmass in to regions, one per civ.
		-- The regional divider requires three arguments:
		-- 1. Number of divisions. (For "Biggest Landmass" this means number of civs in the game).
		-- 2. Fertility table. (This was obtained from the last call.)
		-- 3. Rectangle table. This table includes seven data fields:
		-- westX, southY, width, height, AreaID, fertilityCount, plotCount
		-- This is why we got the fertCount and plotCount from the fertility function.
		--
		-- Assemble the Rectangle data table:
		local rect_table = {iWestX, iSouthY, iWidth, iHeight, iAreaID, fertCount, plotCount};
		-- The data from this call is processed in to self.regionData during the process.
		self:DivideIntoRegions(self.iNumCivs, fert_table, rect_table)
		-- The regions have been defined.
	
	elseif self.method == 3 or self.method == 4 then -- Rectangular
		print("Division Method: rectangular");
		-- Obtain the boundaries of the rectangle to be processed.
		-- If no coords were passed, default to processing the entire map.
		-- Note that it matters if method 3 or 4 is designated, because the difference affects
		-- how city states are placed, whether they look for any uninhabited lands outside the rectangle.
		--
		-- NOTE: these dimensions need to be whole numbers 
		-- or else indexing failures can occur.
		self.inhabited_WestX  = 0;
		self.inhabited_Width  = g_iW;
		
		-- Y bounds as percent of height for avoiding starts near the poles.
		local minY = ScriptData.climateInfo.aspStartMinY or 0;
		local maxY = ScriptData.climateInfo.aspStartMaxY or 0;
		
		-- Convert percentages into plot lengths.
		minY = Round(minY * 0.01 * g_iH);
		maxY = Round(maxY * 0.01 * g_iH);
		
		minY = math.clamp(minY, 0,    g_iH);
		maxY = math.clamp(maxY, minY, g_iH);

		self.inhabited_SouthY = minY;
		self.inhabited_Height = maxY - minY;

		print("SouthY =  "    .. self.inhabited_SouthY .. 
				", Height = " .. self.inhabited_Height);
			
		-- Obtain "Start Placement Fertility" inside the rectangle.
		-- Data returned is: fertility table, sum of all fertility, plot count.
		local fert_table, fertCount, plotCount = self:MeasureStartPlacementFertilityInRectangle(self.inhabited_WestX, 
		                                         self.inhabited_SouthY, self.inhabited_Width, self.inhabited_Height)
		-- Assemble the Rectangle data table:
		local rect_table = {self.inhabited_WestX, self.inhabited_SouthY, self.inhabited_Width, 
		                    self.inhabited_Height, -1, fertCount, plotCount}; -- AreaID -1 means ignore area IDs.
		-- Divide the rectangle.
		self:DivideIntoRegions(self.iNumCivs, fert_table, rect_table)
		-- The regions have been defined.
	
	else -- Continental.
		print("Division Method: continental");
		--[[ Loop through all plots on the map, measuring fertility of each land 
		     plot, identifying its AreaID, building a list of landmass AreaIDs, and
		     tallying the Start Placement Fertility for each landmass. ]]--

		-- region_data: [WestX, EastX, SouthY, NorthY, 
		-- numLandPlotsinRegion, numCoastalPlotsinRegion,
		-- numOceanPlotsinRegion, iRegionNetYield, 
		-- iNumLandAreas, iNumPlotsinRegion]
		local best_areas = {};
		local globalFertilityOfLands = {};

		-- Obtain info on all landmasses for comparision purposes.
		local iGlobalFertilityOfLands = 0;
		local iNumLandPlots = 0;
		local iNumLandAreas = 0;
		local land_area_IDs = {};
		local land_area_plots = {};
		local land_area_fert = {};
		-- Cycle through all plots in the world, checking their Start Placement Fertility and AreaID.
		for x = 0, g_iW - 1 do
			for y = 0, g_iH - 1 do
				local i = y * g_iW + x + 1;
				local plot = Map.GetPlot(x, y);
				if not plot:IsWater() then -- Land plot, process it.
					iNumLandPlots = iNumLandPlots + 1;
					local iArea = plot:GetArea():GetID();
					local plotFertility = self:MeasureStartPlacementFertilityOfPlot(x, y, true); -- Check for coastal land is enabled.
					iGlobalFertilityOfLands = iGlobalFertilityOfLands + plotFertility;
					--
					if TestMembership(land_area_IDs, iArea) == false then -- This plot is the first detected in its AreaID.
						iNumLandAreas = iNumLandAreas + 1;
						table.insert(land_area_IDs, iArea);
						land_area_plots[iArea] = 1;
						land_area_fert[iArea] = plotFertility;
					else -- This AreaID already known.
						land_area_plots[iArea] = land_area_plots[iArea] + 1;
						land_area_fert[iArea] = land_area_fert[iArea] + plotFertility;
					end
				end
			end
		end
		
		-- Sort areas, achieving a list of AreaIDs with best areas first.
		--
		-- Fertility data in land_area_fert is stored with areaID index keys.
		-- Need to generate a version of this table with indices of 1 to n, where n is number of land areas.
		local interim_table = {};
		for loop_index, data_entry in pairs(land_area_fert) do
			table.insert(interim_table, data_entry);
		end
		
		-- Sort the fertility values stored in the interim table. Sort order in Lua is lowest to highest.
		table.sort(interim_table);

		--[[
		for AreaID, fert in ipairs(interim_table) do
			print("Interim Table ID " .. AreaID .. " has fertility of " .. fert);
		end
		print("* * * * * * * * * *");
		--]]

		-- If less players than landmasses, we will ignore the extra landmasses.
		local iNumRelevantLandAreas = math.min(iNumLandAreas, self.iNumCivs);
		-- Now re-match the AreaID numbers with their corresponding fertility values
		-- by comparing the original fertility table with the sorted interim table.
		-- During this comparison, best_areas will be constructed from sorted AreaIDs, richest stored first.
		local best_areas = {};
		-- Currently, the best yields are at the end of the interim table. We need to step backward from there.
		local end_of_interim_table = table.maxn(interim_table);
		-- We may not need all entries in the table. Process only iNumRelevantLandAreas worth of table entries.
		local fertility_value_list = {};
		local fertility_value_tie = false;
		for tableConstructionLoop = end_of_interim_table, (end_of_interim_table - iNumRelevantLandAreas + 1), -1 do
			if TestMembership(fertility_value_list, interim_table[tableConstructionLoop]) == true then
				fertility_value_tie = true;
				print("*** WARNING: Fertility Value Tie exists! ***");
			else
				table.insert(fertility_value_list, interim_table[tableConstructionLoop]);
			end
		end

		if fertility_value_tie == false then -- No ties, so no need of special handling for ties.
			for areaTestLoop = end_of_interim_table, (end_of_interim_table - iNumRelevantLandAreas + 1), -1 do
				for loop_index, AreaID in ipairs(land_area_IDs) do
					if interim_table[areaTestLoop] == land_area_fert[land_area_IDs[loop_index]] then
						table.insert(best_areas, AreaID);
						break
					end
				end
			end
		else -- Ties exist! Special handling required to protect against a shortfall in the number of defined regions.
			local iNumUniqueFertValues = table.maxn(fertility_value_list);
			for fertLoop = 1, iNumUniqueFertValues do
				for AreaID, fert in pairs(land_area_fert) do
					if fert == fertility_value_list[fertLoop] then
						-- Add ties only if there is room!
						local best_areas_length = table.maxn(best_areas);
						if best_areas_length < iNumRelevantLandAreas then
							table.insert(best_areas, AreaID);
						else
							break
						end
					end
				end
			end
		end
				
		-- Debug printout
		print("-"); print("--- Continental Division, Initial Readout ---"); print("-");
		print("- Global Fertility:", iGlobalFertilityOfLands);
		print("- Total Land Plots:", iNumLandPlots);
		print("- Total Areas:", iNumLandAreas);
		print("- Relevant Areas:", iNumRelevantLandAreas); print("-");
		--

		-- Debug printout
		print("* * * * * * * * * *");
		for area_loop, AreaID in ipairs(best_areas) do
			print("Area ID " .. AreaID .. " has fertility of " .. land_area_fert[AreaID]);
		end
		print("* * * * * * * * * *");
		--

		-- Assign continents to receive start plots. Record number of civs assigned to each landmass.
		local inhabitedAreaIDs = {};
		local numberOfCivsPerArea = table.fill(0, iNumRelevantLandAreas); -- Indexed in synch with best_areas. Use same index to match values from each table.
		for civToAssign = 1, self.iNumCivs do
			local bestRemainingArea;
			local bestRemainingFertility = 0;
			local bestAreaTableIndex;
			-- Loop through areas, find the one with the best remaining fertility (civs added 
			-- to a landmass reduces its fertility rating for subsequent civs).
			--
			--print("- - Searching landmasses in order to place Civ #", civToAssign); print("-");
			for area_loop, AreaID in ipairs(best_areas) do
				local thisLandmassCurrentFertility = land_area_fert[AreaID] / (1 + numberOfCivsPerArea[area_loop]);
				if thisLandmassCurrentFertility > bestRemainingFertility then
					bestRemainingArea = AreaID;
					bestRemainingFertility = thisLandmassCurrentFertility;
					bestAreaTableIndex = area_loop;
					--
					--print("- Found new candidate landmass with Area ID#:", bestRemainingArea, " with fertility of ", bestRemainingFertility);
				end
			end
			-- Record results for this pass. (A landmass has been assigned to receive one more start point than it previously had).
			numberOfCivsPerArea[bestAreaTableIndex] = numberOfCivsPerArea[bestAreaTableIndex] + 1;
			if TestMembership(inhabitedAreaIDs, bestRemainingArea) == false then
				table.insert(inhabitedAreaIDs, bestRemainingArea);
			end
			--print("Civ #", civToAssign, "has been assigned to Area#", bestRemainingArea); print("-");
		end
		print("-"); print("--- End of Initial Readout ---"); print("-");
		
		--print("*** Number of Civs per Landmass - Table Readout ***");
		--PrintContentsOfTable(numberOfCivsPerArea)
		--print("--- End of Civs per Landmass readout ***"); print("-"); print("-");
				
		-- Loop through the list of inhabited landmasses, dividing each landmass in to regions.
		-- Note that it is OK to divide a continent with one civ on it: this will assign the whole
		-- of the landmass to a single region, and is the easiest method of recording such a region.
		local iNumInhabitedLandmasses = table.maxn(inhabitedAreaIDs);
		for loop, currentLandmassID in ipairs(inhabitedAreaIDs) do
			-- Obtain the boundaries of and data for this landmass.
			local landmass_data = ObtainLandmassBoundaries(currentLandmassID);
			local iWestX = landmass_data[1];
			local iSouthY = landmass_data[2];
			local iEastX = landmass_data[3];
			local iNorthY = landmass_data[4];
			local iWidth = landmass_data[5];
			local iHeight = landmass_data[6];
			local wrapsX = landmass_data[7];
			local wrapsY = landmass_data[8];
			-- Obtain "Start Placement Fertility" of the current landmass. (Necessary to do this
			-- again because the fert_table can't be built prior to finding boundaries, and we had
			-- to ID the proper landmasses via fertility to be able to figure out their boundaries.
			local fert_table, fertCount, plotCount = self:MeasureStartPlacementFertilityOfLandmass(currentLandmassID, 
		  	                                         iWestX, iEastX, iSouthY, iNorthY, wrapsX, wrapsY);
			-- Assemble the rectangle data for this landmass.
			local rect_table = {iWestX, iSouthY, iWidth, iHeight, currentLandmassID, fertCount, plotCount};
			-- Divide this landmass in to number of regions equal to civs assigned here.
			iNumCivsOnThisLandmass = numberOfCivsPerArea[loop];
			if iNumCivsOnThisLandmass > 0 and iNumCivsOnThisLandmass <= 22 then -- valid number of civs.
			
				-- Debug printout for regional division inputs.
				print("-"); print("- Region #: ", loop);
				print("- Civs on this landmass: ", iNumCivsOnThisLandmass);
				print("- Area ID#: ", currentLandmassID);
				print("- Fertility: ", fertCount);
				print("- Plot Count: ", plotCount); print("-");
				--
			
				self:DivideIntoRegions(iNumCivsOnThisLandmass, fert_table, rect_table)
			else
				print("Invalid number of civs assigned to a landmass: ", iNumCivsOnThisLandmass);
			end
		end
		--
		-- The regions have been defined.
	end
	
	--[[ Printout is for debugging only. Deactivate otherwise.
	local tempRegionData = self.regionData;
	for i, data in ipairs(tempRegionData) do
		print("-");
		print("Data for Start Region #", i);
		print("WestX:  ", data[1]);
		print("SouthY: ", data[2]);
		print("Width:  ", data[3]);
		print("Height: ", data[4]);
		print("AreaID: ", data[5]);
		print("Fertility:", data[6]);
		print("Plots:  ", data[7]);
		print("Fert/Plot:", data[8]);
		print("-");
	end
	--]]
end
------------------------------------------------------------------------------
-- Start of functions tied to ChooseLocations()
------------------------------------------------------------------------------
function GotLakes_ASP:MeasureTerrainInRegions()
	-- This function stores its data in the instance (self) data table.
	for region_loop, region_data_table in ipairs(self.regionData) do
		local iWestX = region_data_table[1];
		local iSouthY = region_data_table[2];
		local iWidth = region_data_table[3];
		local iHeight = region_data_table[4];
		local iAreaID = region_data_table[5];
		
		local totalPlots, areaPlots = 0, 0;
		local waterCount, flatlandsCount, hillsCount, peaksCount = 0, 0, 0, 0;
		local lakeCount, coastCount, oceanCount, iceCount = 0, 0, 0, 0;
		local grassCount, plainsCount, desertCount, tundraCount, snowCount = 0, 0, 0, 0, 0; -- counts flatlands only!
		local forestCount, jungleCount, marshCount, riverCount, floodplainCount, oasisCount = 0, 0, 0, 0, 0, 0;
		local coastalLandCount, nextToCoastCount = 0, 0;

		-- Iterate through the region's plots, getting plotType, terrainType, featureType and river status.
		for region_loop_y = 0, iHeight - 1 do
			for region_loop_x = 0, iWidth - 1 do
				totalPlots = totalPlots + 1;
				local x = (region_loop_x + iWestX) % g_iW;
				local y = (region_loop_y + iSouthY) % g_iH;
				local plot = Map.GetPlot(x, y);
				local area_of_plot = plot:GetArea():GetID();
				-- get plot info
				local terrainType = plot:GetTerrainType()
				local featureType = plot:GetFeatureType()
				
				-- Mountain and Ocean plot types get their own AreaIDs, but we are going to measure them anyway.
				if plot:IsMountain() then
					peaksCount = peaksCount + 1; -- and that's it for Mountain plots. No other stats.
				elseif plot:IsWater() then
					waterCount = waterCount + 1;
					if terrainType == g_TERRAIN_TYPE_COAST then
						if plot:IsLake() then
							lakeCount = lakeCount + 1;
						else
							coastCount = coastCount + 1;
						end
					else
						oceanCount = oceanCount + 1;
					end
					if featureType == g_FEATURE_ICE then
						iceCount = iceCount + 1;
					end

				else
					-- Hills and Flatlands, check plot for region membership. Only process this plot if it is a member.
					if (area_of_plot == iAreaID) or (iAreaID == -1) then
						areaPlots = areaPlots + 1;

						-- set up coastalLand and nextToCoast index
						local i = g_iW * y + x + 1;
			
						-- Record plot data
						if plot:IsHills() then
							hillsCount = hillsCount + 1;

							if self.plotDataIsCoastal[i] then
								coastalLandCount = coastalLandCount + 1;
							elseif self.plotDataIsNextToCoast[i] then
								nextToCoastCount = nextToCoastCount + 1;
							end

							if plot:IsRiver() then
								riverCount = riverCount + 1;
							end

							-- Feature check checking for all types, in case features are not obeying standard allowances.
							if featureType == g_FEATURE_FOREST then
								forestCount = forestCount + 1;
							elseif featureType == g_FEATURE_JUNGLE then
								jungleCount = jungleCount + 1;
							elseif featureType == g_FEATURE_MARSH then
								marshCount = marshCount + 1;
							elseif featureType == g_FEATURE_FLOODPLAIN then
								floodplainCount = floodplainCount + 1;
							elseif featureType == g_FEATURE_OASIS then
								oasisCount = oasisCount + 1;
							end
								
						else -- Flatlands plot
							flatlandsCount = flatlandsCount + 1;
	
							if self.plotDataIsCoastal[i] then
								coastalLandCount = coastalLandCount + 1;
							elseif self.plotDataIsNextToCoast[i] then
								nextToCoastCount = nextToCoastCount + 1;
							end

							if plot:IsRiver() then
								riverCount = riverCount + 1;
							end
				
							if IsBaseTerrainType(g_TERRAIN_TYPE_GRASS, terrainType) then
								grassCount = grassCount + 1;
							elseif IsBaseTerrainType(g_TERRAIN_TYPE_PLAINS, terrainType) then
								plainsCount = plainsCount + 1;
							elseif IsBaseTerrainType(g_TERRAIN_TYPE_DESERT, terrainType) then
								desertCount = desertCount + 1;
							elseif IsBaseTerrainType(g_TERRAIN_TYPE_TUNDRA, terrainType) then
								tundraCount = tundraCount + 1;
							elseif IsBaseTerrainType(g_TERRAIN_TYPE_SNOW, terrainType) then
								snowCount = snowCount + 1;
							end
				
							-- Feature check checking for all types, in case features are not obeying standard allowances.
							if featureType == g_FEATURE_FOREST then
								forestCount = forestCount + 1;
							elseif featureType == g_FEATURE_JUNGLE then
								jungleCount = jungleCount + 1;
							elseif featureType == g_FEATURE_MARSH then
								marshCount = marshCount + 1;
							elseif featureType == g_FEATURE_FLOODPLAIN then
								floodplainCount = floodplainCount + 1;
							elseif featureType == g_FEATURE_OASIS then
								oasisCount = oasisCount + 1;
							end
						end
					end
				end
			end
		end
			
		-- Assemble in to an array the recorded data for this region: 23 variables.
		local regionCounts = {
			totalPlots, areaPlots,
			waterCount, flatlandsCount, hillsCount, peaksCount,
			lakeCount, coastCount, oceanCount, iceCount,
			grassCount, plainsCount, desertCount, tundraCount, snowCount,
			forestCount, jungleCount, marshCount, riverCount, floodplainCount, oasisCount,
			coastalLandCount, nextToCoastCount
			}
		--[[ Table Key:
		
		1) totalPlots
		2) areaPlots                 13) desertCount
		3) waterCount                14) tundraCount
		4) flatlandsCount            15) snowCount
		5) hillsCount                16) forestCount
		6) peaksCount                17) jungleCount
		7) lakeCount                 18) marshCount
		8) coastCount                19) riverCount
		9) oceanCount                20) floodplainCount
		10) iceCount                 21) oasisCount
		11) grassCount               22) coastalLandCount
		12) plainsCount              23) nextToCoastCount   ]]--
			
		-- Add array to the data table.
		table.insert(self.regionTerrainCounts, regionCounts);
		
		--[[ Activate printout only for debugging.
		print("-");
		print("--- Region Terrain Measurements for Region #", region_loop, "---");
		print("Total Plots: ", totalPlots);
		print("Area Plots: ", areaPlots);
		print("-");
		print("Mountains: ", peaksCount, " - Cannot belong to a landmass AreaID.");
		print("Total Water Plots: ", waterCount, " - Cannot belong to a landmass AreaID.");
		print("-");
		print("Lake Plots: ", lakeCount);
		print("Coast Plots: ", coastCount, " - Does not include Lakes.");
		print("Ocean Plots: ", oceanCount);
		print("Icebergs: ", iceCount);
		print("-");
		print("Flatlands: ", flatlandsCount);
		print("Hills: ", hillsCount);
		print("-");
		print("Grass Plots: ", grassCount);
		print("Plains Plots: ", plainsCount);
		print("Desert Plots: ", desertCount);
		print("Tundra Plots: ", tundraCount);
		print("Snow Plots: ", snowCount);
		print("-");
		print("Forest Plots: ", forestCount);
		print("Jungle Plots: ", jungleCount);
		print("Marsh Plots: ", marshCount);
		print("Flood Plains: ", floodplainCount);
		print("Oases: ", oasisCount);
		print("-");
		print("Plots Along Rivers: ", riverCount);
		print("Plots Along Oceans: ", coastalLandCount);
		print("Plots Next To Plots Along Oceans: ", nextToCoastCount);
		print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -");
		]]--
	end
end

function GotLakes_ASP:DetermineRegionTypes()
	-- Determine region type and conditions. Use self.regionTypes to store the results
	--
	-- REGION TYPES
	-- 0. Undefined
	-- 1. Tundra
	-- 2. Jungle
	-- 3. Forest
	-- 4. Desert
	-- 5. Hills
	-- 6. Plains
	-- 7. Grassland
	-- 8. Hybrid

	-- Main loop
	for this_region, terrainCounts in ipairs(self.regionTerrainCounts) do
		-- Set each region to "Undefined Type" as default.
		-- If all efforts fail at determining what type of region this should be, region type will remain Undefined.
		--local totalPlots = terrainCounts[1];
		local areaPlots = terrainCounts[2];
		--local waterCount = terrainCounts[3];
		local flatlandsCount = terrainCounts[4];
		local hillsCount = terrainCounts[5];
		local peaksCount = terrainCounts[6];
		--local lakeCount = terrainCounts[7];
		--local coastCount = terrainCounts[8];
		--local oceanCount = terrainCounts[9];
		local iceCount = terrainCounts[10];
		local grassCount = terrainCounts[11];
		local plainsCount = terrainCounts[12];
		local desertCount = terrainCounts[13];
		local tundraCount = terrainCounts[14];
		local snowCount = terrainCounts[15];
		local forestCount = terrainCounts[16];
		local jungleCount = terrainCounts[17];
		--local marshCount = terrainCounts[18];
		local riverCount = terrainCounts[19];
		--local floodplainCount = terrainCounts[20];
		--local oasisCount = terrainCounts[21];
		--local coastalLandCount = terrainCounts[22];
		--local nextToCoastCount = terrainCounts[23];

		-- If Rectangular regional division, then water plots would be included in area plots.
		-- Let's recalculate area plots based only on flatland and hills plots.
		if self.method == 3 or self.method == 4 then
			areaPlots = flatlandsCount + hillsCount;
		end

		-- Tundra check first.
		if (tundraCount + snowCount) >= areaPlots * 0.3 then
			table.insert(self.regionTypes, 1);
			--print("-");
			--print("Region #", this_region, " has been defined as a Tundra Region.");
		
		-- Jungle check.
		elseif (jungleCount >= areaPlots * 0.30) then
			table.insert(self.regionTypes, 2);
			--print("-");
			--print("Region #", this_region, " has been defined as a Jungle Region.");
		elseif (jungleCount >= areaPlots * 0.20) and (jungleCount + forestCount >= areaPlots * 0.35) then
			table.insert(self.regionTypes, 2);
			--print("-");
			--print("Region #", this_region, " has been defined as a Jungle Region.");
		
		-- Forest check.
		elseif (forestCount >= areaPlots * 0.30) then
			table.insert(self.regionTypes, 3);
			--print("-");
			--print("Region #", this_region, " has been defined as a Forest Region.");
		elseif (forestCount >= areaPlots * 0.20) and (jungleCount + forestCount >= areaPlots * 0.35) then
			table.insert(self.regionTypes, 3);
			--print("-");
			--print("Region #", this_region, " has been defined as a Forest Region.");
		
		-- Desert check.
		elseif (desertCount >= areaPlots * 0.25) then
			table.insert(self.regionTypes, 4);
			--print("-");
			--print("Region #", this_region, " has been defined as a Desert Region.");

		-- Hills check.
		elseif (hillsCount >= areaPlots * 0.415) then
			table.insert(self.regionTypes, 5);
			--print("-");
			--print("Region #", this_region, " has been defined as a Hills Region.");
		
		-- Plains check.
		elseif (plainsCount >= areaPlots * 0.3) and (plainsCount * 0.7 > grassCount) then
			table.insert(self.regionTypes, 6);
			--print("-");
			--print("Region #", this_region, " has been defined as a Plains Region.");
		
		-- Grass check.
		elseif (grassCount >= areaPlots * 0.3) and (grassCount * 0.7 > plainsCount) then
			table.insert(self.regionTypes, 7);
			--print("-");
			--print("Region #", this_region, " has been defined as a Grassland Region.");
		
		-- Hybrid check.
		elseif ((grassCount + plainsCount + desertCount + tundraCount + snowCount + hillsCount + peaksCount) > areaPlots * 0.8) then
			table.insert(self.regionTypes, 8);
			--print("-");
			--print("Region #", this_region, " has been defined as a Hybrid Region.");

		else -- Undefined Region (most likely due to operating on a mod that adds new terrain types.)
			table.insert(self.regionTypes, 0);
			--print("-");
			--print("Region #", this_region, " has been defined as an Undefined Region.");
		
		end
	end
end

function GotLakes_ASP:PlaceImpactAndRipples(x, y)
	-- To establish a bias of 9, we Impact the overlay and Ripple outward 8 times.
	-- Value of 0 in a plot means no influence from existing Impacts in that plot.
	-- Value of 99 means an Impact occurred in that plot and it IS a start point.
	-- Values > 0 and < 99 are "ripples", meaning that plot is near a start point.
	local wrapX = ScriptData.wrapX;
	local wrapY = ScriptData.wrapY;
	local impact_value = 99;
	local ripple_values = {97, 95, 92, 89, 69, 57, 24, 15};
	local odd = self.firstRingYIsOdd;
	local even = self.firstRingYIsEven;
	local nextX, nextY, plot_adjustments;
	-- Now the main data layer, for start points themselves, and the City State data layer.
	-- Place Impact!
	local impactPlotIndex = y * g_iW + x + 1;
	self.distanceData[impactPlotIndex] = impact_value;
	-- Place Ripples
	for ripple_radius, ripple_value in ipairs(ripple_values) do
		-- Moving clockwise around the ring, the first direction to travel will be Northeast.
		-- This matches the direction-based data in the odd and even tables. Each
		-- subsequent change in direction will correctly match with these tables, too.
		--
		-- Locate the plot within this ripple ring that is due West of the Impact Plot.
		local currentX = x - ripple_radius;
		local currentY = y;
		-- Now loop through the six directions, moving ripple_radius number of times
		-- per direction. At each plot in the ring, add the ripple_value for that ring 
		-- to the plot's entry in the distance data table.
		for direction_index = 1, 6 do
			for plot_to_handle = 1, ripple_radius do
				-- Must account for hex factor.
			 	if currentY / 2 > math.floor(currentY / 2) then -- Current Y is odd. Use odd table.
					plot_adjustments = odd[direction_index];
				else -- Current Y is even. Use plot adjustments from even table.
					plot_adjustments = even[direction_index];
				end
				-- Identify the next plot in the ring.
				nextX = currentX + plot_adjustments[1];
				nextY = currentY + plot_adjustments[2];
				-- Make sure the plot exists
				if wrapX == false and (nextX < 0 or nextX >= g_iW) then -- X is out of bounds.
					-- Do not add ripple data to this plot.
				elseif wrapY == false and (nextY < 0 or nextY >= g_iH) then -- Y is out of bounds.
					-- Do not add ripple data to this plot.
				else -- Plot is in bounds, process it.
					-- Handle any world wrap.
					local realX = nextX;
					local realY = nextY;
					if wrapX == true then
						realX = realX % g_iW;
					end
					if wrapY == true then
						realY = realY % g_iH;
					end
					-- Record ripple data for this plot.
					local ringPlotIndex = realY * g_iW + realX + 1;
					if self.distanceData[ringPlotIndex] > 0 then -- This plot is already in range of at least one other civ!
						-- First choose the greater of the two, existing value or current ripple.
						local stronger_value = math.max(self.distanceData[ringPlotIndex], ripple_value);
						-- Now increase it by 1.2x to reflect that multiple civs are in range of this plot.
						local overlap_value = math.min(97, math.floor(stronger_value * 1.2));
						self.distanceData[ringPlotIndex] = overlap_value;
					else
						self.distanceData[ringPlotIndex] = ripple_value;
					end
				end
				currentX, currentY = nextX, nextY;
			end
		end
	end
end

function GotLakes_ASP:MeasureSinglePlot(x, y, region_type)
	local data = table.fill(false, 4);
	-- Note that "Food" is not strictly about tile yield.
	-- Different regions get their food in different ways.
	-- Tundra, Jungle, Forest, Desert, Plains regions will 
	-- get Bonus resource support to cover food shortages.
	--
	-- Data table entries hold results; all begin as false:
	-- [1] "Food"
	-- [2] "Prod"
	-- [3] "Good"
	-- [4] "Junk"
	local plot = Map.GetPlot(x, y);
	local terrainType = plot:GetTerrainType();
	local featureType = plot:GetFeatureType();
	
	if plot:IsMountain() then -- Mountains are Junk.
		data[4] = true;
		return data
	elseif plot:IsWater() then
		if featureType == g_FEATURE_ICE then -- Icebergs are Junk.
			data[4] = true;
			return data
		elseif plot:IsLake() then -- Lakes are Food, Good.
			data[1] = true;
			data[3] = true;
			return data
		elseif self.method == 3 or self.method == 4 then
			if terrainType == g_TERRAIN_TYPE_COAST then -- Shallow water is Good for Archipelago-type maps.
				data[3] = true;
				return data
			end
		end
		-- Other water plots are ignored.
		return data
	end

	if featureType == g_FEATURE_JUNGLE then -- Jungles are Food, Good, except in Grass regions.
		if region_type ~= 7 then -- Region type is not grass.
			data[1] = true;
			data[3] = true;
		elseif plot:IsHills() then -- Jungle hill, in grass region, count as Prod but not Good.
			data[2] = true;
		end
		return data
	elseif featureType == g_FEATURE_FOREST then -- Forests are Prod, Good.
		data[2] = true;
		data[3] = true;
		if region_type == 3 or region_type == 1 then -- In Forest or Tundra Regions, Forests are also Food.
			data[1] = true;
		end
		return data
	elseif featureType == g_FEATURE_OASIS then -- Oases are Food, Good.
		data[1] = true;
		data[3] = true;
		return data
	elseif featureType == g_FEATURE_FLOOD_PLAINS then -- Flood Plains are Food, Good.
		data[1] = true;
		data[3] = true;
		return data
	elseif featureType == g_FEATURE_MARSH then -- Marsh is food.
		data[1] = true; 
		return data
	end

	if plot:IsHills() then -- Hills with no features are Prod, Good.
		data[2] = true;
		data[3] = true;
		return data
	end
	
	-- If we have reached this point in the process, the plot is flatlands.
	if IsBaseTerrainType(g_TERRAIN_TYPE_SNOW, terrainType) then -- Snow is Junk.
		data[4] = true;
		return data
		
	elseif IsBaseTerrainType(g_TERRAIN_TYPE_DESERT, terrainType) then -- Non-Oasis, non-FloodPlain flat deserts are Junk, except in Desert regions.
		if region_type ~= 4 then
			data[4] = true;
		end
		return data

	elseif IsBaseTerrainType(g_TERRAIN_TYPE_TUNDRA, terrainType) then -- Tundra are ignored, except in Tundra Regions where they are Food, Good.
		if region_type == 1 then
			data[1] = true;
			data[3] = true;
		end
		return data

	elseif IsBaseTerrainType(g_TERRAIN_TYPE_PLAINS, terrainType) then -- Plains are Good for all region types, but Food in only about half of them.
		data[3] = true;
		if region_type == 1 or region_type == 4 or region_type == 5 or region_type == 6 or region_type == 8 then
			data[1] = true;
		end
		return data

	elseif IsBaseTerrainType(g_TERRAIN_TYPE_GRASS, terrainType) then -- Grass is Good for all region types, but Food in only about half of them.
		data[3] = true;
		if region_type == 2 or region_type == 3 or region_type == 5 or region_type == 7 or region_type == 8 then
			data[1] = true;
		end
		return data
	end

	-- If we have arrived here, the plot has non-standard terrain.
	print("Encountered non-standard terrain.");
	return data
end

function GotLakes_ASP:EvaluateCandidatePlot(plotIndex, region_type)
	local goodSoFar = true;
	local x = (plotIndex - 1) % g_iW;
	local y = math.floor((plotIndex - x - 1) / g_iW);
	local plot = Map.GetPlot(x, y);
	local isEvenY = true;
	if y / 2 > math.floor(y / 2) then
		isEvenY = false;
	end
	local wrapX = ScriptData.wrapX;
	local wrapY = ScriptData.wrapY;
	local distance_bias = self.distanceData[plotIndex];
	local foodTotal, prodTotal, goodTotal, junkTotal, riverTotal, coastScore = 0, 0, 0, 0, 0, 0;
	local plotScore = 0;
	local search_table = {};
	
	-- Check candidate plot to see if it's adjacent to saltwater.
	if self.plotDataIsCoastal[plotIndex] == true then
		coastScore = 40;
	end
	
	-- Evaluate the plot itself.
	local result = self:MeasureSinglePlot(x, y, region_type);
	if (result[4]) then
		-- The plot itself is junk.
		plotScore = -40;
	end
		
	-- Evaluate First Ring
	if isEvenY then
		search_table = self.firstRingYIsEven;
	else
		search_table = self.firstRingYIsOdd;
	end

	for loop, plot_adjustments in ipairs(search_table) do
		local searchX, searchY;
		if wrapX then
			searchX = (x + plot_adjustments[1]) % g_iW;
		else
			searchX = x + plot_adjustments[1];
		end
		if wrapY then
			searchY = (y + plot_adjustments[2]) % g_iH;
		else
			searchY = y + plot_adjustments[2];
		end
		--
		if searchX < 0 or searchX >= g_iW or searchY < 0 or searchY >= g_iH then
			-- This plot does not exist. It's off the map edge.
			junkTotal = junkTotal + 1;
		else
			local result = self:MeasureSinglePlot(searchX, searchY, region_type)
			if result[4] then
				junkTotal = junkTotal + 1;
			else
				if result[1] then
					foodTotal = foodTotal + 1;
				end
				if result[2] then
					prodTotal = prodTotal + 1;
				end
				if result[3] then
					goodTotal = goodTotal + 1;
				end
				local searchPlot = Map.GetPlot(searchX, searchY);
				if searchPlot:IsRiver() then
					riverTotal = riverTotal + 1;
				end
			end
		end
	end

	-- Now check the results from the first ring against the established targets.
	if foodTotal < self.minFoodInner then
		goodSoFar = false;
	elseif prodTotal < self.minProdInner then
		goodSoFar = false;
	elseif goodTotal < self.minGoodInner then
		goodSoFar = false;
	end

	-- Set up the "score" for this candidate. Inner ring results weigh the heaviest.
	local weightedFoodInner = {0, 8, 14, 19, 22, 24, 25};
	local foodResultInner = weightedFoodInner[foodTotal + 1];
	local weightedProdInner = {0, 10, 16, 20, 20, 12, 0};
	local prodResultInner = weightedProdInner[prodTotal + 1];
	local goodResultInner = goodTotal * 2;
	local innerRingScore = foodResultInner + prodResultInner + goodResultInner + riverTotal - (junkTotal * 3);
	
	-- Evaluate Second Ring
	if isEvenY then
		search_table = self.secondRingYIsEven;
	else
		search_table = self.secondRingYIsOdd;
	end

	for loop, plot_adjustments in ipairs(search_table) do
		local searchX, searchY;
		if wrapX then
			searchX = (x + plot_adjustments[1]) % g_iW;
		else
			searchX = x + plot_adjustments[1];
		end
		if wrapY then
			searchY = (y + plot_adjustments[2]) % g_iH;
		else
			searchY = y + plot_adjustments[2];
		end
		if searchX < 0 or searchX >= g_iW or searchY < 0 or searchY >= g_iH then
			-- This plot does not exist. It's off the map edge.
			junkTotal = junkTotal + 1;
		else
			local result = self:MeasureSinglePlot(searchX, searchY, region_type)
			if result[4] then
				junkTotal = junkTotal + 1;
			else
				if result[1] then
					foodTotal = foodTotal + 1;
				end
				if result[2] then
					prodTotal = prodTotal + 1;
				end
				if result[3] then
					goodTotal = goodTotal + 1;
				end
				if plot:IsRiver() then
					riverTotal = riverTotal + 1;
				end
			end
		end
	end

	-- Check the results from the second ring against the established targets.
	if foodTotal < self.minFoodMiddle then
		goodSoFar = false;
	elseif prodTotal < self.minProdMiddle then
		goodSoFar = false;
	elseif goodTotal < self.minGoodMiddle then
		goodSoFar = false;
	end
	
	-- Update up the "score" for this candidate. Middle ring results weigh significantly.
	local weightedFoodMiddle = {0, 2, 5, 10, 20, 25, 28, 30, 32, 34, 35}; -- 35 for any further values.
	local foodResultMiddle = 35;
	if foodTotal < 10 then
		foodResultMiddle = weightedFoodMiddle[foodTotal + 1];
	end
	local weightedProdMiddle = {0, 10, 20, 25, 30, 35}; -- 35 for any further values.
	local effectiveProdTotal = prodTotal;
	if foodTotal * 2 < prodTotal then
		effectiveProdTotal = math.ceil(foodTotal / 2);
	end
	local prodResultMiddle = 35;
	if effectiveProdTotal < 5 then
		prodResultMiddle = weightedProdMiddle[effectiveProdTotal + 1];
	end
	local goodResultMiddle = goodTotal * 2;
	local middleRingScore = foodResultMiddle + prodResultMiddle + goodResultMiddle + riverTotal - (junkTotal * 3);
	
	-- Evaluate Third Ring
	if isEvenY then
		search_table = self.thirdRingYIsEven;
	else
		search_table = self.thirdRingYIsOdd;
	end

	for loop, plot_adjustments in ipairs(search_table) do
		local searchX, searchY;
		if wrapX then
			searchX = (x + plot_adjustments[1]) % g_iW;
		else
			searchX = x + plot_adjustments[1];
		end
		if wrapY then
			searchY = (y + plot_adjustments[2]) % g_iH;
		else
			searchY = y + plot_adjustments[2];
		end
		if searchX < 0 or searchX >= g_iW or searchY < 0 or searchY >= g_iH then
			-- This plot does not exist. It's off the map edge.
			junkTotal = junkTotal + 1;
		else
			local result = self:MeasureSinglePlot(searchX, searchY, region_type)
			if result[4] then
				junkTotal = junkTotal + 1;
			else
				if result[1] then
					foodTotal = foodTotal + 1;
				end
				if result[2] then
					prodTotal = prodTotal + 1;
				end
				if result[3] then
					goodTotal = goodTotal + 1;
				end
				if plot:IsRiver() then
					riverTotal = riverTotal + 1;
				end
			end
		end
	end

	-- Check the results from the third ring against the established targets.
	if foodTotal < self.minFoodOuter then
		goodSoFar = false;
	elseif prodTotal < self.minProdOuter then
		goodSoFar = false;
	elseif goodTotal < self.minGoodOuter then
		goodSoFar = false;
	end
	if junkTotal > self.maxJunk then
		goodSoFar = false;
	end

	-- Tally the final "score" for this candidate.
	local outerRingScore = foodTotal + prodTotal + goodTotal + riverTotal - (junkTotal * 2);
	local finalScore = plotScore + innerRingScore + middleRingScore + outerRingScore + coastScore;

	-- Check Impact and Ripple data to see if candidate is near an already-placed start point.
	if distance_bias > 0 then
		-- This candidate is near an already placed start. This invalidates its 
		-- eligibility for first-pass placement; but it may still qualify as a 
		-- fallback site, so we will reduce its Score according to the bias factor.
		goodSoFar = false;
		finalScore = finalScore - math.floor(finalScore * distance_bias / 100);
	end
	
	-- Get the size in tiles of the land-chunk that owns this plot.
	-- This can be less than plot:GetArea():GetPlotCount() 
	-- due to mountains making other tiles unreachable by land.
	local plotAreaNumTiles = GetPassableLandArea(plot, self.majorStartMinLandSize);
	
	if (plotAreaNumTiles < self.majorStartMinLandSize) then
		-- This island is less than a small fraction of the biggest land-chunk's size.
		--print("Downgrading tiny-island plot at " .. CoordString(plot:GetX(),plot:GetY()) ..
		--	" which can only reach " .. plotAreaNumTiles .. " tile(s).");
		goodSoFar  = false;
		finalScore = finalScore - 40;	
	end

	--[[ Debug
	print(".");
	print("Plot:", x, y, " Food:", foodTotal, "Prod: ", prodTotal, "Good:", goodTotal, "Junk:", 
	       junkTotal, "River:", riverTotal, "Score:", finalScore);
	print("Plot:", x, y, " Coastal:", self.plotDataIsCoastal[plotIndex], "Distance Bias:", distance_bias);
	--]]
	
	return finalScore, goodSoFar
end

function GotLakes_ASP:IterateThroughCandidatePlotList(plot_list, region_type)
	-- Iterates through a list of candidate plots.
	-- Each plot is identified by its global plot index.
	-- This function assumes all candidate plots can have a city built on them.
	-- Any plots not allowed to have a city should be weeded out when building the candidate list.
	local found_eligible = false;
	local bestPlotScore = -5000;
	local bestPlotIndex;
	local found_fallback = false;
	local bestFallbackScore = -5000;
	local bestFallbackIndex;
	-- Process list of candidate plots.
	for loop, plotIndex in ipairs(plot_list) do
		local score, meets_minimums = self:EvaluateCandidatePlot(plotIndex, region_type)
		-- Test current plot against best known plot.
		if meets_minimums == true then
			found_eligible = true;
			if score > bestPlotScore then
				bestPlotScore = score;
				bestPlotIndex = plotIndex;
			end
		else
			found_fallback = true;
			if score > bestFallbackScore then
				bestFallbackScore = score;
				bestFallbackIndex = plotIndex;
			end
		end
	end
	-- returns table containing six variables: boolean, integer, integer, boolean, integer, integer
	local election_results = {found_eligible, bestPlotScore, bestPlotIndex, found_fallback, bestFallbackScore, bestFallbackIndex};
	return election_results
end

function FindPassableForceStart(iWidth, iHeight, iWestX, iSouthY)
	-- Search a region for a start plot that is passable i.e. not a mountain.
	-- Only do this when there are no fallback plots. We're looking for something
	-- better than the bottom-left corner of the region.
	--
	-- Args:
	--   iWidth: width of the region
	--   iHeight: height of the region
	--   iWestX: west edge of the region
	--   iSouthY: bottom edge of the region
	--
	-- Returns: x, y location of the plot
	
	-- Add some randomness to the start location of the scan.
	local iStartX = TerrainBuilder.GetRandomNumber(iWidth,  "X for passable-start search.");
	local iStartY = TerrainBuilder.GetRandomNumber(iHeight, "Y for passable-start search.");
	
	for region_y = 0, iHeight - 1 do -- When handling global plot indices, process Y first.
		for region_x = 0, iWidth - 1 do
		
			local x = ( ((region_x + iStartX) % iWidth)  + iWestX)  % g_iW;
			local y = ( ((region_y + iStartY) % iHeight) + iSouthY) % g_iH;
			local plotIndex = y * g_iW + x + 1;
			local plot = Map.GetPlot(x, y);
			if IsPassableLand(plot) then
				-- Check if plot can host a city.
				print("Found a passable force-start plot at " .. CoordString(x,y));
				return x, y;
			end
		end
	end
	
	print("An entire civ start region at " .. 
		CoordString(iWestX, iSouthY) .. " with dimensions " .. 
		CoordString(iWidth, iHeight) .. " is impassable!");
		
	local hackPlot = Map.GetPlot(iWestX, iSouthY);	
	TerrainBuilder.SetTerrainType(hackPlot, g_TERRAIN_TYPE_GRASS);
	TerrainBuilder.SetFeatureType(hackPlot, g_FEATURE_NONE);	
		
	return iWestX, iSouthY;
end

function IsFarFromNaturalWonder(plot, minor)
	-- Returns false if the player can start because there is a natural wonder too close.

	-- If Start position config equals legendary you can start near Natural wonders
	if(UserInput.start == START_LEGEND) then
		return true;
	end

	local iMaxNW = 4;
	
	
	if(minor == true) then
		iMaxNW = GlobalParameters.START_DISTANCE_MINOR_NATURAL_WONDER or 3;
	else
		iMaxNW = GlobalParameters.START_DISTANCE_MAJOR_NATURAL_WONDER or 4;
	end


	local plotX = plot:GetX();
	local plotY = plot:GetY();
	for dx = -iMaxNW, iMaxNW do
		for dy = -iMaxNW,iMaxNW do
			local otherPlot = Map.GetPlotXY(plotX, plotY, dx, dy, iMaxNW);
			if(otherPlot and otherPlot:IsNaturalWonder()) then
				return false;
			end
		end
	end 

	return true;
end

function FindClosestPlot(outer_eligible_list, iWidth, iHeight, iWestX, iSouthY)
	-- Helper method to reduce code repetition
	-- and to encapsulate some confusing logic.
	local closestPlot;
	local closestDistance = math.max(g_iW, g_iH);
	local bullseyeX = iWestX + (iWidth / 2);
	if bullseyeX < iWestX then -- wrapped around: un-wrap it for test purposes.
		bullseyeX = bullseyeX + g_iW;
	end
	local bullseyeY = iSouthY + (iHeight / 2);
	if bullseyeY < iSouthY then -- wrapped around: un-wrap it for test purposes.
		bullseyeY = bullseyeY + g_iH;
	end
	if (math.floor(bullseyeY) % 2 == 1) then -- Y coord is odd, add .5 to X coord for hex-shift.
		bullseyeX = bullseyeX + 0.5;
	end
	
	for loop, plotIndex in ipairs(outer_eligible_list) do
		local fx = (plotIndex - 1) % g_iW;
		local fy = (plotIndex - fx - 1) / g_iW;
		local adjusted_x = fx;
		local adjusted_y = fy;
		if (math.floor(fy) % 2 == 1) then -- Y coord is odd, add .5 to X coord for hex-shift.
			adjusted_x = fx + 0.5;
		end
		
		if fx < iWestX then -- wrapped around: un-wrap it for test purposes.
			adjusted_x = adjusted_x + g_iW;
		end
		if fy < iSouthY then -- wrapped around: un-wrap it for test purposes.
			adjusted_y = fy + g_iH;
		end
		local fDistance = math.sqrt( (adjusted_x - bullseyeX)^2 + (adjusted_y - bullseyeY)^2 );
		if fDistance < closestDistance then -- Found new "closer" plot.
			closestPlot = plotIndex;
			closestDistance = fDistance;
		end
	end
	-- Assign the closest eligible plot as the start point.
	local x = (closestPlot - 1) % g_iW;
	local y = math.floor((closestPlot - x - 1) / g_iW);

	return x, y, closestPlot;
end

function GotLakes_ASP:FindStart(region_number)
	-- This function attempts to choose a start position for a single region.
	-- This function returns two boolean flags, indicating the success level of the operation.
	local bSuccessFlag = false; -- Returns true when a start is placed, false when process fails.
	local bForcedPlacementFlag = false; -- Returns true if this region had no eligible starts and one was forced to occur.
	
	-- Obtain data needed to process this region.
	
	local region_data_table = self.regionData[region_number];
	local iWestX = region_data_table[1];
	local iSouthY = region_data_table[2];
	local iWidth = region_data_table[3];
	local iHeight = region_data_table[4];
	local iAreaID = region_data_table[5];
	local iMembershipEastX = iWestX + iWidth - 1;
	local iMembershipNorthY = iSouthY + iHeight - 1;
	--
	local terrainCounts = self.regionTerrainCounts[region_number];
	--
	local region_type = self.regionTypes[region_number];
	-- Done setting up region data.
	-- Set up contingency.
	local fallback_plots = {};
	
	print("Looking for start at " .. CoordString(iWestX, iSouthY) .. 
			" with region size " .. CoordString(iWidth, iHeight));
	
	-- Establish scope of center bias.
	local fCenterWidth = (self.centerBias / 100) * iWidth;
	local iNonCenterWidth = math.floor((iWidth - fCenterWidth) / 2)
	local iCenterWidth = iWidth - (iNonCenterWidth * 2);
	local iCenterWestX = (iWestX + iNonCenterWidth) % g_iW; -- Modulo math to synch coordinate to actual map in case of world wrap.
	local iCenterTestWestX = (iWestX + iNonCenterWidth); -- "Test" values ignore world wrap for easier membership testing.
	local iCenterTestEastX = (iCenterWestX + iCenterWidth - 1);

	local fCenterHeight = (self.centerBias / 100) * iHeight;
	local iNonCenterHeight = math.floor((iHeight - fCenterHeight) / 2)
	local iCenterHeight = iHeight - (iNonCenterHeight * 2);
	local iCenterSouthY = (iSouthY + iNonCenterHeight) % g_iH;
	local iCenterTestSouthY = (iSouthY + iNonCenterHeight);
	local iCenterTestNorthY = (iCenterTestSouthY + iCenterHeight - 1);

	-- Establish scope of "middle donut", outside the center but inside the outer.
	local fMiddleWidth = (self.middleBias / 100) * iWidth;
	local iOuterWidth = math.floor((iWidth - fMiddleWidth) / 2)
	local iMiddleWidth = iWidth - (iOuterWidth * 2);
	local iMiddleWestX = (iWestX + iOuterWidth) % g_iW;
	local iMiddleTestWestX = (iWestX + iOuterWidth);
	local iMiddleTestEastX = (iMiddleTestWestX + iMiddleWidth - 1);

	local fMiddleHeight = (self.middleBias / 100) * iHeight;
	local iOuterHeight = math.floor((iHeight - fMiddleHeight) / 2)
	local iMiddleHeight = iHeight - (iOuterHeight * 2);
	local iMiddleSouthY = (iSouthY + iOuterHeight) % g_iH;
	local iMiddleTestSouthY = (iSouthY + iOuterHeight);
	local iMiddleTestNorthY = (iMiddleTestSouthY + iMiddleHeight - 1); 

	-- Assemble candidates lists.
	local two_plots_from_ocean = {};
	local center_candidates = {};
	local center_river = {};
	local center_coastal = {};
	local center_inland_dry = {};
	local middle_candidates = {};
	local middle_river = {};
	local middle_coastal = {};
	local middle_inland_dry = {};
	local outer_plots = {};
	
	-- Identify candidate plots.
	for region_y = 0, iHeight - 1 do -- When handling global plot indices, process Y first.
		for region_x = 0, iWidth - 1 do
			local x = (region_x + iWestX) % g_iW; -- Actual coords, adjusted for world wrap, if any.
			local y = (region_y + iSouthY) % g_iH; --
			local plotIndex = y * g_iW + x + 1;
			local plot = Map.GetPlot(x, y);
			
			if IsPassableLand(plot) and IsFarFromNaturalWonder(plot,false) then 
				-- Could host a city.
				-- Check if plot is two away from salt water.
				if self.plotDataIsNextToCoast[plotIndex] == true then
					table.insert(two_plots_from_ocean, plotIndex);
				else
					local area_of_plot = plot:GetArea():GetID();
					if area_of_plot == iAreaID or iAreaID == -1 then -- This plot is a member, so it goes on at least one candidate list.
						--
						-- Test whether plot is in center bias, middle donut, or outer donut.
						--
						local test_x = region_x + iWestX; -- "Test" coords, ignoring any world wrap and
						local test_y = region_y + iSouthY; -- reaching in to virtual space if necessary.
						if (test_x >= iCenterTestWestX and test_x <= iCenterTestEastX) and 
						   (test_y >= iCenterTestSouthY and test_y <= iCenterTestNorthY) then -- Center Bias.
							table.insert(center_candidates, plotIndex);
							if plot:IsRiver() then
								table.insert(center_river, plotIndex);
							elseif plot:IsFreshWater() or self.plotDataIsCoastal[plotIndex] == true then
								table.insert(center_coastal, plotIndex);
							else
								table.insert(center_inland_dry, plotIndex);
							end
						elseif (test_x >= iMiddleTestWestX and test_x <= iMiddleTestEastX) and 
						       (test_y >= iMiddleTestSouthY and test_y <= iMiddleTestNorthY) then
							table.insert(middle_candidates, plotIndex);
							if plot:IsRiver() then
								table.insert(middle_river, plotIndex);
							elseif plot:IsFreshWater() or self.plotDataIsCoastal[plotIndex] == true then
								table.insert(middle_coastal, plotIndex);
							else
								table.insert(middle_inland_dry, plotIndex);
							end
						else
							table.insert(outer_plots, plotIndex);
						end
					end
				end
			end
		end
	end

	-- Check how many plots landed on each list.
	local iNumDisqualified = table.maxn(two_plots_from_ocean);
	local iNumCenter = table.maxn(center_candidates);
	local iNumCenterRiver = table.maxn(center_river);
	local iNumCenterCoastLake = table.maxn(center_coastal);
	local iNumCenterInlandDry = table.maxn(center_inland_dry);
	local iNumMiddle = table.maxn(middle_candidates);
	local iNumMiddleRiver = table.maxn(middle_river);
	local iNumMiddleCoastLake = table.maxn(middle_coastal);
	local iNumMiddleInlandDry = table.maxn(middle_inland_dry);
	local iNumOuter = table.maxn(outer_plots);
	
	-- Process lists of candidate plots.
	if iNumCenter + iNumMiddle > 0 then
		local candidate_lists = {};
		if iNumCenterRiver > 0 then -- Process center bias river plots.
			table.insert(candidate_lists, center_river);
		end
		if iNumCenterCoastLake > 0 then -- Process center bias lake or coastal plots.
			table.insert(candidate_lists, center_coastal);
		end
		if iNumCenterInlandDry > 0 then -- Process center bias inland dry plots.
			table.insert(candidate_lists, center_inland_dry);
		end
		if iNumMiddleRiver > 0 then -- Process middle donut river plots.
			table.insert(candidate_lists, middle_river);
		end
		if iNumMiddleCoastLake > 0 then -- Process middle donut lake or coastal plots.
			table.insert(candidate_lists, middle_coastal);
		end
		if iNumMiddleInlandDry > 0 then -- Process middle donut inland dry plots.
			table.insert(candidate_lists, middle_inland_dry);
		end
		--
		for loop, plot_list in ipairs(candidate_lists) do -- Up to six plot lists, processed by priority.
			local election_returns = self:IterateThroughCandidatePlotList(plot_list, region_type)
			-- If any candidates are eligible, choose one.
			local found_eligible = election_returns[1];
			if found_eligible then
				local bestPlotScore = election_returns[2]; 
				local bestPlotIndex = election_returns[3];
				local x = (bestPlotIndex - 1) % g_iW;
				local y = math.floor((bestPlotIndex - x - 1) / g_iW);
				self.startingPlots[region_number] = {x, y, bestPlotScore};
				self:PlaceImpactAndRipples(x, y)
				return true, false
			end
			-- If none eligible, check for fallback plot.
			local found_fallback = election_returns[4];
			if found_fallback then
				local bestFallbackScore = election_returns[5];
				local bestFallbackIndex = election_returns[6];
				local x = (bestFallbackIndex - 1) % g_iW;
				local y = math.floor((bestFallbackIndex - x - 1) / g_iW);
				table.insert(fallback_plots, {x, y, bestFallbackScore});
			end
		end
	end
	-- Reaching this point means no eligible sites in center bias or middle donut subregions!
	
	-- Process candidates from Outer subregion, if any.
	if iNumOuter > 0 then
		local outer_eligible_list = {};
		local found_eligible = false;
		local found_fallback = false;
		local bestFallbackScore = -50;
		local bestFallbackIndex;
		-- Process list of candidate plots.
		for loop, plotIndex in ipairs(outer_plots) do
			local score, meets_minimums = self:EvaluateCandidatePlot(plotIndex, region_type)
			-- Test current plot against best known plot.
			if meets_minimums == true then
				found_eligible = true;
				table.insert(outer_eligible_list, plotIndex);
			else
				found_fallback = true;
				if score > bestFallbackScore then
					bestFallbackScore = score;
					bestFallbackIndex = plotIndex;
				end
			end
		end
		if found_eligible then 
			-- Iterate through eligible plots and choose the one 
			-- closest to the center of the region.
			local x, y, closestPlot = FindClosestPlot(
				outer_eligible_list, iWidth, iHeight, iWestX, iSouthY);
			-- Re-get plot score for inclusion in start plot data.
			local score, meets_minimums = self:EvaluateCandidatePlot(closestPlot, region_type)
			-- Assign this plot as the start for this region.
			self.startingPlots[region_number] = {x, y, score};
			self:PlaceImpactAndRipples(x, y)
			return true, false
		end
		-- Add the fallback plot (best scored plot) from the Outer region to the fallback list.
		if found_fallback then
			local x = (bestFallbackIndex - 1) % g_iW;
			local y = math.floor((bestFallbackIndex - x - 1) / g_iW);
			table.insert(fallback_plots, {x, y, bestFallbackScore});
		end
	end
	-- Reaching here means no plot in the entire region met the minimum standards for selection.
	
	-- The fallback plot contains the best-scored plots from each test area in this region.
	-- We will compare all the fallback plots and choose the best to be the start plot.
	local iNumFallbacks = table.maxn(fallback_plots);
	if iNumFallbacks > 0 then
		local best_fallback_score = 0
		local best_fallback_x;
		local best_fallback_y;
		for loop, plotData in ipairs(fallback_plots) do
			local score = plotData[3];
			if score > best_fallback_score then
				best_fallback_score = score;
				best_fallback_x = plotData[1];
				best_fallback_y = plotData[2];
			end
		end
		-- Assign the start for this region.
		self.startingPlots[region_number] = {best_fallback_x, best_fallback_y, best_fallback_score};
		self:PlaceImpactAndRipples(best_fallback_x, best_fallback_y)
		bSuccessFlag = true;
	else
		-- This region cannot have a start and something has gone way wrong.
		
		-- Look for a passable (non-mountain, etc) tile to avoid breaking mountain pattern 
		-- reachability.
		local plotX, plotY = FindPassableForceStart(iWidth, iHeight, iWestX, iSouthY);
		bSuccessFlag = true;
		bForcedPlacementFlag = true;
		self.startingPlots[region_number] = {plotX, plotY, 0};
		self:PlaceImpactAndRipples(plotX, plotY)
	end

	return bSuccessFlag, bForcedPlacementFlag
end

function GotLakes_ASP:FindCoastalStart(region_number)
	-- This function attempts to choose a start position (which is along an ocean) for a single region.
	-- This function returns two boolean flags, indicating the success level of the operation.
	local bSuccessFlag = false; -- Returns true when a start is placed, false when process fails.
	local bForcedPlacementFlag = false; -- Returns true if this region had no eligible starts and one was forced to occur.
	
	-- Obtain data needed to process this region.
	
	local region_data_table = self.regionData[region_number];
	local iWestX = region_data_table[1];
	local iSouthY = region_data_table[2];
	local iWidth = region_data_table[3];
	local iHeight = region_data_table[4];
	local iAreaID = region_data_table[5];
	local iMembershipEastX = iWestX + iWidth - 1;
	local iMembershipNorthY = iSouthY + iHeight - 1;
	--
	local terrainCounts = self.regionTerrainCounts[region_number];
	local coastalLandCount = terrainCounts[22];
	--
	local region_type = self.regionTypes[region_number];
	-- Done setting up region data.
	-- Set up contingency.
	local fallback_plots = {};
	
	print("Looking for coastal start at " .. CoordString(iWestX, iSouthY) .. 
			" with region size " .. CoordString(iWidth, iHeight));
	
	-- Check region for AlongOcean eligibility.
	if coastalLandCount < 3 then
		-- This region cannot support an Along Ocean start. Try instead to find an inland start for it.
		bSuccessFlag, bForcedPlacementFlag = self:FindStart(region_number)
		if bSuccessFlag == false then
			-- This region cannot have a start and something has gone way wrong.
			-- Look for a passable (non-mountain, etc) tile to avoid breaking mountain pattern 
			-- reachability.
			local plotX, plotY = FindPassableForceStart(iWidth, iHeight, iWestX, iSouthY);
			bSuccessFlag = true;
			bForcedPlacementFlag = true;
			self.startingPlots[region_number] = {plotX, plotY, 0};
			self:PlaceImpactAndRipples(plotX, plotY)
		end
		return bSuccessFlag, bForcedPlacementFlag
	end

	-- Establish scope of center bias.
	local fCenterWidth = (self.centerBias / 100) * iWidth;
	local iNonCenterWidth = math.floor((iWidth - fCenterWidth) / 2)
	local iCenterWidth = iWidth - (iNonCenterWidth * 2);
	local iCenterWestX = (iWestX + iNonCenterWidth) % g_iW; -- Modulo math to synch coordinate to actual map in case of world wrap.
	local iCenterTestWestX = (iWestX + iNonCenterWidth); -- "Test" values ignore world wrap for easier membership testing.
	local iCenterTestEastX = (iCenterWestX + iCenterWidth - 1);

	local fCenterHeight = (self.centerBias / 100) * iHeight;
	local iNonCenterHeight = math.floor((iHeight - fCenterHeight) / 2)
	local iCenterHeight = iHeight - (iNonCenterHeight * 2);
	local iCenterSouthY = (iSouthY + iNonCenterHeight) % g_iH;
	local iCenterTestSouthY = (iSouthY + iNonCenterHeight);
	local iCenterTestNorthY = (iCenterTestSouthY + iCenterHeight - 1);

	-- Establish scope of "middle donut", outside the center but inside the outer.
	local fMiddleWidth = (self.middleBias / 100) * iWidth;
	local iOuterWidth = math.floor((iWidth - fMiddleWidth) / 2)
	local iMiddleWidth = iWidth - (iOuterWidth * 2);
	--local iMiddleDiameterX = (iMiddleWidth - iCenterWidth) / 2;
	local iMiddleWestX = (iWestX + iOuterWidth) % g_iW;
	local iMiddleTestWestX = (iWestX + iOuterWidth);
	local iMiddleTestEastX = (iMiddleTestWestX + iMiddleWidth - 1);

	local fMiddleHeight = (self.middleBias / 100) * iHeight;
	local iOuterHeight = math.floor((iHeight - fMiddleHeight) / 2)
	local iMiddleHeight = iHeight - (iOuterHeight * 2);
	--local iMiddleDiameterY = (iMiddleHeight - iCenterHeight) / 2;
	local iMiddleSouthY = (iSouthY + iOuterHeight) % g_iH;
	local iMiddleTestSouthY = (iSouthY + iOuterHeight);
	local iMiddleTestNorthY = (iMiddleTestSouthY + iMiddleHeight - 1); 

	-- Assemble candidates lists.
	local center_coastal_plots = {};
	local center_plots_on_river = {};
	local center_fresh_plots = {};
	local center_dry_plots = {};
	local middle_coastal_plots = {};
	local middle_plots_on_river = {};
	local middle_fresh_plots = {};
	local middle_dry_plots = {};
	local outer_coastal_plots = {};
	
	-- Identify candidate plots.
	for region_y = 0, iHeight - 1 do -- When handling global plot indices, process Y first.
		for region_x = 0, iWidth - 1 do
			local x = (region_x + iWestX) % g_iW; -- Actual coords, adjusted for world wrap, if any.
			local y = (region_y + iSouthY) % g_iH; --
			local plotIndex = y * g_iW + x + 1;
			if self.plotDataIsCoastal[plotIndex] == true then -- This plot is a land plot next to an ocean.
				local plot = Map.GetPlot(x, y);
				if IsPassableLand(plot) and IsFarFromNaturalWonder(plot,false) then 
					-- Not a mountain plot and sufficiently far from a natural wonder.
					local area_of_plot = plot:GetArea():GetID();
					if area_of_plot == iAreaID or iAreaID == -1 then 
						-- This plot is a member, so it goes on at least one candidate list.
						--
						-- Test whether plot is in center bias, middle donut, or outer donut.
						--
						local test_x = region_x + iWestX; -- "Test" coords, ignoring any world wrap and
						local test_y = region_y + iSouthY; -- reaching in to virtual space if necessary.
						if (test_x >= iCenterTestWestX and test_x <= iCenterTestEastX) and 
						   (test_y >= iCenterTestSouthY and test_y <= iCenterTestNorthY) then
							table.insert(center_coastal_plots, plotIndex);
							if plot:IsRiver() then
								table.insert(center_plots_on_river, plotIndex);
							elseif plot:IsFreshWater() then
								table.insert(center_fresh_plots, plotIndex);
							else
								table.insert(center_dry_plots, plotIndex);
							end
						elseif (test_x >= iMiddleTestWestX and test_x <= iMiddleTestEastX) and 
						       (test_y >= iMiddleTestSouthY and test_y <= iMiddleTestNorthY) then
							table.insert(middle_coastal_plots, plotIndex);
							if plot:IsRiver() then
								table.insert(middle_plots_on_river, plotIndex);
							elseif plot:IsFreshWater() then
								table.insert(middle_fresh_plots, plotIndex);
							else
								table.insert(middle_dry_plots, plotIndex);
							end
						else
							table.insert(outer_coastal_plots, plotIndex);
						end
					end
				end
			end
		end
	end
	-- Check how many plots landed on each list.
	local iNumCenterCoastal = table.maxn(center_coastal_plots);
	local iNumCenterRiver = table.maxn(center_plots_on_river);
	local iNumCenterFresh = table.maxn(center_fresh_plots);
	local iNumCenterDry = table.maxn(center_dry_plots);
	local iNumMiddleCoastal = table.maxn(middle_coastal_plots);
	local iNumMiddleRiver = table.maxn(middle_plots_on_river);
	local iNumMiddleFresh = table.maxn(middle_fresh_plots);
	local iNumMiddleDry = table.maxn(middle_dry_plots);
	local iNumOuterCoastal = table.maxn(outer_coastal_plots);
	
	-- Process lists of candidate plots.
	if iNumCenterCoastal + iNumMiddleCoastal > 0 then
		local candidate_lists = {};
		if iNumCenterRiver > 0 then -- Process center bias river plots.
			table.insert(candidate_lists, center_plots_on_river);
		end
		if iNumCenterFresh > 0 then -- Process center bias fresh water plots that are not rivers.
			table.insert(candidate_lists, center_fresh_plots);
		end
		if iNumCenterDry > 0 then -- Process center bias dry plots.
			table.insert(candidate_lists, center_dry_plots);
		end
		if iNumMiddleRiver > 0 then -- Process middle bias river plots.
			table.insert(candidate_lists, middle_plots_on_river);
		end
		if iNumMiddleFresh > 0 then -- Process middle bias fresh water plots that are not rivers.
			table.insert(candidate_lists, middle_fresh_plots);
		end
		if iNumMiddleDry > 0 then -- Process middle bias dry plots.
			table.insert(candidate_lists, middle_dry_plots);
		end
		--
		for loop, plot_list in ipairs(candidate_lists) do -- Up to six plot lists, processed by priority.
			local election_returns = self:IterateThroughCandidatePlotList(plot_list, region_type)
			-- If any riverside candidates are eligible, choose one.
			local found_eligible = election_returns[1];
			if found_eligible then
				local bestPlotScore = election_returns[2]; 
				local bestPlotIndex = election_returns[3];
				local x = (bestPlotIndex - 1) % g_iW;
				local y = math.floor((bestPlotIndex - x - 1) / g_iW);
				self.startingPlots[region_number] = {x, y, bestPlotScore};
				self:PlaceImpactAndRipples(x, y)
				return true, false
			end
			-- If none eligible, check for fallback plot.
			local found_fallback = election_returns[4];
			if found_fallback then
				local bestFallbackScore = election_returns[5];
				local bestFallbackIndex = election_returns[6];
				local x = (bestFallbackIndex - 1) % g_iW;
				local y = math.floor((bestFallbackIndex - x - 1) / g_iW);
				table.insert(fallback_plots, {x, y, bestFallbackScore});
			end
		end
	end
	-- Reaching this point means no strong coastal sites in center bias or middle donut subregions!
	
	-- Process candidates from Outer subregion, if any.
	if iNumOuterCoastal > 0 then
		local outer_eligible_list = {};
		local found_eligible = false;
		local found_fallback = false;
		local bestFallbackScore = -50;
		local bestFallbackIndex;
		-- Process list of candidate plots.
		for loop, plotIndex in ipairs(outer_coastal_plots) do
			local score, meets_minimums = self:EvaluateCandidatePlot(plotIndex, region_type)
			-- Test current plot against best known plot.
			if meets_minimums == true then
				found_eligible = true;
				table.insert(outer_eligible_list, plotIndex);
			else
				found_fallback = true;
				if score > bestFallbackScore then
					bestFallbackScore = score;
					bestFallbackIndex = plotIndex;
				end
			end
		end
		if found_eligible then 
			-- Iterate through eligible plots and choose the one 
			-- closest to the center of the region.
			local x, y, closestPlot = FindClosestPlot(
				outer_eligible_list, iWidth, iHeight, iWestX, iSouthY);
			-- Re-get plot score for inclusion in start plot data.
			local score, meets_minimums = self:EvaluateCandidatePlot(closestPlot, region_type)
			-- Assign this plot as the start for this region.
			self.startingPlots[region_number] = {x, y, score};
			self:PlaceImpactAndRipples(x, y);
			return true, false
		end
		-- Add the fallback plot (best scored plot) from the Outer region to the fallback list.
		if found_fallback then
			local x = (bestFallbackIndex - 1) % g_iW;
			local y = math.floor((bestFallbackIndex - x - 1) / g_iW);
			table.insert(fallback_plots, {x, y, bestFallbackScore});
		end
	end
	-- Reaching here means no plot in the entire region met the minimum standards for selection.
	
	-- The fallback plot contains the best-scored plots from each test area in this region.
	-- This region must be something awful on food, or had too few coastal plots with none being decent.
	-- We will compare all the fallback plots and choose the best to be the start plot.
	local iNumFallbacks = table.maxn(fallback_plots);
	if iNumFallbacks > 0 then
		local best_fallback_score = 0
		local best_fallback_x;
		local best_fallback_y;
		for loop, plotData in ipairs(fallback_plots) do
			local score = plotData[3];
			if score > best_fallback_score then
				best_fallback_score = score;
				best_fallback_x = plotData[1];
				best_fallback_y = plotData[2];
			end
		end
		-- Assign the start for this region.
		self.startingPlots[region_number] = {best_fallback_x, best_fallback_y, best_fallback_score};
		self:PlaceImpactAndRipples(best_fallback_x, best_fallback_y)
		bSuccessFlag = true;
	else
		-- This region cannot support an Along Ocean start. Try instead to find an Inland start for it.
		bSuccessFlag, bForcedPlacementFlag = self:FindStart(region_number)
		if bSuccessFlag == false then
			-- This region cannot have a start and something has gone way wrong.
			-- Look for a passable (non-mountain, etc) tile to avoid breaking mountain pattern 
			-- reachability.
			local plotX, plotY = FindPassableForceStart(iWidth, iHeight, iWestX, iSouthY);
			bSuccessFlag = true;
			bForcedPlacementFlag = true;
			self.startingPlots[region_number] = {plotX, plotY, 0};
			self:PlaceImpactAndRipples(plotX, plotY);
		end
	end

	return bSuccessFlag, bForcedPlacementFlag
end

function GotLakes_ASP:FindStartWithoutRegardToAreaID(region_number, bMustBeCoast)
	-- This function attempts to choose a start position on the best AreaID section within the Region's rectangle.
	-- This function returns two boolean flags, indicating the success level of the operation.
	local bSuccessFlag = false; -- Returns true when a start is placed, false when process fails.
	local bForcedPlacementFlag = false; -- Returns true if this region had no eligible starts and one was forced to occur.
	
	-- Obtain data needed to process this region.
	local region_data_table = self.regionData[region_number];
	local iWestX = region_data_table[1];
	local iSouthY = region_data_table[2];
	local iWidth = region_data_table[3];
	local iHeight = region_data_table[4];
	
	local wrapX = ScriptData.wrapX;
	local wrapY = ScriptData.wrapY;
	local iMembershipEastX = iWestX + iWidth - 1;
	local iMembershipNorthY = iSouthY + iHeight - 1;
	--
	local region_type = self.regionTypes[region_number];
	local fallback_plots = {};
	-- Done setting up region data.

	print("Looking for start and ignoring area ID at " .. CoordString(iWestX, iSouthY) .. 
			" with region size " .. CoordString(iWidth, iHeight));
	
	-- Obtain info on all landmasses wholly or partially within this region, for comparision purposes.
	local regionalFertilityOfLands = {};
	local iRegionalFertilityOfLands = 0;
	local iNumLandPlots = 0;
	local iNumLandAreas = 0;
	local land_area_IDs = {};
	local land_area_plots = {};
	local land_area_fert = {};
	local land_area_plot_lists = {};
	-- Cycle through all plots in the region, checking their Start Placement Fertility and AreaID.
	for region_y = 0, iHeight - 1 do
		for region_x = 0, iWidth - 1 do
			local x = region_x + iWestX;
			local y = region_y + iSouthY;
			local plot = Map.GetPlot(x, y);
			if IsPassableLand(plot) and IsFarFromNaturalWonder(plot,false) then 
				-- This plot could host a city; process it.
				iNumLandPlots = iNumLandPlots + 1;
				local iArea = plot:GetArea():GetID();
				local plotFertility = self:MeasureStartPlacementFertilityOfPlot(x, y, false); -- Check for coastal land is disabled.
				iRegionalFertilityOfLands = iRegionalFertilityOfLands + plotFertility;
				if TestMembership(land_area_IDs, iArea) == false then -- This plot is the first detected in its AreaID.
					iNumLandAreas = iNumLandAreas + 1;
					table.insert(land_area_IDs, iArea);
					land_area_plots[iArea] = 1;
					land_area_fert[iArea] = plotFertility;
				else -- This AreaID already known.
					land_area_plots[iArea] = land_area_plots[iArea] + 1;
					land_area_fert[iArea] = land_area_fert[iArea] + plotFertility;
				end
			end
		end
	end

	-- Generate empty (non-nil) tables for each Area ID in the plot lists matrix.
	for loop, areaID in ipairs(land_area_IDs) do
		land_area_plot_lists[areaID] = {};
	end
	-- Cycle through all plots in the region again, adding candidates to the applicable AreaID plot list.
	for region_y = 0, iHeight - 1 do
		for region_x = 0, iWidth - 1 do
			local x = region_x + iWestX;
			local y = region_y + iSouthY;
			local i = y * g_iW + x + 1;
			local plot = Map.GetPlot(x, y);
			local iArea = plot:GetArea():GetID();
			if IsPassableLand(plot) and IsFarFromNaturalWonder(plot,false) then
				-- Land plot, process it.
				table.insert(land_area_plot_lists[iArea], i);
			end
		end
	end
	
	local best_areas = {};
	local regionAreaListUnsorted = {};
	local regionAreaListSorted = {}; -- Have to make this a separate table, not merely a pointer to the first table.
	for areaNum, fert in pairs(land_area_fert) do
		table.insert(regionAreaListUnsorted, {areaNum, fert});
		table.insert(regionAreaListSorted, fert);
	end
	table.sort(regionAreaListSorted);
	
	-- Match each sorted fertilty value to the matching unsorted AreaID number and record in sequence.
	local iNumAreas = table.maxn(regionAreaListSorted);
	for area_order = iNumAreas, 1, -1 do -- Best areas are at the end of the list, so run the list backward.
		for loop, data_pair in ipairs(regionAreaListUnsorted) do
			local unsorted_fert = data_pair[2];
			if regionAreaListSorted[area_order] == unsorted_fert then
				local unsorted_area_num = data_pair[1];
				table.insert(best_areas, unsorted_area_num);
				-- HAVE TO remove the entry from the table in case of ties on fert value.
				table.remove(regionAreaListUnsorted, loop);
				break
			end
		end
	end

	--[[ Debug printout.
	print("-");
	print("--- Number of Candidate Plots in each landmass in Region #", region_number, " - Region Type:", region_type, " ---");
	print("-");
	for loop, iAreaID in ipairs(best_areas) do
		local fert_rating = land_area_fert[iAreaID];
		local plotCount = table.maxn(land_area_plot_lists[iAreaID]);
		print("* Area ID#", iAreaID, "has fertility rating of", fert_rating, "and candidate plot count of", plotCount); print("-");
	end
	print("- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -");
	--]]--

	-- Now iterate through areas, from best fertility downward, looking for a site good enough to choose.
	for loop, iAreaID in ipairs(best_areas) do
		local plot_list = land_area_plot_lists[iAreaID];
		local election_returns = self:IterateThroughCandidatePlotList(plot_list, region_type)
		-- If any plots in this area are eligible, choose one.
		local found_eligible = election_returns[1];
		if found_eligible then
			local bestPlotScore = election_returns[2]; 
			local bestPlotIndex = election_returns[3];
			local x = (bestPlotIndex - 1) % g_iW;
			local y = math.floor((bestPlotIndex - x - 1) / g_iW);
			self.startingPlots[region_number] = {x, y, bestPlotScore};
			self:PlaceImpactAndRipples(x, y)
			return true, false
		end
		-- If none eligible, check for fallback plot.
		local found_fallback = election_returns[4];
		if found_fallback then
			local bestFallbackScore = election_returns[5];
			local bestFallbackIndex = election_returns[6];
			local x = (bestFallbackIndex - 1) % g_iW;
			local y = math.floor((bestFallbackIndex - x - 1) / g_iW);
			table.insert(fallback_plots, {x, y, bestFallbackScore});
		end
	end
	-- Reaching this point means no strong sites far enough away from any already-placed start points.

	-- We will compare all the fallback plots and choose the best to be the start plot.
	local iNumFallbacks = table.maxn(fallback_plots);
	if iNumFallbacks > 0 then
		local best_fallback_score = 0
		local best_fallback_x;
		local best_fallback_y;
		for loop, plotData in ipairs(fallback_plots) do
			local score = plotData[3];
			if score > best_fallback_score then
				best_fallback_score = score;
				best_fallback_x = plotData[1];
				best_fallback_y = plotData[2];
			end
		end
		-- Assign the start for this region.
		self.startingPlots[region_number] = {best_fallback_x, best_fallback_y, best_fallback_score};
		self:PlaceImpactAndRipples(best_fallback_x, best_fallback_y)
		bSuccessFlag = true;
	else
		-- This region cannot have a start and something has gone way wrong.
		-- Look for a passable (non-mountain, etc) tile to avoid breaking mountain pattern 
		-- reachability.
		local plotX, plotY = FindPassableForceStart(iWidth, iHeight, iWestX, iSouthY);
		bSuccessFlag = true;
		bForcedPlacementFlag = true;
		self.startingPlots[region_number] = {plotX, plotY, 0};
		self:PlaceImpactAndRipples(plotX, plotY);
	end

	return bSuccessFlag, bForcedPlacementFlag
end

function GotLakes_ASP:ChooseLocations()
	print("Got Lakes ASP: Choosing Start Locations for Civilizations");
	local args = ScriptData.landInfo.aspLocationArgs;
	local mustBeCoast = args.mustBeCoast or false; -- if true, will force all starts on salt water coast if possible
	
	-- Defaults for evaluating potential start plots are assigned in .Create but args
	-- passed in here can override. If args value for a field is nil (no arg) then
	-- these assignments will keep the default values in place.
	self.centerBias = args.centerBias or self.centerBias; -- % of radius from region center to examine first
	self.middleBias = args.middleBias or self.middleBias; -- % of radius from region center to check second
	self.minFoodInner = args.minFoodInner or self.minFoodInner;
	self.minProdInner = args.minProdInner or self.minProdInner;
	self.minGoodInner = args.minGoodInner or self.minGoodInner;
	self.minFoodMiddle = args.minFoodMiddle or self.minFoodMiddle;
	self.minProdMiddle = args.minProdMiddle or self.minProdMiddle;
	self.minGoodMiddle = args.minGoodMiddle or self.minGoodMiddle;
	self.minFoodOuter = args.minFoodOuter or self.minFoodOuter;
	self.minProdOuter = args.minProdOuter or self.minProdOuter;
	self.minGoodOuter = args.minGoodOuter or self.minGoodOuter;
	self.maxJunk = args.maxJunk or self.maxJunk;

	
	-- Get the size in tiles of the biggest chunk of land.
	-- Remember this size when checking for tiny islands.
	local biggest_area = Areas.FindBiggestArea(false);
	local biggestAreaNumTiles = biggest_area:GetPlotCount();
	print("Biggest land chunk has " .. biggestAreaNumTiles .. " tiles.");
	
	local minLandSizeMultiplier = ScriptData.landInfo.aspStartAreaRatio
	
	if (UserInput.Mountains == MtnPatterns.WINDING) then
		print("Reverting to default starting land-size threshold (winding canyons)");
		minLandSizeMultiplier = 0.25;
	end
	
	-- Minimum number of land tiles immediately reachable from the candidate plot.
	-- Any less than this and the candidate becomes a fallback
	-- to avoid starting on tiny islands, mountain-blocked peninsulas, etc.
	self.majorStartMinLandSize = 
		math.clamp(Round(biggestAreaNumTiles * minLandSizeMultiplier), 1, 100);
	
	-- Measure terrain/plot/feature in regions.
	self:MeasureTerrainInRegions()
	
	-- Determine region type.
	self:DetermineRegionTypes()

	-- Set up list of regions (to be processed in this order).
	--
	-- First, make a list of all average fertility values...
	local regionAssignList = {};
	local averageFertilityListUnsorted = {};
	local averageFertilityListSorted = {}; -- Have to make this a separate table, not merely a pointer to the first table.
	for i, region_data in ipairs(self.regionData) do
		local thisRegionAvgFert = region_data[8];
		table.insert(averageFertilityListUnsorted, {i, thisRegionAvgFert});
		table.insert(averageFertilityListSorted, thisRegionAvgFert);
	end
	-- Now sort the copy low to high.
	table.sort(averageFertilityListSorted);
	-- Finally, match each sorted fertilty value to the matching unsorted region number and record in sequence.
	local iNumRegions = table.maxn(averageFertilityListSorted);
	for region_order = 1, iNumRegions do
		for loop, data_pair in ipairs(averageFertilityListUnsorted) do
			local unsorted_fert = data_pair[2];
			if averageFertilityListSorted[region_order] == unsorted_fert then
				local unsorted_reg_num = data_pair[1];
				table.insert(regionAssignList, unsorted_reg_num);
				-- HAVE TO remove the entry from the table in rare case of ties on fert 
				-- value. Or it will just match this value for a second time, then crash 
				-- when the region it was tied with ends up with nil data.
				table.remove(averageFertilityListUnsorted, loop);
				break
			end
		end
	end

	-- main loop
	for assignIndex = 1, iNumRegions do
		local currentRegionNumber = regionAssignList[assignIndex];
		local bSuccessFlag = false;
		local bForcedPlacementFlag = false;
		
		-- Turn region boundaries/dimensions into whole numbers 
		-- so that coordinates are also whole numbers.
		local region_data_table = self.regionData[currentRegionNumber];
		region_data_table[1] = Round(region_data_table[1]);
		region_data_table[2] = Round(region_data_table[2]);
		region_data_table[3] = Round(region_data_table[3]);
		region_data_table[4] = Round(region_data_table[4]);
		
		if self.method == 3 or self.method == 4 then
			bSuccessFlag, bForcedPlacementFlag = self:FindStartWithoutRegardToAreaID(currentRegionNumber, mustBeCoast)
		elseif mustBeCoast == true then
			bSuccessFlag, bForcedPlacementFlag = self:FindCoastalStart(currentRegionNumber)
		else
			bSuccessFlag, bForcedPlacementFlag = self:FindStart(currentRegionNumber)
		end
		
		--[[ Printout for debug only.
		print("- - -");
		print("Start Plot for Region #", currentRegionNumber, " was successful: ", bSuccessFlag);
		print("Start Plot for Region #", currentRegionNumber, " was forced: ", bForcedPlacementFlag);
		]]--		
	end
	--

	--[ Printout of start plots. Debug use only.
	print("-");
	print("--- Table of results, GotLakes_ASP ---");
	for loop, startData in ipairs(self.startingPlots) do
		print("-");
		print("Region#", loop, " has start plot at: ", startData[1], startData[2], "with Fertility Rating of ", startData[3]);
	end
	print("-");
	print("--- Table of results, GotLakes_ASP ---");
	print("-");
	--]]
end
------------------------------------------------------------------------------
-- Overrides for the system that determines civ starting plots.
------------------------------------------------------------------------------
function AssignStartingPlots:__InitStartingData()
	-- Override for improved stability.
	
	if(self.uiMinMajorCivFertility <= 0) then
		self.uiMinMajorCivFertility = 5;
	end

	if(self.uiMinMinorCivFertility <= 0) then
		self.uiMinMinorCivFertility = 5;
	end

	if(self.uiMinBarbarianFertility<= 0) then
		self.uiMinBarbarianFertility = 1;
	end
	
	
	-- Find Default Number of players.
	-- Used for luxury spacing and major/minor civ-start buffers.
	MapSizeTypes = {};
	for row in GameInfo.Maps() do
		MapSizeTypes[row.RowId] = row.DefaultPlayers;
	end
	local sizekey = Map.GetMapSize() + 1;
	local iDefaultNumberPlayers = MapSizeTypes[sizekey] or 8;
	self.iDefaultNumberMajor = iDefaultNumberPlayers;
	self.iDefaultNumberMinor = math.floor(iDefaultNumberPlayers * 1.5);
	
	
	self.iNumMajorCivs = PlayerManager.GetAliveMajorsCount();
	self.iNumMinorCivs = PlayerManager.GetAliveMinorsCount();
	local iMinNumBarbarians = self.iNumMajorCivs / 2;

	local iBonusMajor = math.max(1, math.floor(self.iNumMajorCivs / 2));
	local iBonusMinor = math.max(1, math.floor(self.iNumMinorCivs / 2));

	
	-- Use Custom start positioner.
	local asp = GotLakes_ASP.Create();
	asp:GenerateRegions();
	asp:ChooseLocations();
	
	local startPlots = asp.startingPlots;
	local iMajorCivStartLocs = table.maxn(startPlots);

	if (iMajorCivStartLocs < self.iNumMajorCivs) then
		-- Not enough spots for the major civs. Just bail out.
		print("ERROR: not enough starting spots for " .. self.iNumMajorCivs .. " major civs!");
		return;
	end
	
	-- Export results from our custom start positioner to Civ 6 ASP.
	self.majorStartPlots = {};
	for loop, startData in ipairs(asp.startingPlots) do
		local plot = Map.GetPlot(startData[1], startData[2]);
		table.insert(self.majorStartPlots, plot);
		table.insert(self.majorCopy, plot);
	end
	
	-- Run Civ 6's start positioner for city states.
	-- Magical black box to assign a region for each major and minor civ.
	StartPositioner.DivideMapIntoMajorRegions(
		self.iNumMajorCivs + iBonusMajor,
		self.iNumMinorCivs + iBonusMinor, 
		iMinNumBarbarians, 
		self.uiMinMajorCivFertility, 
		self.uiMinMinorCivFertility, 
		self.uiMinBarbarianFertility);
	
	
	--Begin Start Bias for major
	self.majorList = PlayerManager.GetAliveMajorIDs();
	self:__InitStartBias(false);

	if(UserInput.start == START_BALANCED) then
		print("Tuning resources for balanced starts.");
		self:__AddResourcesBalanced();
		
	elseif (UserInput.start == START_LEGEND) then
		print("Tuning resources for legendary starts.");
		self:__AddResourcesLegendary();
	end
	
	for i = 1, self.iNumMajorCivs do
		local player = Players[self.majorList[i]]
		
		if(player == nil) then
			print("Major start #" .. i .. " FAILED");
		else
			local hasPlot = false;
			for k, v in pairs(self.playerStarts[i]) do
				if(v~= nil and hasPlot == false) then
					hasPlot = true;
					player:SetStartingPlot(v);
					print("Major Start #" .. i .. ": " .. CoordString(v:GetX(),v:GetY()));
				end
			end
		end
	end

	--Place the minor start plots in an array
	self.minorStartPlots = {};
	StartPositioner.DivideUnusedRegions();
	local iMinorCivStartLocs = StartPositioner.GetNumMinorCivStarts();
	local iBarbarianStartLocs = StartPositioner.GetNumBarbarianStarts();
	local i = 0;
	local valid = 0;
	while i <= iMinorCivStartLocs - 1 and valid < self.iNumMinorCivs do
		plots = StartPositioner.GetMinorCivStartPlots(i);
		local startPlot = self:__SetStartMinor(plots);
		info = StartPositioner.GetMinorCivStartInfo(i);
		if(startPlot ~= nil) then
			table.insert(self.minorStartPlots, startPlot);
			--print ("Minor ContinentType: " .. tostring(info.ContinentType));
			--print ("Minor LandmassID: " .. tostring(info.LandmassID));
			--print ("Minor Fertility: " .. tostring(info.Fertility));
			--print ("Minor TotalPlots: " .. tostring(info.TotalPlots));
			--print ("Minor WestEdge: " .. tostring(info.WestEdge));
			--print ("Minor EastEdge: " .. tostring(info.EastEdge));
			--print ("Minor NorthEdge: " .. tostring(info.NorthEdge));
			--print ("Minor SouthEdge: " .. tostring(info.SouthEdge));
			--print("Minor Tried to Start X: ", plot:GetX(), "Minor Tried to Start Y: ", plot:GetY());
			valid = valid + 1;
		else
			--print ("BAAAD Minor ContinentType: " .. tostring(info.ContinentType));
			--print ("BAAAD Minor LandmassID: " .. tostring(info.LandmassID));
			--print ("BAAAD Minor Fertility: " .. tostring(info.Fertility));
			--print ("BAAAD Minor TotalPlots: " .. tostring(info.TotalPlots));
			--print ("BAAAD Minor WestEdge: " .. tostring(info.WestEdge));
			--print ("BAAAD Minor EastEdge: " .. tostring(info.EastEdge));
			--print ("BAAAD Minor NorthEdge: " .. tostring(info.NorthEdge));
			--print ("BAAAD Minor SouthEdge: " .. tostring(info.SouthEdge));
			--print("faILed MINOR");
		end
		
		i = i + 1;
	end

	for k, plot in ipairs(self.minorStartPlots) do
		table.insert(self.minorCopy, plot);
	end

	--Begin Start Bias for minor
	self.minorList = PlayerManager.GetAliveMinorIDs();
	self:__InitStartBias(true);

	for i = 1, self.iNumMinorCivs do
		local player = Players[self.minorList[i]]
		
		if(player == nil) then
			print("Minor start #" .. i .. " FAILED");
		else
			local hasPlot = false;
			for k, v in pairs(self.playerStarts[i + self.iNumMajorCivs]) do
				if(v~= nil and hasPlot == false) then
					hasPlot = true;
					player:SetStartingPlot(v);
					print("Minor Start #" .. i .. ": " .. CoordString(v:GetX(),v:GetY()));
				end
			end
		end
	end
end

function AssignStartingPlots:__GetValidAdjacent(plot, minor)
	-- Override to be more permissive w.r.t. challenging terrain.

	local maxH = 0;
	local minH = 0;
	if(minor == 0) then
		max = math.ceil(g_iH * self.uiStartMaxY / 100);
		min = math.ceil(g_iH * self.uiStartMinY / 100);
	end

	if(plot:GetY() <= minH or plot:GetY() > g_iH - maxH) then
		return false;
	end

	-- Allow some barren tiles next to the start, because a suboptimal start is
	-- better than no start at all, especially on a map with challenging terrain.
	return true;
end
------------------------------------------------------------------------------
-- World Size configs
--
-- Note that the info here includes *suggested* sizes for playing the game; 
-- actual world size might be different in the WB.
------------------------------------------------------------------------------
WorldSizeInfo = 
{
	default =
	{
		-- Map size as {Length, Width}
		size = {84, 54}, -- Civ V: {84, 52}, 
		
		-- The default number of civs for the map size.
		numCivs = 8,
		
		-- Base fractal grain for MLF plate tectonics.
		mlfGrain = 4,
		
		-- Base fractal grain for mountain patterns.
		mtnGrain  = 1, 
		
		-- Number of plates for tectonic mountain pattern.
		numPlates = 18,
		
		-- Number of plates in Fractal World's ridges
		numRPlates = 20,
		
		-- Island Clusters configuration.
		clusters = {},
		
		-- Mini-Donut configuration.
		miniDonuts = {},
		
		-- Latitude variation fractal grain.
		latVariationGrain = 4,
	},

	duel = 
	{
		size = {44, 26}, -- Civ V: {40, 24},
		
		numCivs = 2,
		
		mlfGrain   = 3,
		mtnGrain   = 0,
		numPlates  = 6,
		numRPlates = 4,
		
		miniDonuts = 
		{
			numExtraDonuts = 3,
			radiusMin = 1,
			radiusMax = 1,
		},
		
		latVariationGrain = 3,
	},
	
	tiny = 
	{
		size = {60, 38}, -- Civ V: {52, 32},
		
		numCivs = 4,
		
		mlfGrain   = 3,
		mtnGrain   = 0, 
		numPlates  = 9, 
		numRPlates = 8,
		
		clusters = {targetSizeOffset = 3},
		
		miniDonuts = 
		{
			numExtraDonuts = 6,
			radiusMin = 1,
			radiusMax = 2,
		},
		
		latVariationGrain = 3,
	},
	
	small = 
	{
		size = {74, 46}, -- Civ V: {64, 40}
		
		numCivs = 6,
		
		mlfGrain   = 4,
		mtnGrain   = 1,  
		numPlates  = 12,
		numRPlates = 16,
				
		clusters = {targetSizeOffset = 5},
		
		miniDonuts = 
		{
			numExtraDonuts = 8,
			radiusMin = 1,
			radiusMax = 3,
		},
		
		latVariationGrain = 3,
	},
	
	standard = 
	{
		size = {84, 54}, -- Civ V: {84, 52},
		
		numCivs = 8,
		
		mlfGrain   = 4,
		mtnGrain   = 1, 
		numPlates  = 18,
		numRPlates = 20,
		
		clusters = {targetSizeOffset = 10},	

		latVariationGrain = 4,		
	},
	
	large = 
	{
		size = {96, 60}, -- Civ V: {104, 64},
		
		numCivs = 10,
		
		mlfGrain   = 5,
		mtnGrain   = 2,  
		numPlates  = 24,
		numRPlates = 24,
				
		clusters = {targetSizeOffset = 20},
		
		miniDonuts = 
		{
			numDonutsPerCell = 2,
			numExtraDonuts   = 12,
		},
		
		latVariationGrain = 4,
	},
	
	huge = 
	{
		size = {106, 66}, -- Civ V: {128, 80},
		
		numCivs = 12,
		
		mlfGrain   = 5,
		mtnGrain   = 2, 
		numPlates  = 30, 
		numRPlates = 32,
		
		clusters = {targetSizeOffset = 30},
		
		miniDonuts = 
		{
			numDonutsPerCell = 3,
			numExtraDonuts 	 = 16,
		},
		
		latVariationGrain = 5,
	},
	
	-- Huge for enlarged maps.
	enormous = 
	{
		size = {128, 80}, -- Civ V: {152, 96},
		
		numCivs = 14,
		
		mlfGrain   = 5,
		mtnGrain   = 3,  
		numPlates  = 39, 
		numRPlates = 36,
			
		clusters = {targetSizeOffset = 40},
		
		miniDonuts = 
		{
			numDonutsPerCell = 4,
			numExtraDonuts 	 = 20,
		},
		
		latVariationGrain = 5,
	},
}
------------------------------------------------------------------------------
-- Landmass configs
------------------------------------------------------------------------------
LandInfo = 
{
	default = 
	{
		name        = "unknown land",
		id          = 0,
		description = "",
				
		-- Method to create a fractal world for plot generation.
		-- By default, use FractalWorld.Create().
		CreateFWorld = FractalWorld.Create,
		GenPlotTypes = GotLakes_Fractal.Fractal_GeneratePlotTypes,
		
		-- Args for plot type generation
		fractalInitArgs = {},
		fractalGenArgs  = {},
		mountainArgs    = {terrain_flags={FRAC_POLAR=true}},
		extraArgs       = {},
		coastArgs       = {},
        
		-- Lake grain and sea levels {low, normal, high}
		lake_grain  = 3,
        seas_iceage = {59,64,70},
        seas_normal = {67,72,76},
				
		-- Feature Generation
		-- How many rows of ice to add to the top and bottom of a global map.
		polarIceNumRows = 1,
		
		-- How to apply continents to the map.
		DetermineContinents = Default_DetermineContinents,
		
		-- Starting Plot Database
		aspLandMap  = false, -- true for land-heavy maps
		aspWaterMap = false, -- true for water-heavy maps
		
		-- Major/Minor civ fertility thresholds; 
		-- directly proportional to land/water ratio, 
		-- so raise values for land-heavy maps.
		aspMinMajorFert = 150,
		aspMinMinorFert = 50, 
		
		-- Multiply this by the number of tiles in the biggest landmass
		-- to get the minimum number of reachable tiles for a major 
		-- civ start.
		--
		-- By default, avoid starts that can only reach 1/4 the 
		-- size of the biggest landmass.
		--
		-- A value of 0 effectively disables this.
		-- A value of 1 forces all starts onto the largest landmass.
		--   But beware that mountain ranges can divide land into smaller chunks.
		aspStartAreaRatio = 0.25,
		
		-- Regional division method for civ starts.
		aspDivMethod    = DIVMETHOD_BIGGESTLAND,
		
		-- Args to pass into the ChooseLocations() step of the starting plots system.
		aspLocationArgs = {},
	},
	
	archipelago = 
	{
		name        = "Archipelago",
		id          = Landmasses.ARCHIPELAGO,
		description = "A vast ocean filled with many islands.",
		
		fractalInitArgs = {continent_grain = 4},
		
		mountainArgs = 
		{
			adjust_plates    = 2,
			extra_mountains  = 10,
		},
		
		extraArgs = 
		{
			tectonicGrain  = 4,
		},
        
        seas_iceage = {61,66,73},
        seas_normal = {72,78,83},
				
		aspWaterMap     = true,
		aspMinMajorFert = 75,
		aspMinMinorFert = 5, 
		
		aspDivMethod   = DIVMETHOD_RECTANGULAR,

		-- Forcing starts along the ocean.
		-- Lowering start position minimum eligibility thresholds.
		aspLocationArgs = 
		{
			mustBeCoast   = true,
			minFoodMiddle = 2,
			minFoodOuter  = 3,
			minProdOuter  = 1
		},
	},
	
	bagels = 
	{
		name         = "Bagels",
		id           = Landmasses.BAGELS,
		description  = "A continental breakfast of a world where lands are " ..
					   "round, thick, sliced, and perhaps lightly-toasted. " ..
					   "Contains at least four continent-sized bagel slices.",
		
		CreateFWorld = MultilayeredFractal.Create,
		GenPlotTypes = MultilayeredFractal.Bagels_GeneratePlotsByRegion,
		
		aspDivMethod    = DIVMETHOD_CONTINENTAL,
	},
	
	continents = 
	{
		name         = "Continents",
		id           = Landmasses.CONTINENTS,
		description  = "A few large landmasses and some small islands.",
		
		GenPlotTypes = GotLakes_Fractal.Continents_GeneratePlotTypes,
        
        seas_iceage  = {59,64,70},
        seas_normal  = {67,72,76},
		
		aspDivMethod    = DIVMETHOD_CONTINENTAL,
	},
	
	donut = 
	{
		name        = "Donut",
		id          = Landmasses.DONUT,
		description = 
			"A delicous round pastry of a world whose filling depends on the " ..
			"options that you choose. Hint: world age can make mountains; sea level " ..
			"can make land, normal ocean, or empty ocean; clustered highlands " ..
			"can make hills; region world wrap + tilted axis climate can make " ..
			"desert.",
		
		CreateFWorld = MultilayeredFractal.Create,
		GenPlotTypes = MultilayeredFractal.Donut_GeneratePlotsByRegion,
		mountainArgs = 
		{
			adjust_plates = 0.75,
		},
		
		extraArgs = 
		{
			tectonicFlags  = {FRAC_POLAR = true}, -- keep tectonic islands away from map edges.
			tectonicGrain  = 3,
		}, 
				
		DetermineContinents = Uniform_DetermineContinents,
		
		aspWaterMap = true,
		aspStartAreaRatio = 1.0,
		
		aspDivMethod    = DIVMETHOD_BIGGESTLAND,
		-- Allow non-coastal starts in case the coast is barren terrain.
	},
	
	extras = 
	{
		name        = "Extras",
		id          = Landmasses.EXTRAS,
		description = 
			"A generous sampling of whichever extra islands you have selected, or " ..
			"a random selection if you have selected \"None\". " ..
			"Extra islands: " ..
			"chains = many long chains of small islands, " ..
			"clusters = a dense mega-cluster of small islands, " ..
			"mini-donuts = larger donut-shaped islands, " ..
			"tectonic = islands from many mountains rising up from the sea, " ..
			"tectonic and chains = tectonic with island chains, " ..
			"everything = all types of extra islands.",
		
		CreateFWorld = MultilayeredFractal.Create,
		GenPlotTypes = MultilayeredFractal.Extras_GeneratePlotsByRegion,
		
		mountainArgs = 
		{
			adjust_plates   = 2.5,
			extra_mountains = 10,
		},
		
		-- Enlarge extra islands to compensate for lack of "mainland".
		extraArgs = 
		{
			tectonicGrain    	  = 4,
			tectonicLandOffset    = 1,
			tectonicSeaMultiplier = 2,
			miniDonutROffset      = 2,
		},
        
        seas_iceage = {62,67,74},
        seas_normal = {73,78,82},
		
		-- Coast args: add a 10% chance of coastal expansion
		-- for a higher chance of islands connected by shallow water.
		coastArgs = {expansion_diceroll_table = {10, 4, 4}},
				
		aspWaterMap     = true,
		aspMinMajorFert = 75,
		aspMinMinorFert = 5,
		
		aspDivMethod = DIVMETHOD_RECTANGULAR,
		
		-- Forcing starts along the ocean.
		-- Lowering start position minimum eligibility thresholds.
		aspLocationArgs = 
		{
			mustBeCoast = true,
			minFoodMiddle = 2,
			minProdMiddle = 1,
			minFoodOuter = 2,
			minProdOuter = 1
		},
	},
	
	fractal = 
	{
		name        = "Fractal",
		id          = Landmasses.FRACTAL,
		description = 
			"A highly random map that will often form thin, snaky continents.",
		
		fractalInitArgs = 
		{
			rift_grain      = -1, 
			has_center_rift = false, 
			polar           = true,
		},
		
		fractalGenArgs = {has_center_rift = false},
		
		extraArgs = {tectonicFlags = {FRAC_POLAR = true}}, -- keep tectonic islands away from map edges	
		
        seas_iceage = {57,64,72},
        seas_normal = {65,72,78},
        
		DetermineContinents = Uniform_DetermineContinents,		

		aspMinMajorFert = 175,
		aspMinMinorFert = 50, 
		
		aspDivMethod   = DIVMETHOD_CONTINENTAL,
	},
	
	inlandSea = 
	{
		name        = "Inland Sea",
		id          = Landmasses.INLANDSEA,
		description = 
			"Much of the world's water is in a single inland sea. " ..
			"Sometimes there is a second, smaller inland sea.",
		
		CreateFWorld    = MultilayeredFractal.Create,
		GenPlotTypes    = MultilayeredFractal.InlandSea_GeneratePlotsByRegion,
		mountainArgs    = 
		{
			adjust_plates = 0.5,
		},
		
		extraArgs = 
		{
			-- Don't let extras freshen lakes
			MergeToMainland   = MergeExtrasPreserveSaltLakes, 
			tectonicFlags  	  = {FRAC_POLAR = true}, -- keep tectonic islands close to the lake
			tectonicGrain     = 4,
		},
		
		-- For this landmass, sea level affects lake grains and lake threshold.
		-- Higher grain size meams smaller lakes
		-- Higher threshold means less water overall.
		lake_grain  = 1,
		seas_iceage = {88,78,68},
		seas_normal = {82,72,62},
				
		polarIceNumRows = 2,
		
		DetermineContinents = Uniform_DetermineContinents,
		
		aspLandMap        = true,
		aspMinMajorFert   = 200,
		aspMinMinorFert   = 50, 
		aspStartAreaRatio = 1.0,
		
		aspDivMethod       = DIVMETHOD_BIGGESTLAND,
	},
	
	invertedExtras = 
	{
		name        = "Inverted Extras",
		id          = Landmasses.INVEXTRAS,
		description = 
			"A land-heavy world filled with whichever extra lakes you have " ..
			"selected, or a random selection if you have selected \"None\". " ..
			"Extra lakes: " ..
			"chains = long chains of narrow lakes, " ..
			"clusters = dense clusters of small lakes, " ..
			"mini-donuts = donut-shaped lakes, " ..
			"tectonic = water at the lowest altitudes, " ..
			"tectonic and chains = tectonic with lake chains, " ..
			"everything = all types of extra lakes.",
		
		CreateFWorld    = MultilayeredFractal.Create,
		GenPlotTypes    = MultilayeredFractal.Extras_GeneratePlotsByRegion,
		mountainArgs    = {	adjust_plates = 0.5 },
		
		-- Enlarge extras and make them form water instead of land.
		extraArgs = 
		{	
			invertExtras     	  = true,
			tectonicGrain         = 4,
			tectonicLandOffset    = 1,
			tectonicSeaMultiplier = 2,
			miniDonutROffset 	  = 2,
		},
				
		polarIceNumRows = 2,
		
		DetermineContinents = Uniform_DetermineContinents,
		
		aspLandMap        = true,
		aspMinMajorFert   = 200,
		aspMinMinorFert   = 50, 
		aspStartAreaRatio = 1.0,
		
		aspDivMethod       = DIVMETHOD_RECTANGULAR,
	},
	
	islandPlates = 
	{
		name        = "Island Plates",
		id          = Landmasses.ISLEPLATES,
		description = "A mixture of large and small islands.",
		
		CreateFWorld = MultilayeredFractal.Create,
		GenPlotTypes = MultilayeredFractal.IslandPlates_GeneratePlotsByRegion,		
		
		mountainArgs = 
		{
			adjust_plates   = 1.75,
			extra_mountains = 8,
		},
		
		extraArgs = 
		{
			tectonicGrain = 3,		
		},
				
		aspWaterMap     = true,
		aspMinMajorFert = 85,
		aspMinMinorFert = 5, 
		
		aspDivMethod    = DIVMETHOD_CONTINENTAL,
		
		-- Forcing starts along the ocean.
		-- Lowering start position minimum eligibility thresholds.
		aspLocationArgs = 
		{
			mustBeCoast = true,
			minFoodMiddle = 2,
			minFoodOuter = 3,
			minProdOuter = 1
		},
	},
	
	lakes = 
	{
		name        = "Lakes",
		id          = Landmasses.LAKES,
		description = 
			"A land-heavy world where the water is always fresh. " ..
			"Extra lakes: " ..
			"none = regular lakes" ..
			"chains = long chains of narrow lakes, " ..
			"clusters = dense clusters of small lakes, " ..
			"mini-donuts = donut-shaped lakes, " ..
			"tectonic = water at the lowest altitudes, " ..
			"tectonic and chains = tectonic with lake chains, " ..
			"everything = all types of extra lakes.",
		
		CreateFWorld    = MultilayeredFractal.Create,
		GenPlotTypes    = MultilayeredFractal.Lakes_GeneratePlotsByRegion,
		mountainArgs    = {	adjust_plates     = 0.5   },
		
		-- Make extras form water instead of land.
		extraArgs = 
		{
			invertExtras          = true,
			tectonicGrain 		  = 4,
			tectonicLandOffset    = 1,
			tectonicSeaMultiplier = 2,
			miniDonutBaysHeight   = 85,
			miniDonutROffset 	  = 3,
		},
		coastArgs       = { tectonic_shallows = false },
		
		-- For this landmass, sea level affects lake threshold.
		-- Higher threshold means less water overall.
		-- Note: any lower than 82 results in significant salt-water lakes, which
		-- the plot type function will have to freshen by adding land tiles.
		
		lake_grain  = 5,
		seas_iceage = {85,70,55},
		seas_normal = {80,65,50}, -- Land bridges compensate for extra water.
				
		polarIceNumRows = 0, -- Remove polar ice so that we can freshen lakes.
		
		DetermineContinents = Uniform_DetermineContinents,
		
		aspLandMap        = true,
		aspMinMajorFert   = 250,
		aspMinMinorFert   = 50, 
		aspStartAreaRatio = 1.0,
		
		aspDivMethod       = DIVMETHOD_RECTANGULAR,
	},
	
	largeIslands = 
	{
		name        = "Large Islands",
		id          = Landmasses.LRGISLES,
		description = 
			"Numerous large islands similar in size, but varying in terrain. " ..
			"Civs start on the best available islands.",
		
		CreateFWorld = MultilayeredFractal.Create,
		GenPlotTypes = MultilayeredFractal.LargeIslands_GeneratePlotsByRegion,		
		
		mountainArgs = 
		{
			adjust_plates   = 1.75,
			extra_mountains = 8,
		},
		
		extraArgs = 
		{
			tectonicGrain = 3,		
		},
				
		aspWaterMap     = true,
		aspMinMajorFert = 85,
		aspMinMinorFert = 5, 
		
		aspDivMethod    = DIVMETHOD_CONTINENTAL,
		
		-- Forcing starts along the ocean.
		-- Lowering start position minimum eligibility thresholds.
		aspLocationArgs = 
		{
			mustBeCoast = true,
			minFoodMiddle = 2,
			minFoodOuter = 3,
			minProdOuter = 1
		},
	},
	
	largeLakes = 
	{
		name        = "Large Lakes",
		id          = Landmasses.LRGLAKES,
		description = 
			"A land-dominated world filled with a few large salt-water lakes. " ..
			"There might also be a few fresh-water lakes here and there.",
		
		CreateFWorld    = MultilayeredFractal.Create,
		GenPlotTypes    = MultilayeredFractal.LargeLakes_GeneratePlotsByRegion,
		mountainArgs    = 
		{
			adjust_plates  = 0.5,
		},
		
		extraArgs = 
		{
			-- Don't let extras freshen lakes
			MergeToMainland   = MergeExtrasPreserveSaltLakes, 
			tectonicGrain     = 5,		
		},
		
		-- For this landmass, sea level affects lake grains and lake threshold.
		-- Higher threshold means less water overall.
		lake_grain  = 3, -- grain of 2 at high sea level looks like inland sea
		seas_iceage = {90,84,76},
		seas_normal = {84,76,68},
		
		polarIceNumRows = 2,
		
		DetermineContinents = Uniform_DetermineContinents,
		
		aspLandMap        = true,
		aspMinMajorFert   = 300,
		aspMinMinorFert   = 50, 
		aspStartAreaRatio = 1.0,
		
		aspDivMethod       = DIVMETHOD_BIGGESTLAND,
	},
	
	maze = 
	{
		name         = "Maze",
		id           = Landmasses.MAZE,
		description  = "A sprawling pangea of islands connected by snaky land bridges.",
		CreateFWorld = MultilayeredFractal.Create,
		GenPlotTypes = MultilayeredFractal.Maze_GeneratePlotsByRegion,
		mountainArgs = {adjust_plates = 0.75},
		
		extraArgs = 
		{
			tectonicFlags  = {FRAC_POLAR = true}, -- keep tectonic islands away from map edges.
			tectonicGrain  = 1,			
		},
		
		seas_iceage = {55,60,67},
        seas_normal = {63,68,73},
		
		DetermineContinents = Uniform_DetermineContinents,
		
		aspWaterMap       = true,
		aspMinMajorFert   = 75,
		aspMinMinorFert   = 5, 
		aspStartAreaRatio = 0.75, -- Pangea, but some sections might be blocked off by mountains.
		
		-- Snaky pangea can easily splinter into separate areas, 
		-- so use "continental" regions for starting plots.
		aspDivMethod    = DIVMETHOD_CONTINENTAL,
		aspLocationArgs = {mustBeCoast = true},
	},
	
	oval = 
	{
		name        = "Oval",
		id          = Landmasses.OVAL,
		description = 
			"A large oval-shaped pangea with some bays carved into the edges. Climates that crop the world will also crop the oval.",
		
		CreateFWorld = MultilayeredFractal.Create,
		GenPlotTypes = MultilayeredFractal.Oval_GeneratePlotsByRegion,
		mountainArgs = {adjust_plates = 0.75},
		extraArgs    = 
		{
			tectonicFlags = {FRAC_POLAR = true}, -- keep tectonic islands away from map edges.
			tectonicGrain = 2, -- landmass is so big that we need 2-grain to notice islands	
		},
		coastArgs    = {bExpandCoasts = false},
		
		DetermineContinents = Uniform_DetermineContinents,
				
		aspMinMajorFert   = 200,
		aspMinMinorFert   = 50, 	
		aspStartAreaRatio = 1.0,
		
		aspDivMethod    = DIVMETHOD_BIGGESTLAND,
	},
	
	pangea = 
	{
		name         = "Pangea",
		id           = Landmasses.PANGEA,
		description  = "A single piece of land for all civs to share.",
		GenPlotTypes = GotLakes_Fractal.Pangea_GeneratePlotTypes,
		mountainArgs = {adjust_plates = 0.75},
		
		extraArgs = 
		{
			tectonicFlags  = {FRAC_POLAR = true}, -- keep tectonic islands away from map edges.
			tectonicGrain  = 1,			
		},
		
		seas_iceage = {55,60,67},
        seas_normal = {63,68,73},
				
		DetermineContinents = Uniform_DetermineContinents,
		
		aspMinMajorFert   = 300,
		aspMinMinorFert   = 50, 
		aspStartAreaRatio = 1.0,
		
		aspDivMethod    = DIVMETHOD_BIGGESTLAND,
	},
	
	saltLakes = 
	{
		name        = "Salt Lakes",
		id          = Landmasses.SALTLAKES,
		description = 
			"A land-heavy world filled with small, mostly salt-water lakes.",
		
		CreateFWorld    = MultilayeredFractal.Create,
		GenPlotTypes    = MultilayeredFractal.LargeLakes_GeneratePlotsByRegion,
		mountainArgs    = 
		{
			adjust_plates  = 0.5,
		},
		
		extraArgs = 
		{
			-- Don't let extras freshen lakes
			MergeToMainland = MergeExtrasPreserveSaltLakes, 
			tectonicGrain   = 5,
		},
		
		-- For this landmass, sea level affects lake threshold.
		-- Higher threshold means less water overall.
		-- Note: any lower than 82 results in significant salt-water lakes, which
		-- the plot type function will have to freshen by adding land tiles.
		lake_grain  = 4,
		seas_iceage = {90,82,74},
		seas_normal = {86,78,70}, -- Used to be {93, 90, 85} 
				
		polarIceNumRows = 2,
		
		DetermineContinents = Uniform_DetermineContinents,
		
		aspLandMap        = true,
		aspMinMajorFert   = 300,
		aspMinMinorFert   = 50, 
		aspStartAreaRatio = 1.0,
		
		aspDivMethod       = DIVMETHOD_RECTANGULAR,
	},
	
	smallContinents = 
	{
		name        = "Small Continents",
		id          = Landmasses.SMALLCONTS,
		description = 
			"Medium-Sized lands that are larger than most islands, but smaller than regular continents.",
		
		fractalInitArgs = {continent_grain = 3},
		
        mountainArgs = 
		{
			adjust_plates   = 1.5,
			extra_mountains = 5,
		},
		
		extraArgs = 
		{
			tectonicGrain = 3,
		},
        
        seas_iceage = {60,65,72},
        seas_normal = {69,75,80},
				
		aspMinMajorFert = 125,
		aspMinMinorFert = 40, 
		
		aspDivMethod       = DIVMETHOD_CONTINENTAL,
		aspLocationArgs    = {mustBeCoast = true},
	},
	
	tinyIslands = 
	{
		name        = "Tiny Islands",
		id          = Landmasses.TINYISLES,
		description = "A sprinkling of many small islands over a vast ocean.",
		
		fractalInitArgs = {continent_grain = 5},
		
		mountainArgs = 
		{
			adjust_plates   = 2.5,
			extra_mountains = 15,
		},
		
		extraArgs = 
		{
			tectonicGrain = 5,
		},
        
        seas_iceage = {72,75,78},
        seas_normal = {75,78,81},
		
		-- Coast args: add a 10% chance of coastal expansion
		-- for a higher chance of islands connected by shallow water.
		coastArgs   = {expansion_diceroll_table = {10, 4, 4}},
				
		aspWaterMap     = true,
		aspMinMajorFert = 60,
		aspMinMinorFert = 4,

		aspDivMethod   = DIVMETHOD_RECTANGULAR,
		
		-- Forcing starts along the ocean.
		-- Lowering start position minimum eligibility thresholds.
		aspLocationArgs = 
		{
			mustBeCoast = true,
			minFoodMiddle = 2,
			minProdMiddle = 1,
			minFoodOuter = 2,
			minProdOuter = 1
		},		
	},
};

------------------------------------------------------------------------------
-- Config defaults and climate-specific overrides
------------------------------------------------------------------------------
ClimateInfo = 
{
	default = 
	{
		name        = "unknown climate",
		id          = 0,
		description = "",
			
		-- Fixed latitude +/- variation multiplier for most regions.
		regionLat       = 0.38,  -- Civ V: 0.35
		regLatVariation = 0.1,   -- Civ V: 1.0
		
		-- Lat variables for tilted axis.
		darkSideLat = TILTEDAXIS_TUNDRALAT,	
			
			
		-- Terrain Generation
		terrainArgs = {},
		
		createTGen      = GotLakes_TerrainGenerator.Default_Create,
		TInitFractals   = GotLakes_TerrainGenerator.Default_InitFractals,
		TInitRegion     = GotLakes_TerrainGenerator.Default_InitRegion,
		GenerateTerrain = GotLakes_TerrainGenerator.Default_GenerateTerrain,
		
		GlobeGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.Default_GenerateTerrainAtPlot,
			
		RegionGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.Default_GenerateTerrainAtPlot,
		
		
		-- Latitude lookup
		GlobeGetLatitudeAtPlot  = DefaultGlobe_GetLatitudeAtPlot,
		RegionGetLatitudeAtPlot = DefaultRegion_GetLatitudeAtPlot,
		
		
		-- Feature generation
		featureArgs = {},
		createFGen         = GotLakes_FeatureGenerator.Default_Create,
		GlobeAddIceAtPlot  = GotLakes_FeatureGenerator.Globe_AddIceAtPlot,
		RegionAddIceAtPlot = GotLakes_FeatureGenerator.Region_AddIceAtPlot,
		AddForestsAtPlot   = GotLakes_FeatureGenerator.Default_AddForestsAtPlot,
		AddJunglesAtPlot   = GotLakes_FeatureGenerator.Default_AddJunglesAtPlot,
		AdjustJungleAtPlot = GotLakes_FeatureGenerator.Default_AdjustJungleAtPlot,
		
		
		-- Resource generation
		-- Cold climates can override this flag to allow resources next to ice.
		resourcesAvoidIce = true,
		ValidLuxuryPlots  = ResourceGenerator.__ValidLuxuryPlots,
		ScoreLuxuryPlots  = ResourceGenerator.__ScoreLuxuryPlots,
		
		-- Start plot database
		-- Major/Minor civ fertility multipliers
		-- Reduce for less-fertile maps.
		aspMinMajorFert = 1.0,
		aspMinMinorFert = 1.0, 
		aspStartMinY    = 15, -- Min start latitude as percent of height
		aspStartMaxY    = 85, -- Max start latitude as percent of height
	},
	
	arborea = 
	{
		name        = "Arborea",
		id          = Climates.ARBOREA,
		description = "A heavily-forested world.",
		
		terrainArgs = 
		{
			desert_shift   = 5,
			iDesertPercent = 15, -- Used to be 20
		},
		
		featureArgs = 
		{
			iClumpChange   = 0,
			iForestChange  = 20,
			iForestPercent = 70,
			fMarshPercent  = 10,
			iClumpHeight   = 94,
		},
		
		ValidLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,		
	},
	
	evergreen = 
	{
		name        = "Evergreen",
		id          = Climates.EVERGREEN,
		description = 
			"A colder, heavily-forested world where trees can survive on snow.",
		
		-- Hot: mostly grass with some tundra and desert
		-- Normal: grass/tundra mix, roughly balanced
		-- Cool mostly tundra with some grass and snow
		regionLat       = 0.4,
		regLatVariation = 0.15,
		
		terrainArgs = 
		{
			temperature_shift     = 0.2,  -- default 0.1
			fSnowLatitude         = 0.5,  -- default 0.75
			fTundraLatitude       = 0.2,  -- default 0.6
			fDesertTopLatitude    = 0.1,  -- default 0.5
			fDesertBottomLatitude = 0.0,  -- default 0.2
			fGrassLatitude        = 0.0,  -- default 0.1
		
			desert_shift   = 6,  -- default 8
			iDesertPercent = 12, -- default 32
			iPlainsPercent = 35, -- default 50
		},
		
		featureArgs = 
		{
			-- Add more ice
			iIcePercent = 25,
			iIceChange = 10,
			ice_grain  = 5,
			
			-- Use arborea forest settings.
			iClumpChange   = 0,
			iForestChange  = 20,
			iJungleChange  = 15, -- Arborea 20
			iJunglePercent = 40, -- Arborea 60
			iForestPercent = 70,
			fMarshPercent  = 10,
			iClumpHeight   = 94,
		},
		
		AddForestsAtPlot   = GotLakes_FeatureGenerator.Evergreen_AddForestsAtPlot,
		RegionAddIceAtPlot = GotLakes_FeatureGenerator.IceAgeRegion_AddIceAtPlot,

		resourcesAvoidIce = false,
		ValidLuxuryPlots  = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots  = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
		
		aspMinMajorFert = 0.5,
		aspMinMinorFert = 0.5, 
	},
	
	iceAgeCrop = 
	{
		name        = "Ice Age Crop",
		id          = Climates.ICEAGECROP,
		description = 
			"Ice Age, Cropped: a colder world in which much of the water has become polar " ..
			"ice. Globe maps crop to 2/3 of normal height, eliminating most of the icy poles. Region " ..
			"maps tend to have tundra, but not too much snow or ice.",
		
		-- Hot: mostly grass with some tundra and desert
		-- Normal: grass/tundra mix, roughly balanced
		-- Cool mostly tundra with some grass and snow
		regionLat       = 0.3,
		regLatVariation = 0.3,
		
		createTGen = GotLakes_TerrainGenerator.IceAge_Create,
		createFGen = GotLakes_FeatureGenerator.IceAge_Create,
		
		GlobeGetLatitudeAtPlot = IceAgeGlobe_GetLatitudeAtPlot,
		
		featureArgs = 
		{
			iIcePercent = 10,
			iIceChange = 5,
			ice_grain  = 5
		},
		
		AddForestsAtPlot = 
			GotLakes_FeatureGenerator.IceAge_AddForestsAtPlot,
		
		RegionAddIceAtPlot =
			GotLakes_FeatureGenerator.IceAgeRegion_AddIceAtPlot,
			
		resourcesAvoidIce = false,
		ValidLuxuryPlots  = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots  = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
		
		aspMinMajorFert = 0.5,
		aspMinMinorFert = 0.5,
	},
	
	iceAgeFull = 
	{
		name        = "Ice Age Full",
		id          = Climates.ICEAGEFULL,
		description = 
			"Ice Age, Full: a colder world in which much of the water has become polar " ..
			"ice. For globe maps, the top and bottom 1/6 of the map are ice and snow. " ..
			"Region maps commonly have a mix of tundra and snow.",
		
		-- Hot: mostly grass with some tundra and desert
		-- Normal: grass/tundra mix, roughly balanced
		-- Cool mostly tundra with some grass and snow
		regionLat       = 0.4,
		regLatVariation = 0.13,	
			
		createTGen = GotLakes_TerrainGenerator.IceAge_Create,
		createFGen = GotLakes_FeatureGenerator.IceAge_Create,
		
		featureArgs = 
		{
			-- Add more ice to ice age full
			iIcePercent = 25,
			iIceChange = 10,
			ice_grain  = 5
		},
		
		AddForestsAtPlot = 
			GotLakes_FeatureGenerator.IceAge_AddForestsAtPlot,
		
		RegionAddIceAtPlot =
			GotLakes_FeatureGenerator.IceAgeRegion_AddIceAtPlot,
			
		resourcesAvoidIce = false,
		ValidLuxuryPlots  = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots  = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
		
		aspMinMajorFert = 0.5,
		aspMinMinorFert = 0.5, 
		
		-- Snow extends farther from the poles
		aspStartMinY = 25, 
		aspStartMaxY = 75,
	},
	
	normal = 
	{
		name        = "Normal Climate",
		id          = Climates.NORMAL,
		description = "No special changes to the climate.",
	},
	
	prairie = 
	{
		name        = "Prairie",
		id          = Climates.PRAIRIE,
		description = "A temperate world covered mostly in plains and grasslands.",
		
		terrainArgs = 
		{
			grain_amount   = 5,  -- default 3
			desert_shift   = 10, -- default 16
			iDesertPercent = 20, -- default 32
			iPlainsPercent = 75, -- default 50
		},
		
		featureArgs = 
		{
			iForestChange  = 4, -- default 15
			iJungleChange  = 5, -- default 20
			iJunglePercent = 16, -- default 65
			iForestPercent = 8, -- default 31
		},
		
		ValidLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
	},
	
	rainforest = 
	{
		name        = "Rainforest",
		id          = Climates.RAINFOREST,
		description = "A warmer world covered mostly in jungle.",
		
		regionLat       = 0.25,
		
		terrainArgs = 
		{
			grain_amount       = 3,    -- default 4
			fSnowLatitude      = 0.9,
			fTundraLatitude    = 0.75,
			fDesertTopLatitude = 0.6,
			fGrassLatitude     = 0.2,  -- Normally 0.1
			desert_shift       = 8,    -- Used to be 10
			iDesertPercent     = 16,   -- Used to be 25
			iPlainsPercent     = 5,    -- Used to be 35
		},
		
		featureArgs = 
		{
			iClumpChange   = 0,
			iForestChange  = 10,
			iJungleChange  = 15,
			iJunglePercent = 75,
			iForestPercent = 31,
			fMarshPercent  = 8,
			iClumpHeight   = 94,
		},
		
		AddJunglesAtPlot   = GotLakes_FeatureGenerator.Rainforest_AddJunglesAtPlot,
		
		ValidLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
	},
	
	sandstorm = 
	{
		name        = "Sandstorm",
		id          = Climates.SANDSTORM,
		description = "A warmer world covered mostly in desert.",
		
		regionLat       = 0.25,
		
		terrainArgs = 
		{
			grain_amount   = 3,    -- default 4
			desert_shift   = 10, -- Default 8
			iDesertPercent = 60, -- Default 32
			iPlainsPercent = 90, -- Default 50
		},
		
		TInitFractals = 
			GotLakes_TerrainGenerator.Sandstorm_InitFractals,
			
		GlobeGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.Sandstorm_GenerateTerrainAtPlot,
			
		RegionGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.Sandstorm_GenerateTerrainAtPlot,
					
		ValidLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
		
		aspMinMajorFert = 0.5,
		aspMinMinorFert = 0.5, 
	},
	
	sandyJungle = 
	{
		name        = "Sandy Jungle",
		id          = Climates.SANDYJUNGLE,
		description = "A warmer world covered in overlapping jungles and deserts.",
		
		regionLat       = 0.25,
		
		terrainArgs = 
		{
			fSnowLatitude      = 0.9,
			fTundraLatitude    = 0.75,
			fDesertTopLatitude = 0.6,
			fGrassLatitude     = 0.1, -- Normally 0.1
			desert_shift       = 10,  -- Used to be 10
			iDesertPercent     = 50,  -- Default 32
			iPlainsPercent     = 5,   -- Default 50
		},
		
		featureArgs = 
		{
			iClumpChange   = 0,
			iForestChange  = 10,
			iJunglePercent = 50, -- Slightly less to compensate for more applicable plots
			iJungleChange  = 20,
			iClumpHeight   = 94,
		},
		
		TInitFractals = GotLakes_TerrainGenerator.Sandstorm_InitFractals,
			
		GlobeGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.Sandstorm_GenerateTerrainAtPlot,
			
		RegionGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.Sandstorm_GenerateTerrainAtPlot,
		
		AddJunglesAtPlot   = GotLakes_FeatureGenerator.SandyJungle_AddJunglesAtPlot,
		AdjustJungleAtPlot = GotLakes_FeatureGenerator.SandyJungle_AdjustJungleAtPlot,
		
		ValidLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
		
		aspMinMajorFert = 0.8,
		aspMinMinorFert = 0.8, 
	},
	
	tiltedAxisCrop = 
	{
		name        = "Tilted Axis Crop",
		id          = Climates.TILTEDCROP,
		description = 
			"Tilted Axis, Cropped: a world turned onto its side so that only the " ..
			"southern hemisphere faces the sun. Crops a globe map to 1/2 of its " ..
			"normal height, using the next largest map size to compensate for lost " ..
			"tiles. What remains is tundra to the north, desert to the south, and " ..
			"plenty of tiles to the east and west. For regional maps, a randomly-" ..
			"placed sunlit pole creates zones of hot desert and tundra, leaving a " ..
			"temperate fertile band in between.",
			
		terrainArgs = 
		{
			desert_shift   = 8,    -- Default 16
			iDesertPercent = 16, -- Default 32
		},
		
		featureArgs = 
		{
			jungleMaxLat = TILTEDAXIS_DESERTLAT,
			ice_grain    = 5        -- Default 4		
		},
		
		TInitFractals = 
			GotLakes_TerrainGenerator.TiltedAxis_InitFractals,
		
		TInitRegion = 
			GotLakes_TerrainGenerator.TiltedAxis_InitRegion,
		
		GlobeGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.TiltedAxisGlobe_GenerateTerrainAtPlot,
			
		RegionGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.TiltedAxisRegion_GenerateTerrainAtPlot,

			
		GlobeGetLatitudeAtPlot  = TiltedAxisGlobe_GetLatitudeAtPlot,
		RegionGetLatitudeAtPlot = TiltedAxisRegion_GetLatitudeAtPlot,
		
			
		GlobeAddIceAtPlot =
			GotLakes_FeatureGenerator.TiltedAxisGlobeCropped_AddIceAtPlot,
		
		RegionAddIceAtPlot =
			GotLakes_FeatureGenerator.TiltedAxisRegion_AddIceAtPlot,
		
		-- Avoid some southern desert and some northern tundra/snow
		aspStartMinY = 5, 
		aspStartMaxY = 95,
	},
	
	tiltedAxisFull = 
	{
		name        = "Tilted Axis Full",
		id          = Climates.TILTEDFULL,
		description = 
			"Tilted Axis, Full: a world turned onto its side so that only the " ..
			"southern hemisphere faces the sun. For globe maps, the entire northern " ..
			"hemisphere is a frozen wasteland of snow and ice. What remains is tundra " ..
			"just south of the equator, desert at the south pole, and habitable land " ..
			"somwhere in between. For region maps, a randomly-placed sunlit pole " ..
			"creates zones of hot desert and snow/ice, leaving a temperate fertile " ..
			"band in between.",
			
		terrainArgs = 
		{
			desert_shift   = 8,  -- Default 16
			iDesertPercent = 16, -- Default 32
		},
		
		featureArgs = 
		{
			jungleMaxLat = TILTEDAXIS_DESERTLAT,
		},
		
		darkSideLat = TILTEDAXIS_SNOWLAT,
		
		TInitFractals = 
			GotLakes_TerrainGenerator.TiltedAxis_InitFractals,
		
		TInitRegion = 
			GotLakes_TerrainGenerator.TiltedAxis_InitRegion,
			
		GlobeGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.TiltedAxisGlobe_GenerateTerrainAtPlot,
			
		RegionGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.TiltedAxisRegion_GenerateTerrainAtPlot,
		
		
		GlobeGetLatitudeAtPlot  = TiltedAxisGlobe_GetLatitudeAtPlot,
		RegionGetLatitudeAtPlot = TiltedAxisRegion_GetLatitudeAtPlot,
		
			
		GlobeAddIceAtPlot =
			GotLakes_FeatureGenerator.TiltedAxisGlobeFull_AddIceAtPlot,
		
		RegionAddIceAtPlot =
			GotLakes_FeatureGenerator.TiltedAxisRegion_AddIceAtPlot,
		
		-- Avoid the northern hemisphere.
		aspStartMinY = 5, 
		aspStartMaxY = 50,
	},
	
	twoSuns = 
	{
		name        = "Two Suns",
		id          = Climates.TWOSUNS,
		description = 
			"A world turned onto its side and sandwiched between two stars in a binary " ..
			"star system, so that each pole always faces a star. The coldest regions " ..
			"of the world are at the equator.",
			
		terrainArgs = 
		{
			desert_shift   = 8,       -- Default 16
			iDesertPercent = 16,      -- Default 32
		},
		
		featureArgs = 
		{
			-- Increase jungles for non-desert hot-pole tiles.
			iJunglePercent = 95, -- Default 65
			iJungleChange  = 5,
			jungleMaxLat   = TILTEDAXIS_JUNGLELAT,
		
			ice_grain   = 5,  -- Default 4
			iIcePercent = 25, -- Default 25
			iIceChange  = 10, -- Default 10
		},
		
		darkSideLat = TILTEDAXIS_SNOWLAT,
		
		TInitFractals = 
			GotLakes_TerrainGenerator.TiltedAxis_InitFractals,
		
		TInitRegion = 
			GotLakes_TerrainGenerator.TwoSuns_InitRegion,
			
		GlobeGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.TwoSunsGlobe_GenerateTerrainAtPlot,
		
		RegionGenerateTerrainAtPlot = 
			GotLakes_TerrainGenerator.TwoSunsRegion_GenerateTerrainAtPlot,
		
	
		GlobeGetLatitudeAtPlot  = TwoSunsGlobe_GetLatitudeAtPlot,
		RegionGetLatitudeAtPlot = TwoSunsRegion_GetLatitudeAtPlot,
		
		GlobeAddIceAtPlot =
			GotLakes_FeatureGenerator.TwoSuns_AddIceAtPlot,
		
		RegionAddIceAtPlot =
			GotLakes_FeatureGenerator.TwoSuns_AddIceAtPlot,
		
		-- Avoid hot desert at the poles.
		aspStartMinY = 5, 
		aspStartMaxY = 95,
	},
	
	wetland = 
	{
		name        = "Wetland",
		id          = Climates.WETLAND,
		description = "A temperate world soaked with marshlands and surprisingly-fertile deserts.",
		
		terrainArgs = 
		{
			grain_amount   = 5,  -- default 3
			desert_shift   = 10, -- default 16
			iDesertPercent = 20, -- default 32
			iPlainsPercent = 25, -- default 50
		},
		
		featureArgs = 
		{
			fMarshPercent = 75, -- default 8
			iOasisPercent = 80  -- default 6
		},
		
		ValidLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ValidLuxuryPlots,
		ScoreLuxuryPlots = GotLakes_ResourceGenerator.__Custom_ScoreLuxuryPlots,
	},
};
------------------------------------------------------------------------------
-- Config-state initialization
------------------------------------------------------------------------------
function GetConfigInfo(default, args)
	-- Apply overrides to a default config struct.
	--
	-- Args:
	--   default: a config struct which contains a default value for all 
	--     possible config fields.
	--   args: overrides to apply to the defaults; can be empty or nil
	-- Returns:
	--   config: a shallow copy of default with args overriding defaults
	
	local config = {};
	
	-- This assumes that the default struct has a non-nil value for
	-- all acceptable fields in args.
	for k,v in pairs(default) do
		-- Explicitly check for nil 
		-- in case the default is true and an override is false.
		if (args[k] ~= nil) then
			config[k] = args[k];
		else
			config[k] = v;
		end
	end
	
	return config;
end

function getWorldSizeInfo(args)
	-- Apply defaults and overrides for a single worlds size 
	-- and return the resulting config.
	return GetConfigInfo(WorldSizeInfo.default, args);
end

function AdjustWorldHeight(sizeInfo, adjustment)
	-- Adjust settings for world size info.
	--
	-- Args:
	--   sizeInfo: info to adjust
	--   adjustment: multiply height and data by this much
	--
	-- Return adjusted size info
	
	-- Adjust height, rounding to nearest even integer.
	sizeInfo.size[2] = 2*Round(sizeInfo.size[2]*adjustment/2);
	
	-- Adjust data.
	sizeInfo.mtnGrain  = Round(sizeInfo.mtnGrain*adjustment);
	sizeInfo.numPlates = Round(sizeInfo.numPlates*adjustment);

	return sizeInfo;
end

function InitWorldSizeInfo()
	-- Set the global world size info based on user inputs such as climate.
	--
	-- Returns:
	--   scaled world size info
	
	local worldSize = ScriptData.worldSize;
	
	-- Look for the world-size hash in game info.
	local mapSizeType = MAPSIZE_STANDARD; -- Default val in case hash not found.
	for row in GameInfo.Maps() do
		if(worldSize == row.Hash) then
			mapSizeType = row.MapSizeType;
			break;
		end
	end
	
	print("Map Size is " .. mapSizeType);
	
	-- Original sizes.
	local standardSizes = 
	{
		[MAPSIZE_DUEL]     = WorldSizeInfo.duel,
		[MAPSIZE_TINY]     = WorldSizeInfo.tiny,
		[MAPSIZE_SMALL]    = WorldSizeInfo.small,
		[MAPSIZE_STANDARD] = WorldSizeInfo.standard,
		[MAPSIZE_LARGE]    = WorldSizeInfo.large,
		[MAPSIZE_HUGE]     = WorldSizeInfo.huge    
	};
	local standardInfo = standardSizes[mapSizeType] or WorldSizeInfo.standard;
	standardInfo = getWorldSizeInfo(standardInfo);
	
	-- Enlarge both dimensions by one size level.
	local enlargedSizes = 
	{
		[MAPSIZE_DUEL]     = WorldSizeInfo.tiny,
		[MAPSIZE_TINY]     = WorldSizeInfo.small,
		[MAPSIZE_SMALL]    = WorldSizeInfo.standard,
		[MAPSIZE_STANDARD] = WorldSizeInfo.large,
		[MAPSIZE_LARGE]    = WorldSizeInfo.huge,
		[MAPSIZE_HUGE]     = WorldSizeInfo.enormous  
	};	
	local enlargedInfo = enlargedSizes[mapSizeType] or WorldSizeInfo.large;
	enlargedInfo = getWorldSizeInfo(enlargedInfo);
	
	-- Most maps use standard info.
	local sizeInfo = standardInfo;
	
	if (UserInput.worldWrap == WWRAP_GLOBE) then
		-- Global world wrap: certain climates need cropping / resizing.
		
		if (UserInput.climate == Climates.ICEAGECROP) then
			-- Ice Age cropped
			--
			-- Crop out the polar ice taking over more of the world by 
			-- using a "flatter" grid.  This allows special plot modes
			-- to assign starting plots without having to worry about 
			-- icy regions.
			print("Adjusting map size (global ice age cropped)");
			sizeInfo = AdjustWorldHeight(standardInfo, 2/3);
			
		elseif (UserInput.climate == Climates.TILTEDCROP) then
			-- Tilted Axis cropped
			--
			-- Crop out the polar ice taking over more of the world by 
			-- using a "flatter" grid.  This allows special plot modes
			-- to assign starting plots without having to worry about 
			-- icy regions.
			-- Use larger info to compsensate for wasteland.
			print("Adjusting map size (global tilted axis cropped)");
			sizeInfo = AdjustWorldHeight(enlargedInfo, 0.5);
			
		elseif (UserInput.climate == Climates.TILTEDFULL) then
			-- Increase map size to accomodate frozen wasteland
			print("Increasing map size (global tilted axis full)");
			sizeInfo = enlargedInfo;
		end
	end
	
	ScriptData.worldSizeInfo = sizeInfo;
	print("Grid Size is " .. CoordString(ScriptData.worldSizeInfo.size[1], ScriptData.worldSizeInfo.size[2]));
end

function GetLandInfo(args)
	-- Apply landmass-specific overrides to defaults and return the resulting
	-- config.
	return GetConfigInfo(LandInfo.default, args);
end

function GetClimateInfo(args)
	-- Apply climate-specific overrides to defaults and return the resulting
	-- config.
	return GetConfigInfo(ClimateInfo.default, args);
end
------------------------------------------------------------------------------
-- Configuration and shuffling based on user input
------------------------------------------------------------------------------
function GL_RandSeed(seed)
	-- Seed a native RNG with the map's seed
	-- so that we can randomize map options during map-data-init.
	local seed = seed or os.time();
	
	math.randomseed(seed);
	
	print("Got Lakes map seed: " .. seed);
end

function GL_RandInt(upper, description)
	-- Native RNG that can substitute TerrainBuilder.GetRandomNumber()
	-- Args:
	--   upper: exclusive upper-bound integer
	--   description: ignored, but matches signature of TerrainBuilder.GetRandomNumber()
	-- Returns a random integer in the range [0,upper-1]
	
	local upper = upper or 1;
	
	-- Offset [1,N] to [0,N-1]
	return math.random(upper) - 1;
end

function GetFamilyOption(shuffleName, sortP, familyName, default, infos)
	-- Generate a map shuffle family option, given the following params:
	--   sortP: for sorting different types of options separately
	--   name: family name
	--   default: default value for the family option
	--   infos: references to landmasses or climates in a global info struct
	--
	-- A map shuffle can have multiple families of options, such as continents,
	-- islands, lakes, and pangeas.
	
	local values = {};
	
	-- Generate choices based on info structs.
	for i, info in ipairs(infos) do
	
		local description = "Add " .. string.upper(info.name) .. " to the " .. 
								 string.upper(shuffleName) .. " shuffle.\n\n" ..
								 info.description;
								 
		table.insert(values, {info.name, description});
	end
	
	-- Add all/none choices.
	table.insert(values, 
		{"All", "Add ALL of these choices to the " .. 
				string.upper(shuffleName) .. " shuffle."});
				
	table.insert(values, 
		{"None", "Do not add any of these choices to the " .. 
				string.upper(shuffleName) .. " shuffle."});
	
	local defaultVal = default;
	if (defaultVal < 0) then
		-- Negative default means all or none.
		defaultVal = table.maxn(infos) - default;
	end
	
	local option = 
	{
		Infos        = infos,
		Name         = familyName .. "?",
		NameDebug    = familyName,
		Values       = values,
		DefaultValue = defaultVal,
		SortPriority = sortP
	};
	
	return option;
end

function GetLandmassFamilyOption(sortP, name, default, landInfos)
	-- Generate a landmass family shuffle option, given the following params:
	--   sortP: for sorting different types of options separately
	--   name: name to assign to the option
	--   default: default value for the option
	--   landInfos: references to landmasses in the global landmass
	--     info struct
	
	return GetFamilyOption("LANDMASS", sortP, name, default, landInfos);
end

function GetClimateFamilyOption(sortP, name, default, climateInfos)
	-- Generate a climate shuffle option, given the following params:
	--   sortP: for sorting different types of options separately
	--   name: name to assign to the option
	--   default: default value for the option
	--   climateInfos: references to climates in the global climate
	--     info struct
	
	return GetFamilyOption("CLIMATE", sortP, name, default, climateInfos);
end

function GL_GetCustomOptions()
	-- Return a list containing this map script's options.
	
	return {
			{
				Name         = "World Age",
				Values       = 
				{
					{"3 Billion Years", "A young, rocky world with numerous mountains and " ..
										"hills that have not yet eroded."},
					{"4 Billion Years", "A middle-aged world with some erosion, but still " ..
										"a fair amount of mountains and hills."},
					{"5 Billion Years", "An old, wind-beaten world with fewer mountains and " .. 
										"hills due to erosion."},
					{"Random", "Randomly choose either 3, 4, or 5 billion years."}
				},
				DefaultValue = MapDefaults.WORLDAGE,
				RandSize = 3,
				SortPriority = MapOptions.WORLDAGE
			},
			
			{
				Name         = "Sea Level",
				Values       = 
				{
					{"Low",    "Less water, more land."},
					{"Medium", "The normal balance between land and water."},
					{"High",   "More water, less land."},
					{"Random", "Randomly choose either low, medium, or high sea level."}
				},
				DefaultValue = MapDefaults.SEALEVEL,
				RandSize = 3,
				SortPriority = MapOptions.SEALEVEL
			},
			
			{
				Name         = "Temperature", -- "Temperature Offset" is too long of a name
				Values       = 
				{
					{"Cooler", "A cooler version of selected climate, perhaps " ..
							   "with more tundra/snow/ice, or less desert."},
					{"Normal", "Normal temperature for the selected climate."},
					{"Warmer", "A warmer version of the selected climate, perhaps " .. 
								"with more desert, or less tundra/snow/ice."},
					{"Random", "Randomly choose either cooler, normal, or warmer."}
				},
				DefaultValue = MapDefaults.TEMPERATURE,
				RandSize = 3,
				SortPriority = MapOptions.TEMPERATURE
			},
			
			{
				Name         = "Rainfall",
				Values       = 
				{
					{"Drier",  "A more arid version of selected climate. This " ..
							   "means less forests, jungles, marsh, etc."},
					{"Normal", "Normal rainfall for the selected climate."},
					{"Wetter", "A more humid version of the selected climate. " ..
							   "This means more forests, jungles, marsh, etc."},
					{"Random", "Randomly choose either drier, normal, or wetter."}
				},
				DefaultValue = MapDefaults.RAINFALL,
				RandSize = 3,
				SortPriority = MapOptions.RAINFALL
			},
			
			{
				Name         = "Resources",
				Values       = 
				{
					{"Sparse", "Less resources than a normal world."},
					{"Standard", "The normal distribution of resources."},
					{"Abundant", "More resources than a normal world."},
					{"Random", "Randomly choose from one of the above."}
				},
				DefaultValue = MapDefaults.RESOURCES,
				RandSize = 3,
				SortPriority = MapOptions.RESOURCES
			},
			
			{
				Name         = "Start Position",
				Values       = 
				{
					{"Balanced", "All players have equally good starting spots."},
					{"Standard", "Standard starting position."},
					{"Legendary", "All players have great, but not necessarily equal spots."},
				},
				DefaultValue = MapDefaults.START,
				RandSize = 3,
				SortPriority = MapOptions.START
			},	
		
			{
				Name         = "World Wrap",
				Values       = 
				{
					{"Globe",  "A cylindrical world that has north/south poles and bands " ..
							   "of latitude. Wraps from east to west."},
					{"Region", "A flat world that has no bands of latitude unless you choose " ..
					           "a Tilted Axis climate. Does not wrap in any direction."},
					{"Random", "Randomize whether the world is a globe or a region."}
				},
				DefaultValue = MapDefaults.WORLDWRAP,
				RandSize = 2,
				SortPriority = MapOptions.WORLDWRAP
			},
			
			{
				Name         = "Mountains",
				Values       = 
				{
					{"Tectonic Plates", "Normal, tectonic-style mountain ranges."},
					{"Fractal Ranges", "Snaky, fractal-style mountain ranges."},
					{"Scattered Crags", "Small clumps of mountains and hills sprinkled across " ..
										"the world."},
					{"Clustered Highlands", "A few large clusters of mountains and hills"},
					{"Uncanny Valleys", "Fractal-style mountain ranges forming valleys that " ..
										"are almost realistic, but not quite..."},
					{"Winding Canyons", "An extremely rocky world where canyons can span " ..
										"entire continents!"},
					{"Random", "Randomly choose a mountain pattern."}
				},
				DefaultValue = MapDefaults.MOUNTAINS,
				RandSize = 6,
				SortPriority = MapOptions.MOUNTAINS
			},

			{
				Name         = "Extras",
				Values       = 
				{
					{"None", "No additional islands or lakes."},
					{"Chains", "Add long, thin chains of extras."},
					{"Clusters", "Add clusters of densly-packed islands or lakes."},
					{"Mini Donuts", "Adds donut-shaped extras."},
					{"Tectonic", "Add islands that follow the mountain ranges, or " ..
							     "lakes at the world's lowest altitudes."},
					{"Tectonic and Chains", "Add both chains and tectonic extras."},
					{"Everything", "Add all types of extras."},
					{"Random", "Randomize this option."},
				},
				DefaultValue = MapDefaults.EXTRAS,
				RandSize = 7,
				SortPriority = MapOptions.EXTRAS
			},

			-- Lands
			GetLandmassFamilyOption(MapOptions.CONTINENTS, "Continents", MapDefaults.CONTINENTS,	
			{ 
				LandInfo.bagels,
				LandInfo.continents,
				LandInfo.fractal,
				LandInfo.smallContinents
			}),	

			GetLandmassFamilyOption(MapOptions.ISLANDS, "Islands", MapDefaults.ISLANDS,	
			{ 
				LandInfo.archipelago,
				LandInfo.extras,
				LandInfo.islandPlates,
				LandInfo.largeIslands,
				LandInfo.tinyIslands
			}),					
				
			GetLandmassFamilyOption(MapOptions.LAKES, "Lakes", MapDefaults.LAKES,	
			{ 
				LandInfo.inlandSea,
				LandInfo.invertedExtras,
				LandInfo.lakes,
				LandInfo.largeLakes,
				LandInfo.saltLakes
			}),
				
			GetLandmassFamilyOption(MapOptions.PANGEA, "Pangea", MapDefaults.PANGEA,	
			{ 
				LandInfo.donut,
				LandInfo.maze,
				LandInfo.oval,
				LandInfo.pangea
			}),
			
		
			-- Climate families
			GetClimateFamilyOption(MapOptions.COLD, "Cold", MapDefaults.COLD,	
			{
				ClimateInfo.evergreen,
				ClimateInfo.iceAgeCrop,
				ClimateInfo.iceAgeFull
			}),
				
			GetClimateFamilyOption(MapOptions.TEMPERATE, "Temperate", MapDefaults.TEMPERATE,	
			{
				ClimateInfo.arborea,
				ClimateInfo.normal,
				ClimateInfo.prairie,
				ClimateInfo.wetland
			}),
			
			GetClimateFamilyOption(MapOptions.TILTED, "Tilted", MapDefaults.TILTED,	
			{
				ClimateInfo.tiltedAxisCrop,
				ClimateInfo.tiltedAxisFull,
				ClimateInfo.twoSuns
			}),
			
			GetClimateFamilyOption(MapOptions.WARM, "Warm", MapDefaults.WARM,	
			{
				ClimateInfo.rainforest,
				ClimateInfo.sandstorm,
				ClimateInfo.sandyJungle
			})
		};
end

function GetMapScriptInfo()
	print("Getting map script info (Got Lakes).");
	
	return {
		Name = "Got Lakes?",
		Description = "Quench your thirst for exploration with this incredibly random map script!",
		IsAdvancedMap = 0,
		SupportsMultiplayer = true,
		IconIndex = 13, -- lakes icon
		SortIndex = 8,
		CustomOptions = GL_GetCustomOptions(),
	};
end

function GetOption(customOptions, optIndex)
	return customOptions[optIndex];
end

function GetOptionName(customOption)

	-- Check the debug name first.
	local name = customOption.NameDebug;
	
	if (not name) then
		name = customOption.Name;
	end
	
	return name;
end

function GetOptionValueName(customOption, optIndex, userInput)

	-- Some map options don't have help text.
	local valName = customOption.Values[userInput];
	
	if (type(valName) == "table") then
		-- Got a mapping of name to help text. Isolate the name.
		valName = valName[1];
	end
	
	return valName;
end

function ChooseOption(customOptions, optIndex)
	-- Choose a non-shuffle map option.

	local userInput = MapConfiguration.GetValue(MapOptionNames[optIndex]);
	
	--print("optIndex = " .. optIndex .. ", userInput = " .. NonNilStr(userInput));
	
	local customOption = customOptions[optIndex];
	
	local optName = GetOptionName(customOption);
	local valName = "";
	
	if (not userInput) then
		-- User input not available, so use default value instead.
		userInput = customOption.DefaultValue;
		valName   = "Not Supported --> ";
	end
	
	valName = valName .. GetOptionValueName(customOption, optIndex, userInput);
		
	local randSize = customOption.RandSize;
	
	if (userInput == (randSize + 1)) then
		-- The user selected "Random", so randomize the option.

		userInput = 1 + ScriptData.configRandFcn(randSize, "Random " .. optName .. " - Lua");
		valName = valName .. " --> Chose " .. GetOptionValueName(customOption, optIndex, userInput);
	end
	
	print(optName .. ": " .. valName);
	
	return userInput;
end

function ChooseInitDataOption(customOptions, optIndex, fallback)
	-- Choose an option that initializes in GetMapInitData().
	-- Can only randomize if we have native RNG.
	--
	-- Args:
	--   customOptions: table of all GotLakes map options.
	--   optIndex: which option to initialize
	--   fallback: what to set if user selected random but RNG not available.
	--
	-- Returns: the value to set for the map option
	
	local userInput = MapConfiguration.GetValue(MapOptionNames[optIndex]);
		
	local customOption = customOptions[optIndex];
	
	local optName = GetOptionName(customOption);
	local valName = "";
	
	if (not userInput) then
		-- User input not available, so use default value instead.
		userInput = customOption.DefaultValue;
		valName   = "Not Supported --> ";
	end
	
	valName = valName .. GetOptionValueName(customOption, optIndex, userInput);
		
	local randSize = customOption.RandSize;
	
	if (userInput == (randSize + 1)) then
		-- The user selected "Random", so randomize the option if possible.
		if (ScriptData.configRandFcn == GL_RandInt) then
			-- Got a native RNG, so let's randomize this option.
			userInput = 1 + GL_RandInt(randSize, "Random " .. optName .. " - Lua");
			valName = valName .. " --> Chose " .. GetOptionValueName(customOption, optIndex, userInput);
			
		else
			-- No RNG available because Civ RNG not yet seeded.
			userInput = fallback;
			valName = valName .. " --> RNG Not Available --> " .. 
				GetOptionValueName(customOption, optIndex, userInput);
		end
	end
	
	print(optName .. ": " .. valName);
	
	return userInput;
end

function ChooseFromShuffle(customOptions, firstOpt, lastOpt)
	-- Given a first and last option index, create a "deck" out of 
	-- map-filter user inputs and randomly choose an option.
	--
	-- Args:
	--   customOptions: info about the map menu
	--   firstOpt: the first shuffle family option to consider
	--   lastOpt: the last shuffle family option to consider
	-- 
	-- Returns
	--   selectedInfo: config info corresponding to the user input
	
	local deck = {};
	local deckNames = {};
	local numOpts = lastOpt - firstOpt + 1;

	-- Separate deck containing all options in case the user selects
	-- "none" for everything. This is a fallback.
	local noneDeck = {};

	local familyOpt;
	local familyInfos;

	for iOpt=firstOpt,lastOpt do
		-- For each family option, determine what values to include

		-- Get info about the option
		local familyOpt = GetOption(customOptions, iOpt);
				
		-- Get the value that the user selected.
		local familyChoice = MapConfiguration.GetValue(MapOptionNames[iOpt]);
		
		if (not familyChoice) then
			-- Family option not supported, so use default value instead.
			familyChoice = familyOpt.DefaultValue;
			print(GetOptionName(familyOpt) .. " not supported --> default " .. familyChoice);
		end
		
		-- Get info about the choices for this option
		local familyInfos = familyOpt.Infos or {};
		
		local familyNumInfos = table.maxn(familyInfos);
		
		if (familyChoice == (familyNumInfos - GOT_ALL)) then
			-- ALL selected for this option family.
			for i, info in ipairs(familyInfos) do
				table.insert(deck, info);
				table.insert(deckNames, info.name);
			end
			
		elseif (familyChoice <= familyNumInfos) then
			-- Single landmass selected.
			table.insert(deck,      familyInfos[familyChoice]);
			table.insert(deckNames, familyInfos[familyChoice].name);
			
		else
			-- NONE selected. Populate the fallback deck.
			for i, info in ipairs(familyInfos) do
				table.insert(noneDeck, info);
			end
		end
	end

	local deckSize = table.maxn(deck);
	local selectedInfo;

	if (deckSize == 0) then
		-- Edge case: the deck is empty ("None" for every family!) 
		-- Revert to an even distribution of map types.
		print("Empty deck!  Choosing an option at random.");
		
		selectedInfo = noneDeck[1 + ScriptData.configRandFcn(table.maxn(noneDeck), 
			"Random option, even distribution")];
	else
		-- Choose a random map from the deck.
		print("Choosing from {");
		print(unpack(deckNames));
		print("}");
		selectedInfo = deck[1 + ScriptData.configRandFcn(deckSize, 
			"Random option, weighted by user selection")];
	end
	
	print("Chose " .. selectedInfo.name);

	return selectedInfo;
end

function DetectMods()
	-- Stub to detect other mods that Got Lakes might want to support.
end

function DoShuffle(randFunction)
	-- Obtain options based on user input.
	-- No-op if a previous call to doShuffle() has already generated this info.
	-- Args:
	--   randFunction: call this function to generate a random number in [0,N].
	--                 nil indicates no RNG available
	
	if (not ScriptData.configRandFcn) then
		-- Decide which map to generate; choose lazily in case this 
		-- getter-method gets multiple calls.
		
		-- Set RNG if available.
		local gotRNG = nil;
		if randFunction then
			ScriptData.configRandFcn = randFunction;
			gotRNG = true;
		end
		
		local iNumCivs = PlayerManager.GetAliveMajorsCount();		
		if (iNumCivs == 0) then
			ScriptData.WBDetected = true;
			print("Player info is missing, so we're probably running in the World Builder.");
			print("Pretending that there are " .. ScriptData.WBNumCivs .. " civs.");
		end
		
		-- Load all options into global variables.
		print("Loading options...");
		
		local customOptions = GL_GetCustomOptions();
		
		if gotRNG then
			-- Populate global variables so that other map methods need not invoke MapConfiguration.GetValue().
			UserInput.worldAge    = ChooseOption(customOptions, MapOptions.WORLDAGE);
			UserInput.seaLevel    = ChooseOption(customOptions, MapOptions.SEALEVEL);
			UserInput.temperature = ChooseOption(customOptions, MapOptions.TEMPERATURE);
			UserInput.rainfall    = ChooseOption(customOptions, MapOptions.RAINFALL);
			UserInput.resources   = ChooseOption(customOptions, MapOptions.RESOURCES);
			UserInput.start       = ChooseOption(customOptions, MapOptions.START);
			UserInput.mountains   = ChooseOption(customOptions, MapOptions.MOUNTAINS);
			UserInput.extras      = ChooseOption(customOptions, MapOptions.EXTRAS);
		end
		
		-- Determine world wrap
		-- Note: Civ seeds its RNG *after* initializing world wrap.
		UserInput.worldWrap   = ChooseInitDataOption(
			customOptions, MapOptions.WORLDWRAP, WWRAP_REGION);

		if (UserInput.worldWrap == WWRAP_GLOBE) then
			-- Globe
			ScriptData.wrapX = true;
			ScriptData.wrapY = false;
			
			-- Enable horizontal world-wrap for all fractals that use global flags.
			g_iFlags.FRAC_WRAP_X = true;
		else
			-- Region
			ScriptData.wrapX = false;
			ScriptData.wrapY = false;
			
			-- Avoid edges for all fractals.
			g_iFlags.FRAC_POLAR = true;
		end
		
		if gotRNG then
			-- Shuffle with supplied RNG.
			print("Shuffling landmass and climate settings based on user input.");
			
			ScriptData.landInfo = GetLandInfo(ChooseFromShuffle(
				customOptions, FIRST_LAND_OPT, LAST_LAND_OPT));
				
			ScriptData.climateInfo = GetClimateInfo(ChooseFromShuffle(
				customOptions, FIRST_CLIMATE_OPT, LAST_CLIMATE_OPT));
			
			-- Define some globals to simplify input-checking later on.
			UserInput.landmass = ScriptData.landInfo.id;
			UserInput.climate  = ScriptData.climateInfo.id;
			
			gotExtrasLandmass = 
				((UserInput.landmass == Landmasses.EXTRAS) or 
				 (UserInput.landmass == Landmasses.INVEXTRAS));
			
			if (gotExtrasLandmass and (UserInput.extras   == ExtraIslesLakes.NONE)) then
				print("Landmass is " .. ScriptData.landInfo.name .. 
						"; cannot have \"None\" for extras!");
				-- Reshuffle choice for extras to avoid "None."
				-- Assumes that None == 1.
				UserInput.extras = 2 + ScriptData.configRandFcn(NUM_EXTRAS_OPTS-1, "Reshuffle extras.");
				print("New value for extras: " .. UserInput.extras);
			end
			
			ExtrasGot.miniDonuts = 
				(UserInput.extras == ExtraIslesLakes.DONUTS) or
				(UserInput.extras == ExtraIslesLakes.ALL);
			
			ExtrasGot.chains = 
				(UserInput.extras == ExtraIslesLakes.CHAINS) or
				(UserInput.extras == ExtraIslesLakes.TCHAINS) or
				(UserInput.extras == ExtraIslesLakes.ALL);
				
			ExtrasGot.clusters = 
				(UserInput.extras == ExtraIslesLakes.CLUSTERS) or
				(UserInput.extras == ExtraIslesLakes.ALL);
				
			ExtrasGot.tectonic = 
				(UserInput.extras == ExtraIslesLakes.TECTONIC) or
				(UserInput.extras == ExtraIslesLakes.TCHAINS) or
				(UserInput.extras == ExtraIslesLakes.ALL);
			
			ExtrasGot.tecAndChains	= 
				(UserInput.extras == ExtraIslesLakes.TCHAINS);
				
			ExtrasGot.everything = 
				(UserInput.extras == ExtraIslesLakes.ALL);	
				
			ExtrasGot.anything = 
				(ExtrasGot.miniDonuts == true or 
				ExtrasGot.chains      == true or 
				ExtrasGot.clusters    == true or 
				ExtrasGot.tectonic    == true);
			
			gotLakesFamily = 
				(UserInput.landmass == Landmasses.LAKES) or
				(UserInput.landmass == Landmasses.SALTLAKES) or
				(UserInput.landmass == Landmasses.LRGLAKES) or
				(UserInput.landmass == Landmasses.INLANDSEA);
				
			gotRegionalInlandSea = 
				(UserInput.landmass  == Landmasses.INLANDSEA) and
				(UserInput.worldWrap == WWRAP_REGION);
			
			gotIceAgeFamily = 
				(UserInput.climate == Climates.ICEAGECROP) or 
				(UserInput.climate == Climates.ICEAGEFULL);
				
			gotTiltedFamily = 
				(UserInput.climate == Climates.TILTEDCROP) or 
				(UserInput.climate == Climates.TILTEDFULL) or 
				(UserInput.climate == Climates.TWOSUNS);
		end
		
		-- Detect other mods in case we need to avoid overrides.
		DetectMods();
		
		-- Initialize world-size config.
		InitWorldSizeInfo();
	end
end

function GetMapInitData(worldSize)
	-- This function can reset map grid sizes and world wrap settings.

	print("Running Got Lakes v" .. glVersion);
	
	ScriptData.worldSize = worldSize;
	
	-- DO NOT generate random numbers with TerrainBuilder in this method.
	-- This method happens before initializing the TerrainBuilder's RNG.

	local mapSeed = MapConfiguration.GetValue("RANDOM_SEED");

	if mapSeed then
		-- Use native RNG so that we can randomize world wrap.
		-- Note: multiplayer games between different platforms might not work.
		--
		-- Maybe only do this for single-player games, and 
		-- maybe force multiplayer games to use non-random world wrap.
		--
		-- However, other mods use native RNG in this method and still "support 
		-- multiplayer".
		GL_RandSeed(mapSeed);
		
		-- Shuffle map options using native RNG.
		DoShuffle(GL_RandInt);
	else
		print("Map seed not yet available.");
				
		-- Configure map-init data without using RNG.
		DoShuffle();
	end
	
	-- Set grid size based on user input.
	local grid_size = ScriptData.worldSizeInfo.size;
	
	return {
		Width  = grid_size[1],
		Height = grid_size[2],
		WrapX  = ScriptData.wrapX,
		WrapY  = ScriptData.wrapY
	}; 
end
------------------------------------------------------------------------------
-- Map Generator
------------------------------------------------------------------------------
function GeneratePlotTypes()
	print("Generating Plot Types (" .. ScriptData.landInfo.name ..")");
	local plots;
	
	-- Create the appropriate fractal world for the landmass.
	local fractal_world = ScriptData.landInfo.CreateFWorld();
	
	if (ScriptData.landInfo.CreateFWorld == MultilayeredFractal.Create) then
		------------------------------
		-- MLF landmasses
		------------------------------
		fractal_world.GeneratePlotsByRegion = ScriptData.landInfo.GenPlotTypes;
		
		plots = fractal_world:GeneratePlotsByRegion();
	else
		------------------------------
		-- Fractal World landmasses
		------------------------------
		fractal_world.AddLandAndWater      = GotLakes_Fractal.AddLandAndWater;
		fractal_world.AddMtsHillsExtras    = GotLakes_Fractal.AddMtsHillsExtras;
		fractal_world.GeneratePlotTypes    = ScriptData.landInfo.GenPlotTypes;
	
        if (ScriptData.landInfo.GenPlotTypes == 
				GotLakes_Fractal.Fractal_GeneratePlotTypes) then
			---------------------------
			-- Archipelago
			-- Fractal
			-- Small Continents
			-- Tiny Islands
			--------------------------- 	
			fractal_world:InitFractal(ScriptData.landInfo.fractalInitArgs);
		end
		
		-- Generate plots for a non-MLF fractal world.
		plots = fractal_world:GeneratePlotTypes(ScriptData.landInfo.fractalGenArgs);
	end
	
	--CheckForNilPlotTypes(plots);
	return plots;
end

function GenerateCoasts(plotTypes, terrainTypes)
	print("Setting coasts and oceans...");
	local args = ScriptData.landInfo.coastArgs;
	
	local bExpandCoasts            = args.bExpandCoasts or true;
	local expansion_diceroll_table = args.expansion_diceroll_table or {4, 4};
	
	local shallowWater = g_TERRAIN_TYPE_COAST;
	local deepWater    = g_TERRAIN_TYPE_OCEAN;
		
	for iX = 0, g_iW - 1 do
		for iY = 0, g_iH - 1 do
			local index = (iY * g_iW) + iX + 1;
			if (plotTypes[index] == g_PLOT_TYPE_OCEAN) then
				if (IsAdjacentToLand(plotTypes, iX, iY)) then
					terrainTypes[index] = shallowWater;
				else
					terrainTypes[index] = deepWater;
				end
				
			-- Initialize all land as grass
			elseif (plotTypes[index] == g_PLOT_TYPE_MOUNTAIN) then
			    terrainTypes[index] = g_TERRAIN_TYPE_GRASS_MOUNTAIN;
				
			elseif (plotTypes[index] == g_PLOT_TYPE_HILLS) then
				terrainTypes[index] = g_TERRAIN_TYPE_GRASS_HILLS;
				
			elseif (plotTypes[index] == g_PLOT_TYPE_LAND) then
				terrainTypes[index] = g_TERRAIN_TYPE_GRASS;
			
			else
				print("Unexpected plot type at " .. 
				CoordString(iX, iY) .. ": " .. NonNilStr(plotTypes[index]));
			end
		end
	end
	
	if bExpandCoasts == false then
		return
	end

	print("Expanding coasts");
	for iI = 0, 2 do
		local shallowWaterPlots = {};
		for iX = 0, g_iW - 1 do
			for iY = 0, g_iH - 1 do
				local index = (iY * g_iW) + iX + 1;
				if (terrainTypes[index] == deepWater) then
					-- Chance for each eligible plot to become an expansion is 1 / iExpansionDiceroll.
					-- Default is two passes at 1/4 chance per eligible plot on each pass.
					if (IsAdjacentToShallowWater(terrainTypes, iX, iY) and 
					    TerrainBuilder.GetRandomNumber(4, "add shallows") == 0) then
						table.insert(shallowWaterPlots, index);
					end
				end
			end
		end
		for i, index in ipairs(shallowWaterPlots) do
			terrainTypes[index] = shallowWater;
		end
	end
	
	
	local tectonic_shallows  = args.tectonic_shallows;
	
	if (tectonic_shallows == nil) then 
		tectonic_shallows = true; 
	end
	
	-- The user ultimately decides whether we add tectonic shallow water.
	if ((tectonic_shallows == true) and (ExtrasGot.tectonic == true)) then
		print("Adding submerged tectonic islands...");
	
		-- Higher sea level => higher threshold => less shallow water from mountains
		local seaLevels = {0, 15, 30}; 
		local seaLevel = seaLevels[UserInput.seaLevel];
		
		local mtnThresh = ScriptData.tectonicIslesFractal:GetHeight(15 + seaLevel);
		local mtnCount = 0;
		
		for iX = 0, g_iW - 1 do
			for iY = 0, g_iH - 1 do
				local index = (iY * g_iW) + iX + 1;
				if(terrainTypes[index] == deepWater) then
					-- Check mountain and tectonic island maps, 
					-- decide whether to add shallow water for submerged mountain
					
					if (ScriptData.mtnPlots[i] == g_PLOT_TYPE_MOUNTAIN) then
						-- A mountain from the mountain pattern.
						
						local tectonicVal = ScriptData.tectonicIslesFractal:GetHeight(iX,iY);
						
						if (tectonicVal >= mtnThresh) then
							-- Make shallow water for a submerged mountain.
							terrainTypes[index] = shallowWater;
							mtnCount = mtnCount + 1;
						end
					end
				end
			end
		end
	
		print("Added " .. mtnCount .. " submerged mountains.");
	end	
end

function SetTerrainTypes(terrainTypes)
	-- Set terrain types for the Map/Plot API.
	for y = 0, g_iH - 1, 1 do
		for x = 0, g_iW - 1, 1 do
			local i = y * g_iW + x;
			local plot = Map.GetPlotByIndex(i);
			
			local terrainType = terrainTypes[i+1];
			if terrainType then
				TerrainBuilder.SetTerrainType(plot, terrainType);
			else
				print("Nil terrain type at " .. CoordString(x,y));
			end
		end
	end
	
	-- Each body of water, area of mountains, or area of hills+flatlands 
	-- is independently grouped and tagged.
	AreaBuilder.Recalculate();
end

function AddRivers()

	if (gotRegionalInlandSea) then
		-- Special logic for Inland Sea + Region world shape
		return InlandSea_AddRivers();
	end

	-- Default behavior.
	return Normal_AddRivers();

end

function CanHavePostRiverLake(plot)
	-- Return true iff the plot is a valid candidate for a post-river lake.
	
	-- Avoid plots that are already water because adding a lake to that would have no effect.
	-- Avoid mountains because mountains might be surrounded by other mountains.
	-- Avoid coastal land lest coastal water creeps into the landmass.
	-- Avoid rivers because tiles with rivers already have access to fresh water.
	-- Don't worry about natural wonders--we didn't add those yet.
	return ((plot) and -- Avoid NPE
		   (plot:IsWater() == false) and 
		   (plot:IsMountain() == false) and 
		   (plot:IsCoastalLand() == false) and 
		   (plot:IsRiver() == false and plot:IsRiverAdjacent() == false)); 
end

function AddMoreLake(plot, lakePlots)
	-- Maybe expand a post-river lake plot into a multi-plot lake.
	-- Args: 
	--   plot: lake-plot from which to expand
	--   lakePlots: accumulator of plots to turn into water
	-- Return the number of extra lake tiles added
	local largeLake = 0;
	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		local adjacentPlot = Map.GetAdjacentPlot(plot:GetX(), plot:GetY(), direction);
		
		if CanHavePostRiverLake(adjacentPlot) then
			local r = TerrainBuilder.GetRandomNumber(4 + largeLake, "AddMoreLake");
			if r < 3 then
				--print("Expanding lake at " .. CoordString(adjacentPlot:GetX(),adjacentPlot:GetY()));
				-- Add to lake plots; don't turn land into water yet because that
				-- might prevent expansion to other adjacent plots.
				table.insert(lakePlots, adjacentPlot);
				largeLake = largeLake + 1;
			end		
		end
	end
	return largeLake;
end

function AddLakes()
	if ((UserInput.landmass == Landmasses.LAKES) and 
	    (ExtrasGot.anything == false)) then
		-- The Lakes landmass with fractal-lakes already has plenty of fresh lakes.
		print("Skipping post-river lakes (lakes + no extras).");
		return;
	end

	-- Maybe add some post-river lakes.
	print("Adding post-river lakes.");
			
	local lakePlotRand = GlobalParameters.LAKE_PLOT_RANDOM  or 25;
	
	-- Allow rainfall to raise/lower water-tables for lakes.
	local rainLevel = UserInput.rainfall;
	
	-- Higher rainfall => smaller lake dice => higher frequency of random lakes.
	local lakeMults   = {2.0, 1.0, 0.5};
	local lakeMult    = lakeMults[rainLevel] or 0.5;
	lakePlotRand = lakePlotRand * lakeMult;
	
	-- Turn the dice into a natural number.
	lakePlotRand = math.max(1,Round(lakePlotRand));
		
	print("Chance of lake plot: 1/" .. lakePlotRand);
	
	local numLakeTilesAdded  = 0;
	local numLakesAdded      = 0;
	local numLargeLakesAdded = 0;

	for i = 0, (g_iW * g_iH) - 1, 1 do
		plot = Map.GetPlotByIndex(i);
		
		if CanHavePostRiverLake(plot) then 
			-- This plot has lake potential.
			-- Roll dice to decide whether to make this plot a lake tile.
			local lakeDice = 
				TerrainBuilder.GetRandomNumber(lakePlotRand, "Post-River Lakes");
				
			if (lakeDice == 0) then
				-- Turn this tile and maybe some adjacent tiles into a lake.
				-- Make a list of plots to turn into water so that we can add
				-- more lakes without confusing neighboring tiles with 
				-- "coastal land".
				local lakePlots = {plot};
				
				numLakesAdded     = numLakesAdded + 1;
				
				-- Try to expand the lake into adjacent tiles.
				local numExtraLakesAdded = AddMoreLake(plot, lakePlots);
				if(numExtraLakesAdded > 2) then
					numLargeLakesAdded = numLargeLakesAdded + 1;
				end
				
				-- Turn lake plots into water.
				for i,lakePlot in ipairs(lakePlots) do 
					numLakeTilesAdded = numLakeTilesAdded + 1;
					TerrainBuilder.SetTerrainType(lakePlot, g_TERRAIN_TYPE_COAST);
				end
			end
		end
	end
	
	-- this is a minimalist update because lakes have been added
	if numLakeTilesAdded > 0 then
		print("Lake tiles added:  " .. numLakeTilesAdded)
		print("Total Lakes added: " .. numLakesAdded)
		print("Large Lakes added: " .. numLargeLakesAdded)
		AreaBuilder.Recalculate();
	end
end

function InitLatVariation()
	-- Initialize the latitude variation fractal.
	
	local grain = ScriptData.worldSizeInfo.latVariationGrain or 4;
	
	if (gotTiltedFamily == true) then
		-- Use high-grain for tilted-axis climates.
		-- Makes latitude bands more coherent.
		grain = 5;
	end
	
	print("Latitude variation grain: " .. grain);
	
	local varFrac = Fractal.Create(g_iW, g_iH, grain, g_iFlags, g_fracXExp, g_fracYExp);
	ScriptData.latVariationFrac = varFrac;
	
	ScriptData.latVariationMinHeight  = varFrac:GetHeight(0);
	ScriptData.latVariationMaxHeight  = varFrac:GetHeight(100);
	ScriptData.latVariationHeightDiff = 
		ScriptData.latVariationMaxHeight - ScriptData.latVariationMinHeight;
end

function GenerateTerrain()
	print("Generating Terrain (" .. ScriptData.climateInfo.name .. ")");
	
	-- Get args from climate info.
	local args = ScriptData.climateInfo.terrainArgs;
	
	
	-- Create the terrain generator.
	local terraingen = ScriptData.climateInfo.createTGen(args);
	
	-- Apply overrides based on world shape.
	if (UserInput.worldWrap == WWRAP_REGION) then
		print("Applying overrides for a regional world.");
			
		terraingen.GenerateTerrainAtPlot = 
			ScriptData.climateInfo.RegionGenerateTerrainAtPlot;
	end
	
	-- Generate terrain.
	terraingen:GenerateTerrain();	
end

function GetLatitudeAtPlot(variationFrac, iX, iY)
	-- Override MapUtilities to allow for customization.

	-- Apply overrides based on world shape.
	if (UserInput.worldWrap == WWRAP_REGION) then
		-- Region map => fixed latitude for most climates
		return ScriptData.climateInfo.RegionGetLatitudeAtPlot(variationFrac, iX, iY);
	end
	
	-- Globe map => bands of latitude
	return ScriptData.climateInfo.GlobeGetLatitudeAtPlot(variationFrac, iX, iY);
end

function GenerateNaturalWonders()
	print("Adding natural wonders");
	local wonderArgs = 
	{
		numberToPlace = GameInfo.Maps[Map.GetMapSize()].NumNaturalWonders,
	};
	NaturalWonderGenerator.Create(wonderArgs);
		
	-- Recalculate land areas and check for chokepoints 
	-- before adding cliffs.
	AreaBuilder.Recalculate();
end

function SetCliff(iX, iY)
	local pPlot = Map.GetPlot(iX,iY);

	for direction = 0, DirectionTypes.NUM_DIRECTION_TYPES - 1, 1 do
		local adjacentPlot = Map.GetAdjacentPlot(iX, iY, direction);
		if (adjacentPlot ~= nil) then
			if (adjacentPlot:IsWater() == true) then
				if(direction == DirectionTypes.DIRECTION_NORTHEAST) then
					TerrainBuilder.SetNEOfCliff(adjacentPlot, true);
				elseif(direction == DirectionTypes.DIRECTION_EAST) then
					TerrainBuilder.SetWOfCliff(pPlot, true); 
				elseif(direction == DirectionTypes.DIRECTION_SOUTHEAST) then
					TerrainBuilder.SetNWOfCliff(pPlot, true); 
				elseif(direction == DirectionTypes.DIRECTION_SOUTHWEST) then
					TerrainBuilder.SetNEOfCliff(pPlot, true); 
				elseif(direction == DirectionTypes.DIRECTION_WEST) then
					TerrainBuilder.SetWOfCliff(adjacentPlot, true); 
				elseif(direction == DirectionTypes.DIRECTION_NORTHWEST) then
					TerrainBuilder.SetNWOfCliff(adjacentPlot, true); 
				end
			end
		end
	end
end

function AddCliffs()
	print("Adding cliffs");
	for iX = 0, g_iW - 1 do
		for iY = 0, g_iH - 1 do	
		
			if (MapGetPlotType(iX,iY) == g_PLOT_TYPE_HILLS and 
				-- Allow cliffs next to lakes or coasts
				IsAdjacentToWater(iX, iY) == true and 
				IsAdjacentToIce(iX, iY)   == false and
				IsAdjacentToRiver(iX, iY) == false) then
				
				local plot = Map.GetPlot(iX, iY);
				local area = plot:GetArea();
				
				-- Don't worry about areas not having flat coast.
				-- Mountain reachability will break cliffs if necessary.
				-- Also, Lakes maps might not have any coast, but just fresh lakes.
				if (area:GetPlotCount() > 1) then
					SetCliff(iX, iY);
				end
			end
		end
	end
end

function EnsureTerrainReachability()
	-- Check for areas that are completely surrounded by natural wonders.
	local reachArgs = 
	{
		obstaclesName   = "natural wonders",
		InitReachPlot   = PassMaker.Wonder_InitReachPlot,
		RemoveObstacle  = PassMaker.Wonder_RemoveObstacle,
	};
	EnsureReachabilityWithPassMaker(reachArgs);
	
	-- Reset terrain under natural wonders.
	print("Normalizing terrain under natural wonders.");
	for x = 0, g_iW - 1 do
		for y = 0, g_iH - 1 do
			local plot = Map.GetPlot(x, y);
			
			if plot:IsNaturalWonder() then
				ResetNaturalWonderTerrain(plot);
			end
		end
	end
	
	-- Check for areas that are completely surrounded by mountains and cliffs.
	-- PassMaker's default behavior is mountain reachability.
	reachArgs = { };
	EnsureReachabilityWithPassMaker(reachArgs);
	
	AreaBuilder.Recalculate();
end

function GenerateFeatures()
	
	print("Generating Features (" .. ScriptData.climateInfo.name .. ")");
	
	-- Get args from climate info.
	local args = ScriptData.climateInfo.featureArgs;
	
	-- Create the feature generator.
	local featuregen = ScriptData.climateInfo.createFGen(args);
	
	-- Apply overrides based on world shape.
	if (UserInput.worldWrap == WWRAP_REGION) then
		print("Applying overrides for a regional world.");
		
		featuregen.AddIceAtPlot = 
			ScriptData.climateInfo.RegionAddIceAtPlot;
	end
	
	featuregen:AddFeatures();
	
	-- Ensure that all non-mountain tiles are still reachable despite ice.
	local reachArgs = 
	{
		obstaclesName   = "ice",
		InitReachPlot   = PassMaker.Ice_InitReachPlot,
		RemoveObstacle  = PassMaker.Ice_RemoveObstacle,
	};
	EnsureReachabilityWithPassMaker(reachArgs);
end

function AnalyzeLandmass()
	-- Calculate debug stats for land/water plots
	
	local numTiles = g_iW * g_iH;
	
	local countWater = 0;
	local countDeep  = 0;
	local countMtn   = 0;
	local countHill  = 0;
	local countFlat  = 0;
	
	-- Compute a histogram of plot types.
	for y = 0, g_iH - 1, 1 do
		for x = 0, g_iW - 1, 1 do
			local i = y * g_iW + x;
			local plot = Map.GetPlotByIndex(i);
			
			if not plot:GetTerrainType() then
				print("No terrain at " .. CoordString(x,y));
			end
			
			if(plot:IsWater()) then
				countWater = countWater + 1;
				if(plot:GetTerrainType() == g_TERRAIN_TYPE_OCEAN) then
					countDeep = countDeep + 1;
				end
				
			elseif plot:IsMountain() then
				countMtn = countMtn + 1;
			
			elseif plot:IsHills() then
				countHill = countHill + 1;
			else
				countFlat = countFlat + 1;
			end
		end
	end
	
	local countLand    = numTiles   - countWater;
	local countShallow = countWater - countDeep;
	local countHabitable = countHill + countFlat;

	print("===========================================");
	print("Landmass analysis:");
	print("-------------------------------------------");
	print("Map Elevation:");
	print("  Mountains  ", PercentString(countMtn, numTiles));
	print("  Hills      ", PercentString(countHill, numTiles));
	print("  Flats      ", PercentString(countFlat, numTiles));
	print("  Deep       ", PercentString(countDeep, numTiles));
	print("  Shallow    ", PercentString(countShallow, numTiles));
	print("-------------------------------------------");
	print("Land Elevation:");
	print("  Mountains  ", PercentString(countMtn, countLand));
	print("  Hills      ", PercentString(countHill, countLand));
	print("  Flats      ", PercentString(countFlat, countLand));
	print("-------------------------------------------");
	print("Water Elevation:");
	print("  Deep       ", PercentString(countDeep, countWater));
	print("  Shallow    ", PercentString(countShallow, countWater));
	print("-------------------------------------------");
	print("Misc:");
	print("  Habitable: ", PercentString(countHabitable, numTiles));
	print("  Water:     ", PercentString(countWater, numTiles));
	print("===========================================");
end

function AnalyzeClimate()
	-- Calculate debug stats for terrain and features
	
	local numTiles = g_iW * g_iH;
	
	-- Land and water for terrain/feature coverage stats.
	local countLand   = 0; -- Includes flat land and hills.
	local countWater  = 0; -- Includes ocean, coast, and lakes.
	
	-- Land terrain types.
	local countGrass  = 0;
	local countPlains = 0;
	local countDesert = 0;
	local countTundra = 0;
	local countSnow   = 0;
	
	-- Land features (not including wonders).
	local countFNone  = 0; -- Land without features.
	local countFlood  = 0;
	local countJungle = 0;
	local countForest = 0;
	local countOasis  = 0;
	local countMarsh  = 0;
	
	-- Water features (not including wonders).
	local countIce    = 0;
	
	
	-- Compute a histogram of terrain and features
	for y = 0, g_iH - 1, 1 do
		for x = 0, g_iW - 1, 1 do
			local i = y * g_iW + x;
			local plot = Map.GetPlotByIndex(i);
			
			if (plot:IsMountain() == false) then
				-- Ignore terrain on mountains.

				local featureType = plot:GetFeatureType();

				
				if (plot:IsWater()) then
					-- Water plot.
					countWater = countWater + 1;
					
					if (featureType == g_FEATURE_ICE) then
						countIce = countIce + 1;
					end
					
				else
					-- Land plot.
					countLand = countLand + 1;
				
					local terrainType = plot:GetTerrainType();

					if (IsBaseTerrainType(g_TERRAIN_TYPE_GRASS, terrainType)) then
						countGrass = countGrass + 1;
					
					elseif (IsBaseTerrainType(g_TERRAIN_TYPE_PLAINS, terrainType)) then
						countPlains = countPlains + 1;
					
					elseif (IsBaseTerrainType(g_TERRAIN_TYPE_DESERT, terrainType)) then
						countDesert = countDesert + 1;

					elseif (IsBaseTerrainType(g_TERRAIN_TYPE_TUNDRA, terrainType)) then
						countTundra = countTundra + 1;

					elseif (IsBaseTerrainType(g_TERRAIN_TYPE_SNOW, terrainType)) then
						countSnow = countSnow + 1;
					else
						print("Unexpected terrain at " .. 
						CoordString(x,y) .. ": " .. NonNilStr(terrainType));
					end
						
					if (featureType == g_FEATURE_NONE) then
						countFNone = countFNone + 1;
						
					elseif (featureType == g_FEATURE_FLOODPLAINS) then
						countFlood = countFlood + 1;
						
					elseif (featureType == g_FEATURE_JUNGLE) then
						countJungle = countJungle + 1;
						
					elseif (featureType == g_FEATURE_FOREST) then
						countForest = countForest + 1;
						
					elseif (featureType == g_FEATURE_OASIS) then
						countOasis = countOasis + 1;
						
					elseif (featureType == g_FEATURE_MARSH) then
						countMarsh = countMarsh + 1;
					end
				end	
			end
		end
	end
	
	-- Temperature analysis: count grass as 1/2 temperate, 1/2 warm.
	local countCold      = countTundra        + countSnow;
	local countTemperate = Round(countGrass/2 + countPlains);
	local countWarm      = Round(countGrass/2 + countDesert);
	
	-- Terrain Fertility
	local countBarren  = countSnow + countDesert;
	local countFertile = countGrass + countPlains;
	
	-- Vegetation
	local countTrees = countJungle + countForest;
	local countWet   = countOasis  + countMarsh;
	local countOther = countLand - countTrees - countWet;

	print("===========================================");
	print("Climate analysis:");
	print("-------------------------------------------");
	print("Land Terrain:");
	print("  Grass       ", PercentString(countGrass,		countLand));
	print("  Plains      ", PercentString(countPlains, 		countLand));
	print("  Desert      ", PercentString(countDesert, 		countLand));
	print("  Tundra      ", PercentString(countTundra, 		countLand));
	print("  Snow        ", PercentString(countSnow, 		countLand));
	print("-------------------------------------------");
	print("Land Temperature:");
	print("  Cold        ", PercentString(countCold, 		countLand));
	print("  Temperate   ", PercentString(countTemperate, 	countLand));
	print("  Warm        ", PercentString(countWarm, 		countLand));
	print("-------------------------------------------");
	print("Land Fertility:");
	print("  Barren:     ", PercentString(countBarren, 		countLand));
	print("  Fertile:    ", PercentString(countFertile, 	countLand));
	print("-------------------------------------------");
	print("Land Features:");
	print("  None        ",	PercentString(countFNone, 		countLand));
	print("  Floodplains ", PercentString(countFlood, 		countLand));
	print("  Jungle      ", PercentString(countJungle, 		countLand));
	print("  Forest      ", PercentString(countForest, 		countLand));
	print("  Oasis       ", PercentString(countOasis, 		countLand));
	print("  Marsh       ", PercentString(countMarsh, 		countLand));
	print("-------------------------------------------");
	print("Land Vegetation:");
	print("  Trees       ", PercentString(countTrees, 		countLand));
	print("  Wetlands    ", PercentString(countWet, 		countLand));
	print("  Other       ", PercentString(countOther, 		countLand));
	print("-------------------------------------------");
	print("Water Features:");
	print("  Ice         ", PercentString(countIce, 		countWater));
	print("===========================================");
end

function GeneratePlotsWithFeatures()
	-- Generate plot types, terrain, and features.

	-- Plot types are the core layer of the map, 
	-- determining land or sea, determining flatland, hills or mountains.
	local plotTypes = GeneratePlotTypes();
	
	-- Initialize terrain types.
	local terrainTypes = table.fill(g_TERRAIN_TYPE_GRASS, g_iW * g_iH);
	
	-- Set shallow and deep water as terrain.
	GenerateCoasts(plotTypes, terrainTypes);
	
	-- Civ 6 doesn't "set" plot types, so instead we "set" terrain types 
	-- in order to unbreak rivers and lakes.
	SetTerrainTypes(terrainTypes); 
	
	-- River generation is affected by plot types, originating from 
	-- highlands and preferring to traverse lowlands.
	AddRivers();
	
	-- Lakes would interfere with rivers, causing them to stop and not 
	-- reach the ocean, if placed any sooner.
	AddLakes();
	
	-- Initialize the latitude variation fractal.
	-- Do this before generating terrain.
	InitLatVariation();
	
	-- Terrain covers climate: grassland, plains, desert, tundra, snow.
	GenerateTerrain();
	
	-- Throw in some random natural wonders.
	-- Natural wonders originally came after cliffs.
	-- But nat wonder cliffs don't check whether an area still has flat coast.
	GenerateNaturalWonders();
	
	-- Cliffs also depend on both plot types and terrain types.
	AddCliffs();
	
	-- Check for areas completely surrounded by impassable wonders, mountains, or cliffs.
	EnsureTerrainReachability();
		
	-- Features depend on plot types, terrain types, rivers and lakes 
	-- to help determine their placement.
	GenerateFeatures();
	
	-- Analyze land, hills, mountains, oceans, and coast.
	AnalyzeLandmass();
	
	-- Analyze terrain and features.
	AnalyzeClimate();
end

function DetermineContinents()
	-- Each plot has a continent art type. Mixing and matching these could look
	-- extremely bad, but there is nothing technical to prevent it. The worst 
	-- that will happen is that it can't find a blend and draws red checkerboards.
	--ScriptData.landInfo.DetermineContinents();
	
	-- For Civ 6, just use the default continent stamper.
	print("Using default continent stamper.");
	TerrainBuilder.StampContinents();
end

function GenerateResources()
	print("Generating resources (" .. 
		ScriptData.landInfo.name .. " + " .. 
		ScriptData.climateInfo.name .. ").");

	local resourceArgs = 
	{
		resources = UserInput.resources,
	};
	local resGen = ResourceGenerator.Create(resourceArgs);
end

function DetermineStartingPlots()

	print("Creating start plot database (" .. 
		ScriptData.landInfo.name .. " + " .. 
		ScriptData.climateInfo.name .. ").");

	-- Map options and order of precedence (highest to lowest):
	-- 1) Other mods
	-- 2) Climate options
	-- 3) Landmass options
	
	-- Use both land and climate info to calculate fertility thresholds.
	local majorCivFert = math.max(1,Round(
		ScriptData.landInfo.aspMinMajorFert * 
		ScriptData.climateInfo.aspMinMajorFert));
		
	local minorCivFert = math.max(1,Round( 
		ScriptData.landInfo.aspMinMinorFert * 
		ScriptData.climateInfo.aspMinMinorFert));
	
	-- Ignore the bottom START_MIN_Y % and the top START_MAX_Y % 
	-- of the grid height when placing major civs.
	local startMinY = 0;
	local startMaxY = 0;
	local divMethod = ScriptData.landInfo.aspDivMethod;
	
	if ((UserInput.mountains == MtnPatterns.WINDING) and 
	    (ScriptData.landInfo.aspDivMethod == DIVMETHOD_BIGGESTLAND)) then
		-- Winding canyons can divide a largest landmass into smaller areas,
		-- So use a division method that expects smaller areas
		print("Adjusting regional div method (winding canyons + largest landmass)");
		ScriptData.landInfo.aspDivMethod = DIVMETHOD_CONTINENTAL;
	end
	   
	
	if (UserInput.worldWrap == WWRAP_GLOBE) then
		-- Avoid polar starts for globe maps.
		startMinY = ScriptData.climateInfo.aspStartMinY;
		startMaxY = ScriptData.climateInfo.aspStartMaxY;
		
		if ((UserInput.climate == Climates.ICEAGEFULL) or 
			(UserInput.climate == Climates.TILTEDFULL)) then
			-- Override division method to sub-rectangular 
			-- in order to avoid starts in extreme temperatures.
			ScriptData.landInfo.aspDivMethod = DIVMETHOD_SUBRECT;
		end
	end
	
	-- Create the AssignStartingPlots database.
	-- Even though we might override an init method, use the 
	-- inherited constructor in case there are other 
	-- ASP-overriding mods.
	local args = 
	{
		MIN_MAJOR_CIV_FERTILITY = majorCivFert,
		MIN_MINOR_CIV_FERTILITY = minorCivFert,
		MIN_BARBARIAN_FERTILITY = 1,
		START_MIN_Y             = startMinY,
		START_MAX_Y             = startMaxY,
		LAND                    = ScriptData.landInfo.aspLandMap,
		WATER                   = ScriptData.landInfo.aspWaterMap,
		START_CONFIG            = UserInput.start,
	};
	local asp = AssignStartingPlots.Create(args);
end

function GenerateMap()
	print("Generating GotLakes Map");
	-- This is the core map generation function.
	
	-- Shuffle map options if we haven't already.
	-- Use the TerrainBuilder RNG since that is now available.
	DoShuffle(TerrainBuilder.GetRandomNumber);
	
	-- Set globals
	g_iW, g_iH = Map.GetGridSize();
	g_iFlags   = TerrainBuilder.GetFractalFlags();
	
	-- Generate most of the landmass and climate.
	GeneratePlotsWithFeatures();
	
	-- Note that this does NOT fix wonders that block chokepoints.
	print("Analyzing chokepoints.");
	TerrainBuilder.AnalyzeChokepoints();
	
	-- Determines continental plates and possibly artwork as well.
	DetermineContinents();
	
	-- Add luxuries and whatnot.
	GenerateResources();
	
	-- Assign Starting Plots.
	DetermineStartingPlots();

	-- Goodies depend on not colliding with resources or Natural Wonders, 
	-- or being placed too near to start plots.
	AddGoodies(g_iW, g_iH);
end